/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

// System include(s):
#undef NDEBUG
#include <iostream>
#include <sstream>
#include <cassert>

// Local include(s):
#include "xAODTrigger/gFexJetRoI.h"
#include "xAODTrigger/versions/gFexJetRoIAuxContainer_v1.h"
#include "xAODTrigger/gFexJetRoIAuxContainer.h"
#include "xAODTrigger/gFexJetRoIContainer.h"
#include "xAODTrigger/gFexGlobalRoI.h"
#include "xAODTrigger/versions/gFexGlobalRoIAuxContainer_v1.h"
#include "xAODTrigger/gFexGlobalRoIAuxContainer.h"
#include "xAODTrigger/gFexGlobalRoIContainer.h"
#include "TestTools/FLOATassert.h"

/// Helper macro for testing the code
#define SIMPLE_ASSERT( EXP )                                   \
   do {                                                        \
      const bool result = EXP;                                 \
      if( ! result ) {                                         \
         std::cerr << __FILE__ << ":" << __LINE__              \
                   << " Failed test: " << #EXP << std::endl;   \
         throw std::runtime_error( "Failed " #EXP );           \
      }                                                        \
   } while( 0 )


/// Testing the 32-bit TOB
void testgFexJetRoI() {

    // Create the container that we want to test:
   xAOD::gFexJetRoIAuxContainer aux;
   xAOD::gFexJetRoIContainer c;
   c.setStore( &aux );

   // Create a test object:
   xAOD::gFexJetRoI* obj = new xAOD::gFexJetRoI();
   c.push_back( obj );

   std::cout << "Initializing test object: jet TOB" << std::endl;
   uint32_t word = 213742597;
   int scale = 200.;
   obj->initialize(word, scale);

   SIMPLE_ASSERT( obj->isgRho() == 0 );
   SIMPLE_ASSERT( obj->isgBlockLead() == 0 );
   SIMPLE_ASSERT( obj->isgBlockSub() == 0 );
   SIMPLE_ASSERT( obj->isgJet() == 1 );
   SIMPLE_ASSERT( obj->gFexType() == 3 );
   
    
   SIMPLE_ASSERT( obj->et() == 688800 );
   SIMPLE_ASSERT( obj->eta() == static_cast<float>(-1.7) );
   SIMPLE_ASSERT( obj->etaMin() == static_cast<float>(-1.8) );
   SIMPLE_ASSERT( obj->etaMax() == static_cast<float>(-1.6) );
   SIMPLE_ASSERT( Athena_test::isEqual (obj->phi_gFex(), static_cast<float>(3*(2*M_PI/32)+(2*M_PI/(32*2))) ));
   SIMPLE_ASSERT( obj->phiMin_gFex() == static_cast<float>(3*(2*M_PI/32)) );
   SIMPLE_ASSERT( obj->phiMax_gFex() == static_cast<float>(3*(2*M_PI/32)+(2*M_PI/32)) );


   std::cout << "Test jet TOB completed!" << std::endl;


}

void testgFexGlobalRoI() {

    // Create the container that we want to test:
   xAOD::gFexGlobalRoIAuxContainer aux;
   xAOD::gFexGlobalRoIContainer c;
   c.setStore( &aux );

   // Create a test object:
   xAOD::gFexGlobalRoI* obj = new xAOD::gFexGlobalRoI();
   c.push_back( obj );

   std::cout << "Initializing test object: global TOB" << std::endl;
   uint32_t word = 192035055;
   int scale1 = 200.;
   int scale2 = 800.;
   obj->initialize(word, scale1, scale2);
   
   
   SIMPLE_ASSERT( obj->isgScalar() == 0 );
   SIMPLE_ASSERT( obj->isgMET() == 1 );
   SIMPLE_ASSERT( obj->isgMHT() == 0 );
   SIMPLE_ASSERT( obj->isgMST() == 0 );
   SIMPLE_ASSERT( obj->globalType() == 2 );
   
   SIMPLE_ASSERT( obj->METquantityOne() == 365400 );
   SIMPLE_ASSERT( obj->METquantityTwo() == -361800 );
   SIMPLE_ASSERT( obj->SumEt() == -999 );

   std::cout << "Test global TOB completed!" << std::endl;


}
int main() {

   // Run the tests:
   testgFexJetRoI();
   testgFexGlobalRoI();

   // Return gracefully:
   return 0;


}
