/*
  Copyright (C) 2002-2021 CERN for the benefit of the ATLAS collaboration
*/

// Class: ReadMLP
// Automatically generated by MethodBase::MakeClass
//

#include <vector>
#include <cmath>
#include <string>
#include <iostream>

#ifndef IClassifierReader__def
#define IClassifierReader__def

class IClassifierReader {

 public:

   // constructor
   IClassifierReader() {}
   virtual ~IClassifierReader() {}

   // return classifier response
   virtual double GetMvaValue( const std::vector<double>& inputValues ) const = 0;
};

#endif

class ReadMLP : public IClassifierReader {

 public:

   // constructor
   ReadMLP( std::vector<std::string>& theInputVars ) 
      : IClassifierReader(),
        m_className( "ReadMLP" ),
        m_statusIsClean( true ),
        m_nvars( 6 ),
        m_isNormalised( true )
   {      
      // the training input variables
      const char* inputVars[] = { "ptj", "egcore", "hadcore", "leadpt", "leadptchg", "highm5pt" };

      // sanity checks
      if (theInputVars.size() <= 0) {
         std::cout << "Problem in class \"" << m_className << "\": empty input vector" << std::endl;
         m_statusIsClean = false;
      }

      if (theInputVars.size() != m_nvars) {
         std::cout << "Problem in class \"" << m_className << "\": mismatch in number of input values: "
                   << theInputVars.size() << " != " << m_nvars << std::endl;
         m_statusIsClean = false;
      }

      // validate input variables
      for (size_t ivar = 0; ivar < theInputVars.size(); ivar++) {
         if (theInputVars[ivar] != inputVars[ivar]) {
            std::cout << "Problem in class \"" << m_className << "\": mismatch in input variable names" << std::endl
                      << " for variable [" << ivar << "]: " << theInputVars[ivar].c_str() << " != " << inputVars[ivar] << std::endl;
            m_statusIsClean = false;
         }
      }

      // initialize min and max vectors (for normalisation)
      m_Vmin[0] = 17.0000610351562;
      m_Vmax[0] = 99.9725875854492;
      m_Vmin[1] = 0;
      m_Vmax[1] = 24.9998874664307;
      m_Vmin[2] = 0;
      m_Vmax[2] = 98.9895172119141;
      m_Vmin[3] = 1.05319571495056;
      m_Vmax[3] = 98.9895172119141;
      m_Vmin[4] = 0;
      m_Vmax[4] = 98.9895172119141;
      m_Vmin[5] = 0;
      m_Vmax[5] = 117.28783416748;

      // initialize input variable types
      m_type[0] = 'F';
      m_type[1] = 'F';
      m_type[2] = 'F';
      m_type[3] = 'F';
      m_type[4] = 'F';
      m_type[5] = 'F';

      // initialize constants
      Initialize();

   }

   // destructor
   virtual ~ReadMLP() {
      Clear(); // method-specific
   }

   // the classifier response
   // "inputValues" is a vector of input values in the same order as the 
   // variables given to the constructor
   double GetMvaValue( const std::vector<double>& inputValues ) const;

 private:

   // method-specific destructor
   void Clear();

   // common member variables
   const char* m_className;
   bool m_statusIsClean;

   const size_t m_nvars;
   size_t GetNvar()           const { return m_nvars; }
   char   GetType( int ivar ) const { return m_type[ivar]; }

   // normalisation of input variables
   const bool m_isNormalised;
   bool IsNormalised() const { return m_isNormalised; }
   double m_Vmin[6];
   double m_Vmax[6];
   double NormVariable( double x, double xmin, double xmax ) const {
      // normalise to output range: [-1, 1]
      return 2*(x - xmin)/(xmax - xmin) - 1.0;
   }

   // type of input variable: 'F' or 'I'
   char   m_type[6];

   // initialize internal variables
   void Initialize();
   double GetMvaValue__( const std::vector<double>& inputValues ) const;

   // private members (method specific)

   double ActivationFnc(double x) const;

   int m_layers;
   int m_layerSize[4];
   double m_weightMatrix0to1[12][7];   // weight matrix from layer 0 to 1
   double m_weightMatrix1to2[7][12];   // weight matrix from layer 1 to 2
   double m_weightMatrix2to3[1][7];   // weight matrix from layer 2 to 3

   double * m_weights[4];
};

inline void ReadMLP::Initialize()
{
   // build network structure
   m_layers = 4;
   m_layerSize[0] = 7; m_weights[0] = new double[7]; 
   m_layerSize[1] = 12; m_weights[1] = new double[12]; 
   m_layerSize[2] = 7; m_weights[2] = new double[7]; 
   m_layerSize[3] = 1; m_weights[3] = new double[1]; 
   // weight matrix from layer 0 to 1
   m_weightMatrix0to1[0][0] = -1.35863616970102;
   m_weightMatrix0to1[1][0] = 0.355267825532901;
   m_weightMatrix0to1[2][0] = 1.11968100770938;
   m_weightMatrix0to1[3][0] = 1.63476135793075;
   m_weightMatrix0to1[4][0] = -0.881401114537419;
   m_weightMatrix0to1[5][0] = 1.33099464036524;
   m_weightMatrix0to1[6][0] = -0.687271667189041;
   m_weightMatrix0to1[7][0] = 1.35077317859302;
   m_weightMatrix0to1[8][0] = 0.294200995507164;
   m_weightMatrix0to1[9][0] = -1.48130303337079;
   m_weightMatrix0to1[10][0] = -0.0909154859636571;
   m_weightMatrix0to1[0][1] = 0.959810230401663;
   m_weightMatrix0to1[1][1] = -1.06324092946709;
   m_weightMatrix0to1[2][1] = -5.76336229163837;
   m_weightMatrix0to1[3][1] = -2.07175985540646;
   m_weightMatrix0to1[4][1] = 0.493743211808833;
   m_weightMatrix0to1[5][1] = -1.2891265974912;
   m_weightMatrix0to1[6][1] = 4.24598829927305;
   m_weightMatrix0to1[7][1] = 0.0164430982953178;
   m_weightMatrix0to1[8][1] = 5.48097729924405;
   m_weightMatrix0to1[9][1] = 0.754737219267636;
   m_weightMatrix0to1[10][1] = -0.946890453845568;
   m_weightMatrix0to1[0][2] = 1.18093866713897;
   m_weightMatrix0to1[1][2] = 1.80396001699007;
   m_weightMatrix0to1[2][2] = 1.52717646774325;
   m_weightMatrix0to1[3][2] = 2.33662747894103;
   m_weightMatrix0to1[4][2] = 2.45111772886874;
   m_weightMatrix0to1[5][2] = -2.52604675121866;
   m_weightMatrix0to1[6][2] = -2.08899522797482;
   m_weightMatrix0to1[7][2] = 1.05120556788023;
   m_weightMatrix0to1[8][2] = 5.11881886126749;
   m_weightMatrix0to1[9][2] = 1.38566440548325;
   m_weightMatrix0to1[10][2] = -2.53948815280012;
   m_weightMatrix0to1[0][3] = -3.83920957236234;
   m_weightMatrix0to1[1][3] = -5.43131768659034;
   m_weightMatrix0to1[2][3] = -1.59900627268632;
   m_weightMatrix0to1[3][3] = -6.80972345999108;
   m_weightMatrix0to1[4][3] = 1.55241569423265;
   m_weightMatrix0to1[5][3] = -6.14202954282145;
   m_weightMatrix0to1[6][3] = 5.36065374252205;
   m_weightMatrix0to1[7][3] = 0.0302912087482007;
   m_weightMatrix0to1[8][3] = -2.20277578705225;
   m_weightMatrix0to1[9][3] = 1.76175714498424;
   m_weightMatrix0to1[10][3] = -2.72120334473083;
   m_weightMatrix0to1[0][4] = -0.172322441419825;
   m_weightMatrix0to1[1][4] = -5.13807660097258;
   m_weightMatrix0to1[2][4] = 0.257800572194352;
   m_weightMatrix0to1[3][4] = -2.38940275959441;
   m_weightMatrix0to1[4][4] = 1.66361068160208;
   m_weightMatrix0to1[5][4] = 3.85940163827037;
   m_weightMatrix0to1[6][4] = -2.68691447083723;
   m_weightMatrix0to1[7][4] = 0.420439486625712;
   m_weightMatrix0to1[8][4] = 0.792164705629557;
   m_weightMatrix0to1[9][4] = 1.44406793932154;
   m_weightMatrix0to1[10][4] = 0.482090274915509;
   m_weightMatrix0to1[0][5] = 2.90212135134892;
   m_weightMatrix0to1[1][5] = -0.467329773679144;
   m_weightMatrix0to1[2][5] = -3.44828629211672;
   m_weightMatrix0to1[3][5] = 2.37478884166659;
   m_weightMatrix0to1[4][5] = 1.93095546957089;
   m_weightMatrix0to1[5][5] = -1.81539917249789;
   m_weightMatrix0to1[6][5] = -0.623607995854637;
   m_weightMatrix0to1[7][5] = -0.285335872428147;
   m_weightMatrix0to1[8][5] = -0.803888962558799;
   m_weightMatrix0to1[9][5] = 1.62681417674909;
   m_weightMatrix0to1[10][5] = -7.73343650169979;
   m_weightMatrix0to1[0][6] = 0.196297181875605;
   m_weightMatrix0to1[1][6] = -8.09991145420649;
   m_weightMatrix0to1[2][6] = 2.58514246487067;
   m_weightMatrix0to1[3][6] = -3.16512536957876;
   m_weightMatrix0to1[4][6] = -1.1087033178518;
   m_weightMatrix0to1[5][6] = -3.61844135337377;
   m_weightMatrix0to1[6][6] = -2.17850221642457;
   m_weightMatrix0to1[7][6] = -0.148588223967372;
   m_weightMatrix0to1[8][6] = 1.4548620597591;
   m_weightMatrix0to1[9][6] = -0.428426117424227;
   m_weightMatrix0to1[10][6] = -9.16043701262521;
   // weight matrix from layer 1 to 2
   m_weightMatrix1to2[0][0] = -2.01756746620757;
   m_weightMatrix1to2[1][0] = 1.93315985477623;
   m_weightMatrix1to2[2][0] = 0.586376191536926;
   m_weightMatrix1to2[3][0] = 0.450462742415433;
   m_weightMatrix1to2[4][0] = -0.156965912396646;
   m_weightMatrix1to2[5][0] = 1.56169236818918;
   m_weightMatrix1to2[0][1] = -1.19154563335079;
   m_weightMatrix1to2[1][1] = 0.104425369302282;
   m_weightMatrix1to2[2][1] = 1.11215034967172;
   m_weightMatrix1to2[3][1] = 1.04344037327303;
   m_weightMatrix1to2[4][1] = -1.36111194916732;
   m_weightMatrix1to2[5][1] = -0.5934411867395;
   m_weightMatrix1to2[0][2] = -0.597792885746589;
   m_weightMatrix1to2[1][2] = 0.119601133331117;
   m_weightMatrix1to2[2][2] = 0.0249489435301289;
   m_weightMatrix1to2[3][2] = 0.824322960207608;
   m_weightMatrix1to2[4][2] = -0.652108369211788;
   m_weightMatrix1to2[5][2] = 0.393378677337595;
   m_weightMatrix1to2[0][3] = 0.136847670302133;
   m_weightMatrix1to2[1][3] = -0.641169720832402;
   m_weightMatrix1to2[2][3] = 0.164211018320105;
   m_weightMatrix1to2[3][3] = 0.225216141182485;
   m_weightMatrix1to2[4][3] = -1.82521417318264;
   m_weightMatrix1to2[5][3] = 0.104678584514443;
   m_weightMatrix1to2[0][4] = -0.654253743469733;
   m_weightMatrix1to2[1][4] = -0.737504206242956;
   m_weightMatrix1to2[2][4] = 1.7325685102728;
   m_weightMatrix1to2[3][4] = -0.750263029745144;
   m_weightMatrix1to2[4][4] = 1.68831003559797;
   m_weightMatrix1to2[5][4] = -0.512374593257449;
   m_weightMatrix1to2[0][5] = -1.44137324894219;
   m_weightMatrix1to2[1][5] = 0.386472104891384;
   m_weightMatrix1to2[2][5] = 0.178417459501999;
   m_weightMatrix1to2[3][5] = 0.217291303120707;
   m_weightMatrix1to2[4][5] = -0.524031976509008;
   m_weightMatrix1to2[5][5] = -0.642787303821237;
   m_weightMatrix1to2[0][6] = -0.173391726912647;
   m_weightMatrix1to2[1][6] = -1.30728304244028;
   m_weightMatrix1to2[2][6] = 0.524116377117486;
   m_weightMatrix1to2[3][6] = 0.647137920154457;
   m_weightMatrix1to2[4][6] = -1.94923371512367;
   m_weightMatrix1to2[5][6] = -0.129522211043097;
   m_weightMatrix1to2[0][7] = -0.620232149919463;
   m_weightMatrix1to2[1][7] = 1.23266478704893;
   m_weightMatrix1to2[2][7] = 0.466377470403172;
   m_weightMatrix1to2[3][7] = 1.28246812160841;
   m_weightMatrix1to2[4][7] = -0.396682174654263;
   m_weightMatrix1to2[5][7] = -0.371792672915382;
   m_weightMatrix1to2[0][8] = -0.395261695851815;
   m_weightMatrix1to2[1][8] = -3.06158080081136;
   m_weightMatrix1to2[2][8] = -0.399173564655177;
   m_weightMatrix1to2[3][8] = -0.509866673983801;
   m_weightMatrix1to2[4][8] = 0.132519578949491;
   m_weightMatrix1to2[5][8] = -0.113121939052194;
   m_weightMatrix1to2[0][9] = -1.05589933703286;
   m_weightMatrix1to2[1][9] = -0.348995400582059;
   m_weightMatrix1to2[2][9] = -1.06947911973433;
   m_weightMatrix1to2[3][9] = 0.861430291723568;
   m_weightMatrix1to2[4][9] = 1.11130313790832;
   m_weightMatrix1to2[5][9] = 2.01075385787491;
   m_weightMatrix1to2[0][10] = -1.38574320506867;
   m_weightMatrix1to2[1][10] = 0.297335310833319;
   m_weightMatrix1to2[2][10] = 0.425399695433244;
   m_weightMatrix1to2[3][10] = 0.149270570274458;
   m_weightMatrix1to2[4][10] = -1.52148682746593;
   m_weightMatrix1to2[5][10] = -1.29295042597047;
   m_weightMatrix1to2[0][11] = -3.92413006501098;
   m_weightMatrix1to2[1][11] = 1.53770315964507;
   m_weightMatrix1to2[2][11] = 2.32510532022512;
   m_weightMatrix1to2[3][11] = 2.31673648899074;
   m_weightMatrix1to2[4][11] = -0.936293946316135;
   m_weightMatrix1to2[5][11] = -2.8180691729656;
   // weight matrix from layer 2 to 3
   m_weightMatrix2to3[0][0] = 0.236812517884415;
   m_weightMatrix2to3[0][1] = -0.310842649848729;
   m_weightMatrix2to3[0][2] = -0.17516343751807;
   m_weightMatrix2to3[0][3] = -0.204522042676872;
   m_weightMatrix2to3[0][4] = 0.212815798140234;
   m_weightMatrix2to3[0][5] = 0.259659255800626;
   m_weightMatrix2to3[0][6] = 0.399846765309018;
}

inline double ReadMLP::GetMvaValue__( const std::vector<double>& inputValues ) const
{
   if (inputValues.size() != (unsigned int)m_layerSize[0]-1) {
      std::cout << "Input vector needs to be of size " << m_layerSize[0]-1 << std::endl;
      return 0;
   }

   for (int l=0; l<m_layers; l++)
      for (int i=0; i<m_layerSize[l]; i++) m_weights[l][i]=0;

   for (int l=0; l<m_layers-1; l++)
      m_weights[l][m_layerSize[l]-1]=1;

   for (int i=0; i<m_layerSize[0]-1; i++)
      m_weights[0][i]=inputValues[i];

   // layer 0 to 1
   for (int o=0; o<m_layerSize[1]-1; o++) {
      for (int i=0; i<m_layerSize[0]; i++) {
         double inputVal = m_weightMatrix0to1[o][i] * m_weights[0][i];
         m_weights[1][o] += inputVal;
      }
      m_weights[1][o] = ActivationFnc(m_weights[1][o]);
   }
   // layer 1 to 2
   for (int o=0; o<m_layerSize[2]-1; o++) {
      for (int i=0; i<m_layerSize[1]; i++) {
         double inputVal = m_weightMatrix1to2[o][i] * m_weights[1][i];
         m_weights[2][o] += inputVal;
      }
      m_weights[2][o] = ActivationFnc(m_weights[2][o]);
   }
   // layer 2 to 3
   for (int o=0; o<m_layerSize[3]; o++) {
      for (int i=0; i<m_layerSize[2]; i++) {
         double inputVal = m_weightMatrix2to3[o][i] * m_weights[2][i];
         m_weights[3][o] += inputVal;
      }
   }

   return m_weights[3][0];
}

double ReadMLP::ActivationFnc(double x) const
{
   // hyperbolic tan
   return tanh(x);
}
 
   
// Clean up
inline void ReadMLP::Clear() 
{
   // nothing to clear
}
inline double ReadMLP::GetMvaValue( const std::vector<double>& inputValues ) const
{
   // classifier response value
   double retval = 0;

   // classifier response, sanity check first
   if (!m_statusIsClean) {
      std::cout << "Problem in class \"" << m_className << "\": cannot return classifier response"
                   << " because status is dirty" << std::endl;
      retval = 0;
   }
   else {
      if (IsNormalised()) {
         // normalise variables
         std::vector<double> iV;
         int ivar = 0;
         for (std::vector<double>::const_iterator varIt = inputValues.begin();
              varIt != inputValues.end(); ++varIt, ++ivar) {
            iV.push_back(NormVariable( *varIt, m_Vmin[ivar], m_Vmax[ivar] ));
         }
         retval = GetMvaValue__( iV );
      }
      else {
         retval = GetMvaValue__( inputValues );
      }
   }

   return retval;
}
