"""Define methods to construct configured Pixel overlay algorithms

Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
"""

from AthenaConfiguration.ComponentAccumulator import ComponentAccumulator
from AthenaConfiguration.ComponentFactory import CompFactory


def PixelOverlayAlgCfg(flags, name="PixelOverlay", **kwargs):
    """Return a ComponentAccumulator for PixelOverlay algorithm"""
    acc = ComponentAccumulator()

    kwargs.setdefault("BkgInputKey", f"{flags.Overlay.BkgPrefix}PixelRDOs")
    kwargs.setdefault("SignalInputKey", f"{flags.Overlay.SigPrefix}PixelRDOs")
    kwargs.setdefault("OutputKey", "PixelRDOs")

    # Input setup
    if flags.Overlay.ByteStream:
        from PixelRawDataByteStreamCnv.PixelRawDataByteStreamCnvConfig import PixelRawDataProviderAlgCfg
        acc.merge(PixelRawDataProviderAlgCfg(flags))
    else:
        from SGComps.SGInputLoaderConfig import SGInputLoaderCfg
        acc.merge(SGInputLoaderCfg(flags, [f'PixelRDO_Container#{kwargs["BkgInputKey"]}']))

    # Do Pixel overlay
    acc.addEventAlgo(CompFactory.PixelOverlay(name, **kwargs))

    # Setup output
    if flags.Output.doWriteRDO:
        from OutputStreamAthenaPool.OutputStreamConfig import OutputStreamCfg
        acc.merge(OutputStreamCfg(flags, "RDO", ItemList=[
            "PixelRDO_Container#PixelRDOs"
        ]))

        if not flags.Input.isMC:
            acc.merge(OutputStreamCfg(flags, "RDO", ItemList=[
                "IDCInDetBSErrContainer#PixelByteStreamErrs"
            ]))

    if flags.Output.doWriteRDO_SGNL:
        from OutputStreamAthenaPool.OutputStreamConfig import OutputStreamCfg
        acc.merge(OutputStreamCfg(flags, "RDO_SGNL", ItemList=[
            f"PixelRDO_Container#{flags.Overlay.SigPrefix}PixelRDOs"
        ]))

    # for track overlay, write out the signal RDOs because reco tracking will only run on them
    if flags.Overlay.doTrackOverlay:
        acc.merge(OutputStreamCfg(flags, "RDO", ItemList=[
            f"PixelRDO_Container#{flags.Overlay.SigPrefix}PixelRDOs"
        ]))

    return acc


def PixelTruthOverlayCfg(flags, name="PixelSDOOverlay", **kwargs):
    """Return a ComponentAccumulator for the Pixel SDO overlay algorithm"""
    acc = ComponentAccumulator()

    # We do not need background Pixel SDOs
    kwargs.setdefault("BkgInputKey", "")

    kwargs.setdefault("SignalInputKey", f"{flags.Overlay.SigPrefix}PixelSDO_Map")
    kwargs.setdefault("OutputKey", "PixelSDO_Map")

    # Do Pixel truth overlay
    acc.addEventAlgo(CompFactory.InDetSDOOverlay(name, **kwargs))

    # Setup output
    if flags.Output.doWriteRDO:
        from OutputStreamAthenaPool.OutputStreamConfig import OutputStreamCfg
        acc.merge(OutputStreamCfg(flags, "RDO", ItemList=[
            "InDetSimDataCollection#PixelSDO_Map"
        ]))

    if flags.Output.doWriteRDO_SGNL:
        from OutputStreamAthenaPool.OutputStreamConfig import OutputStreamCfg
        acc.merge(OutputStreamCfg(flags, "RDO_SGNL", ItemList=[
            f"InDetSimDataCollection#{flags.Overlay.SigPrefix}PixelSDO_Map"
        ]))

    return acc


def PixelOverlayCfg(flags):
    """Configure and return a ComponentAccumulator for Pixel overlay"""
    acc = ComponentAccumulator()

    # Add Pixel overlay digitization algorithm
    from PixelDigitization.PixelDigitizationConfig import PixelOverlayDigitizationBasicCfg
    acc.merge(PixelOverlayDigitizationBasicCfg(flags))

    # Add Pixel overlay algorithm
    acc.merge(PixelOverlayAlgCfg(flags))

    # Add Pixel truth overlay
    if flags.Digitization.EnableTruth:
        acc.merge(PixelTruthOverlayCfg(flags))

    return acc
