/*
  Copyright (C) 2002-2025 CERN for the benefit of the ATLAS collaboration
*/

#include "LArCafJobs/DataStore.h"

#include "LArCafJobs/ShapeInfo.h"
#include "LArCafJobs/HistoryContainer.h"
#include "LArCafJobs/DataContainer.h"
#include "LArCafJobs/EventData.h"
#include "LArCafJobs/RunData.h"
#include "LArCafJobs/PersistentAccessor.h"
#include "Identifier/IdentifierHash.h"
#include "TRandom.h"
#include "TMath.h"
#include "TFile.h"
#include "TTree.h"
#include <iostream>

using std::cout;
using std::endl;

using namespace LArSamples;


DataStore::DataStore(): 
   m_cellHistories(nChannels(), nullptr), 
   m_cellHistoriesSC(nChannelsSC(), nullptr)
{
}
  
  
DataStore::~DataStore()
{
  for (HistoryContainer* h : m_cellHistories)
    delete h;
}


HistoryContainer* DataStore::makeNewHistory(const IdentifierHash& hash, CellInfo* info)
{
  HistoryContainer*& histCont = hist_cont(hash);
  if (histCont) return histCont;
  histCont = new HistoryContainer(info);
  return histCont;
}

HistoryContainer* DataStore::makeNewHistorySC(const IdentifierHash& hash, CellInfo* info)
{
  HistoryContainer*& histCont = hist_cont_sc(hash);
  if (histCont) return histCont;
  histCont = new HistoryContainer(info);
  return histCont;
}


bool DataStore::addData(const IdentifierHash& hash, DataContainer* data)
{
  HistoryContainer*& histCont = hist_cont(hash);
  histCont->add(data);
  return true;
}

bool DataStore::addDataSC(const IdentifierHash& hash, DataContainer* data)
{
  HistoryContainer*& histCont = hist_cont_sc(hash);
  histCont->add(data);
  return true;
}


unsigned int DataStore::addEvent(EventData* eventData)
{
  m_events.push_back(eventData);
  return nEvents() - 1;
}


unsigned int DataStore::addRun(RunData* runData)
{
  m_runs.push_back(runData);
  return nRuns() - 1;
}


unsigned int DataStore::size() const
{
  unsigned int size = 0;
 
 for (unsigned int i = 0; i < nChannels(); i++) {
   const HistoryContainer* hist = historyContainer(i);
   if (!hist) continue;
   size += hist->nDataContainers();
 }
 
 for (unsigned int i = 0; i < nChannelsSC(); i++) {
   const HistoryContainer* hist = historyContainerSC(i);
   if (!hist) continue;
   size += hist->nDataContainers();
 }
 
 return size;
}


unsigned int DataStore::nFilledChannels() const
{
  unsigned int n = 0;
 
 for (unsigned int i = 0; i < nChannels(); i++) {
   const HistoryContainer* hist = historyContainer(i);
   if (!hist) continue;
   n++;
 }
 
 return n;
}

unsigned int DataStore::nFilledChannelsSC() const
{
  unsigned int n = 0;
 
 for (unsigned int i = 0; i < nChannelsSC(); i++) {
   const HistoryContainer* hist = historyContainerSC(i);
   if (!hist) continue;
   n++;
 }
 
 return n;
}


double DataStore::footprint() const 
{  
  double fp = sizeof(*this); 
  for (unsigned int i = 0; i < nChannels(); i++) {
    const HistoryContainer* hist = historyContainer(i);
    if (!hist) continue;
    fp += hist->footprint();
  }
    
  for (unsigned int i = 0; i < nChannelsSC(); i++) {
    const HistoryContainer* hist = historyContainerSC(i);
    if (!hist) continue;
    fp += hist->footprint();
  }
    
  return fp; 
}


bool DataStore::writeTrees(const char* fileName)
{
  PersistentAccessor* acc = new PersistentAccessor(fileName);

  for (unsigned int i = 0; i < nRuns(); i++) {
    RunData*& r = runData(i);
    acc->addRun(r);
    delete r; r = nullptr;
  }

  for (unsigned int i = 0; i < nEvents(); i++) {
    EventData*& ev = eventData(i);
    acc->addEvent(ev);
    delete ev; ev = nullptr;
  }
  
  for (unsigned int i = 0; i < nChannels(); i++) {
    HistoryContainer*& hc = hist_cont(i);
    acc->add(hc);	
    delete hc; hc = nullptr;
  }

  for (unsigned int i = 0; i < nChannelsSC(); i++) {
    HistoryContainer*& hc = hist_cont_sc(i);
    acc->addSC(hc);	
    delete hc; hc = nullptr;
  }

  bool result = acc->save();
  delete acc;
  return result;
}

