# Copyright (C) 2002-2022 CERN for the benefit of the ATLAS collaboration

from AthenaConfiguration.AthConfigFlags import AthConfigFlags
from AthenaConfiguration.Enums import BeamType, LHCPeriod, ProductionStep, Project, FlagEnum
    
import re

# Some comments from Ed about existing flags
# MuonCnvExample
# - MuonCnvFlags - not sure we need this - it really only configures single properties of the various cablings - we can just do this directly.
# - MuonCalibFlags - looks like we need this

# MuonByteStream
# - MuonByteStreamFlags.py - AFAICS this is only used here MuonCablings/MuonCablingServers/python/MuonCablingServersConfig.py, MuonCnv/MuonCnvExample/python/MuonCablingConfig.py & duplicates global flag functionality.

# MuonRecExample
# - MuonAlignFlags.py - looks necessary
# - MuonStandaloneFlags.py - necessary, but should be cleaned up (and probably merged with MuonRecFlags i.e. here)
# https://gitlab.cern.ch/atlas/athena/blob/master/MuonSpectrometer/MuonReconstruction/MuonRecExample/python/MuonStandaloneFlags.py
# - MuonRecFlags.py - necessary, but needs cleaning up.
# https://gitlab.cern.ch/atlas/athena/blob/master/MuonSpectrometer/MuonReconstruction/MuonRecExample/python/MuonRecFlags.py


class MMClusterBuilderEnum(FlagEnum):
    """Flag values for Muon.MMClusterCalibRecoTool"""
    Centroid = "Centroid"
    ClusterTimeProjection = "ClusterTimeProjection"

def _muonAlignMode(flags):
    # Small function that determines if the alignment flags should be true or false
    # follows the logic in https://acode-browser1.usatlas.bnl.gov/lxr/source/athena/MuonSpectrometer/MuonReconstruction/MuonRecExample/python/MuonAlignFlags.py
    problematic_tags = ['COMCOND-HLT[A-C]-00[01]-00', 'COMCOND-ES1C-00[01]-00', 'COMCOND-REPC-001', 'COMCOND-SIM-01-00']
    if any(re.match(tag,flags.IOVDb.DatabaseInstance) for tag in problematic_tags):
        return False
    elif flags.Input.isMC:
        return False
    else:
        return True

def createMuonConfigFlags(): 
    mcf=AthConfigFlags()

    # stages of processing
    # 0. Geometry

    ### Load the GeoModel XML detector factory & MuonReaoudGeometryR4    
    from AthenaConfiguration.AutoConfigFlags import DetDescrInfo
    
    mcf.addFlag("Muon.usePhaseIIGeoSetup",lambda prevFlags : DetDescrInfo(prevFlags.GeoModel.AtlasVersion, 
                                                                          prevFlags.GeoModel.SQLiteDB ,
                                                                          prevFlags.GeoModel.SQLiteDBFullPath)["Muon"]["useR4Plugin"] )
    # 1. Digitization
    mcf.addFlag("Muon.doDigitization",True)
    # 2. Reco MuonRecFlags    
    mcf.addFlag("Muon.doTGCClusterSegmentFinding", False) # Run cluster segment finding
    mcf.addFlag("Muon.doRPCClusterSegmentFinding", False) # Run cluster segment finding
    mcf.addFlag("Muon.prdToxAOD", False) # Run clusterization
    mcf.addFlag("Muon.rpcRawToxAOD", False) # Add RPC RDO to xAOD
    mcf.addFlag("Muon.doMSVertex", True) # Run MS vertex (arXiv:1311.7070)
    mcf.addFlag("Muon.doSegmentT0Fit",lambda prevFlags : prevFlags.Beam.Type is not BeamType.Collisions) # Fit MDT segments using a variable t0. Used for cosmics and single beam to compensate for large errors on the trigger time.
    mcf.addFlag("Muon.enableErrorTuning",True) # turn on error tuning to account for misalignments
    mcf.addFlag("Muon.useLooseErrorTuning",False) 
    mcf.addFlag("Muon.useTGCPriorNextBC",False) # Use TGC measurements from Prior and Next Bunch Crossings. These measurements are available in the real data since somewhere in 2008.
    mcf.addFlag("Muon.useAlignmentCorrections",True) # Apply alignment corrections to MuonGeoModel. The corrections are read from a COOL database
    mcf.addFlag("Muon.useWireSagCorrections",False) # tApply wire sag corrections.
    
    mcf.addFlag("Muon.makePRDs",True) # Disable when e.g. re-running from ESD
    
    mcf.addFlag("Muon.enableNRPC",False) # To be enabled for NRPC processing. Disabled by default for backward compatibility

    # MuonStandaloneFlags.py 
    mcf.addFlag("Muon.printSummary", False) # Print out a summary for each event at each reco stage
    mcf.addFlag("Muon.segmentOrigin", "Muon") # Can be 'Muon','TruthTracking'
    # reconstructionMode - I think this is a complete duplication of global.BeamType. Dropping.
    mcf.addFlag("Muon.strategy", []) # CutSeedsOnTracks, CombineSegInStation, DynamicSeeding, PreferOutsideIn, AllowOneSharedHit, DoRefinement, DoAmbiSolving
    mcf.addFlag("Muon.trackBuilder", "Moore") # Allowed: 'Moore','TruthTracking','None'
    mcf.addFlag("Muon.refinementTool", "Moore") # Allowed: Moore TODO surely we can drop this if there is only one option?
    mcf.addFlag("Muon.patternsOnly", False) # TODO probably can be dropped? Just disable later steps.
    mcf.addFlag("Muon.straightLineFitMomentum", 2000.0 ) 
    #mcf.addFlag("Muon.doSegmentsOnly", True) # Also in MuonRecFlags ... redundant in both?
    mcf.addFlag("Muon.Chi2NDofCut", 20.0 )  # chi-squared per degree of freedom cut in fitter.
    mcf.addFlag("Muon.enableCurvedSegmentFinding", False ) # TODO I think this one could possibly be removed, since it really is a Tool level configuration.
    mcf.addFlag("Muon.updateSegmentSecondCoordinate", lambda prevFlags : prevFlags.Beam.Type is BeamType.Collisions) # Do not use for cosmics or singlebeam 
    
    mcf.addFlag("Muon.useSegmentMatching", lambda prevFlags : prevFlags.Beam.Type is BeamType.Collisions) # Do not use for cosmics or singlebeam 
    mcf.addFlag("Muon.useTrackSegmentMatching", True )
    mcf.addFlag("Muon.runCommissioningChain", lambda prevFlags: ( False and (prevFlags.Detector.EnableMM or prevFlags.Detector.EnablesTGC) \
                                                                 and prevFlags.Beam.Type is BeamType.Collisions) )
    
    mcf.addFlag("Muon.applyMMPassivation", lambda prevFlags: prevFlags.Detector.EnableMM and not prevFlags.Common.isOnline and (prevFlags.Common.Project is not Project.AthSimulation \
                                                      and (prevFlags.Common.ProductionStep not in [ProductionStep.Simulation, ProductionStep.FastChain] or prevFlags.Overlay.DataOverlay)))
    # CalibFlags
    mcf.addFlag("Muon.Calib.readMDTCalibFromBlob", True)  # Read mdt tube calibration from blob-folders
    mcf.addFlag("Muon.Calib.correctMdtRtForBField", lambda prevFlags : (prevFlags.Input.isMC is False and prevFlags.Beam.Type is BeamType.Collisions)) # Apply B-field correction to drift times only for collision data (as done in https://acode-browser1.usatlas.bnl.gov/lxr/source/athena/MuonSpectrometer/MuonCnv/MuonCnvExample/python/MuonCalibFlags.py#0028)
    mcf.addFlag("Muon.Calib.correctMdtRtForTimeSlewing", lambda prevFlags : prevFlags.Input.isMC is False) # Apply time slewing correction to drift time only for data (as done in https://acode-browser1.usatlas.bnl.gov/lxr/source/athena/MuonSpectrometer/MuonCnv/MuonCnvExample/python/MuonCalibFlags.py#0028)
    mcf.addFlag("Muon.Calib.useMLRt", True) # use ML-RT functions from COOL
    mcf.addFlag("Muon.Calib.applyRtScaling", False) # TODO - apparently not needed, but currently used in MuonCalibConfig. Set false to match https://acode-browser1.usatlas.bnl.gov/lxr/source/athena/MuonSpectrometer/MuonCnv/MuonCnvExample/python/MuonCalibFlags.py#0072
    mcf.addFlag("Muon.Calib.correctMdtRtWireSag", False) # Apply RT wiresag correction
    mcf.addFlag("Muon.Calib.mdtCalibrationSource", "MDT") # Source for MDT t0s and rts
    mcf.addFlag("Muon.Calib.mdtPropagationSpeedBeta", lambda prevFlags : 1.0 if prevFlags.Input.isMC else 0.85) #
     
    mcf.addFlag("Muon.Calib.CscPedFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscNoiseFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscPSlopeFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscStatusFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscRmsFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscF001FromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscT0BaseFromLocalFile", False)  
    mcf.addFlag("Muon.Calib.CscT0PhaseFromLocalFile", False)  
    
    mcf.addFlag("Muon.Calib.EventTag", "MoMu") 
    
    #  Choose the (MDT) calibration mode.
    # 'ntuple'       : write standard calibration ntuple (up to segments)
    # 'trackNtuple'  : write standard ntuple + tracks
    # 'regionNtuple' : write one ntuple per calibration region
    # 't0Classic'    : do classic t0 calibration
    # 't0MT'         : do MT t0 calibration
    # 'rtClassic'    : do classic rt calibration
    # 'rtAnalytic'   : do analytic rt calibration
    mcf.addFlag("Muon.Calib.mdtMode", "ntuple")


    # for now the T0 calibration in the NSW should be disabled by default until a final calibration is available. Introducing the flags anyhow to allow for studies of the calibration 

    # do not apply NSW T0 calibration if we are running online or MC or a RUN4 geometry, keep only for 23 or 24 
    mcf.addFlag("Muon.Calib.applyMmT0Correction",   lambda prevFlags: prevFlags.GeoModel.Run==LHCPeriod.Run3 and prevFlags.Input.DataYear != 2022 and not prevFlags.Common.isOnline and not prevFlags.Input.isMC )
    mcf.addFlag("Muon.Calib.applysTgcT0Correction", lambda prevFlags: prevFlags.GeoModel.Run == LHCPeriod.Run3  and not prevFlags.Common.isOnline and not prevFlags.Input.isMC and prevFlags.Input.DataYear!=2022) # activated for data offline reco in Run3 but not for the first year of data taking with the NSW.
    mcf.addFlag("Muon.Calib.applyMmBFieldCalib", True) 
    
    # Muon Align flags
    
    mcf.addFlag("Muon.Align.UseALines", lambda prevFlags: (_muonAlignMode(prevFlags)))
    mcf.addFlag("Muon.Align.UseBLines", lambda prevFlags: 'all' if _muonAlignMode(prevFlags) else 'none') # Can be ['none','all','barrel','endcaps']
    mcf.addFlag("Muon.Align.UseILines", lambda prevFlags: (_muonAlignMode(prevFlags)))
    mcf.addFlag("Muon.Align.UseAsBuilt", lambda prevFlags: (_muonAlignMode(prevFlags)))
    mcf.addFlag("Muon.Align.UsesTGCAsBuild",False)

    # Muon Trigger Flags
    mcf.addFlag("Muon.MuonTrigger", False) 
    mcf.addFlag("Muon.SAMuonTrigger", False) 
    mcf.addFlag("Muon.disableNSWForL2SA", True)
    mcf.addFlag("Muon.enableTrigIDtrackReuse", False)

    mcf.addFlag("Muon.enableAlignment",lambda flags: (flags.Common.Project is not Project.AthSimulation \
                                                      and (flags.Common.ProductionStep not in [ProductionStep.Simulation, ProductionStep.FastChain] or flags.Overlay.DataOverlay)))

    # configuration of the DESDM_MCP output format 

    mcf.addFlag("Muon.DESDM_MCP.doAlignmentFormat", False) # Flag to stear the DESDM_MCP format which switches to a looser event selection for toroid off runs used for alignment. 

    # configuration to write out RPC RDO for trigger timing calibration
    mcf.addFlag("Muon.doWriteRpcRDO", True)

    mcf.addFlag("Muon.writeSDOs", lambda prevFlags : prevFlags.Output.doWriteESD and prevFlags.Input.isMC)

    # configure the MM cluster reco method that is used in the cluster calibration step 
    #Use charge weighted only for trigger and 2022 where not sure if the t0 calibrations that we have are ok dor the commissioning phase of the NSW in 2022
    mcf.addFlag("Muon.MMClusterCalibRecoTool", lambda prevFlags: MMClusterBuilderEnum.Centroid if (prevFlags.Common.isOnline or prevFlags.Input.DataYear == 2022 or prevFlags.Beam.Type is not BeamType.Collisions) else MMClusterBuilderEnum.ClusterTimeProjection, type=MMClusterBuilderEnum)

    # use the MDT DCS data to determine if a chamber is alive or not. This is used in the hole search and the region selector. Needs to be false if the job is running online or is the reconstruction of the MDT calib stream
    mcf.addFlag("Muon.useMdtDcsData", lambda prevFlags : not prevFlags.Common.isOnline)






    # TODO - add configuration for above    
        
    return mcf

