# Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration


from AthenaConfiguration.ComponentAccumulator import ComponentAccumulator
from AthenaConfiguration.ComponentFactory import CompFactory
from AthenaConfiguration.Enums import Format



def Run2TriggerTowerMakerCfg(flags, name='Run2TriggerTowerMaker25ns'):
    '''
    Basic setup of tower maker cfg for new JO 
    WARNING: need to add dependencies on digi flags (as above) that are missing as of now
    '''

    acc = ComponentAccumulator()
    # TODO this is only needed when re-running 
    #    from SGComps.AddressRemappingConfig import InputRenameCfg
    #    acc.merge(InputRenameCfg('xAOD::TriggerTowerContainer', 'xAODTriggerTowers_rerun', 'xAODTriggerTowers'))

    # Configure conditions used by R2TTMaker
    from CaloConditions.CaloConditionsConfig import CaloTTIdMapCfg
    from TileConditions.TileInfoLoaderConfig import TileInfoLoaderCfg
    acc.merge(CaloTTIdMapCfg(flags))
    acc.merge(TileInfoLoaderCfg(flags))

    from TrigT1CaloCondSvc.L1CaloCondConfig import L1CaloCondAlgCfg
    acc.merge(L1CaloCondAlgCfg(flags,Physics=True, Calib1=False, Calib2=False))


    # R2TTMaker reads TTL1 containers from input POOL file (RDO, ESD, ...)
    if flags.Input.Format is Format.POOL:
        ttl1Containers = [
            ('LArTTL1Container', 'LArTTL1EM'),
            ('LArTTL1Container', 'LArTTL1HAD'),
            ('TileTTL1Container', 'TileTTL1Cnt'),
        ]
        from SGComps.SGInputLoaderConfig import SGInputLoaderCfg
        acc.merge(SGInputLoaderCfg(flags, Load=ttl1Containers))

    # R2TTMaker reads L1Menu in BeginRun incident, so needs L1ConfigSvc
    from TrigConfigSvc.TrigConfigSvcCfg import L1ConfigSvcCfg
    acc.merge(L1ConfigSvcCfg(flags))

    # R2TTMaker needs the mu information available as EventInfo decoration
    from LumiBlockComps.LumiBlockMuWriterConfig import LumiBlockMuWriterCfg
    acc.merge(LumiBlockMuWriterCfg(flags))

    condalgo = acc.getCondAlgo("L1CaloCondAlg")
    alg = CompFactory.LVL1.Run2TriggerTowerMaker(name,
                                                 DigiEngine = "{}_Digitization".format(name),
                                                 # TODO make these settings flags dependent
                                                 CellType = 3,
                                                 inputTTLocation = 'unused',
                                                 TriggerTowerLocationRerun = 'also_unused',
                                                 ZeroSuppress = True, 
                                                 ChanCalibFolderKey = condalgo.OutputKeyPPr,
                                                 ChanDefaultsFolderKey = condalgo.OutputKeyDef,
                                                 DisabledTowersFolderKey = condalgo.OutputKeyDisTowers,
                                                 DeadChannelsFolderKey = condalgo.OutputKeyPpmDeadChannels,
                                                 TTTool = CompFactory.LVL1.L1TriggerTowerTool("L1TriggerTowerTool",
                                                                                              InputKeyRunParameters=condalgo.OutputKeyRunParameters,
                                                                                              InputKeyPprConditionsRun2=condalgo.OutputKeyPprConditionsRun2,
                                                                                              InputKeyDisabledChannelRun2=condalgo.OutputKeyDisabledChannelRun2,
                                                                                              InputKeyTimeRefs=condalgo.OutputKeyTimeRefs)
                                                 #ExtraInputs = {'LArTTL1Container#LArTTL1EM', 'LArTTL1Container#LArTTL1HAD', 'TileTTL1Container#TileTTL1Cnt'}
                                                 )
    acc.addEventAlgo(alg)

    return acc



def L1CaloLegacySimCfg(flags):
    '''
    Configures Legacy 1 calo in new JO style
    '''
    acc = ComponentAccumulator()
    from AtlasGeoModel.GeoModelConfig import GeoModelCfg
    acc.merge (GeoModelCfg (flags))

    from CaloConditions.CaloConditionsConfig import CaloTriggerTowerCfg,LArTTCellMapCfg,CaloTTIdMapCfg
    acc.merge(CaloTriggerTowerCfg(flags))
    acc.merge(LArTTCellMapCfg(flags))
    acc.merge(CaloTTIdMapCfg(flags))

    from LArGeoAlgsNV.LArGMConfig import LArGMCfg
    from TileGeoModel.TileGMConfig import TileGMCfg
    
    acc.merge(LArGMCfg(flags))
    acc.merge(TileGMCfg(flags))

    # necessary conditions    
    from LArBadChannelTool.LArBadChannelConfig import LArBadChannelCfg, LArBadFebCfg
    acc.merge(LArBadChannelCfg(flags))
    acc.merge(LArBadFebCfg(flags))

    from TileConditions.TileInfoLoaderConfig import TileInfoLoaderCfg
    acc.merge(TileInfoLoaderCfg(flags))

    from TileGeoModel.TileGMConfig import TileGMCfg
    acc.merge(TileGMCfg(flags))

    from TileConditions.TileEMScaleConfig import TileEMScaleCondAlgCfg
    acc.merge( TileEMScaleCondAlgCfg(flags) )

    from TrigConfigSvc.TrigConfigSvcCfg import L1ConfigSvcCfg
    acc.merge(L1ConfigSvcCfg(flags))

    if not flags.Input.isMC:
        from TrigT1CaloByteStream.LVL1CaloRun2ByteStreamConfig import LVL1CaloRun2ReadBSCfg
        acc.merge(LVL1CaloRun2ReadBSCfg(flags))

    if flags.Input.isMC:
        acc.merge(Run2TriggerTowerMakerCfg(flags))
        acc.addEventAlgo(CompFactory.LVL1.Run2CPMTowerMaker('CPMTowerMaker'))
        acc.addEventAlgo(CompFactory.LVL1.Run2JetElementMaker('JetElementMaker'))

    acc.addEventAlgo(CompFactory.LVL1.CPMSim('CPMSim'))
    acc.addEventAlgo(CompFactory.LVL1.JEMJetSim('JEMJetSim'))
    acc.addEventAlgo(CompFactory.LVL1.JEMEnergySim('JEMEnergySim'))
    acc.addEventAlgo(CompFactory.LVL1.CPCMX('CPCMX'))
    acc.addEventAlgo(CompFactory.LVL1.JetCMX('JetCMX'))
    acc.addEventAlgo(CompFactory.LVL1.EnergyCMX('EnergyCMX'))
    acc.addEventAlgo(CompFactory.LVL1.RoIROD('RoIROD'))
    if flags.Input.isMC:
        acc.addEventAlgo(CompFactory.LVL1.TrigT1MBTS())
    return acc

if __name__ == '__main__':
    import sys

    from AthenaConfiguration.MainServicesConfig import MainServicesCfg
    from AthenaConfiguration.AllConfigFlags import initConfigFlags

    flags = initConfigFlags()
    flags.Input.Files = ['/cvmfs/atlas-nightlies.cern.ch/repo/data/data-art/TriggerTest/valid1.410000.PowhegPythiaEvtGen_P2012_ttbar_hdamp172p5_nonallhad.merge.RDO.e4993_s3214_r11315/RDO.17533168._000001.pool.root.1']
    flags.Common.isOnline=False
    flags.Exec.MaxEvents=25
    flags.Concurrency.NumThreads = 1
    flags.Concurrency.NumConcurrentEvents=1
    flags.fillFromArgs()
    flags.lock()

    acc = MainServicesCfg(flags)
    from AthenaPoolCnvSvc.PoolReadConfig import PoolReadCfg
    acc.merge(PoolReadCfg(flags))

    from TrigConfigSvc.TrigConfigSvcCfg import generateL1Menu
    generateL1Menu(flags)

    from AthenaCommon.CFElements import seqAND
    acc.addSequence(seqAND('L1CaloLegacySimSeq'), parentName='AthAlgSeq')
    acc.merge(L1CaloLegacySimCfg(flags), sequenceName='L1CaloLegacySimSeq')

    acc.printConfig(withDetails=True, summariseProps=True, printDefaults=True)
    with open("L1CaloSim.pkl", "wb") as p:
        acc.store(p)
        p.close()

    sys.exit(acc.run().isFailure())
