# Copyright (C) 2002-2020 CERN for the benefit of the ATLAS collaboration

#
# File: CaloClusterCorrection/python/CaloSwEtaoff_v4_1.py
# Created: Jan 2007, sss
# Purpose: Eta offset (s-shape) corrections, v4.1.
#
# These corrections were derived using 12.0.3 simulation and reconstruction,
# for both electrons and photons, 5x5, 3x5, 3x7 clusters.
#
# The difference from v4 is that this is derived with a different
# depth function, which was derived by requiring that the eta position
# resolution be optimal, and the use of large-statistics samples
# for the 50 and 100 GeV energy points.
#
# This correction was in CaloClusterCorrection-00-02-74, in 12.X.0.
#
# See CaloSwEtaoff_v3.h for more details.
#

from CaloClusterCorrection.constants import \
     CALOCORR_EMB1, CALOCORR_EMB2, CALOCORR_EME1, CALOCORR_EME2, \
     EMB1, EME1, EMB2, EME2
from AthenaCommon.SystemOfUnits import GeV

###############################################################################
# Tables for v4_1 sshape correction, 5 GeV electron clusters
#


ele5_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00041221444,    0.0003880279 ],
  [               0,               0,               0,               0,       3.1638417,       1.5267909 ],
  [               0,               0,               0,               0,      -1.7707054,       1.4915055 ],
  [               0,               0,               0,               0,       -5.247785,       1.3097167 ],
  [               0,               0,               0,               0,      0.42856194,     -0.24473727 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       1958.5196,       2741.7193 ],
  [               0,               0,               0,               0,   -0.0028599525,     0.011627152 ],
  [               0,               0,               0,               0,    0.0028602743,    -0.011626715 ],
  [               0,               0,               0,               0,    0.0028590549,    -0.011626578 ],
  [               0,               0,               0,               0,  -4.8012391e-08,  -9.3780224e-09 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0027218002,    0.0080108561 ],
  [               0,               0,               0,               0,      0.07037047,      0.84371892 ],
  [               0,               0,               0,               0,     0.086172315,     -0.68070157 ],
  [               0,               0,               0,               0,     -0.30186016,     -0.64471903 ],
  [               0,               0,               0,               0,     -0.01940861,   -0.0035347405 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,       47302.824,      -32655.116,       6010.7249 ],
  [               0,               0,               0,     0.031937673,     0.046386922,     0.006724709 ],
  [               0,               0,               0,    -0.031877896,    -0.046391105,   -0.0067243895 ],
  [               0,               0,               0,   1.6084151e-05,  -3.2313512e-06,   1.1623984e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     0.091509693,    -0.044025988,     0.017055766 ],
  [               0,               0,               0,      -6.2867245,       3.1964516,       1.4065892 ],
  [               0,               0,               0,       3.8979152,      -1.7565658,     -0.91186376 ],
  [               0,               0,               0,     0.080470816,     0.038176657,     0.007668064 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,      0.42622394,     -0.30544206,     0.063907602 ],
  [               0,               0,               0,       7.7879651,       1.2836201,      0.73740077 ],
  [               0,               0,               0,      0.61429687,      -1.7791832,     -0.60196228 ],
  [               0,               0,               0,      0.39414265,    -0.034704753,    0.0025755695 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele5_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 10 GeV electron clusters
#


ele10_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19013457,      0.16999165 ],
  [               0,               0,               0,               0,   -0.0086751787,      0.28743489 ],
  [               0,               0,               0,               0,    -0.035573789,     -0.26128453 ],
  [               0,               0,               0,               0,    -0.031365039,      -0.2605576 ],
  [               0,               0,               0,               0,          3.9667,       1.2613256 ],
  [               0,               0,               0,               0,    -0.062525284,      0.29627077 ],
  [               0,               0,               0,               0,   4.0542413e-06,    0.0041477301 ],
  [               0,               0,               0,               0,       2.3103194,       2.3810503 ],
 ],
 [
  [               0,               0,               0,     0.010311113,   -0.0082104671,    0.0020588005 ],
  [               0,               0,               0,       14.580366,      -25.743957,       12.489257 ],
  [               0,               0,               0,     -0.24459227,       -1.163953,      0.54639956 ],
  [               0,               0,               0,       17.458078,      -22.979171,       6.9640094 ],
  [               0,               0,               0,       3.0772394,      -3.4754025,       0.8179316 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.96538084,      0.77435082 ],
  [               0,               0,               0,               0,        815.1859,      -648.57165 ],
  [               0,               0,               0,               0,      -562.57111,         447.625 ],
  [               0,               0,               0,               0,       90.236807,      -71.729731 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      -62.821899,       21.241602,       75.211073 ],
  [               0,               0,               0,     -0.03464211,       0.2247228,     -0.14755059 ],
  [               0,               0,               0,     0.034697204,     -0.22478135,      0.14763074 ],
  [               0,               0,               0,     0.034432032,     -0.22419298,      0.14723284 ],
  [               0,               0,               0,  -0.00012708278,   0.00023811262,  -0.00013593153 ],
  [               0,               0,               0,    0.0044545597,   -0.0095336749,     0.005780901 ],
  [               0,               0,               0,       2.1617445,      -5.7445963,       4.5806227 ],
  [               0,               0,               0,      -3.3150684,       3.3402499,     -0.58003286 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0016914266,    0.0086622702,    0.0049310937,    -0.018125364 ],
  [               0,               0,       2.9874166,       4.2803655,       6.1279987,       12.768752 ],
  [               0,               0,      -46.124573,       18.059606,       17.190124,       74.154677 ],
  [               0,               0,        3.425294,       3.7910733,       14.771746,      -39.589578 ],
  [               0,               0,       12.138117,       3.9108523,      -3.6712397,      -39.459229 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00045879333 ],
  [               0,               0,               0,               0,               0,     0.010271582 ],
  [               0,               0,               0,               0,               0,        1.948937 ],
  [               0,               0,               0,               0,               0,       6.7094242 ],
  [               0,               0,               0,               0,               0,       1.5353662 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -858.56227,       764.37801 ],
  [               0,               0,               0,               0,    -0.016302608,     0.021271345 ],
  [               0,               0,               0,               0,     0.016301504,    -0.021268752 ],
  [               0,               0,               0,               0,     0.016305869,    -0.021269212 ],
  [               0,               0,               0,               0,   1.1250306e-06,  -3.0760404e-07 ],
  [               0,               0,               0,               0,  -0.00043106699,   0.00019284888 ],
  [               0,               0,               0,               0,       6.4223753,    -0.040050011 ],
  [               0,               0,               0,               0,       16.579122,      -1.4634772 ],
 ],
 [
  [               0,               0,               0,     0.019174383,    -0.015526962,    0.0038067226 ],
  [               0,               0,               0,       15.643551,      -24.162444,       10.594547 ],
  [               0,               0,               0,      -4.1452664,       4.4079121,      -1.6885955 ],
  [               0,               0,               0,       5.6230215,      -7.7884274,       1.9482427 ],
  [               0,               0,               0,       1.6252177,      -1.8045569,      0.40470179 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.27194958,      0.21912497 ],
  [               0,               0,               0,               0,       576.48776,      -457.35427 ],
  [               0,               0,               0,               0,      -254.13771,       201.42687 ],
  [               0,               0,               0,               0,       148.51229,      -118.06319 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00037265504,  -0.00024899096,    0.0008711242 ],
  [               0,               0,               0,      -58.164348,       111.41036,      -49.183677 ],
  [               0,               0,               0,      -48.806811,       99.616309,      -43.260505 ],
  [               0,               0,               0,       21.402773,      -48.178091,       22.267167 ],
  [               0,               0,               0,      -3.2837531,       2.6579211,      -2.5806783 ],
  [               0,               0,               0,    0.0030057789,   -0.0060567643,    0.0037224753 ],
  [               0,               0,               0,      -1.7318668,       2.4581177,      0.45949767 ],
  [               0,               0,               0,      -1.5875723,       -1.348216,       2.3075116 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0030579866,    0.0030370606,    0.0019993739,    0.0012598922 ],
  [               0,               0,       3.1935941,       4.3791056,       6.0074972,       12.236125 ],
  [               0,               0,       8.1523722,      -77.456462,      -42.427764,       196.14642 ],
  [               0,               0,      -8.2316406,       32.414775,       108.44347,      -197.69446 ],
  [               0,               0,       11.653928,       13.378523,      -14.428438,      -42.365845 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       86904.743 ],
  [               0,               0,               0,               0,               0,    0.0052540714 ],
  [               0,               0,               0,               0,               0,   -0.0052540508 ],
  [               0,               0,               0,               0,               0,   -0.0052540207 ],
  [               0,               0,               0,               0,               0,   7.2260196e-09 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00023509918,   0.00016629675 ],
  [               0,               0,               0,               0,       32.303299,       5.6810692 ],
  [               0,               0,               0,               0,       24.307933,       22.468875 ],
  [               0,               0,               0,               0,       27.375522,       23.674718 ],
  [               0,               0,               0,               0,       1373.0232,       -10.55874 ],
  [               0,               0,               0,               0,     -0.09155849,     0.081019132 ],
  [               0,               0,               0,               0,      0.02400444,     0.011077119 ],
  [               0,               0,               0,               0,       2.7334254,       1.5516503 ],
 ],
 [
  [               0,               0,               0,    -0.053015272,     0.073122812,     -0.01461503 ],
  [               0,               0,               0,       2.2771425,      -3.6477861,        2.124645 ],
  [               0,               0,               0,     -0.51197658,       1.4126631,      -1.2712511 ],
  [               0,               0,               0,     -0.27867503,       1.0541723,       -1.148479 ],
  [               0,               0,               0,    0.0042119315,     -0.01616092,   -0.0028590606 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.1397008,      0.11261723 ],
  [               0,               0,               0,               0,         1543.82,      -1227.3672 ],
  [               0,               0,               0,               0,      -655.04052,       520.74645 ],
  [               0,               0,               0,               0,       285.59784,      -226.99977 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      -66.863437,       55.421611,       35.454073 ],
  [               0,               0,               0,      0.36515097,      -0.4933674,     0.092881387 ],
  [               0,               0,               0,     -0.36543428,      0.49410932,    -0.093242326 ],
  [               0,               0,               0,      -0.3658272,      0.49503903,    -0.093902272 ],
  [               0,               0,               0,  -0.00031262022,   0.00063791162,  -0.00035728394 ],
  [               0,               0,               0,    0.0094966045,    -0.020868561,     0.012082136 ],
  [               0,               0,               0,      -2.9794661,       5.4682765,      -1.2335604 ],
  [               0,               0,               0,      -5.5996682,       8.8928315,       -4.033436 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0010104658,   0.00055396339,  -0.00072949527,    0.0042737159 ],
  [               0,               0,       3.4488029,       4.5412914,       5.8059523,       11.499451 ],
  [               0,               0,      -33.316875,      -12.829884,       22.569554,       92.304888 ],
  [               0,               0,       17.471689,       6.0158338,      -11.667873,      -47.451441 ],
  [               0,               0,       11.742017,       5.1265531,      -8.8328246,      -33.857129 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       103836.04 ],
  [               0,               0,               0,               0,               0,    0.0051323895 ],
  [               0,               0,               0,               0,               0,   -0.0051323698 ],
  [               0,               0,               0,               0,               0,   -0.0051323446 ],
  [               0,               0,               0,               0,               0,   6.0231062e-09 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,     0.023392403,   -0.0099830201,     0.003143927 ],
  [               0,               0,               0,       19.263832,      -7.7566772,       4.0459158 ],
  [               0,               0,               0,       40.904256,      -17.821235,      0.67284747 ],
  [               0,               0,               0,      -21.086604,       9.4012397,     -0.96905383 ],
  [               0,               0,               0,    0.0046062667,   -0.0026591528,    0.0019820785 ],
  [               0,               0,               0,      -2.4543444,      0.88138716,      0.96989983 ],
  [               0,               0,               0,       32.447993,      -13.380574,     -0.12223269 ],
 ],
 [
  [               0,               0,               0,    0.0040684982,   -0.0061284999,    0.0023953981 ],
  [               0,               0,               0,       712.91688,      -872.45357,       263.98859 ],
  [               0,               0,               0,       467.02927,      -422.92188,       88.956319 ],
  [               0,               0,               0,      -92.503691,       92.225353,      -19.410649 ],
  [               0,               0,               0,   -0.0052772859,    0.0059461838,    0.0005763537 ],
  [               0,               0,               0,      -1.7510903,       1.9625266,       0.4878185 ],
  [               0,               0,               0,       4.0164176,      -1.0792836,      -2.2620571 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0038069116 ],
  [               0,               0,               0,               0,               0,       3.1592574 ],
  [               0,               0,               0,               0,               0,      -1.3308446 ],
  [               0,               0,               0,               0,               0,   0.00011391873 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.012041099,    -0.026566025,     0.018616533 ],
  [               0,               0,               0,      -5.1552795,       9.0528467,     -0.76424568 ],
  [               0,               0,               0,      0.97868735,      -1.9830174,     -0.40632933 ],
  [               0,               0,               0,      -1.0596939,       3.0915735,       -1.652309 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0034630612,   -0.0011837134,    0.0022898593,    0.0083366446 ],
  [               0,               0,        3.486797,       4.7381569,       6.2855267,       12.356767 ],
  [               0,               0,      0.57285183,       66.303278,      -50.446287,      -57.445554 ],
  [               0,               0,        30.51946,       34.725177,      -17.718132,      -110.73729 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.040485726,    -0.051417221 ],
  [               0,               0,               0,               0,        439.1042,       -615.8461 ],
  [               0,               0,               0,               0,      -224.29299,       312.89272 ],
  [               0,               0,               0,               0,      -0.2896421,       0.6381007 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     0.086260391,    -0.038307146,      0.01324945 ],
  [               0,               0,               0,       -24.92715,       10.951362,        1.766896 ],
  [               0,               0,               0,       4.3851481,      -2.3001073,     -0.97244476 ],
  [               0,               0,               0,      0.52493169,     0.031164071,    -0.025541816 ],
  [               0,               0,               0,    0.0015717704,    0.0016840668,  -0.00054595909 ],
  [               0,               0,               0,       12.188741,      -9.8385835,       1.8212004 ],
  [               0,               0,               0,       -99.13717,       26.649514,      -4.8135128 ],
 ],
 [
  [               0,               0,               0,    -0.014789902,       0.0232904,   -0.0010747353 ],
  [               0,               0,               0,      -6.8154984,       1.7012397,       4.2494112 ],
  [               0,               0,               0,        3.547862,      -2.5331649,     -0.98053118 ],
  [               0,               0,               0,          1.8844,      -2.4803264,      0.85004236 ],
  [               0,               0,               0,     0.026675644,    -0.028671027,    0.0081170739 ],
  [               0,               0,               0,       2.9050384,      -1.6902728,      0.71650244 ],
  [               0,               0,               0,      -15.643741,       21.669672,      -8.4504894 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.010041784 ],
  [               0,               0,               0,               0,               0,       2.2053263 ],
  [               0,               0,               0,               0,               0,      -1.1837277 ],
  [               0,               0,               0,               0,               0,    0.0052153983 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.00322977,    -0.017695675,     0.023298836 ],
  [               0,               0,               0,       3.4950639,      -5.2991671,       4.2233041 ],
  [               0,               0,               0,      -1.3138872,          2.0189,      -1.9885426 ],
  [               0,               0,               0,      0.73733418,     -0.86673098,      0.26891111 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      -61.559088,       125.27036,        144.9736,      -272.51278 ],
  [               0,               0,      -209.88113,       276.88603,       431.58763,      -571.15415 ],
  [               0,               0,       223.24075,      -331.65363,      -356.66933,       536.95565 ],
  [               0,               0,      -4.0842821,       10.447681,      -5.5800217,      -1.4602417 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,        1207.364,      -1692.7043 ],
  [               0,               0,               0,               0,       3.8059072,      -4.8543955 ],
  [               0,               0,               0,               0,      -1.0198149,       6.5983079 ],
  [               0,               0,               0,               0,       149.81068,      -215.35335 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,        0.188896,    -0.055583339,     0.013412266 ],
  [               0,               0,               0,      -33.618093,       12.153958,       1.8330451 ],
  [               0,               0,               0,       11.066382,       -3.941838,     -0.96525519 ],
  [               0,               0,               0,      -2.6608403,      0.92646925,    -0.040292512 ],
  [               0,               0,               0,    0.0075769642,    0.0021782776,  -0.00035451988 ],
  [               0,               0,               0,      0.82100755,       -3.139634,       1.9702625 ],
  [               0,               0,               0,      -32.909125,       6.8189147,       -4.133101 ],
 ],
 [
  [               0,               0,               0,     -0.13470497,        0.161841,    -0.035968426 ],
  [               0,               0,               0,       14.951703,      -19.723073,       8.5200156 ],
  [               0,               0,               0,      -3.2577744,       4.2192571,      -2.4941926 ],
  [               0,               0,               0,       4.5783142,      -5.2233959,       1.4550695 ],
  [               0,               0,               0,    -0.015912224,     0.018050604,     -0.00417568 ],
  [               0,               0,               0,        39.11824,       -43.78257,       11.929108 ],
  [               0,               0,               0,      -46.160511,       57.906338,      -17.984159 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.010048184 ],
  [               0,               0,               0,               0,               0,       2.2226793 ],
  [               0,               0,               0,               0,               0,       -1.199105 ],
  [               0,               0,               0,               0,               0,     0.011567688 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.008425201,    -0.031827656,     0.032383163 ],
  [               0,               0,               0,       4.1758854,      -5.9654154,       4.1556196 ],
  [               0,               0,               0,       -1.428183,       2.1767254,      -2.0314647 ],
  [               0,               0,               0,       1.0580471,      -1.4816146,      0.56501887 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     -0.65326491,     -0.51818848,      0.37774634,       2.8945095 ],
  [               0,               0,      -276.12423,       371.43848,        548.6881,      -738.53427 ],
  [               0,               0,       303.96512,      -439.87925,      -516.97019,       751.94943 ],
  [               0,               0,      -4.9961505,       14.159894,      -10.428394,       0.5629232 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.43032677,      -0.4674238 ],
  [               0,               0,               0,               0,      -41.845998,       62.351016 ],
  [               0,               0,               0,               0,        40.88372,      -58.507694 ],
  [               0,               0,               0,               0,      0.14087841,     -0.31760898 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele10_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 25 GeV electron clusters
#


ele25_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00035789401,   0.00047668436 ],
  [               0,               0,               0,               0,       1.3533248,       4.0340552 ],
  [               0,               0,               0,               0,      -99.366448,       42.987407 ],
  [               0,               0,               0,               0,      -95.301425,       41.722398 ],
  [               0,               0,               0,               0,        62.11459,      -18.403195 ],
  [               0,               0,               0,               0,    -0.039101268,     0.048585463 ],
  [               0,               0,               0,               0,     -0.31705703,      0.13583647 ],
  [               0,               0,               0,               0,      0.61804932,       1.3891945 ],
 ],
 [
  [               0,               0,               0,     -0.03966172,   -0.0038179499,     0.056188912 ],
  [               0,               0,               0,      -1.2147703,       1.2352423,      0.17285025 ],
  [               0,               0,               0,       1.3295004,      -1.3249085,     -0.12173583 ],
  [               0,               0,               0,       1.3084586,       -1.320366,     -0.12280844 ],
  [               0,               0,               0,    -0.038312083,     0.029567225,   -0.0073470209 ],
  [               0,               0,               0,   0.00073803482,   -0.0012711608,    0.0004803945 ],
  [               0,               0,               0,       50.436486,      -63.012152,        23.40995 ],
  [               0,               0,               0,      -136.40854,       172.10082,      -47.966525 ],
 ],
 [
  [               0,               0,               0,               0,      -8.0091947,       6.4244049 ],
  [               0,               0,               0,               0,       56.103989,      -44.260456 ],
  [               0,               0,               0,               0,      -67.604222,       53.427975 ],
  [               0,               0,               0,               0,       14.857672,      -11.816461 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00066739251,   -0.0011011302,    0.0004711279 ],
  [               0,               0,               0,       284.17012,      -696.98392,       415.91888 ],
  [               0,               0,               0,       82.265419,      -248.76642,       176.40507 ],
  [               0,               0,               0,       341.99613,      -942.61076,       645.48714 ],
  [               0,               0,               0,      -36.457815,       87.825381,      -51.014609 ],
  [               0,               0,               0,    -0.010271383,     0.019272521,   -0.0091533954 ],
  [               0,               0,               0,     0.048566675,      -2.7935405,       4.3883994 ],
  [               0,               0,               0,     -0.29710678,       3.0043597,      -2.0822957 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.04173055,      0.02514905,    -0.022909261,     -0.12951454 ],
  [               0,               0,       137.77477,      -211.91372,       -283.5011,       439.24205 ],
  [               0,               0,       927.38699,      -1335.8889,      -1920.3922,       2764.7066 ],
  [               0,               0,      -362.52792,        504.0573,       707.97159,      -984.96624 ],
  [               0,               0,       71.331307,      -96.533208,      -140.60267,       189.54917 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,       85475.588 ],
  [               0,               0,               0,               0,               0,   -0.0056621256 ],
  [               0,               0,               0,               0,               0,    0.0056620521 ],
  [               0,               0,               0,               0,               0,   6.2257958e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00090599557 ],
  [               0,               0,               0,               0,               0,       39.219195 ],
  [               0,               0,               0,               0,               0,       -4.733301 ],
  [               0,               0,               0,               0,               0,      -5.5503665 ],
  [               0,               0,               0,               0,               0,     -0.55151646 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       166545.14 ],
  [               0,               0,               0,               0,               0,    0.0056322546 ],
  [               0,               0,               0,               0,               0,   -0.0056321925 ],
  [               0,               0,               0,               0,               0,   6.2920033e-10 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.027016406,    -0.039237033,    -0.073058206,      0.10763104 ],
  [               0,               0,       26.269449,      -41.814536,       -94.69083,       154.11968 ],
  [               0,               0,      -1.8441573,       9.6403165,      -18.477729,       11.937113 ],
  [               0,               0,      -4.7769063,       -14.12115,       85.146697,      -80.796974 ],
  [               0,               0,       -3.899164,       13.545363,      -12.280089,       0.9005155 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.010219307,     -0.01500732 ],
  [               0,               0,               0,               0,      -2.4553516,       5.7078383 ],
  [               0,               0,               0,               0,      0.90233821,      -2.3437759 ],
  [               0,               0,               0,               0,      0.96549275,      -2.3839505 ],
  [               0,               0,               0,               0,    -0.034163703,     0.045337867 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00032401924,  -0.00071213741,   0.00019906514,    0.0084784292 ],
  [               0,               0,      -1.5004174,       4.7754207,        4.602546,      -14.806398 ],
  [               0,               0,        5.844199,      -12.693527,      -24.324197,       52.072803 ],
  [               0,               0,       3.7460233,      -7.2686678,      -19.063823,       36.474048 ],
  [               0,               0,     -0.66587081,       1.1344708,       4.1783295,      -7.6405496 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       1283.1396 ],
  [               0,               0,               0,               0,               0,     0.015015072 ],
  [               0,               0,               0,               0,               0,    -0.015014861 ],
  [               0,               0,               0,               0,               0,     8.29434e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00063153758 ],
  [               0,               0,               0,               0,               0,       4.5878428 ],
  [               0,               0,               0,               0,               0,      0.95432812 ],
  [               0,               0,               0,               0,               0,     0.087023263 ],
  [               0,               0,               0,               0,               0,   -0.0093397152 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0039530439,    -0.032075459,     0.062946189 ],
  [               0,               0,               0,       510.10577,      -2488.1924,       3038.4495 ],
  [               0,               0,               0,       83.260182,      -396.02058,       468.84792 ],
  [               0,               0,               0,        16.43253,       68.235183,      -257.44014 ],
  [               0,               0,               0,      0.24997013,    -0.062537319,      -1.2798064 ],
  [               0,               0,               0,       33.915771,      -170.32384,       213.75016 ],
  [               0,               0,               0,       7.1088039,      -29.472387,        26.53977 ],
 ],
 [
  [               0,               0,               0,               0,     0.015095937,    -0.032324564 ],
  [               0,               0,               0,               0,      -1797.3355,        4468.273 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,         -55.078,       134.70045 ],
  [               0,               0,               0,               0,     0.038755179,     -0.31824956 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0016381758,     0.001108542 ],
  [               0,               0,               0,               0,       10.511284,       1.5802974 ],
  [               0,               0,               0,               0,      -1.1976271,      -0.4783496 ],
  [               0,               0,               0,               0,      0.95341024,     -0.41807358 ],
  [               0,               0,               0,               0,     -0.33044982,        0.140938 ],
  [               0,               0,               0,               0,  -0.00074099662,   0.00036618427 ],
  [               0,               0,               0,               0,       5.8967161,    -0.084252166 ],
  [               0,               0,               0,               0,      0.67267418,       1.8812634 ],
 ],
 [
  [               0,               0,               0,     -0.04371854,   -0.0023777664,     0.058009081 ],
  [               0,               0,               0,     -0.99032008,       1.0015451,      0.22687108 ],
  [               0,               0,               0,       1.1159547,       -1.102644,     -0.17398495 ],
  [               0,               0,               0,        1.094282,      -1.0971987,     -0.17522222 ],
  [               0,               0,               0,    -0.037942709,     0.029424629,   -0.0072974779 ],
  [               0,               0,               0,   0.00076160325,   -0.0013008064,   0.00048932696 ],
  [               0,               0,               0,       50.372041,      -62.882593,        23.37678 ],
  [               0,               0,               0,      -137.89659,       173.86678,      -48.454602 ],
 ],
 [
  [               0,               0,               0,               0,      0.10154929,    -0.080757346 ],
  [               0,               0,               0,               0,       21.779223,       22.076338 ],
  [               0,               0,               0,               0,       4570.5055,      -3656.9495 ],
  [               0,               0,               0,               0,      -3036.8255,       2428.9449 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.027955677,    -0.046891377,     0.020197294 ],
  [               0,               0,               0,      -28.208915,       62.045152,      -32.458903 ],
  [               0,               0,               0,       28.760449,       -65.68503,       36.187792 ],
  [               0,               0,               0,       37.305528,      -88.075213,       50.887574 ],
  [               0,               0,               0,      -2.4697658,        5.720733,      -3.2330769 ],
  [               0,               0,               0,   -0.0087146679,     0.016268712,   -0.0077029838 ],
  [               0,               0,               0,       7.6232895,      -19.983646,       13.685487 ],
  [               0,               0,               0,       6.7724362,      -12.247866,       5.7763557 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.73120644,     -0.94255141,      -1.4414529,       1.8614943 ],
  [               0,               0,      -1148.6954,       1595.1711,       2411.1745,      -3347.1428 ],
  [               0,               0,       1320.0023,      -1892.8912,      -2753.1776,       3946.5032 ],
  [               0,               0,      -464.08617,       653.35997,        889.0825,      -1252.5659 ],
  [               0,               0,       241.03241,      -335.58989,      -485.25545,       674.88404 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    -0.004859539,    0.0070610917 ],
  [               0,               0,               0,               0,       12.397158,       21.674179 ],
  [               0,               0,               0,               0,       600.77853,      -785.70643 ],
  [               0,               0,               0,               0,       611.67826,      -888.42325 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00041229844 ],
  [               0,               0,               0,               0,               0,       -39.81755 ],
  [               0,               0,               0,               0,               0,      -6.0492476 ],
  [               0,               0,               0,               0,               0,        6.051534 ],
  [               0,               0,               0,               0,               0,       2.5967665 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.012219475 ],
  [               0,               0,               0,               0,               0,       2.0366547 ],
  [               0,               0,               0,               0,               0,      -1.2163074 ],
  [               0,               0,               0,               0,               0,     0.071821219 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       8.8065422,        -15.9788,      -26.213032,       48.338367 ],
  [               0,               0,      -40.651193,        204.5477,      -342.24798,       190.59615 ],
  [               0,               0,       40.388857,      -203.21341,       339.99932,      -189.33833 ],
  [               0,               0,       39.812645,      -200.42569,       335.51339,      -186.93699 ],
  [               0,               0,     -0.10277514,      0.49887466,     -0.80594365,      0.43332465 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0045133884,   -0.0065793897 ],
  [               0,               0,               0,               0,      -3.6325883,       8.4221804 ],
  [               0,               0,               0,               0,      0.15255474,     -0.86113295 ],
  [               0,               0,               0,               0,      0.34563042,      -1.0552836 ],
  [               0,               0,               0,               0,    -0.080160799,      0.10849868 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0002819129,  -0.00085242206,  -4.5207396e-05,    0.0092348599 ],
  [               0,               0,      -2.1471847,       6.2081721,       7.5422886,      -21.310419 ],
  [               0,               0,       6.4690296,      -14.080971,      -26.974259,       57.969817 ],
  [               0,               0,       4.2097329,      -8.2534938,       -21.21115,       41.027795 ],
  [               0,               0,     -0.68163879,       1.1469526,        4.337274,      -7.8836604 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       2148.8782 ],
  [               0,               0,               0,               0,               0,     0.012643622 ],
  [               0,               0,               0,               0,               0,    -0.012643496 ],
  [               0,               0,               0,               0,               0,   4.9528226e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00078441335 ],
  [               0,               0,               0,               0,               0,        3.730652 ],
  [               0,               0,               0,               0,               0,      0.58519312 ],
  [               0,               0,               0,               0,               0,     -0.24411567 ],
  [               0,               0,               0,               0,               0,    -0.026473067 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00094858373,    0.0039955073,    0.0039655099 ],
  [               0,               0,               0,       64.368962,      -325.79797,       416.14347 ],
  [               0,               0,               0,      -67.172467,       336.41507,      -422.67752 ],
  [               0,               0,               0,       5.3787079,       314.41857,      -781.91638 ],
  [               0,               0,               0,      0.55729713,    0.0046504462,      -3.1907281 ],
  [               0,               0,               0,       21.095436,      -105.49406,       131.86753 ],
  [               0,               0,               0,      -13.597773,       67.775732,      -87.523807 ],
 ],
 [
  [               0,               0,               0,               0,    0.0032019383,   -0.0031933515 ],
  [               0,               0,               0,               0,      -1732.3324,       4306.3216 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -38.820228,       94.414416 ],
  [               0,               0,               0,               0,     0.023997403,      -0.3209254 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0013507647,   0.00096210535 ],
  [               0,               0,               0,               0,       9.8472679,        1.757604 ],
  [               0,               0,               0,               0,     -0.35896629,     -0.56303937 ],
  [               0,               0,               0,               0,       1.4206883,     -0.46102865 ],
  [               0,               0,               0,               0,      -0.4328776,       0.1479781 ],
  [               0,               0,               0,               0,   -0.0005608501,   0.00027913131 ],
  [               0,               0,               0,               0,       6.4640794,     -0.21969056 ],
  [               0,               0,               0,               0,      0.68284682,       1.9295742 ],
 ],
 [
  [               0,               0,               0,    -0.034565707,    -0.006014203,     0.057620026 ],
  [               0,               0,               0,       1.7570296,      -1.9213387,      0.96516901 ],
  [               0,               0,               0,      -1.5389512,       1.7141873,     -0.88360337 ],
  [               0,               0,               0,      -1.5504394,       1.7122201,      -0.8841912 ],
  [               0,               0,               0,    -0.033677785,     0.025511971,   -0.0065700432 ],
  [               0,               0,               0,   0.00011400871,   0.00024880107,  -0.00020751127 ],
  [               0,               0,               0,       94.481129,      -117.89461,       38.616961 ],
  [               0,               0,               0,      -118.32617,       150.22909,      -45.200474 ],
 ],
 [
  [               0,               0,               0,               0,     0.071790792,    -0.039798413 ],
  [               0,               0,               0,               0,      -16.906048,       14.091716 ],
  [               0,               0,               0,               0,      -27.878462,       21.625688 ],
  [               0,               0,               0,               0,       25.380587,      -20.186007 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0040889246,    0.0065286824,  -0.00047121638 ],
  [               0,               0,               0,      -12.100492,       15.384303,      -2.5441185 ],
  [               0,               0,               0,       25.115331,      -43.949092,       17.994268 ],
  [               0,               0,               0,      -7.6110971,       21.445117,      -13.373882 ],
  [               0,               0,               0,      -6.7698147,        13.73464,      -6.8127932 ],
  [               0,               0,               0,    0.0010202792,   -0.0031182108,    0.0017098194 ],
  [               0,               0,               0,       8.9930664,      -20.626741,        13.00131 ],
  [               0,               0,               0,       1.4371258,       2.2249236,       -3.368448 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.035053071,     0.030053003,     -0.01261118,     -0.13497965 ],
  [               0,               0,      -14.911406,        12.44903,       27.966869,      -18.530485 ],
  [               0,               0,        1097.631,      -1581.5928,      -2277.4857,       3279.9399 ],
  [               0,               0,      -397.55188,       557.43139,       773.87687,      -1086.0142 ],
  [               0,               0,       127.82204,      -175.81643,       -254.1561,       348.80849 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,   -0.0047342717,    0.0068842983 ],
  [               0,               0,               0,               0,       12.193654,       21.697133 ],
  [               0,               0,               0,               0,       1636.6223,      -2276.4784 ],
  [               0,               0,               0,               0,       470.61437,      -663.31186 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       89930.243 ],
  [               0,               0,               0,               0,               0,    0.0049143478 ],
  [               0,               0,               0,               0,               0,   -0.0049143328 ],
  [               0,               0,               0,               0,               0,   -0.0049143635 ],
  [               0,               0,               0,               0,               0,   9.1303911e-11 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0052264161 ],
  [               0,               0,               0,               0,               0,       4.3643386 ],
  [               0,               0,               0,               0,               0,      -1.4665149 ],
  [               0,               0,               0,               0,               0,      0.11583178 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     -0.33930299,       1.8027237,      -3.1934929,       1.8863481 ],
  [               0,               0,       224.39508,      -523.46861,      -44.118822,       489.61863 ],
  [               0,               0,       1984.2216,      -9673.6389,       15730.596,      -8529.3264 ],
  [               0,               0,       2499.2095,      -12141.461,       19673.133,      -10629.017 ],
  [               0,               0,      -107.71054,       509.21423,      -802.05947,       420.87011 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0035738105,   -0.0056436587 ],
  [               0,               0,               0,               0,      -4.9070249,       11.388558 ],
  [               0,               0,               0,               0,     -0.88461814,       1.3410196 ],
  [               0,               0,               0,               0,      -4.2959755,       8.1917462 ],
  [               0,               0,               0,               0,     -0.52824946,      0.91946834 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.019447885,    -0.046442674,    -0.093291481,      0.22474687 ],
  [               0,               0,      -8.6507901,       70.482295,      -174.77988,       137.95298 ],
  [               0,               0,       49.664252,      -284.44621,       538.10849,      -336.44808 ],
  [               0,               0,      -15.604062,        114.7627,      -275.00352,       215.01141 ],
  [               0,               0,       18.379464,      -121.95177,       268.65924,      -196.60739 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       4870.6688 ],
  [               0,               0,               0,               0,               0,    0.0097257512 ],
  [               0,               0,               0,               0,               0,   -0.0097256865 ],
  [               0,               0,               0,               0,               0,   2.1647486e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00045512086 ],
  [               0,               0,               0,               0,               0,       5.6742626 ],
  [               0,               0,               0,               0,               0,       2.3340257 ],
  [               0,               0,               0,               0,               0,      0.76163763 ],
  [               0,               0,               0,               0,               0,     -0.10476495 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0047486506,     0.018233085,    -0.008162191 ],
  [               0,               0,               0,       390.02827,      -1913.4257,        2350.888 ],
  [               0,               0,               0,       11.563456,      -48.003846,       46.498321 ],
  [               0,               0,               0,       12.422865,       75.881071,      -252.78918 ],
  [               0,               0,               0,      0.21967231,    -0.020003698,      -1.2090441 ],
  [               0,               0,               0,       50.266787,      -249.61881,       309.90289 ],
  [               0,               0,               0,      -54.033879,       265.62043,      -329.50392 ],
 ],
 [
  [               0,               0,               0,               0,      0.11692137,     -0.28892513 ],
  [               0,               0,               0,               0,      -4402.4121,        10936.64 ],
  [               0,               0,               0,               0,      0.55397227,       237.42506 ],
  [               0,               0,               0,               0,       314.16633,      -778.22625 ],
  [               0,               0,               0,               0,     -0.15359948,    -0.021953043 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,   -0.0046028419,    0.0057428907,     0.008394533 ],
  [               0,               0,               0,       -1.665632,     -0.41598484,       2.9715682 ],
  [               0,               0,               0,     -0.13491465,    -0.040132829,      -1.2283718 ],
  [               0,               0,               0,     0.060246235,      0.11752006,   -0.0089449622 ],
  [               0,               0,               0,    0.0062259558,   -0.0022535032,   0.00015500918 ],
  [               0,               0,               0,      -29.506403,      -1.6842852,       2.5073215 ],
  [               0,               0,               0,      -45.293656,      -14.922573,      0.72803387 ],
 ],
 [
  [               0,               0,               0,    0.0028039758,      0.01152238,   0.00045357374 ],
  [               0,               0,               0,       4.2016105,      -10.284574,       7.3402739 ],
  [               0,               0,               0,        1.740327,      -1.5546788,     -0.93759452 ],
  [               0,               0,               0,     -0.24033055,       0.5580593,     -0.18317688 ],
  [               0,               0,               0,    0.0033390208,   -0.0046812364,    0.0019890788 ],
  [               0,               0,               0,       1.2793575,      0.32186336,      0.37948667 ],
  [               0,               0,               0,       7.0930506,       -6.845824,      0.37208877 ],
 ],
 [
  [               0,               0,               0,               0,       0.1592916,     -0.11808099 ],
  [               0,               0,               0,               0,      -0.7900183,       3.0144648 ],
  [               0,               0,               0,               0,       6.2527232,      -6.2571568 ],
  [               0,               0,               0,               0,      -3.7415511,       3.0207826 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0026993944,   -0.0093909057,      0.01492282 ],
  [               0,               0,               0,     -0.51020366,      0.73714199,       2.3478562 ],
  [               0,               0,               0,       1.4586936,       -3.638916,      0.75933187 ],
  [               0,               0,               0,       47.199559,      -33.102356,       1.5084448 ],
  [               0,               0,               0,       0.1957523,     0.045903302,     -0.11473896 ],
  [               0,               0,               0,     -0.25465661,      0.54668071,      -0.2586117 ],
  [               0,               0,               0,     0.044048067,     -0.11212231,      -3.0697425 ],
 ],
 [
  [               0,               0,               0,    -0.012812879,   -0.0011723229,     0.035528336 ],
  [               0,               0,               0,      -406.51131,       1081.5185,      -716.56638 ],
  [               0,               0,               0,      -259.12357,       730.71778,      -515.88438 ],
  [               0,               0,               0,      -37.339653,       87.859839,      -50.048309 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,       10.262221,      -28.970818,       20.450057 ],
  [               0,               0,               0,      -390.13778,       217.90202,       499.77747 ],
  [               0,               0,               0,      -143.01528,       89.807922,       151.71455 ],
  [               0,               0,               0,      -2.0746218,      0.94022437,       3.7454718 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      -1.0011184,       1.4260073,       1.9459667,       -2.766174 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.61081563,      0.91582214 ],
  [               0,               0,               0,               0,      -59.116413,       88.042303 ],
  [               0,               0,               0,               0,       18.606741,      -28.011877 ],
  [               0,               0,               0,               0,      -11.026245,        16.08035 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0015679171,   -0.0013952569 ],
  [               0,               0,               0,               0,       1647.6036,       -2472.549 ],
  [               0,               0,               0,               0,      -50.967927,       76.106479 ],
  [               0,               0,               0,               0,       90.133326,       -137.5367 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.035553329,     0.055045595,     0.023093307 ],
  [               0,               0,               0,       19.517208,      -56.215274,       40.918543 ],
  [               0,               0,               0,       28.917251,      -98.053992,       82.033019 ],
  [               0,               0,               0,      -12.476945,       41.069323,      -33.754017 ],
  [               0,               0,               0,     0.047344882,     -0.16080204,      0.13738242 ],
  [               0,               0,               0,       1.8391999,      -6.6762384,       7.0783198 ],
  [               0,               0,               0,       3.4451269,      -31.772361,        39.90919 ],
 ],
 [
  [               0,               0,               0,      0.03097767,    -0.058643155,    0.0001012903 ],
  [               0,               0,               0,      -6.4742315,       23.363358,      -18.917617 ],
  [               0,               0,               0,    -0.048804524,       2.5882752,      -6.1823111 ],
  [               0,               0,               0,       2.2928014,      -10.254867,       11.408479 ],
  [               0,               0,               0,    -0.021874189,      0.10216542,     -0.11795318 ],
  [               0,               0,               0,      -5.9537013,       25.096088,      -25.270073 ],
  [               0,               0,               0,      -8.1617441,       27.307062,      -27.133611 ],
 ],
 [
  [               0,               0,               0,     -0.02178334,    0.0017845678,      0.14018752 ],
  [               0,               0,               0,       37.563098,      -174.73911,       204.31932 ],
  [               0,               0,               0,      -5.2547234,       23.323429,      -26.630101 ],
  [               0,               0,               0,      -10.295169,       49.203416,      -58.770329 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,        3835.501 ],
  [               0,               0,               0,               0,               0,     0.015284207 ],
  [               0,               0,               0,               0,               0,    -0.015283293 ],
  [               0,               0,               0,               0,               0,   4.4039094e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,   -0.0028338539,   0.00013595135,    0.0076676194 ],
  [               0,               0,               0,      -2.3830979,       1.5148569,       3.6758498 ],
  [               0,               0,               0,      -1.9061129,      0.71602842,      -1.4033405 ],
  [               0,               0,               0,       209.24715,      -159.99174,       31.877696 ],
  [               0,               0,               0,       1.5950889,      -1.2229525,      0.24442961 ],
  [               0,               0,               0,      0.89170436,     -0.28675203,     0.037057274 ],
  [               0,               0,               0,      0.71257312,      0.82375236,      -3.1356272 ],
 ],
 [
  [               0,               0,               0,    -0.020877016,     0.034626207,   -0.0046578813 ],
  [               0,               0,               0,       6.4017459,      -12.391313,       8.4834122 ],
  [               0,               0,               0,      0.33246822,     -0.36461195,      -1.2306181 ],
  [               0,               0,               0,     -0.93381459,       4.1601625,      -2.6085296 ],
  [               0,               0,               0,     -0.01406065,  -0.00078424795,     0.010276637 ],
  [               0,               0,               0,     -0.43061547,        0.306915,     0.044443604 ],
  [               0,               0,               0,       -11.10968,       11.920954,      -3.5516341 ],
 ],
 [
  [               0,               0,               0,               0,     0.087449055,    -0.059538909 ],
  [               0,               0,               0,               0,       4.6222204,      -1.0429665 ],
  [               0,               0,               0,               0,       4.6584093,      -4.9953739 ],
  [               0,               0,               0,               0,      -3.1009112,       2.4941996 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0016507195,  -0.00046570817,    0.0093198212 ],
  [               0,               0,               0,      -2.6359636,       5.0592877,      0.99192117 ],
  [               0,               0,               0,       2.7476759,      -6.3100556,        2.024887 ],
  [               0,               0,               0,     -0.59033563,       1.6261586,     -0.80172157 ],
  [               0,               0,               0,    0.0062058015,    -0.013919798,    0.0083675134 ],
  [               0,               0,               0,      -4.6280482,       11.458361,      -6.1686087 ],
  [               0,               0,               0,       10.522427,      -20.862262,       7.7835012 ],
 ],
 [
  [               0,               0,               0,      0.18735625,     -0.05312251,     -0.26989477 ],
  [               0,               0,               0,      -3304.8206,       8853.3563,      -5923.9729 ],
  [               0,               0,               0,       688.57506,      -1821.8938,       1201.9916 ],
  [               0,               0,               0,       205.92142,      -555.19344,       374.54811 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0030248011,    0.0011533807,  -0.00043430699 ],
  [               0,               0,               0,      -440.39898,       4.4840356,       901.77027 ],
  [               0,               0,               0,       80.637964,      -261.48065,       208.09705 ],
  [               0,               0,               0,    -0.008078403,     -0.77804219,      -0.7825836 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -2.2437084,       3.2334157,        4.546461,      -6.5479785 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.013935784,     0.020758168 ],
  [               0,               0,               0,               0,      -52.540517,       116.13699 ],
  [               0,               0,               0,               0,       485.44845,      -714.66708 ],
  [               0,               0,               0,               0,      -360.96756,       535.55704 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.11282023,     -0.16748772 ],
  [               0,               0,               0,               0,       349.91113,      -525.24797 ],
  [               0,               0,               0,               0,      -145.03558,       217.85105 ],
  [               0,               0,               0,               0,       121.01747,      -183.09566 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0061936388,    0.0099086453,     0.013155974 ],
  [               0,               0,               0,       1.1686897,      -1.4830923,       1.3669358 ],
  [               0,               0,               0,        1.826115,      -6.8679747,       5.2269498 ],
  [               0,               0,               0,        1.720027,      -5.1571951,       3.8241529 ],
  [               0,               0,               0,    0.0049768805,   -0.0098554623,    0.0029906404 ],
  [               0,               0,               0,       7.7752593,      -33.194277,        34.82301 ],
  [               0,               0,               0,      -13.152215,       49.175915,      -49.337405 ],
 ],
 [
  [               0,               0,               0,      0.16518212,     -0.68898984,        0.725907 ],
  [               0,               0,               0,      -29.077001,       121.42388,      -123.53569 ],
  [               0,               0,               0,     -0.70604099,       2.9726509,        -4.29601 ],
  [               0,               0,               0,      0.81928483,      -3.4751358,       3.6610517 ],
  [               0,               0,               0,    0.0082627542,     -0.03263519,     0.032489878 ],
  [               0,               0,               0,      -7.8921015,       29.883041,      -25.829627 ],
  [               0,               0,               0,       24.797808,      -104.13793,        97.92528 ],
 ],
 [
  [               0,               0,               0,    -0.019578741,    0.0010870806,      0.13012047 ],
  [               0,               0,               0,       40.407007,      -187.97824,       219.93241 ],
  [               0,               0,               0,      -2.6958151,         11.2842,      -12.549909 ],
  [               0,               0,               0,      -10.183623,       48.612465,      -57.996019 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.015708446 ],
  [               0,               0,               0,               0,               0,       1.5375177 ],
  [               0,               0,               0,               0,               0,      -1.0139982 ],
  [               0,               0,               0,               0,               0,     0.034438211 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.061799569,   0.00028111904,     -0.37675108 ],
  [               0,               0,               0,      -3.3938117,      -4.6358454,      -7.5013891 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,        63.30267,      -5.3286158,      -373.79517 ],
  [               0,               0,               0,       4.1191922,      -5.1362282,      -12.678135 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,   -0.0068794557,   0.00046869729,    0.0077978969 ],
  [               0,               0,               0,      -1.1681718,       1.5534751,       3.5866369 ],
  [               0,               0,               0,      -1.8856544,      0.74601786,      -1.4171779 ],
  [               0,               0,               0,       9.9257681,       -7.165002,       1.3482936 ],
  [               0,               0,               0,     0.069236841,    -0.051636523,     0.010437822 ],
  [               0,               0,               0,        5.089451,      -1.4290551,      0.19783649 ],
  [               0,               0,               0,       2.7299699,        2.655704,      -3.0790139 ],
 ],
 [
  [               0,               0,               0,     0.006451479,   -0.0095963391,    0.0088892706 ],
  [               0,               0,               0,      -13.014007,       12.373948,       1.4861898 ],
  [               0,               0,               0,      -2.0728552,      0.96266389,      -1.3261321 ],
  [               0,               0,               0,       1.9426647,       -1.098399,     0.089861786 ],
  [               0,               0,               0,    0.0021104803,   -0.0010084994,    0.0009462313 ],
  [               0,               0,               0,      -3.0391982,       4.4991382,     -0.70819756 ],
  [               0,               0,               0,      0.60115664,      -3.4510012,      0.15651723 ],
 ],
 [
  [               0,               0,               0,               0,     0.092928016,    -0.063005656 ],
  [               0,               0,               0,               0,       5.5742319,      -1.9891654 ],
  [               0,               0,               0,               0,       3.5962655,      -4.1116601 ],
  [               0,               0,               0,               0,      -2.6613887,       2.1370067 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0058647226,    0.0073272629,     0.005875745 ],
  [               0,               0,               0,     -0.85751388,       1.7948086,       2.4092423 ],
  [               0,               0,               0,        1.837067,      -4.4980959,       1.1426346 ],
  [               0,               0,               0,     -0.13887061,      0.75822648,     -0.39205068 ],
  [               0,               0,               0,    0.0064704197,    -0.014280269,    0.0084594548 ],
  [               0,               0,               0,      -4.1232758,       10.227221,      -5.4717075 ],
  [               0,               0,               0,        9.655028,      -19.138815,       6.9556442 ],
 ],
 [
  [               0,       2.3271719,       1.3513467,     -0.31028889,      -2.9241179,      -6.8273337 ],
  [               0,      -79.943037,       35.231427,       161.73018,       156.94244,      -327.31772 ],
  [               0,       107.65565,      -140.90843,      -135.27874,       79.887162,       127.30769 ],
  [               0,       13.717943,       -26.81046,      -10.130597,       36.411071,      -10.227771 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.032185119,    0.0014719109,      0.07143499 ],
  [               0,               0,               0,      -444.50319,       5.3615316,       908.36545 ],
  [               0,               0,               0,       49.457507,      -125.12143,       78.502816 ],
  [               0,               0,               0,     -0.16922031,     -0.58821472,     -0.89519817 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele25_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,     -0.49618137,      0.76049866,      0.91168059,       -1.403593 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0056273077,   -0.0074495721 ],
  [               0,               0,               0,               0,        12.32593,       21.405534 ],
  [               0,               0,               0,               0,      -50.596663,       72.275782 ],
  [               0,               0,               0,               0,      -72.943597,       109.00641 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.45613464,     -0.68381798 ],
  [               0,               0,               0,               0,       460.10108,      -691.19002 ],
  [               0,               0,               0,               0,       99.824987,      -152.83378 ],
  [               0,               0,               0,               0,      -61.164847,       92.452116 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.16175899,      0.54540386,      -0.4466292 ],
  [               0,               0,               0,       42.183537,      -141.86925,       121.12939 ],
  [               0,               0,               0,       -3.792641,       12.456907,      -11.354258 ],
  [               0,               0,               0,     -0.34234333,       1.6898318,      -1.8427889 ],
  [               0,               0,               0,    -0.016375903,     0.060534133,    -0.054798931 ],
  [               0,               0,               0,      -14.468108,       44.878597,      -33.610613 ],
  [               0,               0,               0,       11.967264,      -47.943176,       44.018769 ],
 ],
 [
  [               0,               0,               0,     0.024693056,    -0.036208879,    -0.019377362 ],
  [               0,               0,               0,      -4.7873429,       17.310247,      -13.341629 ],
  [               0,               0,               0,       8.3657335,      -36.610804,       39.084356 ],
  [               0,               0,               0,      -3.4535849,       15.427278,      -17.235386 ],
  [               0,               0,               0,     0.030522346,     -0.13613871,      0.15202767 ],
  [               0,               0,               0,       2.4481994,      -8.6782298,       8.6625028 ],
  [               0,               0,               0,       28.387215,      -107.17049,       80.924405 ],
 ],
 [
  [               0,               0,               0,    -0.025900125,    0.0016554596,      0.16716633 ],
  [               0,               0,               0,       37.147551,      -172.07108,       200.46521 ],
  [               0,               0,               0,      -3.4314016,       14.608448,      -16.269399 ],
  [               0,               0,               0,      -9.1968976,        43.88806,      -52.343273 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.018146391 ],
  [               0,               0,               0,               0,               0,       1.4263021 ],
  [               0,               0,               0,               0,               0,     -0.96748627 ],
  [               0,               0,               0,               0,               0,     0.026969659 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       7.5921159,      -18.755376 ],
  [               0,               0,               0,               0,      -20.169784,       51.184063 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,      -9.2894892,       22.086891 ],
  [               0,               0,               0,               0,     0.076594906,       -0.147383 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 50 GeV electron clusters (large statistics)
#


ele50large_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,       25396.731,      -154.06516,     -0.38741061,    0.0014576486 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00010084789,   0.00083361157 ],
  [               0,               0,               0,               0,     -0.16479326,       2.9740469 ],
  [               0,               0,               0,               0,       1.4753981,       -1.194324 ],
  [               0,               0,               0,               0,      -1.7970963,     -0.28682726 ],
  [               0,               0,               0,               0,     -0.71520142,      0.14254882 ],
  [               0,               0,               0,               0,  -0.00033603968,   0.00010269292 ],
  [               0,               0,               0,               0,      0.22329751,       1.5382042 ],
  [               0,               0,               0,               0,     -0.44662489,       2.9852162 ],
 ],
 [
  [               0,               0,               0,    0.0016427398,   -0.0027350333,    0.0015137127 ],
  [               0,               0,               0,       2.1626409,      -3.1679721,       4.4329367 ],
  [               0,               0,               0,      -77.545913,       19.002464,       5.0165202 ],
  [               0,               0,               0,       -79.21678,       18.065265,       5.4576417 ],
  [               0,               0,               0,      -14.949612,       11.079864,      -2.1547248 ],
  [               0,               0,               0,    -0.031370436,   -0.0041641319,    0.0071538636 ],
  [               0,               0,               0,      0.84817471,      -1.4057663,      0.81623067 ],
  [               0,               0,               0,     -0.21772956,      0.35929037,       1.4740715 ],
 ],
 [
  [               0,               0,               0,               0,     0.040336003,   -0.0084129672 ],
  [               0,               0,               0,               0,      -20.143575,       16.604469 ],
  [               0,               0,               0,               0,       -1.233627,      0.44356331 ],
  [               0,               0,               0,               0,       17.789005,      -14.142398 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.040232396,    0.0074147077,     0.081068999 ],
  [               0,               0,               0,    -0.034485571,     0.010716789,       0.3884062 ],
  [               0,               0,               0,     0.025797543,    0.0022252135,      -0.3784668 ],
  [               0,               0,               0,     0.046325228,    -0.040293703,     -0.35955974 ],
  [               0,               0,               0,   -0.0034993944,     0.010285676,   -0.0087219514 ],
  [               0,               0,               0,    0.0005071139,   -0.0012679541,   0.00080407833 ],
  [               0,               0,               0,       19.008176,      -34.058615,       15.136774 ],
  [               0,               0,               0,       22.392823,      -51.435174,       28.092103 ],
 ],
 [
  [               0,               0,               0,               0,               0,      0.23170193 ],
  [               0,               0,               0,               0,               0,      0.24831587 ],
  [               0,               0,               0,               0,               0,     -0.24341894 ],
  [               0,               0,               0,               0,               0,   -0.0015297561 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0058966192,    0.0026997651,   -0.0039457766,     -0.01576684 ],
  [               0,               0,       2.3809428,       5.0977292,       6.8938188,       13.307441 ],
  [               0,               0,      -10.512739,      -13.031985,      -15.896221,       -19.32144 ],
  [               0,               0,   0.00013898156,    0.0075954231,    -0.079748752,      -1.0549097 ],
  [               0,               0,       6.7960144,       9.5911654,       13.548008,       18.910743 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      -49.154274,       206.44282,      -288.99409,       134.85235 ],
  [               0,               0,      -488.85556,       934.43457,       306.97336,      -915.97489 ],
  [               0,               0,      -210.57359,       389.37976,       65.923695,      -275.54965 ],
  [               0,               0,       243.82023,      -554.63154,      -18.918734,       445.25253 ],
  [               0,               0,       2086.4351,      -4018.6369,       -1251.205,       3904.7476 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,       5170.0339,       -7343.602 ],
  [               0,               0,               0,               0,      -1.4909593,       2.1061991 ],
  [               0,               0,               0,               0,       1.4788222,      -2.0887224 ],
  [               0,               0,               0,               0,    -0.001744469,    0.0025467361 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       21.679876,       13.825983 ],
  [               0,               0,               0,               0,       8.5011477,      -12.460248 ],
  [               0,               0,               0,               0,      -8.5093405,       13.070117 ],
  [               0,               0,               0,               0,      -8.4645627,       13.004783 ],
  [               0,               0,               0,               0,       -47.81927,       69.935657 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.099069884,      0.15047862 ],
  [               0,               0,               0,               0,      -26.576466,       45.493685 ],
  [               0,               0,               0,               0,      -9.8367856,       13.762115 ],
  [               0,               0,               0,               0,       121.27949,      -180.20857 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00065277061 ],
  [               0,               0,               0,               0,               0,   -0.0062749279 ],
  [               0,               0,               0,               0,               0,     -0.96542411 ],
  [               0,               0,               0,               0,               0,       1.2354375 ],
  [               0,               0,               0,               0,               0,       6.3336789 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       1.4759118,      -2.7526174,      -4.0087804,       7.6384691 ],
  [               0,               0,      -36.078848,       181.24295,      -302.47582,       168.06208 ],
  [               0,               0,       32.781357,      -164.91231,       275.60765,      -153.36463 ],
  [               0,               0,       33.810974,      -170.00273,       283.98668,      -157.95659 ],
  [               0,               0,     0.064246922,     -0.34143521,      0.59908305,     -0.34766712 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0034649005,   -0.0055479017 ],
  [               0,               0,               0,               0,      -6.9912989,        16.28395 ],
  [               0,               0,               0,               0,       -2.511851,       4.1084142 ],
  [               0,               0,               0,               0,      -2.6163256,       4.3847356 ],
  [               0,               0,               0,               0,     0.094141484,     -0.25344673 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0099152288,     -0.02408234,    -0.048205681,      0.11864175 ],
  [               0,               0,       6.6328419,      -18.987218,      -1.5602615,       27.561581 ],
  [               0,               0,       7.4837065,      -34.319688,       45.578318,      -14.656941 ],
  [               0,               0,     -0.19315769,       10.255893,      -39.272266,       38.116709 ],
  [               0,               0,       1.1920672,      -10.820395,       29.842878,      -25.988246 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0023034966,    0.0081231339 ],
  [               0,               0,               0,               0,      -41.789537,       97.070756 ],
  [               0,               0,               0,               0,       25.390877,      -59.214931 ],
  [               0,               0,               0,               0,     -0.73301008,       1.7671045 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   3.2696367e-05,    0.0012930908 ],
  [               0,               0,               0,               0,      -5.4143016,       16.047491 ],
  [               0,               0,               0,               0,       1.0087907,      -3.1270506 ],
  [               0,               0,               0,               0,       0.6743735,      -2.2671445 ],
  [               0,               0,               0,               0,     -0.44738997,       1.0522123 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.068516037,      0.16452757 ],
  [               0,               0,               0,               0,      -231.99476,       567.59512 ],
  [               0,               0,               0,               0,      -2443.6883,       5853.4956 ],
  [               0,               0,               0,               0,      -2141.8122,       5122.7052 ],
  [               0,               0,               0,               0,      0.57338612,      -6.5804805 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0085695479,     0.015931998,     0.019492308 ],
  [               0,               0,               0,         1.00844,       -12.95065,       30.853208 ],
  [               0,               0,               0,      -85.276694,       404.30226,      -480.46656 ],
  [               0,               0,               0,      -55.960365,       281.96043,       -354.6079 ],
  [               0,               0,               0,     0.010037164,    -0.015104156,    -0.022046406 ],
  [               0,               0,               0,       1.7371254,      -5.2601523,       3.2924488 ],
  [               0,               0,               0,      -36.212464,       163.93932,       -193.1591 ],
 ],
 [
  [               0,               0,               0,               0,    0.0059566624,    -0.006418628 ],
  [               0,               0,               0,               0,       -3327.253,       8274.9733 ],
  [               0,               0,               0,               0,      -3.1338001,      -8.4016304 ],
  [               0,               0,               0,               0,       67.499396,      -167.76537 ],
  [               0,               0,               0,               0,      0.21345467,      0.34456763 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,      -9393.7914,       210.42167,      -1.3697046,    0.0016728731 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0019158769,    0.0011749818 ],
  [               0,               0,               0,               0,       3.8858496,        2.252384 ],
  [               0,               0,               0,               0,      -92.057594,         11.3993 ],
  [               0,               0,               0,               0,       -92.06016,       11.343399 ],
  [               0,               0,               0,               0,       9.6320091,      -2.6729557 ],
  [               0,               0,               0,               0,    -0.072829825,     0.046205724 ],
  [               0,               0,               0,               0,     -0.74676969,     0.098999177 ],
  [               0,               0,               0,               0,      0.25251826,       1.5029318 ],
 ],
 [
  [               0,               0,               0,    0.0016350226,   -0.0026963131,    0.0014981473 ],
  [               0,               0,               0,      0.98371712,      -1.8877861,       4.0866137 ],
  [               0,               0,               0,      -84.885513,       16.033844,       7.4020943 ],
  [               0,               0,               0,      -86.132973,       14.698161,       7.9305308 ],
  [               0,               0,               0,      -15.934811,       11.463334,      -2.1469773 ],
  [               0,               0,               0,    -0.039494539,   -0.0056672971,    0.0091013138 ],
  [               0,               0,               0,      0.78814278,      -1.2916379,      0.74539924 ],
  [               0,               0,               0,     -0.15054979,      0.25687308,       1.5093214 ],
 ],
 [
  [               0,               0,               0,               0,     0.042461733,    -0.010510769 ],
  [               0,               0,               0,               0,       -20.36312,       16.782987 ],
  [               0,               0,               0,               0,      -1.3963774,      0.56992202 ],
  [               0,               0,               0,               0,       18.092683,       -14.38382 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.015276948,    -0.011559522,      0.10011082 ],
  [               0,               0,               0,     0.074670845,    -0.034828555,      0.40233116 ],
  [               0,               0,               0,    -0.066663978,     0.041367125,     -0.39195366 ],
  [               0,               0,               0,    -0.024174333,    -0.051844035,     -0.34683383 ],
  [               0,               0,               0,   -0.0088634901,     0.020521199,    -0.015086972 ],
  [               0,               0,               0,   0.00065792285,   -0.0016478415,    0.0010416977 ],
  [               0,               0,               0,       18.963291,      -34.115471,       15.286769 ],
  [               0,               0,               0,       21.125579,      -48.337741,       26.145741 ],
 ],
 [
  [               0,               0,               0,               0,               0,        10.86059 ],
  [               0,               0,               0,               0,               0,     0.070460928 ],
  [               0,               0,               0,               0,               0,    -0.070334551 ],
  [               0,               0,               0,               0,               0,  -3.2980402e-05 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0023734013,    0.0011189265,   -0.0015573221,   -0.0064448215 ],
  [               0,               0,      -5.1159267,      -5.0818862,      -6.9304742,      -5.4767872 ],
  [               0,               0,       -15.54894,      -12.279105,      -73.689831,      -9.1230685 ],
  [               0,               0,    -0.095824082,       6.5331647,       333995.51,      -1.8001593 ],
  [               0,               0,       28.807487,       10.346132,       22.154346,       7.3149798 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.044545641,     -0.11357162,     -0.12746925,      0.29207016 ],
  [               0,               0,       402.25053,       -576.4016,      -771.16629,       1109.1789 ],
  [               0,               0,      -191.32816,       343.01317,       54.711117,       -220.9385 ],
  [               0,               0,       370.32709,      -477.57988,       -1010.964,       1336.2558 ],
  [               0,               0,       2004.1468,      -2853.8539,      -4070.0737,       5794.3517 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    0.0025504848,   -0.0030943663 ],
  [               0,               0,               0,               0,      -19.497523,      -10.561298 ],
  [               0,               0,               0,               0,      -404.18914,       590.14705 ],
  [               0,               0,               0,               0,      -281.19616,       402.11152 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0029810014,     0.013485212 ],
  [               0,               0,               0,               0,      -5.6801343,       9.9359162 ],
  [               0,               0,               0,               0,       27.529924,      -41.653542 ],
  [               0,               0,               0,               0,      -20.209155,       28.295516 ],
  [               0,               0,               0,               0,     0.074364119,     -0.14862572 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.064702597,     0.097606694 ],
  [               0,               0,               0,               0,       899.18493,      -1326.6886 ],
  [               0,               0,               0,               0,      -395.54477,       588.17907 ],
  [               0,               0,               0,               0,       538.53301,      -800.61193 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00025633102 ],
  [               0,               0,               0,               0,               0,      -39.526036 ],
  [               0,               0,               0,               0,               0,       14.272358 ],
  [               0,               0,               0,               0,               0,       2.2859528 ],
  [               0,               0,               0,               0,               0,       9.3689669 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       4.6338207,      -8.0570735,      -15.864656,       28.017649 ],
  [               0,               0,      -20.581001,        105.3621,      -179.06656,        101.2595 ],
  [               0,               0,       19.928139,       -102.0904,       173.62305,      -98.249489 ],
  [               0,               0,       20.394321,      -104.40055,       177.43436,      -100.34292 ],
  [               0,               0,     0.025766381,     -0.13411256,      0.23126686,     -0.13224428 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0034512318,   -0.0055152577 ],
  [               0,               0,               0,               0,      -6.9088221,       16.109006 ],
  [               0,               0,               0,               0,      -2.4520533,       3.9919833 ],
  [               0,               0,               0,               0,      -2.5838681,       4.3207215 ],
  [               0,               0,               0,               0,     0.088313889,     -0.24198964 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.009800377,    -0.023811993,    -0.047688102,      0.11742186 ],
  [               0,               0,       6.6433385,      -18.868385,      -2.2920906,       28.476042 ],
  [               0,               0,       7.4688147,      -34.486325,       46.493342,      -15.704937 ],
  [               0,               0,     -0.32153105,       10.710983,      -39.460854,       37.695984 ],
  [               0,               0,       1.2242503,      -11.027409,       30.287241,      -26.306597 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0025394194,    0.0073732294 ],
  [               0,               0,               0,               0,      -42.186927,       97.991645 ],
  [               0,               0,               0,               0,       25.704389,      -59.939984 ],
  [               0,               0,               0,               0,     -0.82227878,       1.9734903 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   3.1665998e-05,    0.0013004176 ],
  [               0,               0,               0,               0,       -5.354307,       15.904158 ],
  [               0,               0,               0,               0,      0.97802778,      -3.0584191 ],
  [               0,               0,               0,               0,      0.61846119,      -2.1410251 ],
  [               0,               0,               0,               0,     -0.45405745,       1.0678333 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -7.880411e-05,   0.00032917625 ],
  [               0,               0,               0,               0,       2.3138584,       34.409465 ],
  [               0,               0,               0,               0,       47.033512,      -114.60486 ],
  [               0,               0,               0,               0,      -51.919943,       116.64476 ],
  [               0,               0,               0,               0,      -1.7817721,      -6.6515676 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0011764089,      0.02641929,    -0.048939786 ],
  [               0,               0,               0,      -53.990746,       236.21238,      -250.08633 ],
  [               0,               0,               0,      -100.19288,       485.12873,      -588.67458 ],
  [               0,               0,               0,       -3.228317,       20.996965,      -31.935174 ],
  [               0,               0,               0,   -0.0093997521,   0.00037923626,     0.055221095 ],
  [               0,               0,               0,       5.0388136,      -10.345974,      -4.4290563 ],
  [               0,               0,               0,      -10.872073,       79.763039,      -137.03246 ],
 ],
 [
  [               0,               0,               0,               0,    0.0024769889,    0.0022078209 ],
  [               0,               0,               0,               0,      -3349.8317,       8330.8585 ],
  [               0,               0,               0,               0,      -2.7704891,      -8.5810965 ],
  [               0,               0,               0,               0,       57.768736,      -143.71352 ],
  [               0,               0,               0,               0,      0.20816654,      0.31354079 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,       265.92445,       109.06518,      -1.0759462,     0.001518294 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   4.8984048e-05,   0.00099942556 ],
  [               0,               0,               0,               0,     -0.90902636,       2.9817638 ],
  [               0,               0,               0,               0,       0.9663303,      -1.1508266 ],
  [               0,               0,               0,               0,     -0.69834796,     -0.67576154 ],
  [               0,               0,               0,               0,     -0.42592939,     0.071374046 ],
  [               0,               0,               0,               0,  -0.00023988217,   8.3069825e-05 ],
  [               0,               0,               0,               0,     -0.63414752,        1.686911 ],
  [               0,               0,               0,               0,     -0.24017433,       2.4144661 ],
 ],
 [
  [               0,               0,               0,    0.0071208526,   -0.0049498251,      0.00175649 ],
  [               0,               0,               0,     -0.80025819,      -2.6446506,       4.1177739 ],
  [               0,               0,               0,       2.2126501,      -3.1344988,      0.23584478 ],
  [               0,               0,               0,   -0.0044805457,     -0.70081445,     -0.59017419 ],
  [               0,               0,               0,      -0.7584969,      0.73907408,      -0.2796027 ],
  [               0,               0,               0,   -0.0039463033,    0.0041040765,   -0.0010673633 ],
  [               0,               0,               0,       6.2023839,      -7.5301915,       2.7999134 ],
  [               0,               0,               0,      -41.979555,       60.743443,      -19.725342 ],
 ],
 [
  [               0,               0,               0,               0,     0.041609563,    -0.012157766 ],
  [               0,               0,               0,               0,      -22.780037,       18.731755 ],
  [               0,               0,               0,               0,      -2.3203948,       1.2848219 ],
  [               0,               0,               0,               0,        20.23853,      -16.089921 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.027177118,    -0.009012646,     0.011260477 ],
  [               0,               0,               0,     0.060946058,     -0.40335267,      0.95612908 ],
  [               0,               0,               0,    -0.039974436,      0.28751836,     -0.79789258 ],
  [               0,               0,               0,     0.019273551,      0.17534952,     -0.75851178 ],
  [               0,               0,               0,   -0.0048303091,     0.021280019,    -0.024150563 ],
  [               0,               0,               0,   0.00059672283,   -0.0013634172,   0.00076532133 ],
  [               0,               0,               0,       15.755532,      -28.151741,       12.541072 ],
  [               0,               0,               0,       81.472935,      -184.20033,       99.226601 ],
 ],
 [
  [               0,               0,               0,               0,               0,       4.0440168 ],
  [               0,               0,               0,               0,               0,     0.093617592 ],
  [               0,               0,               0,               0,               0,    -0.093334044 ],
  [               0,               0,               0,               0,               0,  -9.6930021e-05 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0034753324,    0.0015565274,   -0.0023873973,   -0.0091285435 ],
  [               0,               0,      -5.1569792,      -5.0496294,      -6.8675218,       -5.362491 ],
  [               0,               0,      -8.8503746,      -24.236239,      -19.122101,      -25.171494 ],
  [               0,               0,  -2.7856732e+12,  -3.7514027e+12,   3.2631635e+12,    1.399358e+12 ],
  [               0,               0,       8.0178904,       11.779797,       24.634001,       16.261451 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.040022001,     -0.11395703,      -0.1231085,      0.30008635 ],
  [               0,               0,       361.17262,      -517.47744,      -688.50645,       990.57068 ],
  [               0,               0,      -286.60339,       480.98628,        251.3397,      -505.63188 ],
  [               0,               0,       347.61156,      -444.68792,      -968.17639,       1274.2459 ],
  [               0,               0,       1990.3622,      -2834.5714,      -4042.5415,       5755.8761 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,    0.0056459183,   -0.0075283064 ],
  [               0,               0,               0,               0,      -19.471723,      -9.7076656 ],
  [               0,               0,               0,               0,      -347.24371,        505.9153 ],
  [               0,               0,               0,               0,      -206.47094,       294.80932 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.24608216,      0.36485814 ],
  [               0,               0,               0,               0,       173.97986,      -253.96738 ],
  [               0,               0,               0,               0,       -273.4155,       400.89535 ],
  [               0,               0,               0,               0,      -44.325472,         65.7249 ],
  [               0,               0,               0,               0,      0.17856558,     -0.45815762 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.17077263,      0.25828389 ],
  [               0,               0,               0,               0,       155.21782,      -225.75469 ],
  [               0,               0,               0,               0,      0.26267975,      -1.5631996 ],
  [               0,               0,               0,               0,       48.630654,      -72.166732 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0002447008 ],
  [               0,               0,               0,               0,               0,       10.868872 ],
  [               0,               0,               0,               0,               0,       7.3320973 ],
  [               0,               0,               0,               0,               0,     -0.59723699 ],
  [               0,               0,               0,               0,               0,       11.258229 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       3.5272858,      -6.4609101,      -10.689372,       19.872331 ],
  [               0,               0,       -36.45547,        184.3951,       -309.9765,       173.43429 ],
  [               0,               0,       35.205769,      -178.15876,       299.63781,        -167.736 ],
  [               0,               0,       36.045811,      -182.31687,       306.48959,      -171.49468 ],
  [               0,               0,     0.053003367,     -0.27243446,       0.4645088,     -0.26289852 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0032142733,   -0.0051600798 ],
  [               0,               0,               0,               0,      -7.5687945,       17.544615 ],
  [               0,               0,               0,               0,      -2.9087468,       4.9377059 ],
  [               0,               0,               0,               0,        -3.16494,       5.4791367 ],
  [               0,               0,               0,               0,      0.18003683,     -0.44733354 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0056364145,    -0.012622153,    -0.034116506,      0.07757097 ],
  [               0,               0,       13.344638,       -56.76889,       67.941416,       -13.74718 ],
  [               0,               0,       30.917532,      -186.63739,       375.59129,      -252.93981 ],
  [               0,               0,      -17.023554,       115.34109,        -257.166,       188.17058 ],
  [               0,               0,       13.200524,      -87.816589,        194.2598,      -142.93214 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0013394148,    0.0045090119 ],
  [               0,               0,               0,               0,       -63.26358,       146.77062 ],
  [               0,               0,               0,               0,       35.776229,      -83.271563 ],
  [               0,               0,               0,               0,      -3.2111978,       7.5256856 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    2.686832e-05,    0.0012229488 ],
  [               0,               0,               0,               0,      -4.7106912,       14.505673 ],
  [               0,               0,               0,               0,      0.63180733,       -2.181257 ],
  [               0,               0,               0,               0,      0.63361321,      -2.1450555 ],
  [               0,               0,               0,               0,     -0.42351521,       0.9802105 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -7.5136881e-06,   0.00019971776 ],
  [               0,               0,               0,               0,       6.6504651,       23.452105 ],
  [               0,               0,               0,               0,       110.37566,      -266.04892 ],
  [               0,               0,               0,               0,      -167.79872,       394.11475 ],
  [               0,               0,               0,               0,      -1.7834588,      -4.5478465 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.014376671,      0.06922336,    -0.075217763 ],
  [               0,               0,               0,      -9.8741625,       32.226073,       -14.97481 ],
  [               0,               0,               0,       133.15817,      -645.64491,       780.83492 ],
  [               0,               0,               0,      -135.96795,        666.4879,      -816.43319 ],
  [               0,               0,               0,     0.010909198,     -0.05023155,     0.059100115 ],
  [               0,               0,               0,       26.431616,      -140.92598,       187.20296 ],
  [               0,               0,               0,      -223.67578,       1091.5833,      -1366.1839 ],
 ],
 [
  [               0,               0,               0,               0,     -0.05124219,      0.13561014 ],
  [               0,               0,               0,               0,      -3435.0439,       8541.7551 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,        36.41668,      -90.966561 ],
  [               0,               0,               0,               0,      0.19102211,      0.24999398 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.022161758,      0.01039781,     0.007297853 ],
  [               0,               0,               0,       2.3656522,      -1.3870657,       3.4527144 ],
  [               0,               0,               0,      0.24120151,     -0.22480812,      -1.2421672 ],
  [               0,               0,               0,      0.72264296,      -0.1852416,    -0.011605027 ],
  [               0,               0,               0,   -0.0032283691,    0.0012097809,   0.00025739471 ],
  [               0,               0,               0,       24.659379,      -10.788793,       1.3251102 ],
  [               0,               0,               0,      -22.620159,         9.09653,      -1.6377466 ],
 ],
 [
  [               0,               0,               0,   -0.0089515877,     0.024772055,   -0.0043479327 ],
  [               0,               0,               0,       2.0868728,      -7.2530701,       6.9840848 ],
  [               0,               0,               0,        0.448501,     -0.16417184,      -1.4887689 ],
  [               0,               0,               0,      0.65210523,     -0.59791686,      0.24049667 ],
  [               0,               0,               0,    0.0099659139,     -0.01456518,     0.005630608 ],
  [               0,               0,               0,      -3.0270466,       2.1506991,      0.60642592 ],
  [               0,               0,               0,      -6.2140742,       3.6366738,      -2.0960277 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0066198721,      0.01402848 ],
  [               0,               0,               0,               0,       14.091191,      -8.4974649 ],
  [               0,               0,               0,               0,    -0.068882459,      -1.2146335 ],
  [               0,               0,               0,               0,      -2.6738396,       2.1566764 ],
  [               0,               0,               0,               0,     0.019459079,    -0.015314043 ],
  [               0,               0,               0,               0,      -13.336052,       12.310488 ],
  [               0,               0,               0,               0,       -151.8969,       90.370609 ],
 ],
 [
  [               0,               0,               0,   0.00054018944,    -0.002296643,     0.010723448 ],
  [               0,               0,               0,      0.94715512,      -2.7423798,       4.4730177 ],
  [               0,               0,               0,       0.4254486,      -1.2032749,     -0.60309224 ],
  [               0,               0,               0,       100.21766,      -5.9600907,      -26.008622 ],
  [               0,               0,               0,      0.70344057,      0.24983651,     -0.34250203 ],
  [               0,               0,               0,    -0.021710157,     0.041669106,   -0.0066308071 ],
  [               0,               0,               0,     0.045067443,     -0.10714849,      -3.0742924 ],
 ],
 [
  [               0,               0,               0,               0,      0.17271288,     -0.22510545 ],
  [               0,               0,               0,               0,      -41.136579,       58.530582 ],
  [               0,               0,               0,               0,       20.761704,      -29.759311 ],
  [               0,               0,               0,               0,      -14.122859,       19.582179 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.40981391,     -0.75054301,      0.24749162 ],
  [               0,               0,               0,      -174.53219,      -2.5754608,       385.68644 ],
  [               0,               0,               0,      -11883.378,       33598.475,      -23755.174 ],
  [               0,               0,               0,      -18.184994,       29.650276,      -4.0305625 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      0.67899815,     -0.93722361,      -1.5585184,       2.1611892 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0001589221,   0.00014097577,   4.4102684e-05,   0.00028334688 ],
  [               0,               0,      -2.7939871,     -0.29069771,       1.5640127,       11.252627 ],
  [               0,               0,      -1.9217704,      -1.4298109,      0.80919761,       6.3841743 ],
  [               0,               0,       9.2531908,       4.3823254,      -6.9983328,      -30.147056 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.054745236,    -0.081807886 ],
  [               0,               0,               0,               0,       38.893258,      -56.023998 ],
  [               0,               0,               0,               0,      -338.36634,       509.71689 ],
  [               0,               0,               0,               0,       369.15314,      -558.33623 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.11850151,      0.41030081,     -0.34407111 ],
  [               0,               0,               0,        24.98257,      -85.199693,       74.720122 ],
  [               0,               0,               0,      -3.6324126,       12.297649,      -11.498227 ],
  [               0,               0,               0,      0.26101021,     -0.76953372,      0.53106855 ],
  [               0,               0,               0,   -0.0056992421,     0.020083858,    -0.017777795 ],
  [               0,               0,               0,      -9.4847581,       40.722343,      -41.138185 ],
  [               0,               0,               0,       15.478074,      -54.886012,       46.620797 ],
 ],
 [
  [               0,               0,               0,    0.0093927373,    -0.039710908,     0.042215343 ],
  [               0,               0,               0,      -392.55252,       1620.4792,      -1654.1045 ],
  [               0,               0,               0,       68.157765,      -270.24005,        264.9423 ],
  [               0,               0,               0,      -46.269757,        186.3076,      -186.27265 ],
  [               0,               0,               0,   -0.0001069906,  -0.00052991305,    0.0040277008 ],
  [               0,               0,               0,      0.15681035,     -0.75650299,       1.9022139 ],
  [               0,               0,               0,       2.3303167,      -8.8153552,       6.6401064 ],
 ],
 [
  [               0,               0,               0,    -0.013063717,  -0.00097268083,     0.087856706 ],
  [               0,               0,               0,       60.065018,        -280.274,       328.65191 ],
  [               0,               0,               0,        -1.11533,       3.9206148,      -4.0935785 ],
  [               0,               0,               0,      -9.2864427,       53.730699,      -74.376233 ],
  [               0,               0,               0,     0.019614012,   0.00018829132,     -0.10083153 ],
  [               0,               0,               0,      -2.6157306,       12.170562,      -14.116624 ],
  [               0,               0,               0,      -9.4949008,       54.954473,      -79.837893 ],
 ],
 [
  [               0,               0,               0,               0,     -0.13864758,      0.34576389 ],
  [               0,               0,               0,               0,      -11.330244,       30.550519 ],
  [               0,               0,               0,               0,      -90.367633,       221.11152 ],
  [               0,               0,               0,               0,       115.50328,      -283.80772 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00097228493,   -0.0019933803,    0.0027220036,   -0.0092501108 ],
  [               0,               0,     -0.69182736,      -1.3789288,      -3.4154663,      -4.4508373 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       7.2993303,       5.5708768,      -5.9925831,      -30.092103 ],
  [               0,               0,        1.505067,       1.6611374,       20.036978,       28.161527 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.037502356,     0.018078565,    0.0053850376 ],
  [               0,               0,               0,       11.296555,      -5.7995699,       4.8656419 ],
  [               0,               0,               0,      -1.9992007,      0.76911193,      -1.4011216 ],
  [               0,               0,               0,       1.1197803,     -0.37779833,     0.023717102 ],
  [               0,               0,               0,    0.0061109007,   -0.0030625001,   0.00094283781 ],
  [               0,               0,               0,       14.840856,      -6.5846821,       1.1111259 ],
  [               0,               0,               0,      -12.800263,       5.1240212,      -1.8893541 ],
 ],
 [
  [               0,               0,               0,   -0.0085208407,    0.0089288061,    0.0047625473 ],
  [               0,               0,               0,      -4.5651923,       3.9688452,       3.4343248 ],
  [               0,               0,               0,      -2.4279793,       2.4347697,      -2.0154359 ],
  [               0,               0,               0,        2.657962,      -2.8688505,      0.85206729 ],
  [               0,               0,               0,    0.0098225103,    -0.012046364,    0.0043509153 ],
  [               0,               0,               0,       1.4169417,     -0.24270503,      0.22495586 ],
  [               0,               0,               0,      -5.5196155,       6.6667142,      -10.259309 ],
 ],
 [
  [               0,               0,               0,               0,      0.02501148,    -0.011806351 ],
  [               0,               0,               0,               0,       13.448015,      -7.2546921 ],
  [               0,               0,               0,               0,       6.6614918,       -6.784916 ],
  [               0,               0,               0,               0,       0.1831013,       0.1405662 ],
  [               0,               0,               0,               0,     0.042231519,    -0.031563347 ],
  [               0,               0,               0,               0,       7.9374734,      -6.6916731 ],
  [               0,               0,               0,               0,      -4.7899453,      -18.408203 ],
 ],
 [
  [               0,               0,               0,   -0.0023485334,    0.0059423695,    0.0044578277 ],
  [               0,               0,               0,       1.5227634,      -4.6202939,       6.5308964 ],
  [               0,               0,               0,      0.24255168,     -0.82300436,     -0.85188115 ],
  [               0,               0,               0,      -1.6899074,      0.15865884,       4.0600361 ],
  [               0,               0,               0,    -0.016461558,    0.0060757857,     0.030051963 ],
  [               0,               0,               0,      0.15069208,     -0.11407336,     0.030840338 ],
  [               0,               0,               0,       1.1536519,      -2.6535493,      -1.5179073 ],
 ],
 [
  [               0,               0,      0.87977023,      -1.0801058,      -1.5864588,       1.9454281 ],
  [               0,               0,      -6100.6917,       8869.3353,       10031.081,      -14695.804 ],
  [               0,               0,      -2360.7756,       3271.2705,       4409.3096,      -6112.2635 ],
  [               0,               0,       602.18095,      -822.90476,      -1132.0306,       1547.3944 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0088768701,    0.0015953978,     0.028634735 ],
  [               0,               0,               0,      -392.12369,       4.4985675,       807.39842 ],
  [               0,               0,               0,       88.257114,      -250.54892,       176.91394 ],
  [               0,               0,               0,      0.12542807,     -0.43241753,     -0.47747022 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,     -0.63773815,      0.93521891,        1.166245,      -1.7130381 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.019813476,    -0.001412673,     -0.02231342,    -0.024656491 ],
  [               0,               0,       96.225294,      -117.70105,       -179.2209,       213.77169 ],
  [               0,               0,      -203.15508,       306.68312,       447.13171,      -676.31496 ],
  [               0,               0,      -126.79945,       179.97651,       270.89064,      -383.84573 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -1.2123133,       1.8554328 ],
  [               0,               0,               0,               0,       40.512425,      -59.560612 ],
  [               0,               0,               0,               0,       18.290435,      -28.659701 ],
  [               0,               0,               0,               0,      -23.490101,       35.537686 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.033140927,       0.1125457,     -0.08442665 ],
  [               0,               0,               0,       11.386498,      -38.270952,       34.616557 ],
  [               0,               0,               0,     -0.99883721,       3.2740837,      -3.8368243 ],
  [               0,               0,               0,     -0.14532718,      0.56856507,     -0.55280636 ],
  [               0,               0,               0,     0.001061346,    -0.002835299,    0.0017222091 ],
  [               0,               0,               0,       41.318009,      -139.90528,       119.76889 ],
  [               0,               0,               0,      -42.842507,      -63.545043,      -21.702114 ],
 ],
 [
  [               0,               0,               0,     0.040802727,     -0.16681625,       0.1794931 ],
  [               0,               0,               0,      -6.7221674,       26.097714,      -22.486075 ],
  [               0,               0,               0,       1.9586746,        -7.86881,        6.673106 ],
  [               0,               0,               0,      -27.051509,       112.19196,      -116.02571 ],
  [               0,               0,               0,     0.063254576,     -0.25915411,      0.27586671 ],
  [               0,               0,               0,    -0.088386494,       0.3881641,     -0.40669767 ],
  [               0,               0,               0,       22.870429,       -94.86909,       96.571342 ],
 ],
 [
  [               0,               0,               0,    -0.010896733,     0.017871636,     0.028854047 ],
  [               0,               0,               0,       63.250267,      -296.39034,        349.5707 ],
  [               0,               0,               0,       8.9957076,      -43.489609,       51.365098 ],
  [               0,               0,               0,      -3.1463024,       26.249913,      -45.585254 ],
  [               0,               0,               0,    -0.026830463,   -0.0010790441,      0.16556251 ],
  [               0,               0,               0,      -7.9592062,       37.186071,      -43.461161 ],
  [               0,               0,               0,      -41.338488,       202.90493,      -248.16707 ],
 ],
 [
  [               0,               0,               0,               0,    -0.060438311,      0.15232039 ],
  [               0,               0,               0,               0,      -2.5842425,       9.7800015 ],
  [               0,               0,               0,               0,      -151.12313,       370.95498 ],
  [               0,               0,               0,               0,        167.3502,      -411.93626 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00068079644,    0.0011109269,   -0.0034130686,   -0.0087194931 ],
  [               0,               0,      -1.2073208,      -1.3689254,      -3.3949295,      -4.4198087 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      -191.72022,      -158.04332,       407.71925,       2980.7616 ],
  [               0,               0,      0.43413309,       1.4836362,       11.443414,       35.785376 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.036025696,     0.017396289,    0.0054555558 ],
  [               0,               0,               0,       10.732348,      -5.4934571,       4.8223722 ],
  [               0,               0,               0,      -1.9391196,      0.74341142,      -1.3989401 ],
  [               0,               0,               0,       1.1244756,     -0.37981268,     0.024422279 ],
  [               0,               0,               0,    0.0049244142,   -0.0025228581,   0.00092671861 ],
  [               0,               0,               0,       14.626283,       -6.550415,       1.1023331 ],
  [               0,               0,               0,      -12.251492,       4.8928576,      -1.8797099 ],
 ],
 [
  [               0,               0,               0,    -0.010279341,     0.011009745,    0.0041814991 ],
  [               0,               0,               0,      -4.1004891,       3.3902954,       3.5968557 ],
  [               0,               0,               0,      -2.3633507,       2.3694177,      -1.9999273 ],
  [               0,               0,               0,       2.6431014,      -2.8704723,      0.86212802 ],
  [               0,               0,               0,     0.010354229,    -0.012854991,    0.0046775821 ],
  [               0,               0,               0,       1.8519001,      -0.7384411,      0.34068682 ],
  [               0,               0,               0,      -4.4653098,        5.510238,      -9.9545863 ],
 ],
 [
  [               0,               0,               0,               0,     0.046494751,     -0.02894457 ],
  [               0,               0,               0,               0,       9.1256496,      -3.8208257 ],
  [               0,               0,               0,               0,       5.7705397,      -6.0704353 ],
  [               0,               0,               0,               0,      -7.4779511,       6.7189959 ],
  [               0,               0,               0,               0,    0.0027916997,    0.0036650347 ],
  [               0,               0,               0,               0,      -1.1806921,       1.1411935 ],
  [               0,               0,               0,               0,        1.575409,      -23.097631 ],
 ],
 [
  [               0,               0,               0,  -0.00055752084,    0.0023100953,    0.0063251744 ],
  [               0,               0,               0,       1.1934626,      -3.9715103,       6.1903348 ],
  [               0,               0,               0,     0.033384877,     -0.31869035,       -1.152258 ],
  [               0,               0,               0,      -86.790086,       17.895046,       212.12958 ],
  [               0,               0,               0,     -0.96408075,      0.83804084,         1.28234 ],
  [               0,               0,               0,      0.03236749,    -0.051526082,     0.028477087 ],
  [               0,               0,               0,     0.031191553,    -0.079538834,      -3.0770884 ],
 ],
 [
  [               0,               0,     -0.16318008,      0.37664409,      0.41168522,     -0.84038815 ],
  [               0,               0,      -2021.7021,       2963.2408,       3111.9219,      -4624.2188 ],
  [               0,               0,       -101.3651,       151.66299,       197.75883,      -296.53346 ],
  [               0,               0,       134.27986,      -179.89466,      -250.43278,       335.81692 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0077057357,    0.0024759457,   -0.0075098924 ],
  [               0,               0,               0,      -407.74134,       4.1122328,        838.7922 ],
  [               0,               0,               0,       26.140478,      -84.899516,       66.755458 ],
  [               0,               0,               0,    -0.029284003,     -0.25428302,     -0.48991296 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele50large_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,     -0.65264932,      0.95756941,       1.2009068,      -1.7649924 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.013508929,    0.0069351405,    -0.009154969,    -0.042298354 ],
  [               0,               0,      -374.10953,       575.49364,        851.9062,      -1305.6839 ],
  [               0,               0,      -108.85673,       164.90049,       240.39859,      -365.49367 ],
  [               0,               0,      -87.585022,        122.8413,       185.75954,      -259.86182 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.56472472,       0.8656957 ],
  [               0,               0,               0,               0,       66.496553,      -98.984798 ],
  [               0,               0,               0,               0,      -30.134779,       44.661014 ],
  [               0,               0,               0,               0,       17.870685,      -27.015152 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.099618062,     -0.34085938,      0.29884364 ],
  [               0,               0,               0,      -19.422214,       66.949482,      -54.441059 ],
  [               0,               0,               0,       3.8576534,      -13.407244,       10.361316 ],
  [               0,               0,               0,      -2.1539025,       7.4501535,      -6.3777102 ],
  [               0,               0,               0,   -0.0058061172,     0.020384959,    -0.017053452 ],
  [               0,               0,               0,      -11.135156,       37.904945,      -31.793625 ],
  [               0,               0,               0,       15.110676,      -50.942847,       40.659547 ],
 ],
 [
  [               0,               0,               0,     0.043757772,     -0.17955046,      0.19317956 ],
  [               0,               0,               0,      -7.3981883,       29.046528,      -25.688302 ],
  [               0,               0,               0,       2.0581161,      -8.3011805,       7.1432157 ],
  [               0,               0,               0,      -35.364145,       146.91894,       -151.9387 ],
  [               0,               0,               0,     0.055400256,     -0.22644199,      0.24543868 ],
  [               0,               0,               0,     -0.11713137,      0.50567274,     -0.52946606 ],
  [               0,               0,               0,       22.964446,      -95.414428,        97.12153 ],
 ],
 [
  [               0,               0,               0,   -0.0077078914,   -0.0044606556,     0.063845687 ],
  [               0,               0,               0,       76.439807,      -357.36462,       419.98917 ],
  [               0,               0,               0,       9.1274935,      -44.320974,        52.59456 ],
  [               0,               0,               0,      -1.8021129,        20.95273,      -40.669207 ],
  [               0,               0,               0,    -0.030209477,  -0.00066046596,      0.18519737 ],
  [               0,               0,               0,       -5.004022,       23.191425,      -26.888839 ],
  [               0,               0,               0,      -27.236264,       136.28046,      -169.40182 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0038492683,     0.017505066 ],
  [               0,               0,               0,               0,      -32.010063,       81.488667 ],
  [               0,               0,               0,               0,      -9.7176534,       22.710971 ],
  [               0,               0,               0,               0,       22.300075,      -54.836292 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0016555801,  -0.00070630165,  -0.00052974949,    -0.019423102 ],
  [               0,               0,      -1.2018113,      -1.3737702,      -3.4058099,      -4.4437598 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       108.64181,       86.420827,      -214.88242,      -1614.7387 ],
  [               0,               0,       0.3139116,       3.6779848,      0.94734387,       2.9720122 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 75 GeV electron clusters
#


ele75_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00028408254,   0.00037376028 ],
  [               0,               0,               0,               0,     -0.85228879,       4.4810557 ],
  [               0,               0,               0,               0,       1.8644245,       0.3438361 ],
  [               0,               0,               0,               0,        2.387998,      0.18112377 ],
  [               0,               0,               0,               0,     -0.39926413,    -0.041475814 ],
  [               0,               0,               0,               0,  -0.00026316699,  -1.5571036e-05 ],
  [               0,               0,               0,               0,      -1.7634862,       2.2300876 ],
  [               0,               0,               0,               0,     -0.96635089,       1.6500375 ],
 ],
 [
  [               0,               0,               0,     0.023947773,    -0.029771116,     0.010296775 ],
  [               0,               0,               0,      -16.878414,       15.132245,      0.11336151 ],
  [               0,               0,               0,      -4.8671166,       8.6646397,      -4.4693215 ],
  [               0,               0,               0,      -2.7166989,       5.0216533,      -3.0219513 ],
  [               0,               0,               0,        1.150339,      -1.7400516,      0.54697697 ],
  [               0,               0,               0,   -0.0037877425,    0.0052902364,   -0.0018644082 ],
  [               0,               0,               0,       8.8780645,      -9.5791498,       4.3610021 ],
  [               0,               0,               0,       27.750062,      -24.604904,       4.1249211 ],
 ],
 [
  [               0,               0,               0,               0,     0.023311807,    0.0045478441 ],
  [               0,               0,               0,               0,       3.2463855,      -2.0037077 ],
  [               0,               0,               0,               0,       -19.77181,       15.204298 ],
  [               0,               0,               0,               0,       17.384316,      -13.824017 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0024201445,    0.0016144695,     0.012293161 ],
  [               0,               0,               0,       4.0618797,      -8.4281587,       5.0956832 ],
  [               0,               0,               0,      -2.9545684,       6.1642165,      -3.8224642 ],
  [               0,               0,               0,      -3.7568606,        7.843231,      -4.6807443 ],
  [               0,               0,               0,     -0.13234262,      0.27613418,     -0.15581742 ],
  [               0,               0,               0,   -0.0020405852,    0.0042924092,   -0.0021485531 ],
  [               0,               0,               0,       8.9140206,      -19.860626,       12.851273 ],
  [               0,               0,               0,       10.760115,      -21.508588,       14.348888 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.037204531,     0.029698396,    -0.014834827,     -0.13653062 ],
  [               0,               0,       35.634736,      -69.509603,       -82.95362,       160.42741 ],
  [               0,               0,       903.47439,      -1302.6129,      -1870.3815,       2695.1076 ],
  [               0,               0,       -469.1965,       663.05648,       921.61968,      -1303.7608 ],
  [               0,               0,       87.889576,      -119.73044,      -174.43465,       236.90712 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,       243706.23 ],
  [               0,               0,               0,               0,               0,   -0.0036886292 ],
  [               0,               0,               0,               0,               0,    0.0036885991 ],
  [               0,               0,               0,               0,               0,   1.6278971e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00089962187 ],
  [               0,               0,               0,               0,               0,       30.067121 ],
  [               0,               0,               0,               0,               0,       2.4270194 ],
  [               0,               0,               0,               0,               0,       4.1065615 ],
  [               0,               0,               0,               0,               0,     -0.84888737 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.2233484,      0.33849144 ],
  [               0,               0,               0,               0,      -72.933252,       112.42495 ],
  [               0,               0,               0,               0,       34.797872,      -52.913105 ],
  [               0,               0,               0,               0,       18.774449,      -27.745815 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.036593225,    -0.065182743,     -0.10475702,      0.19019811 ],
  [               0,               0,      -92.232388,       456.91082,      -751.00905,       410.96265 ],
  [               0,               0,      -41.457904,        200.4686,      -323.18178,       172.78637 ],
  [               0,               0,       3.9733983,      -21.060196,       36.174672,      -21.151425 ],
  [               0,               0,      -3.1136034,       15.051997,      -24.242784,       13.006932 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0045944233,   -0.0069444627 ],
  [               0,               0,               0,               0,      -5.1846294,       12.148209 ],
  [               0,               0,               0,               0,      0.40606828,      -1.7342705 ],
  [               0,               0,               0,               0,      0.23950419,      -1.3903786 ],
  [               0,               0,               0,               0,     0.023898558,    -0.089875701 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.015328343,     -0.03681251,    -0.073586402,      0.17865001 ],
  [               0,               0,       6.6725261,      -20.082798,       3.0356475,        22.45135 ],
  [               0,               0,       13.057853,      -65.782747,       103.62455,      -49.553884 ],
  [               0,               0,      -7.7364828,       54.992573,      -127.74625,       96.378097 ],
  [               0,               0,       4.0855992,      -29.596157,       70.353926,      -55.052926 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00022610445,    0.0030927426 ],
  [               0,               0,               0,               0,      -147.52938,       341.09292 ],
  [               0,               0,               0,               0,       70.557899,      -163.65788 ],
  [               0,               0,               0,               0,      -2.8642521,       6.9427074 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00082755153,    0.0025061398 ],
  [               0,               0,               0,               0,     -0.77157193,       3.4631823 ],
  [               0,               0,               0,               0,     -0.41535693,     0.036760264 ],
  [               0,               0,               0,               0,       1.5362287,      -4.5650681 ],
  [               0,               0,               0,               0,      0.67832489,      -1.5919273 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.002182912,   0.00067582465,     0.019015049 ],
  [               0,               0,               0,      0.72403187,      -10.374464,       26.449674 ],
  [               0,               0,               0,      -1.4202425,       4.3170427,      -3.8185628 ],
  [               0,               0,               0,       17.363953,       20.784244,      -103.59196 ],
  [               0,               0,               0,     0.047549557,     0.051997858,    -0.023803721 ],
  [               0,               0,               0,       4.3682032,      -21.404936,       26.254419 ],
  [               0,               0,               0,       1.9724025,      -9.5222026,        8.366409 ],
 ],
 [
  [               0,               0,               0,      0.20719124,     0.012847925,       -1.215596 ],
  [               0,               0,               0,       45.775931,      -1.5838117,      -276.12569 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,      -26.109295,      -1.0419652,        162.2913 ],
  [               0,               0,               0,     0.030148413,     0.075331701,      -0.2041746 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00043732557,     0.001113792 ],
  [               0,               0,               0,               0,     -0.88885131,       3.2521578 ],
  [               0,               0,               0,               0,      0.12124584,      -0.5944119 ],
  [               0,               0,               0,               0,     -0.79216689,     -0.45378632 ],
  [               0,               0,               0,               0,     -0.32641128,     0.036555833 ],
  [               0,               0,               0,               0,  -0.00061807863,   0.00038222684 ],
  [               0,               0,               0,               0,      -1.2572592,      0.96203227 ],
  [               0,               0,               0,               0,     -0.24179732,       1.6853278 ],
 ],
 [
  [               0,               0,               0,     0.023688357,    -0.029476848,     0.010211085 ],
  [               0,               0,               0,      -17.475545,       15.824023,    -0.070887263 ],
  [               0,               0,               0,      -4.6979458,        8.477872,      -4.4195868 ],
  [               0,               0,               0,      -2.2611138,        4.522024,      -2.8904039 ],
  [               0,               0,               0,       1.1866755,      -1.7820748,      0.55824903 ],
  [               0,               0,               0,   -0.0037548793,    0.0052467426,   -0.0018505279 ],
  [               0,               0,               0,       8.9800422,      -9.6004316,       4.3520775 ],
  [               0,               0,               0,        27.31673,      -24.262692,       4.0586658 ],
 ],
 [
  [               0,               0,               0,               0,     0.023311807,    0.0045478441 ],
  [               0,               0,               0,               0,       3.2463855,      -2.0037077 ],
  [               0,               0,               0,               0,       -19.77181,       15.204298 ],
  [               0,               0,               0,               0,       17.384316,      -13.824017 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0030628134,    0.0020217046,     0.014969841 ],
  [               0,               0,               0,        3.716803,      -7.7163118,       4.6748242 ],
  [               0,               0,               0,      -2.8088658,        5.860062,      -3.6313825 ],
  [               0,               0,               0,      -3.4765589,         7.25779,      -4.3457079 ],
  [               0,               0,               0,     -0.11055805,       0.2305536,      -0.1298384 ],
  [               0,               0,               0,   -0.0020217143,    0.0042561638,   -0.0021295425 ],
  [               0,               0,               0,       8.9944594,      -20.048528,       12.955619 ],
  [               0,               0,               0,       10.925286,      -21.876749,       14.566184 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.043483348,    -0.023678146,     0.029275355,      0.14355047 ],
  [               0,               0,        513.0838,      -733.60579,      -994.32377,       1425.1678 ],
  [               0,               0,      -306.46922,       512.28298,       298.92494,      -579.60811 ],
  [               0,               0,       342.93111,      -432.32322,      -965.69872,       1259.2891 ],
  [               0,               0,       1832.0468,      -2607.6562,      -3727.8015,       5304.8229 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,     0.002608342 ],
  [               0,               0,               0,               0,               0,      -39.984192 ],
  [               0,               0,               0,               0,               0,       1.0317197 ],
  [               0,               0,               0,               0,               0,      0.01407727 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0017943114 ],
  [               0,               0,               0,               0,               0,       10.180222 ],
  [               0,               0,               0,               0,               0,      -0.4080014 ],
  [               0,               0,               0,               0,               0,     -0.29788534 ],
  [               0,               0,               0,               0,               0,     -0.40314913 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.48739216,      0.73517877 ],
  [               0,               0,               0,               0,       44.027499,      -61.992731 ],
  [               0,               0,               0,               0,       54.797386,      -82.779367 ],
  [               0,               0,               0,               0,      -34.703024,       51.733128 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.01976553,    -0.038649577,    -0.059287078,      0.11842049 ],
  [               0,               0,      -45.882814,       247.75863,      -437.78508,        255.3612 ],
  [               0,               0,       -30.26137,       139.98717,      -215.58606,       109.62714 ],
  [               0,               0,       4.6156103,       -27.33768,        51.39801,      -32.059065 ],
  [               0,               0,      -3.4419196,       16.548741,      -26.523762,       14.169494 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0045951914,   -0.0069479395 ],
  [               0,               0,               0,               0,      -5.2330416,       12.239909 ],
  [               0,               0,               0,               0,      0.42751694,      -1.7737631 ],
  [               0,               0,               0,               0,      0.24533789,      -1.4007224 ],
  [               0,               0,               0,               0,     0.023052479,    -0.088347875 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.015327149,    -0.036814943,    -0.073577664,       0.1786549 ],
  [               0,               0,       6.5357223,      -19.631418,       2.9730148,       21.861324 ],
  [               0,               0,       13.051701,       -65.47084,       102.38405,      -48.268784 ],
  [               0,               0,      -7.6414102,       54.649214,       -127.5737,       96.648202 ],
  [               0,               0,        4.061645,      -29.446085,       70.040866,      -54.835464 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00021565236,    0.0031137407 ],
  [               0,               0,               0,               0,      -148.01816,       342.21931 ],
  [               0,               0,               0,               0,       70.596253,      -163.74668 ],
  [               0,               0,               0,               0,      -2.8218735,       6.8454491 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00082049073,    0.0024931068 ],
  [               0,               0,               0,               0,     -0.79218179,       3.5164828 ],
  [               0,               0,               0,               0,     -0.41116498,     0.026677279 ],
  [               0,               0,               0,               0,       1.5526559,      -4.6038153 ],
  [               0,               0,               0,               0,       0.6820034,      -1.6005649 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0028029191,    0.0011973946,     0.021489173 ],
  [               0,               0,               0,      -8.3844342,       35.865024,      -32.172314 ],
  [               0,               0,               0,       26.387189,      -131.34297,       161.63417 ],
  [               0,               0,               0,       32.132507,      -58.940335,       25.151258 ],
  [               0,               0,               0,     0.051215321,    -0.078141391,     -0.67419939 ],
  [               0,               0,               0,      -2.3207503,       11.392619,      -14.008724 ],
  [               0,               0,               0,      -3.5802869,       17.340488,      -21.007289 ],
 ],
 [
  [               0,               0,               0,               0,     -0.43373018,        1.090507 ],
  [               0,               0,               0,               0,      -3531.4969,       8785.3854 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -35.191635,       86.504129 ],
  [               0,               0,               0,               0,      0.23763336,       0.2499317 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00081297404,    0.0010191207 ],
  [               0,               0,               0,               0,      0.37974666,       3.1947019 ],
  [               0,               0,               0,               0,      -5.1473083,      0.87985764 ],
  [               0,               0,               0,               0,      -6.1671299,       1.0032436 ],
  [               0,               0,               0,               0,     -0.60249451,     0.091845636 ],
  [               0,               0,               0,               0,   -0.0047005558,    0.0014143631 ],
  [               0,               0,               0,               0,     -0.41747535,      0.53218057 ],
  [               0,               0,               0,               0,    -0.088237556,       1.6394601 ],
 ],
 [
  [               0,               0,               0,    0.0003992681,    0.0010999482,   0.00016836304 ],
  [               0,               0,               0,       4.6685813,       -10.94199,       7.9701308 ],
  [               0,               0,               0,      -7.3786104,       9.5225221,      -3.7975371 ],
  [               0,               0,               0,       10.589128,      -13.449036,       3.2231687 ],
  [               0,               0,               0,       3.2908543,      -4.1931412,       1.1789863 ],
  [               0,               0,               0,   -0.0012228421,     0.001489048,   -0.0004809333 ],
  [               0,               0,               0,      -2.2105884,      -1.9232951,        3.426941 ],
  [               0,               0,               0,      -56.371198,       51.230738,      -12.254963 ],
 ],
 [
  [               0,               0,               0,               0,     0.023997049,    0.0050447885 ],
  [               0,               0,               0,               0,       5.0774925,       -3.471859 ],
  [               0,               0,               0,               0,      -19.942959,       15.348758 ],
  [               0,               0,               0,               0,       16.582228,      -13.186135 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0011034496,   -0.0051106116,     0.014862682 ],
  [               0,               0,               0,       6.9876015,      -14.960217,       8.5405349 ],
  [               0,               0,               0,      -5.3188377,       11.437363,       -6.595076 ],
  [               0,               0,               0,      -6.5503885,       14.052213,      -7.9411974 ],
  [               0,               0,               0,        -0.19269,      0.40839086,     -0.22594929 ],
  [               0,               0,               0,   -0.0029285948,    0.0062842948,   -0.0031992827 ],
  [               0,               0,               0,       10.143411,      -22.105637,       13.759699 ],
  [               0,               0,               0,       7.7729231,      -15.320517,       11.406499 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.042008171,     -0.02085549,     0.029994251,      0.13218997 ],
  [               0,               0,        394.5751,      -564.95611,      -753.31401,       1082.3159 ],
  [               0,               0,      -329.01572,        544.2207,       346.62881,      -647.19275 ],
  [               0,               0,       331.79585,      -415.26027,      -947.69163,       1231.1949 ],
  [               0,               0,       1831.8482,      -2607.2631,      -3727.6294,       5304.3477 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,    0.0067086906 ],
  [               0,               0,               0,               0,               0,      -2.7044016 ],
  [               0,               0,               0,               0,               0,       1.2261216 ],
  [               0,               0,               0,               0,               0,     -0.13327027 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0048763155 ],
  [               0,               0,               0,               0,               0,       3.5884648 ],
  [               0,               0,               0,               0,               0,      -1.1648794 ],
  [               0,               0,               0,               0,               0,      -1.0732419 ],
  [               0,               0,               0,               0,               0,     -0.15303677 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.030241668,     0.061298296 ],
  [               0,               0,               0,               0,      -17.794447,       29.375504 ],
  [               0,               0,               0,               0,        69.46338,      -104.65904 ],
  [               0,               0,               0,               0,       -56.73077,       84.470732 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.018135485,    -0.035515784,    -0.054648355,      0.10936561 ],
  [               0,               0,      -47.872361,       259.10452,      -458.91389,       268.28817 ],
  [               0,               0,      -30.827839,       141.53309,      -215.99246,       108.64656 ],
  [               0,               0,        4.717666,      -28.445274,       54.266654,      -34.233507 ],
  [               0,               0,      -3.5922834,       17.248501,      -27.620539,       14.747246 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0053834375,   -0.0083777634 ],
  [               0,               0,               0,               0,      -4.9243737,       11.586052 ],
  [               0,               0,               0,               0,      0.12660187,      -1.1913514 ],
  [               0,               0,               0,               0,     -0.17445116,      -0.6053379 ],
  [               0,               0,               0,               0,     0.035020324,     -0.12045666 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00022809612,  -0.00037317743,   0.00036724841,    0.0057988591 ],
  [               0,               0,       3.1920503,      -6.5169964,       -14.87875,       32.786555 ],
  [               0,               0,       1.0931071,      -2.3343655,      -4.0866425,       7.6551208 ],
  [               0,               0,      -2.1773663,       5.0637991,       8.9635115,      -21.992009 ],
  [               0,               0,     -0.83422047,       1.6364094,       4.4789542,      -8.9008148 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.14886719,      0.34588676 ],
  [               0,               0,               0,               0,       -97.90347,        226.9775 ],
  [               0,               0,               0,               0,         31.7016,      -74.309865 ],
  [               0,               0,               0,               0,       14.639939,      -33.301947 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00072139114,    0.0022454185 ],
  [               0,               0,               0,               0,      -2.4150013,       7.4305391 ],
  [               0,               0,               0,               0,      0.53632332,      -2.2056298 ],
  [               0,               0,               0,               0,       2.5174302,      -6.8846237 ],
  [               0,               0,               0,               0,      0.73055667,      -1.7170044 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0093561005,   0.00017965112,     0.063216061 ],
  [               0,               0,               0,      0.36608812,       7.5438384,      -15.297409 ],
  [               0,               0,               0,      -13.357648,       60.950074,      -70.865019 ],
  [               0,               0,               0,       3.8510306,       5.8032963,      -33.797886 ],
  [               0,               0,               0,   -0.0049398211,   -0.0034350995,     0.012586982 ],
  [               0,               0,               0,       -5.529482,       26.715946,       -32.14523 ],
  [               0,               0,               0,      -2.9621841,         17.9057,      -26.001411 ],
 ],
 [
  [               0,               0,               0,               0,     -0.22469364,      0.56913104 ],
  [               0,               0,               0,               0,      -3696.8553,       9194.8771 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -23.816837,       58.318097 ],
  [               0,               0,               0,               0,      0.22764498,      0.23478446 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,     0.041554437,     -0.01090539,    0.0086009336 ],
  [               0,               0,               0,      -10.147225,         2.43571,       3.4356132 ],
  [               0,               0,               0,       1.1657371,     -0.41945045,      -1.2895062 ],
  [               0,               0,               0,       6.8177974,      -4.7378863,      0.84730094 ],
  [               0,               0,               0,     0.061539187,    -0.042029079,    0.0072746996 ],
  [               0,               0,               0,       11.824435,      -2.2958989,      0.17238531 ],
  [               0,               0,               0,      -26.088425,       7.5872939,      -3.2740347 ],
 ],
 [
  [               0,               0,               0,   -0.0022684608,    0.0034002325,    0.0071293804 ],
  [               0,               0,               0,      -3.6741412,       3.1121583,       2.7819282 ],
  [               0,               0,               0,      -2.1891505,       2.3220409,       -1.982582 ],
  [               0,               0,               0,        120.6892,      -91.737568,       18.191235 ],
  [               0,               0,               0,       1.0114547,     -0.76928614,       0.1521203 ],
  [               0,               0,               0,      0.98282837,      -1.3833731,      0.52037709 ],
  [               0,               0,               0,      0.66029312,      -1.0161552,      -2.7359906 ],
 ],
 [
  [               0,               0,               0,               0,       0.1915821,     -0.14738018 ],
  [               0,               0,               0,               0,      -34.052965,        31.17454 ],
  [               0,               0,               0,               0,       13.989497,      -12.858762 ],
  [               0,               0,               0,               0,      -1.7755714,       1.5126457 ],
  [               0,               0,               0,               0,   -0.0024691445,    0.0031897762 ],
  [               0,               0,               0,               0,      -34.659737,       28.635122 ],
  [               0,               0,               0,               0,      -108.41879,       83.998423 ],
 ],
 [
  [               0,               0,               0,   -0.0022961722,    0.0080817014,     0.002574266 ],
  [               0,               0,               0,       1.9657146,      -5.9660315,       6.9788215 ],
  [               0,               0,               0,     -0.14508925,     0.040891392,      -1.2943664 ],
  [               0,               0,               0,     -0.13772439,      -1.2337335,       2.7355841 ],
  [               0,               0,               0,   -0.0057778113,   -0.0001054713,     0.016258697 ],
  [               0,               0,               0,      0.54052554,     -0.83123421,      0.37589068 ],
  [               0,               0,               0,      0.93540356,      -2.2867978,       -1.724653 ],
 ],
 [
  [               0,               0,               0,    -0.012170562,  -9.4762262e-05,     0.035368353 ],
  [               0,               0,               0,      -108.01163,       272.62377,      -168.61049 ],
  [               0,               0,               0,      -138.98653,       394.10238,       -280.3097 ],
  [               0,               0,               0,      -26.293768,       61.110822,      -34.195632 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.48975295,     -0.67667505 ],
  [               0,               0,               0,               0,      -318.34247,       485.68018 ],
  [               0,               0,               0,               0,      -42.002275,       56.332393 ],
  [               0,               0,               0,               0,       3.6468636,      -4.4526809 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,       1.2187465,      -1.6925484,      -2.6579048,       3.6984071 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -2.2952781e-05,   -0.0034366642,   -0.0025075232,     0.011954386 ],
  [               0,               0,       115.99631,       56.462392,      -82.841204,      -354.29043 ],
  [               0,               0,      -13.520401,       3.4141788,       17.073514,       9.3311846 ],
  [               0,               0,       11.406851,       7.1352365,      -7.3224963,      -42.637557 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.042731486,    -0.060901792 ],
  [               0,               0,               0,               0,      -162.66741,       248.65506 ],
  [               0,               0,               0,               0,      -11.242694,       16.040115 ],
  [               0,               0,               0,               0,       50.583348,      -76.452887 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0024251793,   0.00090629961,    0.0014521967 ],
  [               0,               0,               0,       4.7299539,      -17.448132,       18.346332 ],
  [               0,               0,               0,     -0.82678622,        2.853087,      -3.5959536 ],
  [               0,               0,               0,      0.11895345,     -0.27780995,      0.11776036 ],
  [               0,               0,               0,    0.0022294011,   -0.0082222144,    0.0074838829 ],
  [               0,               0,               0,      0.48818365,       1.4125991,       -1.112558 ],
  [               0,               0,               0,      0.11504688,      -4.2176632,      -3.2019458 ],
 ],
 [
  [               0,               0,               0,     0.030242621,     -0.12373536,      0.12917146 ],
  [               0,               0,               0,      -35.905225,       144.94593,       -142.1302 ],
  [               0,               0,               0,       27.514815,       -116.0415,       120.55646 ],
  [               0,               0,               0,      -14.184713,       59.969079,      -63.050798 ],
  [               0,               0,               0,    0.0035907165,    -0.015462533,     0.018193168 ],
  [               0,               0,               0,       1.4801409,      -6.1285062,       7.2895908 ],
  [               0,               0,               0,       17.774846,       -75.67626,       78.591096 ],
 ],
 [
  [               0,               0,               0,     -0.01585126,   -0.0063762779,      0.11625473 ],
  [               0,               0,               0,       99.741598,      -465.61734,       545.08441 ],
  [               0,               0,               0,      -13.768154,       63.640097,      -74.554577 ],
  [               0,               0,               0,      -4.1093955,         49.9077,      -88.670088 ],
  [               0,               0,               0,     0.001410806,     0.018105233,     0.031558738 ],
  [               0,               0,               0,       1.0488065,      -5.0766882,       6.1460871 ],
  [               0,               0,               0,       7.6013961,      -28.310158,       21.551119 ],
 ],
 [
  [               0,               0,               0,               0,    0.0055198402,     0.011244323 ],
  [               0,               0,               0,               0,      -107.57671,       266.72538 ],
  [               0,               0,               0,               0,        65.76399,      -162.85545 ],
  [               0,               0,               0,               0,      -31.485696,       77.634496 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0050046663,   -0.0058425945,    -0.011900025,    -0.010712024 ],
  [               0,               0,     0.053050441,       -1.209493,      -3.5599514,       -4.998647 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       1.4190939,       2.4351437,       3.3747589,       0.7392107 ],
  [               0,               0,      0.63623214,       1.5709685,       3.8990522,       9.3701319 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    0.0030149631,   0.00084120165,    0.0071091115 ],
  [               0,               0,               0,     -0.61631725,     -0.96016798,       4.5894192 ],
  [               0,               0,               0,      -1.3955879,      0.72693953,       -1.461072 ],
  [               0,               0,               0,       26.534061,      -9.3859834,       1.4702949 ],
  [               0,               0,               0,       0.1871655,    -0.063007785,      0.01053256 ],
  [               0,               0,               0,       1.7460785,     -0.95746859,      0.19426306 ],
  [               0,               0,               0,      -12.210186,       4.9943182,      -3.1491638 ],
 ],
 [
  [               0,               0,               0,    -0.016760962,     0.019843579,    0.0015329279 ],
  [               0,               0,               0,     -0.48082131,     -0.82930316,       4.8830116 ],
  [               0,               0,               0,      -3.7427456,       3.9092193,      -2.4263922 ],
  [               0,               0,               0,       1.5158201,       -1.206837,      0.29600514 ],
  [               0,               0,               0,    0.0016782969,  -0.00060428128,   0.00057416759 ],
  [               0,               0,               0,       6.8330628,      -8.2252804,       2.9483686 ],
  [               0,               0,               0,       5.2092251,      -8.1442745,      -5.6087427 ],
 ],
 [
  [               0,               0,               0,               0,      0.10799861,    -0.078209463 ],
  [               0,               0,               0,               0,      -7.4346824,       9.8345213 ],
  [               0,               0,               0,               0,       11.089066,      -10.470626 ],
  [               0,               0,               0,               0,       -4.700052,       3.9943116 ],
  [               0,               0,               0,               0,    -0.034522625,     0.030201618 ],
  [               0,               0,               0,               0,       17.479925,      -13.479351 ],
  [               0,               0,               0,               0,      -67.292629,       50.976853 ],
 ],
 [
  [               0,               0,               0,    -0.000248812,    0.0027995817,    0.0052148325 ],
  [               0,               0,               0,       1.4483295,      -4.9940315,       7.2435192 ],
  [               0,               0,               0,     -0.43846229,      0.55542323,      -1.5414892 ],
  [               0,               0,               0,     -0.31477537,       1.1023408,     -0.54390823 ],
  [               0,               0,               0,    0.0023068743,   -0.0045611277,    0.0036580122 ],
  [               0,               0,               0,       1.1265513,      -1.7776155,      0.88586676 ],
  [               0,               0,               0,       2.7163065,      -7.3698722,        2.049419 ],
 ],
 [
  [               0,               0,               0,      0.48496132,     -0.18121488,      -0.6387254 ],
  [               0,               0,               0,       1288.1607,      -3506.4029,       2388.4783 ],
  [               0,               0,               0,      -46.594284,       160.17285,      -133.43033 ],
  [               0,               0,               0,       1229.5882,      -3368.6809,       2307.3441 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.034474254,    0.0022732539,    -0.059091286 ],
  [               0,               0,               0,      -372.53901,       3.6428638,       768.28296 ],
  [               0,               0,               0,       105.80885,      -290.01267,       197.37936 ],
  [               0,               0,               0,       0.1519669,     -0.39465971,     -0.43344798 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -0.9796559,       1.4249422,       1.8569306,      -2.7022287 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     -0.04035682,      0.00475159,     0.048222438,     0.054895191 ],
  [               0,               0,       1215.1754,      -1762.7373,      -2640.1904,       3831.4272 ],
  [               0,               0,      -212.13124,       307.55058,       454.32898,      -659.55292 ],
  [               0,               0,      -142.41208,       210.19286,        309.4108,      -456.55748 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0032706366,    0.0014735431 ],
  [               0,               0,               0,               0,       100.11423,      -148.55727 ],
  [               0,               0,               0,               0,       9.7288848,      -16.040643 ],
  [               0,               0,               0,               0,        0.224028,     -0.26395227 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.055928779,     -0.18246951,      0.15478791 ],
  [               0,               0,               0,      -12.307358,       40.161742,      -28.955886 ],
  [               0,               0,               0,       7.5456038,      -26.496499,       21.894422 ],
  [               0,               0,               0,      -3.5212252,       12.810801,      -11.542489 ],
  [               0,               0,               0,     0.012558361,    -0.042584977,     0.036879356 ],
  [               0,               0,               0,      -8.7521015,       27.961354,      -21.520243 ],
  [               0,               0,               0,       21.123622,      -76.748826,       67.411828 ],
 ],
 [
  [               0,               0,               0,      0.03193093,     -0.13453786,      0.15002496 ],
  [               0,               0,               0,      -10.098385,       41.285667,      -39.185587 ],
  [               0,               0,               0,        1.350285,      -5.5171031,       4.3848431 ],
  [               0,               0,               0,      -50.435106,       208.50539,      -213.31269 ],
  [               0,               0,               0,    -0.015960499,     0.058411669,    -0.027576008 ],
  [               0,               0,               0,    -0.078593569,      0.33973182,     -0.35467961 ],
  [               0,               0,               0,       18.574871,      -76.800994,       77.009683 ],
 ],
 [
  [               0,               0,               0,   -0.0069471278,   -0.0034499667,     0.056797141 ],
  [               0,               0,               0,       82.334777,      -385.15595,       452.85181 ],
  [               0,               0,               0,       1.8118371,      -9.1700457,       10.341867 ],
  [               0,               0,               0,      -5.3017217,       38.461893,      -52.150596 ],
  [               0,               0,               0,     -0.03893121,    0.0060880962,      0.29311048 ],
  [               0,               0,               0,     -0.35770212,       1.4560946,      -1.4148156 ],
  [               0,               0,               0,        16.67414,      -76.415498,       84.420108 ],
 ],
 [
  [               0,               0,               0,               0,       2.8134474,      -6.8888616 ],
  [               0,               0,               0,               0,      -9.4452087,       25.194765 ],
  [               0,               0,               0,               0,       55.641766,      -138.20598 ],
  [               0,               0,               0,               0,        -45.7909,       112.78073 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     -0.01282718,     0.031791375,     0.089568917,     -0.22198799 ],
  [               0,               0,       160.11344,       146.70439,       625.74361,       -4879.478 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       -266.2066,      -221.91979,       589.65219,       4165.4149 ],
  [               0,               0,        1.399377,       10.779892,       25.598208,       14.362249 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    0.0015027925,  -0.00056235207,    0.0074118841 ],
  [               0,               0,               0,      0.36979666,     -0.36556673,       4.4409916 ],
  [               0,               0,               0,      -1.0973406,       0.6279684,       -1.451529 ],
  [               0,               0,               0,       844.04116,      -219.02698,       24.333428 ],
  [               0,               0,               0,       6.2841756,      -1.6276239,       0.1803365 ],
  [               0,               0,               0,      0.28897409,     -0.22723529,     0.048899525 ],
  [               0,               0,               0,       -2.568067,        1.247791,      -3.1445215 ],
 ],
 [
  [               0,               0,               0,     -0.01942779,     0.022376817,   0.00094210709 ],
  [               0,               0,               0,      0.61272845,      -1.9070064,       5.1470096 ],
  [               0,               0,               0,      -4.3336535,       4.5633064,       -2.611547 ],
  [               0,               0,               0,       1.6617984,      -1.3754715,      0.35337499 ],
  [               0,               0,               0,    0.0011736874,   2.6304407e-05,   0.00043905652 ],
  [               0,               0,               0,       7.1824024,      -8.5043484,       2.9809082 ],
  [               0,               0,               0,       4.7681665,      -7.2404243,      -5.9989885 ],
 ],
 [
  [               0,               0,               0,               0,      0.12332383,    -0.090520023 ],
  [               0,               0,               0,               0,      -10.612274,       12.398037 ],
  [               0,               0,               0,               0,       13.172084,      -12.162435 ],
  [               0,               0,               0,               0,      -5.2917739,        4.495119 ],
  [               0,               0,               0,               0,    -0.043775229,     0.038048644 ],
  [               0,               0,               0,               0,       16.158791,      -12.424092 ],
  [               0,               0,               0,               0,      -69.870298,        53.03483 ],
 ],
 [
  [               0,               0,               0,   0.00039049132,    0.0017998775,    0.0056184178 ],
  [               0,               0,               0,       1.4088325,      -5.0073486,       7.2803613 ],
  [               0,               0,               0,     -0.39751518,      0.51465964,      -1.5445337 ],
  [               0,               0,               0,     -0.45083775,       1.3511621,     -0.28487486 ],
  [               0,               0,               0,   0.00070651975,  -0.00036150791,    0.0039619823 ],
  [               0,               0,               0,      0.62628384,        -1.00682,      0.51155655 ],
  [               0,               0,               0,       1.2743227,      -3.5856816,     -0.52130613 ],
 ],
 [
  [               0,               0,               0,      0.50051805,     -0.18481888,     -0.66081088 ],
  [               0,               0,               0,       1215.7131,      -3315.8829,       2263.3261 ],
  [               0,               0,               0,      -42.691813,       148.79673,      -125.09747 ],
  [               0,               0,               0,       1387.9653,      -3804.7503,       2607.4657 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.011020785,    0.0029654466,    -0.012702204 ],
  [               0,               0,               0,      -391.53287,       4.1154222,       805.56995 ],
  [               0,               0,               0,       37.471075,      -95.718009,       59.492632 ],
  [               0,               0,               0,    -0.005439587,     -0.21499541,     -0.47063404 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele75_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,     -0.47657048,      0.70837188,      0.82710868,      -1.2358329 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0024901901,     -0.01251302,    -0.011525696,     0.039289989 ],
  [               0,               0,       176.56955,      -239.98154,      -373.63664,       510.61089 ],
  [               0,               0,      -263.99764,       385.50289,       572.03372,      -836.60872 ],
  [               0,               0,       -134.6985,       198.70203,       293.27057,       -432.4871 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.017646814,    -0.026258793 ],
  [               0,               0,               0,               0,       1812.1349,      -2723.1424 ],
  [               0,               0,               0,               0,      -1214.1802,       1831.2134 ],
  [               0,               0,               0,               0,       1519.2427,      -2296.5414 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0036317717,   -0.0095409752,     0.011800852 ],
  [               0,               0,               0,       5.8479522,      -20.830896,       22.217619 ],
  [               0,               0,               0,       8.9307019,      -30.791522,       25.112616 ],
  [               0,               0,               0,      -6.0122725,       20.765106,      -17.825987 ],
  [               0,               0,               0,   -0.0021381301,    0.0070919099,    -0.004907017 ],
  [               0,               0,               0,     -0.34752619,       1.1377343,     -0.12700266 ],
  [               0,               0,               0,        36.09945,      -123.69584,       103.79578 ],
 ],
 [
  [               0,               0,               0,     0.035769572,     -0.15204521,      0.16996501 ],
  [               0,               0,               0,      -11.129943,       45.941431,      -44.431829 ],
  [               0,               0,               0,       1.4654035,      -6.0330417,       4.9621998 ],
  [               0,               0,               0,      -45.154266,       186.89425,      -190.58814 ],
  [               0,               0,               0,    -0.086829854,      0.34605674,     -0.31828802 ],
  [               0,               0,               0,     -0.21976016,      0.94171158,     -0.98902311 ],
  [               0,               0,               0,       17.086741,      -70.835552,       70.705983 ],
 ],
 [
  [               0,               0,               0,   -0.0097167462,   -0.0041541795,     0.074201563 ],
  [               0,               0,               0,       103.11794,      -480.75105,       562.65581 ],
  [               0,               0,               0,       1.4413629,      -7.9362704,       9.5034607 ],
  [               0,               0,               0,      -12.641677,       70.116648,      -84.578961 ],
  [               0,               0,               0,    -0.083706215,    0.0021784707,      0.56902735 ],
  [               0,               0,               0,     0.072807353,     -0.51239972,      0.82788007 ],
  [               0,               0,               0,       19.258023,      -88.375967,       98.256239 ],
 ],
 [
  [               0,               0,               0,               0,       2.8138091,      -6.8910354 ],
  [               0,               0,               0,               0,      -4.8965284,       14.033282 ],
  [               0,               0,               0,               0,       57.982123,      -143.99879 ],
  [               0,               0,               0,               0,      -47.447121,       116.86153 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00082762144,    0.0014245571,   -0.0020685054,    -0.016209728 ],
  [               0,               0,      -1.1967281,      -1.3806497,      -3.4218995,      -4.4810856 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      -354.97526,      -284.37455,       736.18476,       5416.5465 ],
  [               0,               0,       1.2375404,       1.1158975,       15.269181,       8.2550292 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 100 GeV electron clusters (large stats)
#


ele100large_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,      -9738.6855,       190.43562,      -1.1762939,    0.0013381985 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00029288817,   0.00063473641 ],
  [               0,               0,               0,               0,     -0.76512659,       3.8063442 ],
  [               0,               0,               0,               0,     0.004913302,     -0.66277599 ],
  [               0,               0,               0,               0,     -0.44031748,     -0.71897912 ],
  [               0,               0,               0,               0,      0.52893547,     -0.34800136 ],
  [               0,               0,               0,               0,  -0.00052581512,   0.00022117836 ],
  [               0,               0,               0,               0,      -1.9967001,      0.35730167 ],
  [               0,               0,               0,               0,      0.83473041,     -0.20884408 ],
 ],
 [
  [               0,               0,               0,    0.0051681434,   -0.0016047475,   0.00018002848 ],
  [               0,               0,               0,       3.8524122,      -11.492313,       8.5193319 ],
  [               0,               0,               0,       16.608338,      -22.691586,       6.6873708 ],
  [               0,               0,               0,       1.1009117,      -1.8288805,     -0.28094052 ],
  [               0,               0,               0,      -3.6101543,       4.9711195,      -1.7817805 ],
  [               0,               0,               0,    0.0031986242,   -0.0042165157,    0.0013318748 ],
  [               0,               0,               0,       4.6818685,      -3.0476501,       1.6163339 ],
  [               0,               0,               0,      -2.9329663,       5.2109086,      -2.8628941 ],
 ],
 [
  [               0,               0,               0,               0,     0.025739399,    0.0017780927 ],
  [               0,               0,               0,               0,      -13.649671,       11.461423 ],
  [               0,               0,               0,               0,       -12.94771,       9.7500084 ],
  [               0,               0,               0,               0,       19.782612,      -15.729405 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   1.5824214e-05,   7.9096399e-06,   0.00028018539 ],
  [               0,               0,               0,      -9.1018725,       14.288871,      -1.6488177 ],
  [               0,               0,               0,      -20.176363,       1.3715288,      -32.350945 ],
  [               0,               0,               0,      -17.039169,      -2.9629637,      -32.323215 ],
  [               0,               0,               0,     -0.74697039,       8.0997271,      -6.8369366 ],
  [               0,               0,               0,    0.0027575062,    -0.020655129,   -0.0044127728 ],
  [               0,               0,               0,      0.13506034,     -0.22309747,      0.32488433 ],
  [               0,               0,               0,     0.036332936,     -0.15136336,       1.6688612 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00012912753 ],
  [               0,               0,               0,               0,               0,       39.968509 ],
  [               0,               0,               0,               0,               0,      -8.0757409 ],
  [               0,               0,               0,               0,               0,      -1.9697527 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.035642069,     0.035003062,     0.028309038,     0.011380636 ],
  [               0,               0,       170.97678,       78.544555,      -110.96412,      -453.48833 ],
  [               0,               0,       4.4511629,       1.7617413,      -3.1475844,      -11.779896 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,    -0.066760088,      -0.1013217,     -0.12883243,     -0.43227421 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.094665715,      -0.1537772,     -0.20717742,      0.34120714 ],
  [               0,               0,       525.28148,      -757.67778,      -1046.7065,       1514.8243 ],
  [               0,               0,       -40.16026,       141.75244,      -316.17644,       278.90335 ],
  [               0,               0,       471.76625,      -626.35797,      -1203.9063,       1619.6277 ],
  [               0,               0,       2093.5229,      -2981.1828,      -4249.6117,        6050.018 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,  -0.00046245288,   0.00066913375 ],
  [               0,               0,               0,               0,        1964.459,       -2816.366 ],
  [               0,               0,               0,               0,       -3198.448,       4263.1737 ],
  [               0,               0,               0,               0,       56043.884,      -79347.434 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.022414987,    -0.032396419 ],
  [               0,               0,               0,               0,        603.1753,      -849.09577 ],
  [               0,               0,               0,               0,      -1888.5254,       2785.7057 ],
  [               0,               0,               0,               0,       1640.9506,      -2361.2407 ],
  [               0,               0,               0,               0,     -0.32678851,      -1.9103207 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.2675662,      0.40528939 ],
  [               0,               0,               0,               0,      -109.34964,       166.26739 ],
  [               0,               0,               0,               0,       27.247108,      -41.593723 ],
  [               0,               0,               0,               0,       20.042929,      -29.653548 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0040197533 ],
  [               0,               0,               0,               0,               0,      -1.0438864 ],
  [               0,               0,               0,               0,               0,       1.8196944 ],
  [               0,               0,               0,               0,               0,       0.9004827 ],
  [               0,               0,               0,               0,               0,      0.67812965 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0026395487,    -0.001852674,    0.0043202028,     0.025976389 ],
  [               0,               0,      -9.4445504,       48.225458,      -81.584078,       46.580083 ],
  [               0,               0,      -1.0098576,       4.3741653,      -6.3788934,       2.4352979 ],
  [               0,               0,      -2.3248725,       9.9595099,       -14.01326,       5.7444797 ],
  [               0,               0,       -1.134905,       5.2893246,      -8.1820584,       4.2003556 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.004615078,   -0.0065298262 ],
  [               0,               0,               0,               0,      -2.4233299,       6.7778943 ],
  [               0,               0,               0,               0,     -0.55190022,    -0.012882269 ],
  [               0,               0,               0,               0,    -0.071586228,     -0.88446243 ],
  [               0,               0,               0,               0,     0.064829296,      -0.1426568 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.009131964,    -0.022592721,    -0.045501582,      0.11474911 ],
  [               0,               0,       3.3111863,      -12.200558,       12.359101,     -0.86976673 ],
  [               0,               0,       13.827573,      -81.114369,       157.27231,      -101.69517 ],
  [               0,               0,      -6.1810873,       43.193315,      -99.577955,       74.832941 ],
  [               0,               0,       7.3004446,      -48.555036,       107.42026,      -79.056155 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       0.6820407,      -1.5611619 ],
  [               0,               0,               0,               0,       -133.9909,       309.63579 ],
  [               0,               0,               0,               0,       75.337032,      -174.37889 ],
  [               0,               0,               0,               0,      -16.203007,       37.447084 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00033012086,    0.0014661694 ],
  [               0,               0,               0,               0,      -3.4280738,       10.924814 ],
  [               0,               0,               0,               0,      0.87629965,      -3.1608367 ],
  [               0,               0,               0,               0,       1.8474675,      -5.3648739 ],
  [               0,               0,               0,               0,      0.15813522,     -0.34620658 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       1.6207353,      -3.8802488 ],
  [               0,               0,               0,               0,       -298.9263,       718.15281 ],
  [               0,               0,               0,               0,       188.61623,      -453.11053 ],
  [               0,               0,               0,               0,      -432.34911,       1034.5966 ],
  [               0,               0,               0,               0,      0.40770346,      -1.3308246 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0031661148,    0.0014516547,    -0.014341901 ],
  [               0,               0,               0,      -6.2255642,       3.7735533,       33.339042 ],
  [               0,               0,               0,      -3.7774377,        10.69122,      -5.2079313 ],
  [               0,               0,               0,       1.2145273,       2.6118145,      -13.476419 ],
  [               0,               0,               0,    -0.012135874,  -9.5469779e-05,     0.072259588 ],
  [               0,               0,               0,      -1.6479064,     -0.16328668,       9.3175533 ],
  [               0,               0,               0,       20.812902,       1.2815903,      -129.18951 ],
 ],
 [
  [               0,               0,               0,               0,      0.01361661,     0.016215994 ],
  [               0,               0,               0,               0,       7.1668916,       21.799095 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       19.899384,      -49.373798 ],
  [               0,               0,               0,               0,      0.34341562,      0.58616624 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,        15199.69,       -20.18936,     -0.68674043,   0.00095273602 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -1.2933859e-05,   0.00073138046 ],
  [               0,               0,               0,               0,      0.11794693,       3.5212526 ],
  [               0,               0,               0,               0,      -5.6464621,       1.0419921 ],
  [               0,               0,               0,               0,      -6.0245027,      0.99009065 ],
  [               0,               0,               0,               0,       1.8737031,     -0.53520083 ],
  [               0,               0,               0,               0,  -8.3312114e-05,    0.0020012631 ],
  [               0,               0,               0,               0,     -0.70644451,      0.21462322 ],
  [               0,               0,               0,               0,       0.7979882,       1.3778404 ],
 ],
 [
  [               0,               0,               0,    0.0048340389,   -0.0015682358,   0.00020389703 ],
  [               0,               0,               0,        8.480704,      -17.587395,       10.496899 ],
  [               0,               0,               0,       14.635221,      -20.437103,       6.1522936 ],
  [               0,               0,               0,      -1.3361031,       1.6759005,      -1.5094845 ],
  [               0,               0,               0,      -4.1480612,       5.8140245,      -2.1197528 ],
  [               0,               0,               0,    0.0028075245,   -0.0038199137,    0.0012607382 ],
  [               0,               0,               0,       13.586592,      -12.732125,       4.0344613 ],
  [               0,               0,               0,     -0.19494386,       2.8152085,      -2.2606469 ],
 ],
 [
  [               0,               0,               0,               0,     0.025894167,    0.0017023997 ],
  [               0,               0,               0,               0,      -13.687817,       11.491226 ],
  [               0,               0,               0,               0,      -12.844407,       9.6683094 ],
  [               0,               0,               0,               0,       19.737958,      -15.693912 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   2.9774527e-05,   1.1334274e-05,   0.00027368846 ],
  [               0,               0,               0,      -10.128038,       16.539801,      -2.9265623 ],
  [               0,               0,               0,      -17.948232,      -1.1924778,      -28.750158 ],
  [               0,               0,               0,      -14.760779,      -5.5803916,      -28.692648 ],
  [               0,               0,               0,     -0.11948382,        6.649079,      -6.0968641 ],
  [               0,               0,               0,   0.00035741439,     -0.01839317,   -0.0033093305 ],
  [               0,               0,               0,      0.13878728,     -0.24286116,      0.34433725 ],
  [               0,               0,               0,     0.033521023,     -0.14834016,        1.668831 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00012912753 ],
  [               0,               0,               0,               0,               0,       39.968509 ],
  [               0,               0,               0,               0,               0,      -8.0757409 ],
  [               0,               0,               0,               0,               0,      -1.9697527 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0052962432,   -0.0023742072,    0.0036727715,     0.015353588 ],
  [               0,               0,       391.47789,       174.69172,      -260.04588,      -1041.6278 ],
  [               0,               0,       114.33856,       49.640179,      -80.262782,      -315.73348 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,        1.147928,       1.3995578,       2.3102125,       4.2944862 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.036620916,    -0.090316768,     -0.10213059,      0.23141548 ],
  [               0,               0,       683.28088,      -979.69072,      -1337.4715,       1922.1706 ],
  [               0,               0,      -81.817004,       205.99265,      -256.97845,       184.74163 ],
  [               0,               0,       503.89248,      -667.22166,      -1294.5402,       1738.7201 ],
  [               0,               0,       2296.1881,      -3269.6332,      -4662.5964,       6637.7213 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    0.0033611449,   -0.0038599025 ],
  [               0,               0,               0,               0,      -633.88088,       882.01496 ],
  [               0,               0,               0,               0,      -11.201485,       20.435972 ],
  [               0,               0,               0,               0,      -314.58341,       451.31833 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.013920632,    -0.019955492 ],
  [               0,               0,               0,               0,       1584.7446,      -2297.4096 ],
  [               0,               0,               0,               0,      -2103.6085,       3101.4757 ],
  [               0,               0,               0,               0,       2908.0873,      -4220.2741 ],
  [               0,               0,               0,               0,     -0.21228061,      -1.9798037 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.021379753,     0.048948755 ],
  [               0,               0,               0,               0,      -18.459379,       30.228625 ],
  [               0,               0,               0,               0,       56.307258,      -85.044598 ],
  [               0,               0,               0,               0,      -45.554093,       67.837981 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00046898539 ],
  [               0,               0,               0,               0,               0,   -0.0025487924 ],
  [               0,               0,               0,               0,               0,        3.285267 ],
  [               0,               0,               0,               0,               0,      0.97994062 ],
  [               0,               0,               0,               0,               0,       5.2622117 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0026937406,   -0.0018288236,    0.0044323586,     0.026045942 ],
  [               0,               0,      -9.4085768,       48.077948,       -81.38957,       46.497658 ],
  [               0,               0,      -0.9795741,       4.1966642,      -6.0417926,       2.2272225 ],
  [               0,               0,      -2.3204639,       9.9128722,      -13.897228,        5.661086 ],
  [               0,               0,      -1.1366963,       5.2994947,      -8.2010422,       4.2120279 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0045925427,   -0.0064964487 ],
  [               0,               0,               0,               0,      -2.4359731,       6.8064914 ],
  [               0,               0,               0,               0,     -0.55791201,   -0.0011488795 ],
  [               0,               0,               0,               0,    -0.069978892,     -0.88701175 ],
  [               0,               0,               0,               0,     0.066198884,     -0.14535062 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0092241696,    -0.022813973,     -0.04593218,      0.11578356 ],
  [               0,               0,       3.3978394,      -12.727827,       13.450807,      -1.6411921 ],
  [               0,               0,       13.901334,      -81.494165,       157.88603,      -101.99227 ],
  [               0,               0,      -6.1201144,       42.865959,      -99.016016,       74.531829 ],
  [               0,               0,       7.2933207,      -48.513111,       107.33755,      -79.001362 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.67481546,      -1.5445825 ],
  [               0,               0,               0,               0,      -134.26788,       310.27643 ],
  [               0,               0,               0,               0,       75.226026,       -174.1255 ],
  [               0,               0,               0,               0,       -16.17629,       37.386537 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00033196865,     0.001468377 ],
  [               0,               0,               0,               0,      -3.4169781,       10.893092 ],
  [               0,               0,               0,               0,      0.88157842,      -3.1735472 ],
  [               0,               0,               0,               0,       1.8443835,      -5.3583491 ],
  [               0,               0,               0,               0,      0.15832621,     -0.34676769 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00030943917,    0.0036087906 ],
  [               0,               0,               0,               0,      -5.5606405,       14.602285 ],
  [               0,               0,               0,               0,       19.003231,      -46.508009 ],
  [               0,               0,               0,               0,       97.027785,      -233.75963 ],
  [               0,               0,               0,               0,    -0.052221114,     -0.40366009 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.017273277,    -0.057997463,     0.047163307 ],
  [               0,               0,               0,       69.735836,      -365.71547,       482.48174 ],
  [               0,               0,               0,      -19.565179,       98.591591,      -125.65629 ],
  [               0,               0,               0,      -64.099112,       312.72915,      -381.30225 ],
  [               0,               0,               0,   0.00059797597,    -0.030515658,     0.071160385 ],
  [               0,               0,               0,       8.4856024,      -29.077356,       21.602202 ],
  [               0,               0,               0,       1.0451914,       19.080224,      -55.077447 ],
 ],
 [
  [               0,               0,               0,               0,     -0.91440909,       2.2891397 ],
  [               0,               0,               0,               0,      -2864.6591,       7130.0606 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       -21.31067,       52.125918 ],
  [               0,               0,               0,               0,      0.82900837,      -1.1441982 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,       10482.202,       16.645331,     -0.75948881,   0.00098016419 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00011705645,   0.00075313122 ],
  [               0,               0,               0,               0,      0.47724338,       3.4450188 ],
  [               0,               0,               0,               0,      -5.9059557,       1.1650123 ],
  [               0,               0,               0,               0,      -6.3405006,       1.1253665 ],
  [               0,               0,               0,               0,       2.0716584,     -0.55383268 ],
  [               0,               0,               0,               0,  -0.00038469492,    0.0023673654 ],
  [               0,               0,               0,               0,     -0.65166552,      0.20106215 ],
  [               0,               0,               0,               0,      0.76074792,       1.3966739 ],
 ],
 [
  [               0,               0,               0,     0.005725108,   -0.0023699121,   0.00036923858 ],
  [               0,               0,               0,       8.9039953,      -18.378042,       10.813204 ],
  [               0,               0,               0,       15.163564,      -21.196008,       6.4205514 ],
  [               0,               0,               0,     -0.75486477,      0.93067264,      -1.2783426 ],
  [               0,               0,               0,      -4.1817946,       5.8943261,      -2.1583344 ],
  [               0,               0,               0,    0.0028487102,   -0.0038844568,    0.0012853203 ],
  [               0,               0,               0,       12.727175,      -11.814466,       3.8043763 ],
  [               0,               0,               0,       1.2616705,       1.2338158,      -1.8884318 ],
 ],
 [
  [               0,               0,               0,               0,     0.025595821,    0.0017866241 ],
  [               0,               0,               0,               0,      -13.427154,       11.285713 ],
  [               0,               0,               0,               0,      -13.087497,        9.860238 ],
  [               0,               0,               0,               0,       19.830437,      -15.767387 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00021245392,   -0.0008680667,    0.0034581023 ],
  [               0,               0,               0,      0.22658784,     0.038800037,       1.4000027 ],
  [               0,               0,               0,     -0.68301107,       1.4144769,      -1.6396196 ],
  [               0,               0,               0,     0.070824231,     -0.44788589,     -0.68733077 ],
  [               0,               0,               0,      0.12569571,     -0.32328662,      0.11662432 ],
  [               0,               0,               0,   0.00029861884,  -0.00078964314,   0.00049614419 ],
  [               0,               0,               0,        -16.1744,       39.310613,      -20.504776 ],
  [               0,               0,               0,      -1.7967048,       5.6448741,    -0.098687904 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00013361831 ],
  [               0,               0,               0,               0,               0,       39.957375 ],
  [               0,               0,               0,               0,               0,      -7.7922907 ],
  [               0,               0,               0,               0,               0,       -1.936781 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0055362107,    0.0062691083,    0.0067160437,    0.0070363151 ],
  [               0,               0,       139.14137,        62.79513,       -92.05852,      -368.92172 ],
  [               0,               0,        22.65566,       9.8882398,      -15.611352,       -61.96552 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,     0.079700469,     0.069442503,      0.16484964,      0.19616577 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.042948731,     -0.10053625,     -0.11601209,      0.25368252 ],
  [               0,               0,       661.87058,      -948.93954,      -1294.9702,       1861.1277 ],
  [               0,               0,      -123.44873,       265.71799,      -170.41603,       60.577339 ],
  [               0,               0,       500.10777,      -661.37882,      -1289.0829,       1730.0335 ],
  [               0,               0,       2298.9628,      -3273.7443,      -4668.3468,       6646.2384 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,     0.033051411,    -0.045929664 ],
  [               0,               0,               0,               0,      -20.085554,      -10.543988 ],
  [               0,               0,               0,               0,      -136.86951,       200.89777 ],
  [               0,               0,               0,               0,      -100.96338,       144.18799 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.15065257,     -0.21500759 ],
  [               0,               0,               0,               0,       66.760078,      -94.901396 ],
  [               0,               0,               0,               0,      -56.672458,       82.343852 ],
  [               0,               0,               0,               0,       155.55145,      -226.98569 ],
  [               0,               0,               0,               0,     0.033724753,     -0.21443825 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.022685181,     0.052581069 ],
  [               0,               0,               0,               0,      -23.782557,       38.049993 ],
  [               0,               0,               0,               0,       57.403966,      -86.656836 ],
  [               0,               0,               0,               0,      -46.842549,       69.751647 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0005466141 ],
  [               0,               0,               0,               0,               0,   -0.0022445717 ],
  [               0,               0,               0,               0,               0,       3.1524297 ],
  [               0,               0,               0,               0,               0,      0.99805584 ],
  [               0,               0,               0,               0,               0,       4.7252468 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,        6.061083,      -27.603101,       41.915778,      -21.219332 ],
  [               0,               0,      -3.0929055,       26.613632,       -67.12084,       52.994113 ],
  [               0,               0,       -34.29769,       171.96552,      -284.61928,       154.81469 ],
  [               0,               0,      -31.342997,       156.39814,      -257.35594,       138.94518 ],
  [               0,               0,     -0.26093871,       1.2012386,      -1.8246054,      0.91265676 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.004508213,   -0.0062447037 ],
  [               0,               0,               0,               0,      -2.0930882,       6.0954942 ],
  [               0,               0,               0,               0,      -0.5345012,    -0.035968987 ],
  [               0,               0,               0,               0,     -0.12140645,     -0.79415083 ],
  [               0,               0,               0,               0,     0.070401504,      -0.1589183 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0096801614,    -0.023890768,    -0.048139713,      0.12099735 ],
  [               0,               0,       4.4328684,      -18.657126,       24.677527,      -8.6750683 ],
  [               0,               0,       14.686814,      -86.121962,       166.94331,      -107.87204 ],
  [               0,               0,      -5.7644836,       40.389089,      -93.339512,       70.240432 ],
  [               0,               0,       7.3482596,      -48.950785,       108.42985,      -79.877839 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0011457387,     0.003979995 ],
  [               0,               0,               0,               0,      -80.682569,       187.01499 ],
  [               0,               0,               0,               0,       29.247765,      -68.369418 ],
  [               0,               0,               0,               0,       1.0220463,      -2.1640581 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00036423218,    0.0015042744 ],
  [               0,               0,               0,               0,       -3.022207,       9.8735267 ],
  [               0,               0,               0,               0,      0.69186223,      -2.7276358 ],
  [               0,               0,               0,               0,       1.8428547,      -5.3769852 ],
  [               0,               0,               0,               0,      0.18053768,     -0.40737832 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.27006401,      0.64972265 ],
  [               0,               0,               0,               0,       633.21448,      -1516.3615 ],
  [               0,               0,               0,               0,       37.048368,      -90.594716 ],
  [               0,               0,               0,               0,      -383.67684,        918.2865 ],
  [               0,               0,               0,               0,     -0.21122278,     -0.11510432 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.063789857,      0.24923139,      -0.2217945 ],
  [               0,               0,               0,      -271.75493,       1344.0772,      -1655.3015 ],
  [               0,               0,               0,      -151.52075,       721.67397,      -860.57093 ],
  [               0,               0,               0,      -3.0207739,       25.545515,      -44.109188 ],
  [               0,               0,               0,    0.0047579015,    -0.058754964,      0.11395687 ],
  [               0,               0,               0,      -2.1003776,       11.202771,      -13.851549 ],
  [               0,               0,               0,      -43.103196,       201.72933,      -228.27607 ],
 ],
 [
  [               0,               0,               0,               0,    0.0030385318,   -0.0014837572 ],
  [               0,               0,               0,               0,      -2802.9578,       6977.1573 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       404.77216,      -1002.0054 ],
  [               0,               0,               0,               0,      0.32886204,        1.039711 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.021642704,     0.010875912,    0.0069524112 ],
  [               0,               0,               0,        3.310336,      -1.9079805,       3.7976965 ],
  [               0,               0,               0,     -0.25913619,     0.053575155,      -1.2935311 ],
  [               0,               0,               0,      0.54526418,     -0.14125023,   -0.0075767564 ],
  [               0,               0,               0,    0.0009931887,  -0.00083675796,   0.00040740881 ],
  [               0,               0,               0,       21.136193,      -9.2733934,       1.2943808 ],
  [               0,               0,               0,      -23.730897,       9.5867275,      -1.8891713 ],
 ],
 [
  [               0,               0,               0,   -0.0047507927,    0.0058396271,    0.0055888312 ],
  [               0,               0,               0,      -3.5903629,       3.0979847,       3.0795818 ],
  [               0,               0,               0,      -0.5828778,      0.32642095,      -1.4010396 ],
  [               0,               0,               0,       1.0129404,     -0.79612871,      0.20014776 ],
  [               0,               0,               0,    0.0010475139,   -0.0010737147,   0.00071289118 ],
  [               0,               0,               0,       1.8387829,       -2.288389,       1.4962284 ],
  [               0,               0,               0,       1.0744846,      -2.4235182,      -1.0212622 ],
 ],
 [
  [               0,               0,               0,               0,     0.078884755,    -0.057548263 ],
  [               0,               0,               0,               0,       5.5313897,     -0.69215977 ],
  [               0,               0,               0,               0,      -3.5029134,       1.4211556 ],
  [               0,               0,               0,               0,       5.5447605,      -4.3994558 ],
  [               0,               0,               0,               0,     0.013468393,   -0.0095614028 ],
  [               0,               0,               0,               0,      -22.484818,       18.976163 ],
  [               0,               0,               0,               0,      -49.615238,       25.447552 ],
 ],
 [
  [               0,               0,               0,   0.00014832653,   0.00017920297,    0.0084047626 ],
  [               0,               0,               0,        1.066629,      -3.2315754,       5.1077222 ],
  [               0,               0,               0,      0.46333638,      -1.2530204,     -0.60420268 ],
  [               0,               0,               0,       2.3552031,       2.9714561,      -2.5609867 ],
  [               0,               0,               0,      0.02788323,    0.0093208041,    -0.013998484 ],
  [               0,               0,               0,     -0.12067163,      0.24202954,    -0.054592361 ],
  [               0,               0,               0,      0.21616942,       -0.565578,      -2.7511535 ],
 ],
 [
  [               0,               0,    -0.014468591,    -0.026121258,   -0.0025471036,      0.10081067 ],
  [               0,               0,       425.36354,      -585.60846,      -812.44841,       1120.6437 ],
  [               0,               0,      -131.91597,       188.21366,       237.39918,       -340.5468 ],
  [               0,               0,      -40.872333,       53.033832,       72.798871,      -93.998169 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.25342962,   0.00044719826,       -0.495511 ],
  [               0,               0,               0,       5.7686025,        8.337925,       16.050125 ],
  [               0,               0,               0,      -1623.3269,       4557.5229,      -3202.6659 ],
  [               0,               0,               0,      -2.1267675,       3.7522917,    -0.032684474 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      0.80729139,      -1.1050574,      -1.8028803,       2.4773547 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,    -0.005517198,   -0.0018575843,    0.0034084963,    0.0094419095,     0.013993477 ],
  [               0,        157.3123,      -99.019151,        -338.164,      -285.11354,       731.26948 ],
  [               0,       3.9210888,      -4.9952944,      -11.662878,      -6.3937702,       31.414664 ],
  [               0,       3.1705816,      -1.4038513,      -5.5313625,      -5.0847554,       8.8574322 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0080926386,   -0.0057425564 ],
  [               0,               0,               0,               0,       54.356833,      -79.875638 ],
  [               0,               0,               0,               0,      -10.008495,       14.009867 ],
  [               0,               0,               0,               0,       3.2370462,      -4.8827446 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.065273258,      0.20794981,     -0.15610093 ],
  [               0,               0,               0,       16.594242,      -54.212231,       46.817226 ],
  [               0,               0,               0,     -0.73922312,       2.3138214,       -2.970579 ],
  [               0,               0,               0,      0.18070759,     -0.52454321,      0.36294505 ],
  [               0,               0,               0,      0.02143257,    -0.069103658,     0.055628893 ],
  [               0,               0,               0,       24.642338,      -91.149057,       83.292941 ],
  [               0,               0,               0,       4.2212456,      -15.724314,       9.8199276 ],
 ],
 [
  [               0,               0,               0,     0.021974619,    -0.084847166,     0.091080663 ],
  [               0,               0,               0,       -1.365873,       3.4973539,      0.94906866 ],
  [               0,               0,               0,      0.96383984,      -3.6341043,       2.2231749 ],
  [               0,               0,               0,      -8.5871765,       35.634964,      -36.846006 ],
  [               0,               0,               0,    0.0066134068,    -0.025636151,     0.027207682 ],
  [               0,               0,               0,     0.085909763,     -0.36997017,      0.43552314 ],
  [               0,               0,               0,       33.081324,      -137.36857,       140.65384 ],
 ],
 [
  [               0,               0,               0,    -0.093505965,      0.40675662,     -0.42714333 ],
  [               0,               0,               0,       58.702929,      -275.57543,       325.31277 ],
  [               0,               0,               0,      -5.6567642,       25.674833,      -30.176344 ],
  [               0,               0,               0,       21.960221,      -101.57331,        117.4518 ],
  [               0,               0,               0,     0.028520199,     -0.16147677,      0.22152864 ],
  [               0,               0,               0,       18.642921,      -89.785123,       107.92867 ],
  [               0,               0,               0,      -94.810812,       467.27334,      -575.88704 ],
 ],
 [
  [               0,               0,               0,               0,     -0.29525964,      0.73430603 ],
  [               0,               0,               0,               0,     -0.13087068,       2.5655291 ],
  [               0,               0,               0,               0,      -42.716491,       103.93271 ],
  [               0,               0,               0,               0,       58.669677,      -144.09497 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00040242621,    0.0051062681,   -0.0098453817,    -0.013070216 ],
  [               0,               0,      -1.1863674,      -1.3858478,      -3.4335284,      -4.5064851 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      -265.62346,      -208.55854,       559.48425,       4046.0028 ],
  [               0,               0,       1.2003408,        2.767867,       24.286219,      -6.9427097 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.023714942,     0.013539177,    0.0056167877 ],
  [               0,               0,               0,       7.1895351,      -4.5538575,       5.0936753 ],
  [               0,               0,               0,      -1.4719288,      0.55190175,      -1.3864526 ],
  [               0,               0,               0,       1.2668403,     -0.39057434,    0.0076913864 ],
  [               0,               0,               0,    0.0048906086,   -0.0028864378,   0.00092201564 ],
  [               0,               0,               0,       11.757295,      -5.6915987,        1.122668 ],
  [               0,               0,               0,      -20.099263,       7.2669833,      -1.9009011 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0070485119 ],
  [               0,               0,               0,               0,               0,       4.4261691 ],
  [               0,               0,               0,               0,               0,      -1.4244986 ],
  [               0,               0,               0,               0,               0,     0.069159929 ],
  [               0,               0,               0,               0,               0,    0.0006307083 ],
  [               0,               0,               0,               0,               0,      0.60624507 ],
  [               0,               0,               0,               0,               0,      -8.1631126 ],
 ],
 [
  [               0,               0,               0,               0,    -0.064854625,      0.06008381 ],
  [               0,               0,               0,               0,       35.836165,      -24.816827 ],
  [               0,               0,               0,               0,       1.0205429,       -2.232225 ],
  [               0,               0,               0,               0,      -15.146745,        12.52132 ],
  [               0,               0,               0,               0,     -0.11415432,     0.094770144 ],
  [               0,               0,               0,               0,       14.674639,      -11.500819 ],
  [               0,               0,               0,               0,       15.166311,      -27.567667 ],
 ],
 [
  [               0,               0,               0,   -0.0033677885,    0.0089502325,    0.0021794526 ],
  [               0,               0,               0,       2.6213071,      -7.1206107,       8.2327385 ],
  [               0,               0,               0,     0.020407448,     -0.28458591,      -1.1846509 ],
  [               0,               0,               0,      0.26980643,    -0.010060825,      -0.1120248 ],
  [               0,               0,               0,    0.0045669184,   -0.0077829989,    0.0038982807 ],
  [               0,               0,               0,       0.4083218,      -1.3743667,       1.4785896 ],
  [               0,               0,               0,       1.9679912,      -5.9817172,       1.6256285 ],
 ],
 [
  [               0,               0,       0.1559957,    -0.033553948,      -0.1689458,    -0.095114417 ],
  [               0,               0,      -2416.7239,       3417.5736,       4300.9151,      -6087.6556 ],
  [               0,               0,      -1178.2541,       1635.3295,       2217.4808,       -3079.414 ],
  [               0,               0,       698.43033,      -956.08353,      -1309.5546,       1792.8235 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0012757071,     0.002552462,     0.014413445 ],
  [               0,               0,               0,      -342.46004,       4.5952915,       708.39063 ],
  [               0,               0,               0,        51.83884,      -150.63906,       108.12885 ],
  [               0,               0,               0,     0.069591443,     -0.19588916,     -0.36389981 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -1.2935518,       1.8771259,       2.5070115,      -3.6390488 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,      0.04246024,   -0.0082324247,    -0.059112787,    -0.068210651,     0.062651733 ],
  [               0,      -45.482648,       14.650569,       134.42457,       166.37132,      -367.45682 ],
  [               0,        387.0526,      -282.72093,      -847.59425,      -634.52412,       1853.4975 ],
  [               0,      -363.54508,       269.20638,       789.53653,       578.55728,      -1713.7109 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0022489385,     0.003938741 ],
  [               0,               0,               0,               0,       70.902063,      -104.44699 ],
  [               0,               0,               0,               0,        3.441608,       -6.462909 ],
  [               0,               0,               0,               0,       -1.651545,       2.5609995 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00033134942,   0.00077456646,    0.0069509334 ],
  [               0,               0,               0,       4.7462402,      -16.527549,       17.672875 ],
  [               0,               0,               0,     -0.90191929,       3.0350342,      -3.7869164 ],
  [               0,               0,               0,      0.25906536,     -0.81839425,      0.64249367 ],
  [               0,               0,               0,  -0.00018304081,   0.00071159445,  -0.00041781415 ],
  [               0,               0,               0,       1.1684363,      -4.3135278,       5.6657864 ],
  [               0,               0,               0,         5.37836,        -19.2058,       12.173828 ],
 ],
 [
  [               0,               0,               0,     0.014630357,    -0.055746358,     0.061502873 ],
  [               0,               0,               0,      -1.8933714,       5.3035153,     0.083144056 ],
  [               0,               0,               0,      0.63653117,      -2.3305609,      0.85193251 ],
  [               0,               0,               0,      -55.028458,       213.03737,       -197.6918 ],
  [               0,               0,               0,   -0.0016898722,   -0.0056330334,      0.10506653 ],
  [               0,               0,               0,    -0.012472405,      0.04837466,    -0.042933282 ],
  [               0,               0,               0,       6.8444996,      -26.838299,       23.788012 ],
 ],
 [
  [               0,               0,               0,    -0.025447912,     0.082736021,    -0.043275501 ],
  [               0,               0,               0,       79.192262,       -369.8343,       434.23097 ],
  [               0,               0,               0,       6.6600065,      -32.692436,       38.864443 ],
  [               0,               0,               0,     -0.17640976,       33.542355,      -77.131093 ],
  [               0,               0,               0,     0.005910664,    -0.029060358,     0.062926997 ],
  [               0,               0,               0,     -0.96821738,       4.3692953,      -4.9274315 ],
  [               0,               0,               0,      -15.273278,        83.94345,      -111.07827 ],
 ],
 [
  [               0,               0,               0,               0,     -0.14755897,      0.36775531 ],
  [               0,               0,               0,               0,       28.049509,      -66.256188 ],
  [               0,               0,               0,               0,      -148.06945,       363.58357 ],
  [               0,               0,               0,               0,       155.02845,      -381.61641 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0010760985,   0.00050789463,   -0.0025698271,    -0.013123556 ],
  [               0,               0,      -1.1907423,      -1.3865718,      -3.4354279,      -4.5116535 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       379.78336,       319.02358,      -763.59502,      -5667.5093 ],
  [               0,               0,       3.6442934,       7.4910329,       4.0608698,       16.419649 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.022666969,     0.013129807,    0.0056503476 ],
  [               0,               0,               0,       6.7443365,      -4.3708512,       5.0765021 ],
  [               0,               0,               0,      -1.4299461,      0.53097128,       -1.384155 ],
  [               0,               0,               0,        1.317074,     -0.40520478,    0.0079353593 ],
  [               0,               0,               0,    0.0043026132,   -0.0025949195,    0.0009171595 ],
  [               0,               0,               0,       11.478558,      -5.6292776,       1.1141536 ],
  [               0,               0,               0,      -19.918402,       7.1669044,      -1.8939941 ],
 ],
 [
  [               0,               0,               0,   -0.0025123863,    0.0026829316,    0.0068677814 ],
  [               0,               0,               0,      -3.9761237,        3.692505,       3.5053192 ],
  [               0,               0,               0,     -0.85847503,      0.70130062,      -1.5078782 ],
  [               0,               0,               0,      -10.747297,       15.605358,      -3.4765111 ],
  [               0,               0,               0,    -0.080635446,      0.11047624,    -0.021554247 ],
  [               0,               0,               0,      0.29839441,     -0.22997312,     0.077996048 ],
  [               0,               0,               0,       3.3554019,      -5.0847666,      -7.2917384 ],
 ],
 [
  [               0,               0,               0,               0,     0.060135839,    -0.037485347 ],
  [               0,               0,               0,               0,      -1.1859413,        4.014652 ],
  [               0,               0,               0,               0,       4.4427927,      -4.8389193 ],
  [               0,               0,               0,               0,      -2.9266928,       2.3463882 ],
  [               0,               0,               0,               0,   -0.0057254613,    0.0044334074 ],
  [               0,               0,               0,               0,      -35.551372,       31.655813 ],
  [               0,               0,               0,               0,       14.640919,      -28.656213 ],
 ],
 [
  [               0,               0,               0,   -0.0039437214,    0.0094358326,    0.0026054891 ],
  [               0,               0,               0,       2.4855705,      -6.6093875,       7.7723451 ],
  [               0,               0,               0,      0.13087038,     -0.58635569,     -0.97469812 ],
  [               0,               0,               0,     -0.74228188,      0.15679162,       1.7360045 ],
  [               0,               0,               0,   -0.0070259963,    0.0014584518,     0.014384861 ],
  [               0,               0,               0,      0.19799571,     -0.10849723,      0.01300866 ],
  [               0,               0,               0,        1.503056,      -3.4513978,       -1.000723 ],
 ],
 [
  [               0,               0,      0.14485351,     -0.02261281,     -0.15084827,     -0.11209332 ],
  [               0,               0,      -3609.0992,       5108.4855,       6380.9737,      -9044.6773 ],
  [               0,               0,      -1251.3404,       1736.3511,       2357.6487,      -3273.1418 ],
  [               0,               0,       715.94541,       -980.6391,      -1342.3646,        1838.825 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00028834861,    0.0026996543,     0.010723531 ],
  [               0,               0,               0,      -354.74967,       4.4508095,       732.73124 ],
  [               0,               0,               0,       44.622859,      -129.80124,       93.119655 ],
  [               0,               0,               0,     0.053809203,     -0.19292844,     -0.37292642 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele100large_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -1.2329538,       1.7925303,       2.3846992,      -3.4686076 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.076506609,    -0.018568891,     -0.11112409,     -0.12187975,      0.13086023 ],
  [               0,        41.05186,      -59.289106,       -45.04273,        62.52687,       3.3887266 ],
  [               0,       387.75192,      -283.77841,      -849.62176,      -635.08338,       1858.9458 ],
  [               0,      -355.28363,       263.34013,        771.5506,       564.75391,      -1674.5435 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.15488152,     -0.22610767 ],
  [               0,               0,               0,               0,       54.866093,      -80.139317 ],
  [               0,               0,               0,               0,       25.157656,       -39.38315 ],
  [               0,               0,               0,               0,      -18.348953,       27.812207 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00024249721,   0.00083667283,    0.0072332739 ],
  [               0,               0,               0,       4.6421633,      -16.099022,       17.223589 ],
  [               0,               0,               0,     -0.92133268,       3.0994488,      -3.8359043 ],
  [               0,               0,               0,      0.26524344,     -0.84257338,      0.66310514 ],
  [               0,               0,               0,  -0.00011072751,   0.00044242484,  -0.00015523583 ],
  [               0,               0,               0,      0.62038515,       -2.337876,       3.8874928 ],
  [               0,               0,               0,       4.9985357,      -17.853984,       10.976448 ],
 ],
 [
  [               0,               0,               0,    0.0072425368,    -0.025229831,     0.030151223 ],
  [               0,               0,               0,     -0.46531427,     -0.56232162,       6.0741844 ],
  [               0,               0,               0,      0.50157695,      -1.7747368,      0.28299993 ],
  [               0,               0,               0,      -21.141158,       77.792088,      -67.034912 ],
  [               0,               0,               0,   -0.0079186093,   -0.0043553956,     0.078830086 ],
  [               0,               0,               0,    -0.022309377,     0.083251811,    -0.065676032 ],
  [               0,               0,               0,       9.4009887,      -36.907776,       33.541011 ],
 ],
 [
  [               0,               0,               0,   -0.0071096677,   -0.0050635389,     0.062446732 ],
  [               0,               0,               0,        77.52983,      -362.78739,       426.75804 ],
  [               0,               0,               0,      -1.0500245,       4.0371341,      -4.8494991 ],
  [               0,               0,               0,      -14.749727,       46.892946,      -41.727298 ],
  [               0,               0,               0,    -0.068905552,   -0.0026542352,      0.56165699 ],
  [               0,               0,               0,     -0.37158021,       1.7900851,      -2.1525761 ],
  [               0,               0,               0,      -9.8270491,       46.528626,      -55.646975 ],
 ],
 [
  [               0,               0,               0,               0,     -0.20017585,      0.49854481 ],
  [               0,               0,               0,               0,       21.246406,      -49.686447 ],
  [               0,               0,               0,               0,      -98.364729,       241.18388 ],
  [               0,               0,               0,               0,        106.0134,      -260.94576 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.031820891,    -0.047451877,    -0.060988747,    -0.041530986 ],
  [               0,               0,      -1.1584801,      -1.4141155,      -3.4939375,      -4.5646534 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      0.58536576,      0.78343598,      0.42852575,      -3.8593558 ],
  [               0,               0,     0.073732289,      0.24597371,      0.75757276,       1.6911635 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 200 GeV electron clusters
#


ele200_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00020864523,   0.00090255945 ],
  [               0,               0,               0,               0,       1.6353821,       2.5238522 ],
  [               0,               0,               0,               0,     -0.56662908,     -0.69527266 ],
  [               0,               0,               0,               0,     -0.94945443,     -0.69663845 ],
  [               0,               0,               0,               0,     -0.20327235,   -0.0082676933 ],
  [               0,               0,               0,               0,  -0.00014014948,    5.990219e-05 ],
  [               0,               0,               0,               0,       2.8372791,       2.0616771 ],
  [               0,               0,               0,               0,      0.60019296,      -1.8112601 ],
 ],
 [
  [               0,               0,               0,    0.0014274318,   -0.0024032602,    0.0018389176 ],
  [               0,               0,               0,       -29.80205,       30.712366,      -4.6407743 ],
  [               0,               0,               0,       1.8818555,     -0.53192378,      -1.1009322 ],
  [               0,               0,               0,       15.222616,      -14.648978,       2.3804592 ],
  [               0,               0,               0,       1.7628092,      -2.1220592,      0.49441234 ],
  [               0,               0,               0,   -0.0029155176,    0.0028993313,  -0.00071807447 ],
  [               0,               0,               0,      -5.2544969,       3.5163171,       1.9441898 ],
  [               0,               0,               0,      -8.8721967,       14.802017,      -5.0226697 ],
 ],
 [
  [               0,               0,               0,               0,     0.024850161,    0.0020201405 ],
  [               0,               0,               0,               0,      -11.190721,       9.5248352 ],
  [               0,               0,               0,               0,      -19.057572,       14.592507 ],
  [               0,               0,               0,               0,       20.500554,      -16.298031 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.027755964,     0.015583658,     0.085180306 ],
  [               0,               0,               0,      0.24691528,      0.28628281,      -1.1097987 ],
  [               0,               0,               0,     -0.16954514,     -0.36167832,       1.0501496 ],
  [               0,               0,               0,      -0.3152413,    -0.024115278,      0.85055717 ],
  [               0,               0,               0,    -0.046868814,      0.10697675,    -0.063194157 ],
  [               0,               0,               0,   -0.0015698649,   0.00096222977,    0.0029039781 ],
  [               0,               0,               0,      0.62282439,      -1.7870809,       2.0568453 ],
  [               0,               0,               0,       1.0381172,      -2.4983584,      -0.1206729 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       0.2687012,     -0.44661359,      -0.5972888,       0.9902349 ],
  [               0,               0,       521.00823,      -748.40041,      -1022.6304,       1473.7667 ],
  [               0,               0,      -766.25734,       1179.2685,       1166.0508,      -1838.7063 ],
  [               0,               0,       354.02187,      -414.29531,      -1149.0971,       1451.3867 ],
  [               0,               0,       2725.5051,      -3883.7608,      -5538.3501,        7890.272 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,      0.06613342,    -0.094017078 ],
  [               0,               0,               0,               0,       1415.4547,      -2017.9185 ],
  [               0,               0,               0,               0,       485.00979,      -699.66285 ],
  [               0,               0,               0,               0,      -782.38846,       1130.4458 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.006943056,     0.010532226 ],
  [               0,               0,               0,               0,      -174.16566,       257.04443 ],
  [               0,               0,               0,               0,       90.488724,       -133.6394 ],
  [               0,               0,               0,               0,      -96.908045,       139.37293 ],
  [               0,               0,               0,               0,      0.20933732,     -0.34965299 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.12416994,      0.18762497 ],
  [               0,               0,               0,               0,      -259.93743,       393.77266 ],
  [               0,               0,               0,               0,      -167.02063,       247.31743 ],
  [               0,               0,               0,               0,       245.16941,      -364.20317 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00039841988,   0.00061285217,    0.0033595442,     0.014674976 ],
  [               0,               0,      -4.6089935,       23.740378,      -41.068185,        24.55091 ],
  [               0,               0,      -1.7914753,        8.340711,      -12.526884,       5.3713493 ],
  [               0,               0,      -1.2819912,       5.5389751,      -7.4765722,       2.3820578 ],
  [               0,               0,     -0.12327923,      0.41531504,     -0.37395755,     0.041832516 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0059512481,   -0.0093983413 ],
  [               0,               0,               0,               0,      -5.9945699,       14.074904 ],
  [               0,               0,               0,               0,     -0.30012973,     -0.56284266 ],
  [               0,               0,               0,               0,    -0.066161609,     -0.94981327 ],
  [               0,               0,               0,               0,    -0.072817448,      0.12970167 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00048655536,  -0.00030580686,    0.0010893308,    0.0070926812 ],
  [               0,               0,      0.80533276,     -0.80187515,       -4.667598,       8.1389576 ],
  [               0,               0,       1.6137796,      -3.4596731,      -7.1407258,       14.155969 ],
  [               0,               0,    -0.085902593,      0.79080527,      -1.7172487,     -0.16356961 ],
  [               0,               0,      -0.3434208,      0.55681519,       2.3792587,      -4.2628803 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00093717491,    0.0035425817 ],
  [               0,               0,               0,               0,       -96.90763,       224.44682 ],
  [               0,               0,               0,               0,       32.234304,      -75.324753 ],
  [               0,               0,               0,               0,       2.0298814,      -4.4299004 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0008052123,    0.0024536398 ],
  [               0,               0,               0,               0,     -0.72489538,       3.7265355 ],
  [               0,               0,               0,               0,     -0.47358652,    -0.017298114 ],
  [               0,               0,               0,               0,      0.94685561,      -3.3021917 ],
  [               0,               0,               0,               0,      0.11124329,     -0.24310117 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.06037488,       0.1559068 ],
  [               0,               0,               0,               0,       45.924812,      -105.94601 ],
  [               0,               0,               0,               0,      -5.1027039,        10.91987 ],
  [               0,               0,               0,               0,        4.162498,      -10.073524 ],
  [               0,               0,               0,               0,    -0.015665142,     0.037932419 ],
  [               0,               0,               0,               0,       11.178275,      -24.955026 ],
  [               0,               0,               0,               0,       22.698929,      -57.572496 ],
 ],
 [
  [               0,               0,               0,               0,               0,      0.02677002 ],
  [               0,               0,               0,               0,               0,       10.992314 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.50372363 ],
  [               0,               0,               0,               0,               0,      0.97992197 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0001843055,   0.00090265165 ],
  [               0,               0,               0,               0,       1.4776585,       2.6185664 ],
  [               0,               0,               0,               0,      -0.4793838,     -0.72972994 ],
  [               0,               0,               0,               0,     -0.84677991,     -0.73738634 ],
  [               0,               0,               0,               0,     -0.20317302,   -0.0089402526 ],
  [               0,               0,               0,               0,  -0.00012383968,   5.2898799e-05 ],
  [               0,               0,               0,               0,        3.227192,       2.0367207 ],
  [               0,               0,               0,               0,      0.58567175,      -1.8248244 ],
 ],
 [
  [               0,               0,               0,    0.0024079586,    -0.003253841,    0.0020066967 ],
  [               0,               0,               0,       -25.75441,       26.389469,      -3.4728539 ],
  [               0,               0,               0,      -0.7171008,       2.0328419,      -1.7251051 ],
  [               0,               0,               0,       11.739026,      -11.154595,        1.512967 ],
  [               0,               0,               0,       1.8391847,      -2.1820832,      0.50398168 ],
  [               0,               0,               0,   -0.0025977024,    0.0025933898,  -0.00064555539 ],
  [               0,               0,               0,      -4.4949216,       2.5405406,       2.2544098 ],
  [               0,               0,               0,      -9.7209642,       16.187175,      -5.6493554 ],
 ],
 [
  [               0,               0,               0,               0,     0.025651984,    0.0027425007 ],
  [               0,               0,               0,               0,      -10.836574,       9.2286483 ],
  [               0,               0,               0,               0,      -17.614878,       13.456016 ],
  [               0,               0,               0,               0,        19.29577,      -15.340234 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.025131671,     0.013194496,     0.079767981 ],
  [               0,               0,               0,      0.19441352,      0.47390064,      -1.3549132 ],
  [               0,               0,               0,      -0.1176043,     -0.49922553,       1.2107687 ],
  [               0,               0,               0,     -0.27082136,     -0.14400901,       1.0003175 ],
  [               0,               0,               0,    -0.049016106,      0.11213807,    -0.066476867 ],
  [               0,               0,               0,   -0.0018580195,   0.00075031139,    0.0041579033 ],
  [               0,               0,               0,      0.46033008,      -1.5151935,       1.9436765 ],
  [               0,               0,               0,      0.74642203,      -1.8071779,     -0.51701022 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.13404056,     -0.28406105,     -0.34183569,      0.68753871 ],
  [               0,               0,       466.47689,       -672.6641,      -886.63453,       1283.4377 ],
  [               0,               0,      -738.69598,        1147.647,       1087.3027,      -1742.0323 ],
  [               0,               0,       275.26611,      -306.36766,      -984.85559,       1226.0979 ],
  [               0,               0,       2753.5701,      -3923.2981,      -5596.7172,       7972.5311 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,     0.017064497,    -0.022713069 ],
  [               0,               0,               0,               0,      -16.634275,      -15.646365 ],
  [               0,               0,               0,               0,      -112.97413,       164.75246 ],
  [               0,               0,               0,               0,      -58.183906,       83.924472 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0058992969,     0.019152829 ],
  [               0,               0,               0,               0,      -146.81665,       216.80004 ],
  [               0,               0,               0,               0,       85.180617,      -125.89506 ],
  [               0,               0,               0,               0,      -42.184868,       59.645422 ],
  [               0,               0,               0,               0,      0.17656626,     -0.28602248 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.11500272,      0.17328085 ],
  [               0,               0,               0,               0,      -56.539245,       92.798625 ],
  [               0,               0,               0,               0,       -319.4779,       474.16604 ],
  [               0,               0,               0,               0,       396.28566,      -588.83132 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00037893212,    0.0005673667,    0.0033223275,     0.014794179 ],
  [               0,               0,      -4.6391695,       23.887787,      -41.305274,       24.678179 ],
  [               0,               0,      -1.8034223,       8.4010317,      -12.630114,       5.4298319 ],
  [               0,               0,      -1.2890403,       5.5761475,      -7.5438204,       2.4222138 ],
  [               0,               0,     -0.12450521,      0.42045245,     -0.38100246,     0.044986481 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0059485809,   -0.0093944417 ],
  [               0,               0,               0,               0,        -6.00012,       14.086637 ],
  [               0,               0,               0,               0,     -0.30102012,     -0.56110655 ],
  [               0,               0,               0,               0,    -0.067001633,     -0.94822663 ],
  [               0,               0,               0,               0,    -0.072833034,      0.12970608 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00049484858,   -0.0003133935,     0.001102688,    0.0072025317 ],
  [               0,               0,      0.86708515,     -0.91823217,      -4.9870132,       8.7382852 ],
  [               0,               0,        1.607489,      -3.4484083,      -7.1060462,       14.092917 ],
  [               0,               0,    -0.093702611,      0.81114717,       -1.707306,     -0.20111241 ],
  [               0,               0,     -0.34451798,      0.55740791,       2.3900114,       -4.277836 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00091413843,    0.0035965199 ],
  [               0,               0,               0,               0,      -96.884418,       224.39326 ],
  [               0,               0,               0,               0,       32.248545,       -75.35744 ],
  [               0,               0,               0,               0,       2.0270112,      -4.4233365 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.029942317,    -0.065871923 ],
  [               0,               0,               0,               0,      -7.5581992,       19.781874 ],
  [               0,               0,               0,               0,      0.42190127,      -2.1184048 ],
  [               0,               0,               0,               0,      0.97753919,       -3.372493 ],
  [               0,               0,               0,               0,    -0.090677769,      0.23146721 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.05924766,       0.1531093 ],
  [               0,               0,               0,               0,       45.125382,      -103.94891 ],
  [               0,               0,               0,               0,      -5.4391827,       11.749958 ],
  [               0,               0,               0,               0,       4.2796066,      -10.361979 ],
  [               0,               0,               0,               0,    -0.014570404,     0.035234865 ],
  [               0,               0,               0,               0,       10.496353,      -23.263974 ],
  [               0,               0,               0,               0,       25.584587,         -64.683 ],
 ],
 [
  [               0,               0,               0,               0,               0,      0.02677002 ],
  [               0,               0,               0,               0,               0,       10.992314 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.50372363 ],
  [               0,               0,               0,               0,               0,      0.97992197 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0001851758,   0.00087655898 ],
  [               0,               0,               0,               0,       1.5223844,       2.6623851 ],
  [               0,               0,               0,               0,      -0.4727287,     -0.71766836 ],
  [               0,               0,               0,               0,     -0.89645683,     -0.71379746 ],
  [               0,               0,               0,               0,     -0.21708346,   -0.0076440058 ],
  [               0,               0,               0,               0,  -0.00012682336,   5.2928175e-05 ],
  [               0,               0,               0,               0,       3.1286952,       2.0387093 ],
  [               0,               0,               0,               0,      0.78322312,      -1.8507904 ],
 ],
 [
  [               0,               0,               0,    0.0078199428,    -0.003164944,   0.00065749797 ],
  [               0,               0,               0,      -4.8106598,      -5.6859323,       7.2578628 ],
  [               0,               0,               0,       19.326719,      -20.576724,       4.5208251 ],
  [               0,               0,               0,        23.16345,      -24.091865,       5.1062679 ],
  [               0,               0,               0,      0.30281823,    -0.035163039,     -0.17791943 ],
  [               0,               0,               0,   -0.0040880955,    0.0040792032,   -0.0010125023 ],
  [               0,               0,               0,      -4.5047882,     0.034716268,       4.0712402 ],
  [               0,               0,               0,      -22.379344,       32.295411,      -10.375468 ],
 ],
 [
  [               0,               0,               0,               0,     0.024439723,     0.002164008 ],
  [               0,               0,               0,               0,      -11.105135,       9.4596282 ],
  [               0,               0,               0,               0,      -19.289329,       14.774789 ],
  [               0,               0,               0,               0,       20.659262,      -16.424256 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00091657715,   -0.0025814253,     0.001867433 ],
  [               0,               0,               0,       14.249614,      -40.375978,       25.670942 ],
  [               0,               0,               0,      -6.5553087,       28.682466,       -17.90629 ],
  [               0,               0,               0,       22.707604,      -26.891141,       4.5219466 ],
  [               0,               0,               0,      -3.8270027,       5.5880045,      -3.1583251 ],
  [               0,               0,               0,   5.3199354e-05,  -2.2570352e-05,   0.00037965076 ],
  [               0,               0,               0,     -0.90862452,       1.9773862,      0.17009983 ],
  [               0,               0,               0,       1.2998141,      -3.9877663,       1.1786728 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.14531019,     -0.30058836,     -0.36529973,      0.72191214 ],
  [               0,               0,       552.35433,      -795.21719,      -1061.4698,       1532.9075 ],
  [               0,               0,      -757.79404,       1174.9913,       1125.8748,      -1797.2502 ],
  [               0,               0,       292.99053,      -331.46084,      -1021.1112,       1277.4228 ],
  [               0,               0,       2769.6566,      -3946.3249,      -5629.3169,       8019.1894 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,     0.021760725,    -0.031952173 ],
  [               0,               0,               0,               0,        1918.844,      -2749.8801 ],
  [               0,               0,               0,               0,      -103.60451,       151.95391 ],
  [               0,               0,               0,               0,       481.26824,      -695.05552 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       1.1890499,      -1.7001746 ],
  [               0,               0,               0,               0,      -119.15086,       176.10161 ],
  [               0,               0,               0,               0,       78.552269,      -116.20884 ],
  [               0,               0,               0,               0,      -24.018827,       33.321901 ],
  [               0,               0,               0,               0,      0.18823688,     -0.29468348 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.1155373,      0.17411184 ],
  [               0,               0,               0,               0,       -91.76976,       144.91559 ],
  [               0,               0,               0,               0,       -318.3515,       472.43653 ],
  [               0,               0,               0,               0,       391.78238,       -582.1164 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.44269151,     -0.78527635,      -1.2272614,       2.2106084 ],
  [               0,               0,      -8.5272416,       37.611903,      -53.836859,       25.353516 ],
  [               0,               0,      -1.2990122,       10.248633,      -23.578518,       16.275087 ],
  [               0,               0,     -0.21418521,        4.646125,      -13.982847,       10.824222 ],
  [               0,               0,     -0.18211215,      0.76800028,      -1.0532047,      0.46591211 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0062358615,   -0.0098213004 ],
  [               0,               0,               0,               0,      -5.5650572,       13.175294 ],
  [               0,               0,               0,               0,     -0.20495255,     -0.74890935 ],
  [               0,               0,               0,               0,    0.0058837176,      -1.1075969 ],
  [               0,               0,               0,               0,    -0.056017717,     0.093727172 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00056770018,  -0.00029578259,    0.0013078626,    0.0075474302 ],
  [               0,               0,       1.1074379,      -1.3251547,      -6.4177727,       11.240531 ],
  [               0,               0,       1.6642866,      -3.5340026,      -7.5307875,       14.842075 ],
  [               0,               0,     -0.14479932,      0.95581592,      -1.7295668,     -0.31819311 ],
  [               0,               0,     -0.35457958,       0.5732886,       2.4409235,      -4.3646848 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00094141287,    0.0036180245 ],
  [               0,               0,               0,               0,      -94.942301,       219.91337 ],
  [               0,               0,               0,               0,       32.287977,      -75.445853 ],
  [               0,               0,               0,               0,       1.9776093,       -4.311566 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00087700089,    0.0026506057 ],
  [               0,               0,               0,               0,     -0.92273949,       4.1053641 ],
  [               0,               0,               0,               0,     -0.22320772,     -0.58825067 ],
  [               0,               0,               0,               0,      0.87553803,      -3.1343849 ],
  [               0,               0,               0,               0,     0.059599055,     -0.12693973 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.062515857,       0.1611128 ],
  [               0,               0,               0,               0,       46.467329,      -107.31741 ],
  [               0,               0,               0,               0,      -5.3447193,       11.529653 ],
  [               0,               0,               0,               0,       4.2792293,      -10.365238 ],
  [               0,               0,               0,               0,    -0.015274278,     0.036936079 ],
  [               0,               0,               0,               0,        11.82008,      -26.460948 ],
  [               0,               0,               0,               0,       24.565405,      -62.198002 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.027466048 ],
  [               0,               0,               0,               0,               0,       11.054272 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,      -0.4950241 ],
  [               0,               0,               0,               0,               0,      0.98932142 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.032701463,     0.015772403,    0.0062147696 ],
  [               0,               0,               0,       10.682942,      -5.2763524,       4.4317744 ],
  [               0,               0,               0,      0.34301953,     -0.45737208,      -1.2049569 ],
  [               0,               0,               0,       1.1646112,     -0.25731378,    -0.051433393 ],
  [               0,               0,               0,   -0.0015402994,  -4.2222642e-05,    0.0005180915 ],
  [               0,               0,               0,       14.942702,      -7.5313717,       1.2610508 ],
  [               0,               0,               0,      -38.106248,        13.35779,      -1.5494964 ],
 ],
 [
  [               0,               0,               0,   -0.0003798032,   0.00091306231,    0.0082153954 ],
  [               0,               0,               0,      -5.2110163,       5.0678655,       2.3561442 ],
  [               0,               0,               0,      -1.5355173,       1.4489577,      -1.6653956 ],
  [               0,               0,               0,       455.12865,       39.284362,      -62.767283 ],
  [               0,               0,               0,        4.133056,      0.14493106,     -0.49976487 ],
  [               0,               0,               0,     0.071700021,    -0.078750569,     0.028219006 ],
  [               0,               0,               0,      0.14923922,       -0.220671,      -9.3381803 ],
 ],
 [
  [               0,               0,               0,               0,     0.029790879,    -0.019320729 ],
  [               0,               0,               0,               0,       9.1500818,      -2.8641746 ],
  [               0,               0,               0,               0,       2.8840659,      -3.6449493 ],
  [               0,               0,               0,               0,      -3.6558081,       2.8903283 ],
  [               0,               0,               0,               0,    0.0038472166,   -0.0017717166 ],
  [               0,               0,               0,               0,     -0.39525136,       1.3625501 ],
  [               0,               0,               0,               0,      -8.2487675,       5.0329012 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00015245785,     0.008823566 ],
  [               0,               0,               0,               0,     -0.74473748,       3.9212499 ],
  [               0,               0,               0,               0,     -0.36432585,      -1.0239109 ],
  [               0,               0,               0,               0,       6.7662763,       942.49677 ],
  [               0,               0,               0,               0,    -0.088824552,       8.3193042 ],
  [               0,               0,               0,               0,    0.0052069499,   -0.0017851401 ],
  [               0,               0,               0,               0,   0.00032771039,      -3.1399216 ],
 ],
 [
  [               0,               0,    -0.017198493,    -0.023052426,    0.0021717959,     0.095837521 ],
  [               0,               0,       389.56328,      -536.44937,      -740.44913,       1021.9235 ],
  [               0,               0,      -109.95086,       156.59595,       199.35271,      -285.56086 ],
  [               0,               0,      -49.337975,       64.552835,       89.394222,      -116.61711 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      -0.2966163,    -0.022372629,      0.60827888 ],
  [               0,               0,               0,      -7.0792524,      -10.205991,      -10.715572 ],
  [               0,               0,               0,      -1051.6672,       2996.2326,      -2130.7457 ],
  [               0,               0,               0,      -1.5415898,       1.9974035,     -0.52552978 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      0.36138984,     -0.46331169,     -0.87878114,       1.1473578 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.024721833,   -0.0086053695,    -0.039250371 ],
  [               0,               0,               0,       1063.6504,      -3105.7354,       2270.3172 ],
  [               0,               0,               0,       1157.3394,      -3500.7045,       2644.9755 ],
  [               0,               0,               0,       378.35705,      -1080.4675,       770.10089 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0020605405,    0.0058664783 ],
  [               0,               0,               0,               0,       163.03349,      -242.97915 ],
  [               0,               0,               0,               0,      -137.58237,       206.90147 ],
  [               0,               0,               0,               0,       126.14395,      -190.70787 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0014184501,  -0.00020703897,    0.0061371087 ],
  [               0,               0,               0,       4.9755521,      -17.983934,       18.822312 ],
  [               0,               0,               0,     -0.73534259,       2.4620416,      -3.2456709 ],
  [               0,               0,               0,      0.27927747,     -0.81968909,      0.59264879 ],
  [               0,               0,               0,   -0.0035740274,     0.011587917,   -0.0092853735 ],
  [               0,               0,               0,     -0.45456616,     -0.63794186,       3.3529291 ],
  [               0,               0,               0,       10.980813,      -77.759183,      -12.110532 ],
 ],
 [
  [               0,               0,               0,     0.017603292,    -0.055357909,     0.046600432 ],
  [               0,               0,               0,      -1.6757895,       3.4255152,       2.7790692 ],
  [               0,               0,               0,       2.1090362,      -8.3970933,       7.0867169 ],
  [               0,               0,               0,      -0.6610519,       2.7663666,       -2.827589 ],
  [               0,               0,               0,     0.014125399,     -0.06155094,     0.067483459 ],
  [               0,               0,               0,      -4.0610138,        15.28655,      -13.392238 ],
  [               0,               0,               0,      -6.5650917,       24.708822,      -24.785391 ],
 ],
 [
  [               0,               0,               0,   0.00066437762,   -0.0065202758,      0.02361249 ],
  [               0,               0,               0,       30.009574,      -143.08138,       172.54539 ],
  [               0,               0,               0,       -4.775466,       22.266139,      -27.060527 ],
  [               0,               0,               0,       4.3215183,      -15.961307,       13.797156 ],
  [               0,               0,               0,     0.010164558,   -0.0026918459,    -0.047695064 ],
  [               0,               0,               0,        12.95679,      -60.929409,       71.673147 ],
  [               0,               0,               0,       11.563864,      -64.248581,       84.746906 ],
 ],
 [
  [               0,               0,               0,               0,    0.0062316539,   -0.0013131138 ],
  [               0,               0,               0,               0,       -35.69988,       89.797616 ],
  [               0,               0,               0,               0,       9.1398365,      -23.719049 ],
  [               0,               0,               0,               0,       1.9279975,      -4.5473094 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0035356017,    0.0027179326,   -0.0072182679,    -0.052598315 ],
  [               0,               0,      -1.1974876,      -1.3818781,      -3.4241345,      -4.4846012 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       128.06011,       108.55626,      -252.12276,      -1927.0799 ],
  [               0,               0,      0.74965637,       1.7579869,       4.5715807,       12.716917 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.030352853,     0.015968773,    0.0054601715 ],
  [               0,               0,               0,       15.204114,      -8.2032887,       5.7095167 ],
  [               0,               0,               0,     -0.67848221,    -0.035618874,      -1.2812819 ],
  [               0,               0,               0,       1.9251941,     -0.52107768,    -0.038169449 ],
  [               0,               0,               0,    0.0022618989,   -0.0018102675,   0.00086599467 ],
  [               0,               0,               0,       10.985392,      -5.5518878,       1.1011205 ],
  [               0,               0,               0,      -34.123827,       11.960992,      -1.7775844 ],
 ],
 [
  [               0,               0,               0,   -0.0091869775,     0.011186982,    0.0045651703 ],
  [               0,               0,               0,      -1.1628455,      0.16519001,       4.6395526 ],
  [               0,               0,               0,      -2.1328944,       2.1293643,      -1.8875922 ],
  [               0,               0,               0,       41.137868,       -64.15733,       62.157462 ],
  [               0,               0,               0,    0.0030936933,     -0.12318301,      0.37975116 ],
  [               0,               0,               0,      0.18653553,      -0.1535167,     0.034828432 ],
  [               0,               0,               0,       1.2076836,      -1.8031695,      -2.4406313 ],
 ],
 [
  [               0,               0,               0,               0,     0.065754624,    -0.044861253 ],
  [               0,               0,               0,               0,      0.27848694,        3.966933 ],
  [               0,               0,               0,               0,       13.155247,      -11.952573 ],
  [               0,               0,               0,               0,      -9.1976454,       7.4271857 ],
  [               0,               0,               0,               0,    -0.033649887,     0.027492161 ],
  [               0,               0,               0,               0,      -6.8188487,       5.9980496 ],
  [               0,               0,               0,               0,       16.560612,      -15.566372 ],
 ],
 [
  [               0,               0,               0,               0,    -0.000151891,    0.0081801782 ],
  [               0,               0,               0,               0,     -0.80763572,       4.7228748 ],
  [               0,               0,               0,               0,     -0.32247156,      -1.1098665 ],
  [               0,               0,               0,               0,     0.087640895,      0.23544055 ],
  [               0,               0,               0,               0,   -0.0013220962,    0.0035578024 ],
  [               0,               0,               0,               0,      0.37352122,     -0.13881795 ],
  [               0,               0,               0,               0,   -0.0031729248,      -2.8293679 ],
 ],
 [
  [               0,               0,      0.19231302,    -0.030808015,     -0.20061283,     -0.14628962 ],
  [               0,               0,        678.8703,      -918.69061,      -1232.5848,       1670.9882 ],
  [               0,               0,      -405.45965,       571.01792,       757.49108,      -1068.6424 ],
  [               0,               0,       560.67416,      -764.32065,      -1054.2022,       1437.2501 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.033108679,    0.0024449292,     0.082076424 ],
  [               0,               0,               0,      -322.01818,       5.6654781,       669.04181 ],
  [               0,               0,               0,       87.480448,      -247.48233,       173.73153 ],
  [               0,               0,               0,      0.13964914,     -0.19805498,     -0.29756156 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -2.3506881,       3.4038739,       4.6515739,      -6.7360766 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.023166891,   0.00096590664,      0.05374125 ],
  [               0,               0,               0,       30.478996,       9.0744883,      -75.813694 ],
  [               0,               0,               0,      -13.426873,       19.783439,      -1.5745133 ],
  [               0,               0,               0,     -0.79819062,       6.8036249,      -8.1008413 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00013767215,     0.002917321 ],
  [               0,               0,               0,               0,       215.41646,         -321.12 ],
  [               0,               0,               0,               0,      -170.61063,        256.7035 ],
  [               0,               0,               0,               0,       169.58031,      -256.29891 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0009245607,   0.00053520711,    0.0052065968 ],
  [               0,               0,               0,       5.9470343,      -22.025068,       23.908628 ],
  [               0,               0,               0,     -0.59492896,        2.034115,       -3.014034 ],
  [               0,               0,               0,     0.055955141,     -0.11529032,     0.047762824 ],
  [               0,               0,               0,  -0.00027322349,    0.0007227395,  -0.00016913168 ],
  [               0,               0,               0,     -0.13387168,      0.36649738,        1.631435 ],
  [               0,               0,               0,     0.036758414,     -0.73080136,      -3.7081195 ],
 ],
 [
  [               0,               0,               0,      0.01992918,    -0.070516454,     0.067807315 ],
  [               0,               0,               0,      -3.5138283,       10.822603,      -4.0229906 ],
  [               0,               0,               0,       2.4000433,      -9.7695364,       8.6119803 ],
  [               0,               0,               0,     -0.79488406,       3.4112444,      -3.5848775 ],
  [               0,               0,               0,     0.011583338,     -0.04978335,     0.054202229 ],
  [               0,               0,               0,      -3.5749787,       13.205902,      -11.219396 ],
  [               0,               0,               0,      -2.3365922,       7.6024662,      -7.5048495 ],
 ],
 [
  [               0,               0,               0,   -0.0068778999,   -0.0038124867,     0.058069027 ],
  [               0,               0,               0,       68.045376,      -318.05257,       374.11661 ],
  [               0,               0,               0,       2.8531491,      -14.410966,       16.922627 ],
  [               0,               0,               0,      -2.5383076,       40.480403,      -81.684313 ],
  [               0,               0,               0,    -0.011027165,    0.0028933146,     0.093668166 ],
  [               0,               0,               0,     -0.48976837,       2.4064083,      -2.9463867 ],
  [               0,               0,               0,       17.511393,      -90.746422,       115.14702 ],
 ],
 [
  [               0,               0,               0,               0,   0.00082653982,     0.010231939 ],
  [               0,               0,               0,               0,      -12.060245,       32.069287 ],
  [               0,               0,               0,               0,       6.8859175,      -18.166373 ],
  [               0,               0,               0,               0,     -0.26393059,      0.70893854 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.022163848,    -0.033349783,    -0.041501228,    -0.029613737 ],
  [               0,               0,      -1.1812921,      -1.3832046,      -3.4333848,      -4.4973234 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      0.96935501,        1.268067,      0.43967253,      -6.7519841 ],
  [               0,               0,      0.14219461,      0.38607017,       1.0603916,       2.4045629 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.029878962,     0.015695769,    0.0054957565 ],
  [               0,               0,               0,       14.926213,      -8.0473335,       5.6911015 ],
  [               0,               0,               0,     -0.76212443,    0.0079849071,      -1.2879784 ],
  [               0,               0,               0,        2.027064,     -0.56538416,    -0.033627052 ],
  [               0,               0,               0,    0.0019074241,   -0.0015969725,   0.00085617708 ],
  [               0,               0,               0,       11.026234,      -5.5546903,       1.0918077 ],
  [               0,               0,               0,      -33.640899,       11.795111,      -1.7857267 ],
 ],
 [
  [               0,               0,               0,   -0.0084610066,     0.010352943,    0.0048076065 ],
  [               0,               0,               0,      -1.5136954,      0.55662627,       4.5288916 ],
  [               0,               0,               0,      -2.1563249,         2.13651,      -1.8871809 ],
  [               0,               0,               0,        29.42608,      -45.319817,       64.054871 ],
  [               0,               0,               0,     -0.14827242,      0.10080079,      0.37156263 ],
  [               0,               0,               0,      0.16218024,     -0.13139058,     0.030121248 ],
  [               0,               0,               0,        0.889693,      -1.3442648,      -2.6091613 ],
 ],
 [
  [               0,               0,               0,               0,     0.069559755,    -0.048062733 ],
  [               0,               0,               0,               0,       1.2240068,       3.2366218 ],
  [               0,               0,               0,               0,       12.647677,      -11.542957 ],
  [               0,               0,               0,               0,      -8.0316496,       6.4755027 ],
  [               0,               0,               0,               0,    -0.028099948,     0.022998932 ],
  [               0,               0,               0,               0,      -8.1989161,       7.2034718 ],
  [               0,               0,               0,               0,       13.866668,      -13.203669 ],
 ],
 [
  [               0,               0,               0,               0,   0.00014816191,    0.0079282288 ],
  [               0,               0,               0,               0,     -0.88062279,       4.7838856 ],
  [               0,               0,               0,               0,     -0.32074086,      -1.1104826 ],
  [               0,               0,               0,               0,      -1.9310898,       6.6458414 ],
  [               0,               0,               0,               0,    -0.017064914,     0.054776985 ],
  [               0,               0,               0,               0,     0.079939864,    -0.025504936 ],
  [               0,               0,               0,               0,    -0.032184265,       -3.032569 ],
 ],
 [
  [               0,               0,      0.19522526,     0.010082538,     -0.17387933,     -0.26188373 ],
  [               0,               0,       825.21009,      -1118.9294,      -1520.6933,       2064.5087 ],
  [               0,               0,       95.123015,      -116.17379,      -188.08201,       229.55676 ],
  [               0,               0,        515.3077,      -703.07131,      -970.05969,       1323.6105 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.026693716,    0.0025308071,     0.068846601 ],
  [               0,               0,               0,      -333.67341,       5.4756324,       692.42988 ],
  [               0,               0,               0,       89.358239,      -253.45799,       178.39819 ],
  [               0,               0,               0,       0.1465714,     -0.20415765,     -0.29504551 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele200_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -2.9541203,       4.2682429,       5.8740192,      -8.4867752 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,       15.692453,      -46.533935,        34.50147 ],
  [               0,               0,               0,      -391.43598,       1269.3065,       -1017.324 ],
  [               0,               0,               0,       1466.7623,      -4360.3585,       3238.7326 ],
  [               0,               0,               0,      -680.93028,       2013.7629,      -1488.6358 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.022735374,     0.037671405 ],
  [               0,               0,               0,               0,       226.16907,      -337.87615 ],
  [               0,               0,               0,               0,      -129.49344,       194.58362 ],
  [               0,               0,               0,               0,       127.82298,      -193.19022 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0011348018,   0.00038286461,    0.0049398708 ],
  [               0,               0,               0,       5.8756574,      -22.008711,       24.077759 ],
  [               0,               0,               0,      -0.5627123,       1.9408504,      -2.9462283 ],
  [               0,               0,               0,     0.049123143,    -0.094200727,     0.030043807 ],
  [               0,               0,               0,  -0.00027110765,   0.00067491837,   -8.265342e-05 ],
  [               0,               0,               0,    0.0084802088,     -0.31051515,       2.3618897 ],
  [               0,               0,               0,    -0.064330964,     -0.37465542,      -4.0238578 ],
 ],
 [
  [               0,               0,               0,     0.027865421,     -0.10228173,     0.099308032 ],
  [               0,               0,               0,      -3.9977847,       12.518724,      -5.4137291 ],
  [               0,               0,               0,       3.0185179,      -12.341451,       11.271869 ],
  [               0,               0,               0,      -1.2880739,       5.4575383,      -5.6999139 ],
  [               0,               0,               0,    0.0088045684,    -0.038612056,     0.043085191 ],
  [               0,               0,               0,      -4.0049143,       15.082469,      -13.240327 ],
  [               0,               0,               0,     -0.67060686,      0.74769449,     -0.47362766 ],
 ],
 [
  [               0,               0,               0,   -0.0057331568,   -0.0010548575,     0.045501548 ],
  [               0,               0,               0,       57.591888,      -270.44916,       319.96261 ],
  [               0,               0,               0,      0.10860311,     -0.63664812,     -0.32519846 ],
  [               0,               0,               0,      -5.7928091,       37.057956,      -49.019457 ],
  [               0,               0,               0,    -0.012024565,   0.00040075307,      0.13422461 ],
  [               0,               0,               0,     0.049690613,     -0.57276203,        1.112765 ],
  [               0,               0,               0,       12.252767,      -55.784627,       60.338428 ],
 ],
 [
  [               0,               0,               0,               0,   0.00068972388,     0.010402366 ],
  [               0,               0,               0,               0,      -13.955904,       36.770014 ],
  [               0,               0,               0,               0,        7.882902,       -20.62658 ],
  [               0,               0,               0,               0,      -0.2476234,      0.67136141 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0017745959,   -0.0012096617,   -0.0029375435,    -0.012334614 ],
  [               0,               0,      -1.2027967,      -1.3840392,      -3.4304238,      -4.4848247 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       8.2764859,       19.620799,       13.590817,      -127.29683 ],
  [               0,               0,         3.48177,       7.7912286,       8.4933695,       3.7409569 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 500 GeV electron clusters
#


ele500_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00042126807,    0.0011363511 ],
  [               0,               0,               0,               0,       1.1131351,       2.0017952 ],
  [               0,               0,               0,               0,      -0.4305346,     -0.71467226 ],
  [               0,               0,               0,               0,     -0.85186053,     -0.68807665 ],
  [               0,               0,               0,               0,     -0.19814568,   -0.0026538293 ],
  [               0,               0,               0,               0,  -9.5009627e-05,   8.9247652e-05 ],
  [               0,               0,               0,               0,       1.2613549,       2.1093481 ],
  [               0,               0,               0,               0,     -0.30424007,      -1.6298879 ],
 ],
 [
  [               0,               0,               0,   0.00096085901,    0.0012360652,   0.00011862024 ],
  [               0,               0,               0,      -9.1105143,       6.0840389,       2.5021258 ],
  [               0,               0,               0,       7.9637193,      -9.4880738,       1.9002125 ],
  [               0,               0,               0,       10.298756,      -12.040692,       2.3355373 ],
  [               0,               0,               0,    -0.035358483,     0.084808395,     -0.16449634 ],
  [               0,               0,               0,    0.0013610589,   -0.0015717289,   0.00046683549 ],
  [               0,               0,               0,      -2.9221014,       3.2366533,       1.6134426 ],
  [               0,               0,               0,       17.258534,      -20.335544,       3.6971767 ],
 ],
 [
  [               0,               0,               0,               0,     0.024510211,    0.0070805293 ],
  [               0,               0,               0,               0,      -5.6899199,       5.1021059 ],
  [               0,               0,               0,               0,      -17.269827,       13.207654 ],
  [               0,               0,               0,               0,       16.647717,      -13.237757 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00079613679,    0.0014795808,  -0.00025676855 ],
  [               0,               0,               0,     -0.55528262,    -0.042533446,       2.7415372 ],
  [               0,               0,               0,      0.78427407,      -1.8000896,     0.095067752 ],
  [               0,               0,               0,       1.6937366,      -3.7122112,       0.8021164 ],
  [               0,               0,               0,     -0.08363038,      0.29430574,     -0.34056264 ],
  [               0,               0,               0,   0.00016947022,  -0.00049327266,   0.00034114493 ],
  [               0,               0,               0,       6.2100337,      -7.9721067,       4.3390015 ],
  [               0,               0,               0,        12.20406,      -29.225524,       20.067926 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       0.1860884,     -0.33051188,     -0.42954645,      0.75425147 ],
  [               0,               0,        820.5077,      -1165.3166,      -1613.1632,        2295.782 ],
  [               0,               0,       -564.9132,       917.48316,       659.40011,      -1167.5588 ],
  [               0,               0,       159.97045,      -105.36367,      -887.85354,       1013.0574 ],
  [               0,               0,       3092.4222,       -4406.899,      -6284.9091,       8954.4328 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,    0.0082082058,     0.013094912 ],
  [               0,               0,               0,               0,      -4.6764007,       8.2581519 ],
  [               0,               0,               0,               0,       32.948128,      -48.530546 ],
  [               0,               0,               0,               0,      -33.884481,       48.979779 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0026297926,     0.018382155 ],
  [               0,               0,               0,               0,      -48.941069,       74.070653 ],
  [               0,               0,               0,               0,       45.547085,      -68.208977 ],
  [               0,               0,               0,               0,      -80.969724,       116.36445 ],
  [               0,               0,               0,               0,      0.13907967,     -0.21406038 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.034193645,      -0.0411256 ],
  [               0,               0,               0,               0,      -198.30088,        298.8442 ],
  [               0,               0,               0,               0,       38.333991,      -58.390323 ],
  [               0,               0,               0,               0,      -6.4007114,       9.6989776 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00086158162,  -0.00096828671,   -0.0013622964,    0.0075971579 ],
  [               0,               0,      -12.901125,       65.166784,      -109.85353,        63.03043 ],
  [               0,               0,      -5.5877666,       27.453074,      -44.728448,       23.232607 ],
  [               0,               0,      -2.4222583,       10.945173,      -16.243811,       6.9728013 ],
  [               0,               0,     -0.82502263,       3.7064174,      -5.5187911,         2.72641 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0056663539,   -0.0082347638 ],
  [               0,               0,               0,               0,      -2.5516945,       7.2040054 ],
  [               0,               0,               0,               0,     -0.22137517,     -0.74696448 ],
  [               0,               0,               0,               0,     -0.38848837,     -0.36855712 ],
  [               0,               0,               0,               0,     -0.15718124,      0.29977163 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0004577755,   -0.0003500168,    0.0010352347,    0.0074797929 ],
  [               0,               0,       1.4375518,      -2.5444752,      -6.5367504,       13.926808 ],
  [               0,               0,       1.1410436,      -2.7453575,      -3.9965126,       8.7484637 ],
  [               0,               0,     -0.91649713,       2.3182941,       3.0412255,      -9.2189112 ],
  [               0,               0,     -0.45421435,      0.75411051,       3.0963848,      -5.6026638 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00045378021,    0.0030827281 ],
  [               0,               0,               0,               0,       -143.2103,       331.35464 ],
  [               0,               0,               0,               0,       50.688148,      -117.99552 ],
  [               0,               0,               0,               0,       1.3629431,      -2.7681249 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00044094583,    0.0015600873 ],
  [               0,               0,               0,               0,     -0.28062819,       3.7229486 ],
  [               0,               0,               0,               0,     -0.87046527,      0.82280485 ],
  [               0,               0,               0,               0,      0.68320136,      -2.6823273 ],
  [               0,               0,               0,               0,      0.08483264,     -0.15226224 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00017936397,     0.030990498,    -0.067001002 ],
  [               0,               0,               0,      -22.893078,       73.098373,      -36.700063 ],
  [               0,               0,               0,       110.09488,      -537.59274,       654.73232 ],
  [               0,               0,               0,      -120.99844,       592.60292,       -725.4604 ],
  [               0,               0,               0,     0.001581587,   -0.0050939479,     0.003229239 ],
  [               0,               0,               0,      -8.2928768,      -5.3535749,       66.807659 ],
  [               0,               0,               0,       1.2359992,      -3.6391846,      0.91921545 ],
 ],
 [
  [               0,               0,               0,               0,     0.016765293,   -0.0080750311 ],
  [               0,               0,               0,               0,      0.74703435,        9.669896 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -18.837635,       46.105066 ],
  [               0,               0,               0,               0,      0.25550693,      0.31337743 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00026195869,     0.001508417 ],
  [               0,               0,               0,               0,    -0.071352579,       2.0709285 ],
  [               0,               0,               0,               0,      0.16447959,     -0.93545735 ],
  [               0,               0,               0,               0,     -0.12828332,     -0.92942745 ],
  [               0,               0,               0,               0,     -0.15462442,   -0.0033986518 ],
  [               0,               0,               0,               0,   4.3412701e-05,   5.0619749e-05 ],
  [               0,               0,               0,               0,      0.29587328,       2.4383109 ],
  [               0,               0,               0,               0,     -0.18632969,      -1.6514026 ],
 ],
 [
  [               0,               0,               0,    0.0041005512,   -0.0011185917,   0.00062681291 ],
  [               0,               0,               0,       -3.334576,      -1.9202034,        4.803925 ],
  [               0,               0,               0,       7.1654491,      -8.0814954,       1.3517099 ],
  [               0,               0,               0,       9.3240036,      -10.278815,       1.6764069 ],
  [               0,               0,               0,     0.024826459,     0.081258257,     -0.16423962 ],
  [               0,               0,               0,     0.001220427,   -0.0012923958,   0.00036443492 ],
  [               0,               0,               0,     -0.45833438,     -0.75031616,       3.0733398 ],
  [               0,               0,               0,       2.1825133,      -1.9516682,      -1.5357324 ],
 ],
 [
  [               0,               0,               0,               0,     0.024510211,    0.0070805293 ],
  [               0,               0,               0,               0,      -5.6899199,       5.1021059 ],
  [               0,               0,               0,               0,      -17.269827,       13.207654 ],
  [               0,               0,               0,               0,       16.647717,      -13.237757 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00097995933,    0.0011622248,   -0.0001345838 ],
  [               0,               0,               0,     -0.51665538,     -0.17545884,       2.8436794 ],
  [               0,               0,               0,      0.79412977,      -1.8236501,      0.10875157 ],
  [               0,               0,               0,       1.6981477,      -3.7144688,      0.79874653 ],
  [               0,               0,               0,     -0.08609385,      0.30439511,     -0.34885842 ],
  [               0,               0,               0,   0.00016201586,  -0.00047601855,   0.00033133955 ],
  [               0,               0,               0,       6.4535739,      -8.5041967,       4.6178615 ],
  [               0,               0,               0,       12.485665,      -29.825934,       20.367077 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.005332728,     -0.11719723,    -0.076776223,      0.37442462 ],
  [               0,               0,      -218.45965,       325.63278,       515.95042,      -759.31682 ],
  [               0,               0,      -91.157552,       239.66341,       -242.2424,        122.7251 ],
  [               0,               0,      -154.43901,       329.60448,      -234.10822,       108.11951 ],
  [               0,               0,       2863.4207,      -4082.0078,      -5822.2645,       8298.1944 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    -0.015645366,     0.022557731 ],
  [               0,               0,               0,               0,      -61.179808,       49.197915 ],
  [               0,               0,               0,               0,      -258.14304,       345.37398 ],
  [               0,               0,               0,               0,      -112.02694,       193.85658 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0013550287,     0.017313918 ],
  [               0,               0,               0,               0,      -211.84822,       312.78189 ],
  [               0,               0,               0,               0,       79.165724,      -117.11016 ],
  [               0,               0,               0,               0,      -126.61461,         182.582 ],
  [               0,               0,               0,               0,      0.10185022,     -0.17530466 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0068563947,     0.023315512 ],
  [               0,               0,               0,               0,      -81.114851,       124.18631 ],
  [               0,               0,               0,               0,       50.903791,      -77.119981 ],
  [               0,               0,               0,               0,      -34.440698,       51.363447 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00091122953,   -0.0012318812,    0.0012777921,     0.014087137 ],
  [               0,               0,      -8.5428084,       44.296802,      -75.893502,       44.103604 ],
  [               0,               0,      -3.7252485,       17.610874,      -27.792585,       13.774176 ],
  [               0,               0,      -2.1405034,       9.1808741,      -13.007486,         5.22243 ],
  [               0,               0,     -0.43178494,       1.8228336,      -2.5120266,       1.1254853 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0057261175,   -0.0083275588 ],
  [               0,               0,               0,               0,      -2.5208323,       7.1330874 ],
  [               0,               0,               0,               0,     -0.22083174,      -0.7468524 ],
  [               0,               0,               0,               0,     -0.38394678,     -0.37716789 ],
  [               0,               0,               0,               0,     -0.15537606,      0.29622888 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0063999169,    -0.016448172,     -0.03245812,     0.085954689 ],
  [               0,               0,       3.1094824,      -12.505742,       15.380612,      -3.8387339 ],
  [               0,               0,       18.194591,      -110.24224,       221.50554,      -148.66312 ],
  [               0,               0,      -11.025083,       69.612233,      -146.07616,       100.77903 ],
  [               0,               0,       9.0953053,      -60.250844,       132.86704,      -97.524835 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0004524262,    0.0030844539 ],
  [               0,               0,               0,               0,          -143.2,       331.33169 ],
  [               0,               0,               0,               0,       50.700204,      -118.02349 ],
  [               0,               0,               0,               0,       1.3520225,       -2.742916 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00044108303,    0.0015602536 ],
  [               0,               0,               0,               0,     -0.27984312,       3.7207294 ],
  [               0,               0,               0,               0,     -0.87208443,      0.82662659 ],
  [               0,               0,               0,               0,      0.68446518,      -2.6853779 ],
  [               0,               0,               0,               0,     0.085467352,     -0.15377632 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00052294154,     0.030905113,    -0.068817903 ],
  [               0,               0,               0,      -23.006254,       72.442853,      -34.437418 ],
  [               0,               0,               0,       114.60827,      -559.69431,       681.79112 ],
  [               0,               0,               0,      -121.05607,       592.91072,      -725.86911 ],
  [               0,               0,               0,     0.001420179,    -0.004152863,    0.0018966506 ],
  [               0,               0,               0,      -7.2846566,      -2.9398793,       54.786824 ],
  [               0,               0,               0,      0.39424136,       -3.464204,       5.5169061 ],
 ],
 [
  [               0,               0,               0,               0,     0.013252798,   0.00059477847 ],
  [               0,               0,               0,               0,        0.784982,       9.4800301 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -19.155843,       46.889041 ],
  [               0,               0,               0,               0,      0.25468937,      0.31100297 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00024962139,    0.0016281819 ],
  [               0,               0,               0,               0,    -0.047930631,       1.9860574 ],
  [               0,               0,               0,               0,      0.15860159,     -0.94069539 ],
  [               0,               0,               0,               0,     -0.11961629,     -0.93416554 ],
  [               0,               0,               0,               0,      -0.1453007,   -0.0029704612 ],
  [               0,               0,               0,               0,   4.3889543e-05,   5.1534492e-05 ],
  [               0,               0,               0,               0,      0.24631769,       2.4652384 ],
  [               0,               0,               0,               0,     -0.17057265,      -1.6532426 ],
 ],
 [
  [               0,               0,               0,    0.0040853546,   -0.0011542828,   0.00064123087 ],
  [               0,               0,               0,      -3.3119806,      -1.9652983,       4.8450122 ],
  [               0,               0,               0,       7.0901437,      -8.0089943,       1.3347841 ],
  [               0,               0,               0,       9.2242208,      -10.170513,       1.6438443 ],
  [               0,               0,               0,     0.019023774,     0.088638989,     -0.16781597 ],
  [               0,               0,               0,    0.0012056862,   -0.0012763705,   0.00035918899 ],
  [               0,               0,               0,     -0.53459583,     -0.68832159,       3.0757706 ],
  [               0,               0,               0,       2.2350315,      -1.9676252,      -1.5560581 ],
 ],
 [
  [               0,               0,               0,               0,     -0.17452147,      0.16713582 ],
  [               0,               0,               0,               0,      -5.2947587,         4.77679 ],
  [               0,               0,               0,               0,      -16.844486,       12.878409 ],
  [               0,               0,               0,               0,       15.953296,      -12.685661 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.001400461,  -0.00061066134,    0.0012319447 ],
  [               0,               0,               0,      -1.0235993,       1.2444382,       1.9248546 ],
  [               0,               0,               0,      0.57250211,      -1.1948846,      -0.3317984 ],
  [               0,               0,               0,       1.4505546,      -3.2488302,      0.60028343 ],
  [               0,               0,               0,    -0.026682094,      0.11416824,     -0.20786155 ],
  [               0,               0,               0,  -0.00021559574,   0.00033013802,  -7.9529785e-05 ],
  [               0,               0,               0,     -0.43212479,      -3.3087823,       5.3333189 ],
  [               0,               0,               0,        16.50867,      -33.879857,       20.572883 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0058048806,     -0.11695811,    -0.076285711,      0.37469917 ],
  [               0,               0,      -175.78158,        263.5759,       428.56733,      -632.33198 ],
  [               0,               0,      -150.98303,       326.65471,      -126.25626,      -46.087617 ],
  [               0,               0,       -145.5481,       319.34212,      -257.93909,       137.18972 ],
  [               0,               0,       2873.9039,       -4096.818,      -5843.7491,       8328.5506 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,      -0.3993943,      0.57726333 ],
  [               0,               0,               0,               0,      -36.468595,       13.067826 ],
  [               0,               0,               0,               0,       -71.95933,       105.57797 ],
  [               0,               0,               0,               0,        67.17486,      -96.757025 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0028498491,    0.0075526504 ],
  [               0,               0,               0,               0,      -6.7831671,       12.722261 ],
  [               0,               0,               0,               0,      0.47069139,       -1.916468 ],
  [               0,               0,               0,               0,      -49.936638,       71.488054 ],
  [               0,               0,               0,               0,      0.11687347,     -0.17938595 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.41700847,     -0.60469578 ],
  [               0,               0,               0,               0,      -115.27702,       174.68514 ],
  [               0,               0,               0,               0,       64.860013,      -97.861577 ],
  [               0,               0,               0,               0,      -41.453627,       61.779766 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00074456744,   -0.0011609563,     0.001142573,     0.013622353 ],
  [               0,               0,      -8.7356284,       44.945744,      -76.567853,       44.306632 ],
  [               0,               0,       -3.726202,       17.782378,       -28.29774,       14.148377 ],
  [               0,               0,      -2.1667812,       9.4316848,      -13.572245,       5.5927105 ],
  [               0,               0,     -0.42856182,       1.8070481,      -2.4870599,       1.1126691 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0057902722,   -0.0083572513 ],
  [               0,               0,               0,               0,       -2.368422,       6.7976401 ],
  [               0,               0,               0,               0,     -0.23156282,     -0.72197874 ],
  [               0,               0,               0,               0,     -0.32177458,     -0.50051204 ],
  [               0,               0,               0,               0,      -0.1348835,      0.25451947 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00044226251,  -0.00035042307,    0.0010000243,    0.0074077523 ],
  [               0,               0,       1.4636879,      -2.6284417,      -6.5465189,       14.070395 ],
  [               0,               0,       1.0839278,      -2.6739243,      -3.5594312,       8.0514862 ],
  [               0,               0,     -0.96004702,       2.4228807,       3.1844573,      -9.5780345 ],
  [               0,               0,     -0.44885761,      0.72918476,       3.1372763,      -5.6307181 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00043492577,    0.0029789494 ],
  [               0,               0,               0,               0,      -150.32546,       347.77265 ],
  [               0,               0,               0,               0,       51.067951,      -118.89576 ],
  [               0,               0,               0,               0,       2.0464536,      -4.3245209 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00043910754,    0.0015476577 ],
  [               0,               0,               0,               0,     -0.26072362,       3.6925737 ],
  [               0,               0,               0,               0,     -0.88367032,      0.85841709 ],
  [               0,               0,               0,               0,      0.70687492,      -2.7392832 ],
  [               0,               0,               0,               0,     0.080221396,     -0.14332972 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0012468188,     0.027565127,    -0.064917197 ],
  [               0,               0,               0,      -19.888574,       59.215035,      -20.811277 ],
  [               0,               0,               0,       114.92394,      -561.41796,       684.12165 ],
  [               0,               0,               0,      -120.77495,       591.63998,      -724.44477 ],
  [               0,               0,               0,   -0.0015835887,     0.011224479,    -0.017733886 ],
  [               0,               0,               0,      -7.5550832,      -1.3038355,       52.389825 ],
  [               0,               0,               0,     -0.67551113,     -0.36217709,       4.3625753 ],
 ],
 [
  [               0,               0,               0,               0,      0.01650422,    -0.003809024 ],
  [               0,               0,               0,               0,      0.99168176,       10.719572 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,      -17.019672,       41.663297 ],
  [               0,               0,               0,               0,      0.26917149,       0.3489894 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.027679731,     0.014662533,    0.0064534212 ],
  [               0,               0,               0,       6.6979078,      -4.2437028,       4.5248759 ],
  [               0,               0,               0,     0.070989777,     0.003168407,      -1.3134809 ],
  [               0,               0,               0,     -0.20460364,      0.11783638,    -0.026854225 ],
  [               0,               0,               0,   0.00063887492,   -0.0018598127,   0.00042092003 ],
  [               0,               0,               0,       3.7151038,      -7.5479277,        1.343015 ],
  [               0,               0,               0,       -7.926954,       3.0725499,      -1.6220849 ],
 ],
 [
  [               0,               0,               0,               0,    0.0033825171,    0.0064413563 ],
  [               0,               0,               0,               0,      -2.0286378,       4.9534997 ],
  [               0,               0,               0,               0,     -0.13576056,      -1.2443375 ],
  [               0,               0,               0,               0,      0.27035158,     -0.15149655 ],
  [               0,               0,               0,               0,   -0.0006149484,   0.00051246362 ],
  [               0,               0,               0,               0,      0.58711806,      0.27404701 ],
  [               0,               0,               0,               0,      -7.3738968,      -28.851835 ],
 ],
 [
  [               0,               0,               0,               0,     0.034818079,    -0.024000876 ],
  [               0,               0,               0,               0,      -14.658027,       16.851406 ],
  [               0,               0,               0,               0,       21.927817,      -19.093771 ],
  [               0,               0,               0,               0,      -12.872562,       10.341106 ],
  [               0,               0,               0,               0,   -0.0046404609,    0.0055255201 ],
  [               0,               0,               0,               0,       2.7526307,      -1.1474092 ],
  [               0,               0,               0,               0,       9.6705903,      -9.5042585 ],
 ],
 [
  [               0,               0,               0,   -0.0021057023,     0.005738655,    0.0047122097 ],
  [               0,               0,               0,       1.6788675,      -4.5768961,       6.3198612 ],
  [               0,               0,               0,      0.36342713,      -1.0397144,     -0.68630776 ],
  [               0,               0,               0,    -0.081005104,      0.63422573,     -0.46876104 ],
  [               0,               0,               0,    0.0020994886,   -0.0028706252,    0.0011689322 ],
  [               0,               0,               0,     -0.64013275,      0.95912556,       0.1906189 ],
  [               0,               0,               0,       4.7670085,      -12.281817,       4.9634372 ],
 ],
 [
  [               0,               0,   -0.0029951329,    -0.037764706,    -0.021594012,      0.11850629 ],
  [               0,               0,        576.2179,      -793.91163,      -1089.2455,       1503.3668 ],
  [               0,               0,      -82.335308,       123.54118,       135.14983,      -206.61213 ],
  [               0,               0,       -72.89785,       97.035154,       132.05464,      -175.46539 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.26378686,     -0.34523248 ],
  [               0,               0,               0,               0,       12.328618,       22.279003 ],
  [               0,               0,               0,               0,      -18.782227,       23.973095 ],
  [               0,               0,               0,               0,      0.92212415,     -0.58484504 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_55_e = [
 [
  [     0.050393251,    0.0068888963,    -0.059538967,     -0.15394586,     -0.28032418,      0.56000021 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     -0.04665212,    -0.021890484,     0.036882441,      0.15682153 ],
  [               0,               0,      -16.439434,       35.137403,       45.734368,      -89.771531 ],
  [               0,               0,       81.055226,      -125.15287,      -179.05587,       275.44996 ],
  [               0,               0,        25.93205,      -37.448908,      -55.741334,       80.452094 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.028204417,    -0.040571311 ],
  [               0,               0,               0,               0,      -23.695059,       40.196607 ],
  [               0,               0,               0,               0,      -263.30497,       396.99072 ],
  [               0,               0,               0,               0,       254.42545,      -384.66717 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.022049514,     0.077124946,    -0.058232202 ],
  [               0,               0,               0,       11.058657,       -38.70174,       36.770251 ],
  [               0,               0,               0,     -0.99443334,       3.3490015,      -4.0605306 ],
  [               0,               0,               0,       104.76866,      -450.37407,       477.92931 ],
  [               0,               0,               0,     -0.24407705,     0.098519493,      0.70013668 ],
  [               0,               0,               0,      0.12381446,      -0.4176098,      0.35260913 ],
  [               0,               0,               0,      -1.1345242,       6.3083585,      -10.031421 ],
 ],
 [
  [               0,               0,               0,    0.0033828222,   -0.0072655317,     0.010541058 ],
  [               0,               0,               0,       1.1786228,      -7.3179708,       12.613795 ],
  [               0,               0,               0,     -0.14250259,      0.88499761,      -2.4344428 ],
  [               0,               0,               0,      -9.7510316,        51.45809,      -58.077779 ],
  [               0,               0,               0,     0.031818744,     0.011421744,    -0.092657017 ],
  [               0,               0,               0,      0.33668213,      -1.3422169,       1.3434732 ],
  [               0,               0,               0,       1.7649196,      -8.4747398,       7.0250273 ],
 ],
 [
  [               0,               0,               0,    -0.062889196,      0.25472099,     -0.23782091 ],
  [               0,               0,               0,        53.63666,      -251.71173,        297.2992 ],
  [               0,               0,               0,      -1.8624958,       8.1192623,       -9.901661 ],
  [               0,               0,               0,      -10.175174,       48.328014,      -57.349506 ],
  [               0,               0,               0,       0.0132662,    -0.054550236,     0.054888254 ],
  [               0,               0,               0,        39.18506,      -193.12496,       238.19433 ],
  [               0,               0,               0,       145.90342,      -685.01786,       802.07678 ],
 ],
 [
  [               0,               0,               0,               0,    0.0040121272,    0.0023881988 ],
  [               0,               0,               0,               0,      -45.221584,       113.53365 ],
  [               0,               0,               0,               0,        8.942199,      -23.310402 ],
  [               0,               0,               0,               0,        1.587897,      -3.6859787 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.020960946,    -0.044802305,     -0.13572407,      0.29247385 ],
  [               0,               0,      -1.2101484,      -1.3679299,       -3.389893,      -4.4012295 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       438.25367,       379.62005,      -859.63145,      -6693.0413 ],
  [               0,               0,       1.2663248,        7.904026,       12.886767,        28.87579 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.034780439,     0.017227903,    0.0053985947 ],
  [               0,               0,               0,        13.20151,      -7.3916813,       5.8532197 ],
  [               0,               0,               0,      0.75857178,     -0.24835316,      -1.3024818 ],
  [               0,               0,               0,     -0.20508325,     0.080388413,    -0.050314176 ],
  [               0,               0,               0,     0.010967957,   -0.0051259822,   0.00095032888 ],
  [               0,               0,               0,       12.809266,      -5.4750835,       1.1882749 ],
  [               0,               0,               0,      -13.851297,       6.3872383,      -1.6460856 ],
 ],
 [
  [               0,               0,               0,               0,    0.0021840032,    0.0064792768 ],
  [               0,               0,               0,               0,      -2.1718311,       5.9197965 ],
  [               0,               0,               0,               0,     -0.13367087,       -1.281051 ],
  [               0,               0,               0,               0,       0.3789305,     -0.23347389 ],
  [               0,               0,               0,               0,  -0.00082007459,   0.00095912397 ],
  [               0,               0,               0,               0,     0.079376328,      0.31081419 ],
  [               0,               0,               0,               0,      -4.0565951,      -24.180202 ],
 ],
 [
  [               0,               0,               0,               0,     0.018015336,   -0.0096752109 ],
  [               0,               0,               0,               0,     -0.83611015,       6.3516257 ],
  [               0,               0,               0,               0,       22.916281,       -19.85493 ],
  [               0,               0,               0,               0,      -12.950306,       10.402677 ],
  [               0,               0,               0,               0,   -0.0051197187,    0.0057210038 ],
  [               0,               0,               0,               0,       2.5000506,     -0.94799028 ],
  [               0,               0,               0,               0,       16.599719,      -15.016368 ],
 ],
 [
  [               0,               0,               0,    -0.007481651,     0.020778027,   -0.0064419499 ],
  [               0,               0,               0,       5.2501768,      -13.909079,       13.160994 ],
  [               0,               0,               0,      0.13478914,     -0.42356218,      -1.1409497 ],
  [               0,               0,               0,     -0.03670087,      0.35750529,     -0.26148225 ],
  [               0,               0,               0,     0.004189412,   -0.0098206433,    0.0062819902 ],
  [               0,               0,               0,     -0.81743354,      0.92845105,      0.72837666 ],
  [               0,               0,               0,      -1.2339224,      0.85450109,      -1.4973238 ],
 ],
 [
  [               0,               0,      0.19424289,    -0.031760987,      -0.2031421,     -0.14595022 ],
  [               0,               0,       492.05979,      -636.97341,      -891.75189,       1156.4511 ],
  [               0,               0,      -341.13796,        487.8859,       630.13125,      -903.60496 ],
  [               0,               0,       649.31854,      -884.75993,      -1221.6277,       1664.7041 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.39971275,     -0.55669616 ],
  [               0,               0,               0,               0,      -868.93995,         1265.14 ],
  [               0,               0,               0,               0,      -27.387215,       35.707154 ],
  [               0,               0,               0,               0,     -0.16487188,      0.10503325 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_35_e = [
 [
  [        77.20888,      -155.71281,      -124.33614,        182.8486,       446.66213,      -459.60217 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0021087714,   -0.0001110296,    0.0034258821,    0.0095845951 ],
  [               0,               0,       -6.797546,       14.737116,       18.294249,      -34.349563 ],
  [               0,               0,      -29.152669,       41.768568,       62.497819,      -90.782987 ],
  [               0,               0,      -7.1391033,         10.0864,       15.301176,      -21.479722 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.034861476,    -0.048734322 ],
  [               0,               0,               0,               0,       56.191836,      -80.853315 ],
  [               0,               0,               0,               0,      -104.94783,       157.37363 ],
  [               0,               0,               0,               0,       103.61361,      -156.58896 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00032028334,   0.00023437831,    0.0071113717 ],
  [               0,               0,               0,       6.2510396,      -22.342055,       23.559297 ],
  [               0,               0,               0,     -0.25622727,      0.75368832,      -1.8705198 ],
  [               0,               0,               0,     -0.16081684,     -0.80146451,       2.2405228 ],
  [               0,               0,               0,   -0.0033911267,  -0.00038262736,     0.013900081 ],
  [               0,               0,               0,       1.4909446,      -4.5930036,       3.6286242 ],
  [               0,               0,               0,       2.5996325,       -8.359372,       3.9214601 ],
 ],
 [
  [               0,               0,               0,     0.028430349,     -0.11638406,      0.12802292 ],
  [               0,               0,               0,      -6.3781852,       24.684612,      -20.576488 ],
  [               0,               0,               0,       1.4779627,      -5.8551977,       4.4647145 ],
  [               0,               0,               0,      -3.3913006,       14.041063,      -14.347482 ],
  [               0,               0,               0,    -0.023258967,       0.0979251,     -0.10153509 ],
  [               0,               0,               0,       2.5208179,      -10.987218,        12.19457 ],
  [               0,               0,               0,       7.6009217,      -30.125223,       27.208372 ],
 ],
 [
  [               0,               0,               0,   -0.0019419727,    -0.014064398,     0.055238811 ],
  [               0,               0,               0,       28.351527,      -131.97609,       156.12522 ],
  [               0,               0,               0,       3.3546148,      -16.418326,       18.856763 ],
  [               0,               0,               0,      -49.428019,       233.39983,      -275.38634 ],
  [               0,               0,               0,    0.0089755882,    -0.012475178,    -0.018778191 ],
  [               0,               0,               0,       2.8974275,       -14.27877,        17.64869 ],
  [               0,               0,               0,      -20.303709,       112.93126,        -155.797 ],
 ],
 [
  [               0,               0,               0,               0,     -0.22556216,      0.56037335 ],
  [               0,               0,               0,               0,       -342.2754,       846.43885 ],
  [               0,               0,               0,               0,       2.1853099,      -5.7380148 ],
  [               0,               0,               0,               0,       95.629958,      -235.38956 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00026394452,    0.0045097182,   0.00035230818,    -0.024596032 ],
  [               0,               0,      -1.2248723,      -1.3543714,      -3.3625893,      -4.4669324 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       5.7258565,       11.310764,       2.6943173,       55.659914 ],
  [               0,               0,       3.1271465,       1.4815638,     -0.50058473,       129.96434 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.034195745,     0.016950105,    0.0054475142 ],
  [               0,               0,               0,       12.871139,      -7.2255128,       5.8268346 ],
  [               0,               0,               0,      0.78985152,     -0.26018171,      -1.3026178 ],
  [               0,               0,               0,     -0.19396402,     0.073663648,    -0.049414394 ],
  [               0,               0,               0,     0.010508604,   -0.0049092228,    0.0009343505 ],
  [               0,               0,               0,       13.108763,      -5.6125598,       1.1880802 ],
  [               0,               0,               0,       -13.38107,       6.1855549,      -1.6354341 ],
 ],
 [
  [               0,               0,               0,               0,    0.0021108403,    0.0065439685 ],
  [               0,               0,               0,               0,      -2.1359857,       5.8912686 ],
  [               0,               0,               0,               0,     -0.12369934,      -1.2875281 ],
  [               0,               0,               0,               0,      0.40346891,     -0.24936819 ],
  [               0,               0,               0,               0,  -0.00086759773,    0.0010460854 ],
  [               0,               0,               0,               0,     0.075848362,      0.27254144 ],
  [               0,               0,               0,               0,      -3.9577111,      -24.227702 ],
 ],
 [
  [               0,               0,               0,               0,     0.016818316,   -0.0086516599 ],
  [               0,               0,               0,               0,       1.1356052,       4.7159605 ],
  [               0,               0,               0,               0,       22.513978,      -19.513328 ],
  [               0,               0,               0,               0,      -12.759369,       10.239986 ],
  [               0,               0,               0,               0,   -0.0043437683,     0.005085008 ],
  [               0,               0,               0,               0,       2.4508649,     -0.91088717 ],
  [               0,               0,               0,               0,       16.855424,      -15.210009 ],
 ],
 [
  [               0,               0,               0,   -0.0070682305,     0.019715606,   -0.0057599606 ],
  [               0,               0,               0,       4.9953545,      -13.288792,       12.783117 ],
  [               0,               0,               0,      0.12280507,      -0.4143103,      -1.1392411 ],
  [               0,               0,               0,    -0.029342258,      0.35089821,     -0.26069693 ],
  [               0,               0,               0,    0.0039622023,   -0.0092477522,    0.0059439723 ],
  [               0,               0,               0,     -0.72174643,       0.7558739,      0.79764438 ],
  [               0,               0,               0,      -1.1071168,      0.61522415,      -1.3893705 ],
 ],
 [
  [               0,               0,       0.1971201,    -0.018842009,     -0.19659365,     -0.18513647 ],
  [               0,               0,       512.22081,      -670.76259,      -932.68887,       1223.7468 ],
  [               0,               0,       -185.2471,        273.8457,       336.81209,      -500.82493 ],
  [               0,               0,       640.74352,      -873.62002,      -1205.9854,       1644.3957 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.52664938,     -0.73361984 ],
  [               0,               0,               0,               0,       11.028496,       24.148006 ],
  [               0,               0,               0,               0,      -57.885145,       78.476349 ],
  [               0,               0,               0,               0,       1.3412358,      -1.4052756 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele500_layer2_etaoffv4_1_37_e = [
 [
  [       72.143655,      -145.55918,      -116.25424,       171.05735,       417.96114,      -430.25099 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0013164973,   0.00010341369,    0.0027670198,    0.0078206496 ],
  [               0,               0,       -5.159829,       11.270218,       13.896271,      -25.609612 ],
  [               0,               0,      -26.680015,       38.047424,       57.006228,      -82.525133 ],
  [               0,               0,      -6.5877851,       9.3736555,       14.165564,      -20.030581 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.014117823,    -0.017269396 ],
  [               0,               0,               0,               0,       88.888329,      -130.38584 ],
  [               0,               0,               0,               0,      -106.11186,       159.16246 ],
  [               0,               0,               0,               0,       103.75144,      -156.80489 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00097126719,  -0.00091315785,    0.0072395247 ],
  [               0,               0,               0,       6.8209956,      -24.586466,       25.709118 ],
  [               0,               0,               0,      -1.3311961,       4.3468362,      -4.8426424 ],
  [               0,               0,               0,      0.26654521,       4.5408911,      -8.1451759 ],
  [               0,               0,               0,     0.093835004,     -0.30141521,       0.2465672 ],
  [               0,               0,               0,       2.0349959,      -6.8235207,       5.7512835 ],
  [               0,               0,               0,       14.319554,      -56.637815,         52.8012 ],
 ],
 [
  [               0,               0,               0,     0.026323417,     -0.10731926,      0.11838225 ],
  [               0,               0,               0,      -5.8361348,        22.39064,      -18.163312 ],
  [               0,               0,               0,       1.3879354,      -5.4746862,       4.0651478 ],
  [               0,               0,               0,      -3.5662023,       14.759902,      -15.075715 ],
  [               0,               0,               0,     -0.02422918,      0.10205139,     -0.10579745 ],
  [               0,               0,               0,        2.251861,       -9.851022,       10.983277 ],
  [               0,               0,               0,       8.4125811,      -33.496091,       30.695655 ],
 ],
 [
  [               0,               0,               0,   -0.0018455129,    -0.015151019,     0.057321564 ],
  [               0,               0,               0,       29.030813,      -135.11358,       159.74835 ],
  [               0,               0,               0,       3.3989881,      -16.643044,       19.138908 ],
  [               0,               0,               0,      -52.417819,       247.38713,       -291.7377 ],
  [               0,               0,               0,    0.0096257749,    -0.013837733,    -0.019027269 ],
  [               0,               0,               0,       3.4902318,      -17.084927,       20.966976 ],
  [               0,               0,               0,      -19.103511,       107.11365,      -148.72752 ],
 ],
 [
  [               0,               0,               0,               0,     -0.22912225,      0.56915942 ],
  [               0,               0,               0,               0,      -342.11298,       846.05475 ],
  [               0,               0,               0,               0,         1.41296,      -3.8399086 ],
  [               0,               0,               0,               0,       96.632656,      -237.85342 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.043126652,    -0.068123136,    -0.078292555,    -0.043655954 ],
  [               0,               0,      -1.1815946,      -1.4124837,      -3.4678362,      -4.4949086 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      0.32431715,      0.52713193,       0.8808594,     -0.14580347 ],
  [               0,               0,     0.044838656,      0.16981589,      0.63428539,       1.4072843 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 1000 GeV electron clusters
#


ele1000_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00018848747,    0.0014886733 ],
  [               0,               0,               0,               0,     -0.54920636,       2.6041696 ],
  [               0,               0,               0,               0,      0.49933016,      -1.1276591 ],
  [               0,               0,               0,               0,      0.24325911,      -1.1470581 ],
  [               0,               0,               0,               0,     -0.17417635,   -0.0064150376 ],
  [               0,               0,               0,               0,   8.0277273e-05,   2.3708408e-05 ],
  [               0,               0,               0,               0,      -1.7281074,       3.1379636 ],
  [               0,               0,               0,               0,      -1.3999581,      -1.3394899 ],
 ],
 [
  [               0,               0,               0,               0,   0.00034605623,   0.00078608066 ],
  [               0,               0,               0,               0,     -0.66715468,       3.4284507 ],
  [               0,               0,               0,               0,     0.024564069,     -0.81874969 ],
  [               0,               0,               0,               0,     -0.11658327,     -0.93720314 ],
  [               0,               0,               0,               0,     -0.21183047,    -0.018612226 ],
  [               0,               0,               0,               0,   -6.984658e-05,   9.4130057e-05 ],
  [               0,               0,               0,               0,     -0.30036854,       2.3571313 ],
  [               0,               0,               0,               0,      -2.1500382,     -0.92228282 ],
 ],
 [
  [               0,               0,               0,               0,      0.02637508,    -0.001583144 ],
  [               0,               0,               0,               0,      -4.0983288,       3.8935256 ],
  [               0,               0,               0,               0,      -25.777258,       19.930398 ],
  [               0,               0,               0,               0,        23.90898,      -19.013983 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00043568766,   0.00059092507,    0.0012948551 ],
  [               0,               0,               0,     -0.12280226,     -0.57907696,       2.8094443 ],
  [               0,               0,               0,    -0.076339095,      0.36050767,      -1.2572925 ],
  [               0,               0,               0,      0.46877512,     -0.88152113,     -0.86646998 ],
  [               0,               0,               0,    -0.058150268,      0.14751327,     -0.20926325 ],
  [               0,               0,               0,   0.00021113494,  -0.00046459114,   0.00025627042 ],
  [               0,               0,               0,       2.0582136,     -0.42796786,        1.779875 ],
  [               0,               0,               0,      -4.0871289,     0.098789674,    -0.092591631 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.32786595,     -0.55665405,     -0.73916149,       1.2463794 ],
  [               0,               0,       649.86498,      -928.35617,       -1253.952,       1796.0921 ],
  [               0,               0,      -921.48088,       1481.9407,       1208.4895,      -2064.2789 ],
  [               0,               0,      -69.246874,       244.98692,      -536.41576,       464.01096 ],
  [               0,               0,       3533.1541,      -5031.9634,       -7191.541,       10240.165 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,     0.059055216,    -0.060113996 ],
  [               0,               0,               0,               0,      -1.8059879,       3.8952805 ],
  [               0,               0,               0,               0,       11.940203,       -18.28163 ],
  [               0,               0,               0,               0,      -17.999737,       26.143424 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0047420078,   -0.0046202071 ],
  [               0,               0,               0,               0,       20.251845,      -22.057724 ],
  [               0,               0,               0,               0,      -397.79336,       584.83821 ],
  [               0,               0,               0,               0,       252.14976,      -366.04885 ],
  [               0,               0,               0,               0,      0.18314199,     -0.20753297 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.20694893,      0.31379886 ],
  [               0,               0,               0,               0,      -202.47775,       307.00737 ],
  [               0,               0,               0,               0,      -42.762021,        62.18761 ],
  [               0,               0,               0,               0,       65.234885,      -96.746563 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.025642975,    -0.047452895,    -0.088654282,      0.16649272 ],
  [               0,               0,      -5.5750395,       32.735889,       -59.23604,       34.958712 ],
  [               0,               0,      -2.0055429,       11.087253,      -21.313888,       13.022059 ],
  [               0,               0,       -1.155212,       7.1394866,      -15.105667,       9.7388693 ],
  [               0,               0,     -0.24341229,      0.87521386,        -0.90309,      0.20832427 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.007507784,    -0.011852641 ],
  [               0,               0,               0,               0,       -5.362723,       12.761359 ],
  [               0,               0,               0,               0,      0.06416712,      -1.3310254 ],
  [               0,               0,               0,               0,     -0.28116339,     -0.57722095 ],
  [               0,               0,               0,               0,     -0.15162335,      0.29840536 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00053467367,  -0.00038938012,     0.001177086,    0.0081781262 ],
  [               0,               0,       1.8846811,      -3.3388848,      -8.8904243,       18.305926 ],
  [               0,               0,      0.84954683,      -2.0394098,      -2.7519967,       5.6739364 ],
  [               0,               0,     -0.48194824,       1.4854026,      0.65576759,      -4.5881184 ],
  [               0,               0,     -0.32950801,      0.52254308,       2.3577046,      -4.1810075 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19717496,      0.45686181 ],
  [               0,               0,               0,               0,      -122.11116,       283.06174 ],
  [               0,               0,               0,               0,      -1.5645455,       1.9814622 ],
  [               0,               0,               0,               0,       38.302102,      -87.565554 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00055407904,    0.0019526158 ],
  [               0,               0,               0,               0,      -2.0806832,       7.6066778 ],
  [               0,               0,               0,               0,      0.16856105,      -1.6244368 ],
  [               0,               0,               0,               0,       0.7869834,      -2.9740624 ],
  [               0,               0,               0,               0,     0.090472886,     -0.17674145 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00016144502,   0.00088763174 ],
  [               0,               0,               0,               0,      -1.4648185,       5.8662063 ],
  [               0,               0,               0,               0,    -0.091383654,     -0.59742642 ],
  [               0,               0,               0,               0,     -0.21012591,      -1.5704531 ],
  [               0,               0,               0,               0,     -0.18576403,     -0.74542129 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.013584028,    0.0048611775,     -0.08439855 ],
  [               0,               0,               0,      -34.367742,       110.32962,      -57.778695 ],
  [               0,               0,               0,      -135.85647,       667.34214,      -821.14229 ],
  [               0,               0,               0,       -28.00908,       136.64334,        -166.492 ],
  [               0,               0,               0,    0.0084429563,    0.0030211483,    -0.058058804 ],
  [               0,               0,               0,       2.0359031,      -3.2414141,      -2.9710076 ],
  [               0,               0,               0,       3.8403807,      0.73648606,       -23.26926 ],
 ],
 [
  [               0,               0,               0,               0,     0.027388523,    0.0038636027 ],
  [               0,               0,               0,               0,      -2.0565298,       45.089301 ],
  [               0,               0,               0,               0,   6.3420093e-05,      -0.6083803 ],
  [               0,               0,               0,               0,     -0.93938397,       2.0361924 ],
  [               0,               0,               0,               0,       9.2120918,      -21.548183 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00016058158,    0.0013739249 ],
  [               0,               0,               0,               0,     -0.55973501,       2.7356087 ],
  [               0,               0,               0,               0,      0.50085312,      -1.1187945 ],
  [               0,               0,               0,               0,      0.22737892,      -1.1398732 ],
  [               0,               0,               0,               0,     -0.18761052,   -0.0069746758 ],
  [               0,               0,               0,               0,   7.9187426e-05,   2.1736701e-05 ],
  [               0,               0,               0,               0,      -1.6540479,       3.0951679 ],
  [               0,               0,               0,               0,      -1.4411677,      -1.3379651 ],
 ],
 [
  [               0,               0,               0,               0,   0.00034597654,   0.00078620898 ],
  [               0,               0,               0,               0,     -0.66786668,       3.4288123 ],
  [               0,               0,               0,               0,     0.024933361,     -0.81903383 ],
  [               0,               0,               0,               0,     -0.11631559,     -0.93744725 ],
  [               0,               0,               0,               0,     -0.21184136,    -0.018595989 ],
  [               0,               0,               0,               0,  -6.9760163e-05,   9.4065432e-05 ],
  [               0,               0,               0,               0,     -0.29948238,       2.3567275 ],
  [               0,               0,               0,               0,      -2.1505411,     -0.92212934 ],
 ],
 [
  [               0,               0,               0,               0,      0.02637508,    -0.001583144 ],
  [               0,               0,               0,               0,      -4.0983288,       3.8935256 ],
  [               0,               0,               0,               0,      -25.777258,       19.930398 ],
  [               0,               0,               0,               0,        23.90898,      -19.013983 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00040976603,    0.0005878209,     0.001331075 ],
  [               0,               0,               0,     -0.17155118,     -0.44201274,       2.7173895 ],
  [               0,               0,               0,    -0.065576321,      0.33758886,      -1.2453541 ],
  [               0,               0,               0,      0.48486972,      -0.9237629,     -0.83958064 ],
  [               0,               0,               0,    -0.056821305,      0.14153194,     -0.20421596 ],
  [               0,               0,               0,   0.00021042444,  -0.00046341198,   0.00025592039 ],
  [               0,               0,               0,       2.0583783,     -0.41464175,        1.758594 ],
  [               0,               0,               0,      -4.1197194,      0.11114535,    -0.046385815 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.098362107,     -0.29479425,     -0.31406434,      0.77492421 ],
  [               0,               0,      -181.91551,       265.02027,       455.27857,       -655.8757 ],
  [               0,               0,       -15.17692,        192.2359,      -581.46667,       482.73431 ],
  [               0,               0,      -386.04685,        677.5135,       142.94063,      -465.55302 ],
  [               0,               0,       3237.9123,      -4613.3755,       -6592.929,       9391.5805 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,   -0.0078209965,    0.0002070627,     0.016993786 ],
  [               0,               0,               0,      -23.093837,      -9.1755957,        22.18613 ],
  [               0,               0,               0,        -26.5875,     -0.15842739,       54.945904 ],
  [               0,               0,               0,      -15.667372,      0.48394724,       33.957433 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.22600655,     -0.32749646 ],
  [               0,               0,               0,               0,       800.93638,      -1164.1434 ],
  [               0,               0,               0,               0,      -83.544748,         121.498 ],
  [               0,               0,               0,               0,      -287.39682,       422.03885 ],
  [               0,               0,               0,               0,      0.19623547,     -0.22401378 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.21091681,      0.31966613 ],
  [               0,               0,               0,               0,       -180.5349,       274.38161 ],
  [               0,               0,               0,               0,      -47.104883,       68.637411 ],
  [               0,               0,               0,               0,       67.433401,      -100.01534 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00088219366,   -0.0031276761,   -0.0038135372,     0.014359325 ],
  [               0,               0,      -26.887777,       134.54209,      -222.38804,       123.08866 ],
  [               0,               0,      -11.188592,       56.887144,      -96.579201,       53.672129 ],
  [               0,               0,      -2.1871026,       11.916348,      -21.797884,         12.3377 ],
  [               0,               0,      -2.6140243,       12.217514,      -18.969983,       9.7955565 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0075078514,    -0.011852694 ],
  [               0,               0,               0,               0,      -5.3632553,        12.76236 ],
  [               0,               0,               0,               0,     0.064379372,       -1.331425 ],
  [               0,               0,               0,               0,     -0.28088343,     -0.57777095 ],
  [               0,               0,               0,               0,     -0.15159952,      0.29835637 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0005346107,  -0.00038973611,    0.0011762561,    0.0081787454 ],
  [               0,               0,       1.8894037,      -3.3472497,      -8.9141106,       18.349407 ],
  [               0,               0,        0.850534,       -2.041845,      -2.7548087,       5.6814332 ],
  [               0,               0,     -0.48167466,       1.4860258,      0.65006675,      -4.5814304 ],
  [               0,               0,     -0.32969837,      0.52269188,       2.3596997,      -4.1840747 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19717496,      0.45686181 ],
  [               0,               0,               0,               0,      -122.11116,       283.06174 ],
  [               0,               0,               0,               0,      -1.5645455,       1.9814622 ],
  [               0,               0,               0,               0,       38.302102,      -87.565554 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00055480698,    0.0019536816 ],
  [               0,               0,               0,               0,      -2.0860934,       7.6178059 ],
  [               0,               0,               0,               0,      0.17179491,      -1.6319145 ],
  [               0,               0,               0,               0,      0.78792879,      -2.9763676 ],
  [               0,               0,               0,               0,     0.090065313,     -0.17584287 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00016144502,   0.00088763174 ],
  [               0,               0,               0,               0,      -1.4648185,       5.8662063 ],
  [               0,               0,               0,               0,    -0.091383654,     -0.59742642 ],
  [               0,               0,               0,               0,     -0.21012591,      -1.5704531 ],
  [               0,               0,               0,               0,     -0.18576403,     -0.74542129 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0031391298,  -0.00017358028,     0.027146766 ],
  [               0,               0,               0,      -7.1833709,        14.72895,       13.648212 ],
  [               0,               0,               0,      -61.410056,       291.55833,      -347.43908 ],
  [               0,               0,               0,      -3.6095622,       23.284599,      -35.190225 ],
  [               0,               0,               0,    0.0037074228,   -0.0010508823,    -0.018809884 ],
  [               0,               0,               0,      -12.752085,        1.181138,       74.602522 ],
  [               0,               0,               0,      -23.940682,      -1.7983542,       145.50319 ],
 ],
 [
  [               0,               0,               0,               0,     0.027311698,    0.0038168341 ],
  [               0,               0,               0,               0,       -2.034734,       45.035315 ],
  [               0,               0,               0,               0,   2.3578855e-05,     -0.48413489 ],
  [               0,               0,               0,               0,     -0.66169055,       1.3490804 ],
  [               0,               0,               0,               0,       9.1925471,      -21.499095 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00015071812,    0.0013589499 ],
  [               0,               0,               0,               0,     -0.56904449,        2.759396 ],
  [               0,               0,               0,               0,      0.49858493,      -1.1181955 ],
  [               0,               0,               0,               0,      0.22582413,      -1.1408226 ],
  [               0,               0,               0,               0,     -0.18868969,   -0.0073730862 ],
  [               0,               0,               0,               0,   7.7948629e-05,   2.1276507e-05 ],
  [               0,               0,               0,               0,      -1.6775454,         3.10281 ],
  [               0,               0,               0,               0,      -1.4468101,      -1.3370102 ],
 ],
 [
  [               0,               0,               0,               0,   0.00033805994,   0.00078752461 ],
  [               0,               0,               0,               0,     -0.67530202,       3.4251048 ],
  [               0,               0,               0,               0,     0.033138281,     -0.81827635 ],
  [               0,               0,               0,               0,     -0.10227362,     -0.93781457 ],
  [               0,               0,               0,               0,      -0.2132992,    -0.018020729 ],
  [               0,               0,               0,               0,  -6.7650058e-05,   9.4011678e-05 ],
  [               0,               0,               0,               0,     -0.30270542,        2.352482 ],
  [               0,               0,               0,               0,      -2.0771474,     -0.95407134 ],
 ],
 [
  [               0,               0,               0,               0,     0.025904312,  -0.00063220793 ],
  [               0,               0,               0,               0,      -4.3398002,       4.0784962 ],
  [               0,               0,               0,               0,      -24.894533,       19.233635 ],
  [               0,               0,               0,               0,       23.279648,      -18.513505 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0011722551,    -0.002410998,    0.0028904692 ],
  [               0,               0,               0,       1.6363858,      -5.2305609,       6.0004565 ],
  [               0,               0,               0,      -2.2453474,       5.4110227,       -4.091417 ],
  [               0,               0,               0,      -2.3968826,        5.030558,      -3.8565987 ],
  [               0,               0,               0,     -0.25532578,      0.42157711,     -0.31291931 ],
  [               0,               0,               0,   -0.0014515161,    0.0027548242,   -0.0013445562 ],
  [               0,               0,               0,       2.0704896,      -7.0471903,       6.6712113 ],
  [               0,               0,               0,      -9.9871215,       26.588071,      -22.292393 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.098599373,     -0.29523938,     -0.31463006,      0.77595847 ],
  [               0,               0,      -189.96277,       276.42983,       471.55575,      -678.95132 ],
  [               0,               0,      -21.917929,       201.86938,      -567.68352,       463.03686 ],
  [               0,               0,      -384.75577,       675.57447,       140.62371,      -462.04947 ],
  [               0,               0,       3237.8768,      -4613.3267,      -6592.8567,       9391.4821 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,     0.019181184,   -0.0010158844,    -0.037078341 ],
  [               0,               0,               0,       32.668277,      -10.227702,      -91.677818 ],
  [               0,               0,               0,       14.428581,      -7.0369091,      -17.271478 ],
  [               0,               0,               0,      -97.593605,      -4.6398168,        210.4261 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.11530302,     -0.16573852 ],
  [               0,               0,               0,               0,       830.30711,      -1207.4704 ],
  [               0,               0,               0,               0,      -159.88509,       234.04662 ],
  [               0,               0,               0,               0,      -54.217357,       81.542272 ],
  [               0,               0,               0,               0,      0.22659467,      -0.2704883 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.21296352,      0.32267102 ],
  [               0,               0,               0,               0,      -183.28696,       278.45329 ],
  [               0,               0,               0,               0,      -49.426776,       72.086982 ],
  [               0,               0,               0,               0,       70.369891,       -104.3708 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.027009559,    -0.050879291,    -0.090018507,      0.17222768 ],
  [               0,               0,      -5.7986266,       33.622168,      -60.439368,       35.511526 ],
  [               0,               0,      -2.1338576,       11.555387,      -21.812686,         13.1563 ],
  [               0,               0,      -1.1898817,        7.094844,       -14.68889,       9.3363648 ],
  [               0,               0,      -0.2759158,       1.0377701,      -1.1748058,      0.35977455 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0079028727,    -0.012490067 ],
  [               0,               0,               0,               0,      -5.1522335,       12.296762 ],
  [               0,               0,               0,               0,     0.082527374,      -1.3637635 ],
  [               0,               0,               0,               0,     -0.19896169,      -0.7419514 ],
  [               0,               0,               0,               0,     -0.13595427,      0.26621206 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00054345368,  -0.00040947658,    0.0011830273,    0.0083989111 ],
  [               0,               0,       1.9238007,      -3.4192812,      -9.0739264,       18.659379 ],
  [               0,               0,      0.85973661,      -2.1308466,      -2.5195976,       5.4982345 ],
  [               0,               0,     -0.51888892,       1.6139732,      0.59907826,      -4.6926611 ],
  [               0,               0,     -0.32872034,      0.49102586,       2.4830297,       -4.313309 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00029900562,    0.0031385703 ],
  [               0,               0,               0,               0,      -193.78361,       447.78248 ],
  [               0,               0,               0,               0,       52.203003,      -121.56726 ],
  [               0,               0,               0,               0,       6.0642028,      -13.503619 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00055268103,    0.0019485621 ],
  [               0,               0,               0,               0,      -2.0676546,        7.582465 ],
  [               0,               0,               0,               0,      0.19448633,      -1.6841079 ],
  [               0,               0,               0,               0,        0.778117,      -2.9538223 ],
  [               0,               0,               0,               0,     0.078052583,     -0.14836779 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.012531352,     0.023479608,     0.024945577 ],
  [               0,               0,               0,      -25.764074,       132.33224,      -162.50048 ],
  [               0,               0,               0,       -85.64861,       397.34577,      -461.57153 ],
  [               0,               0,               0,       -17.40681,       98.003027,      -135.15582 ],
  [               0,               0,               0,     0.011295671,  -0.00064006128,    -0.064215109 ],
  [               0,               0,               0,      -2.1834807,       10.608393,      -11.938842 ],
  [               0,               0,               0,      -4.8602632,       9.3403022,       3.6536505 ],
 ],
 [
  [               0,               0,               0,               0,     0.023864478,     0.015518455 ],
  [               0,               0,               0,               0,       1.3986261,       36.520353 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,     -0.61117948,       1.0353963 ],
  [               0,               0,               0,               0,       24.583034,      -59.778312 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.086002558,     0.037940486,    0.0045659585 ],
  [               0,               0,               0,       35.634301,      -16.038582,       5.6626395 ],
  [               0,               0,               0,      -1.9085945,       1.0048908,      -1.4307994 ],
  [               0,               0,               0,      0.32616447,     -0.25432793,     0.024235904 ],
  [               0,               0,               0,     0.018850745,   -0.0086559896,    0.0010530463 ],
  [               0,               0,               0,       43.052485,      -18.229301,       1.6074259 ],
  [               0,               0,               0,      -8.9879412,       5.0917389,       -2.082507 ],
 ],
 [
  [               0,               0,               0,   -0.0070467753,    0.0046485152,    0.0077436439 ],
  [               0,               0,               0,       4.4019802,      -5.3205885,       5.6319887 ],
  [               0,               0,               0,       -3.695125,       3.7237681,      -2.2026094 ],
  [               0,               0,               0,       2.2198405,      -2.2230545,      0.51544145 ],
  [               0,               0,               0,    0.0084419178,   -0.0089643825,    0.0025410533 ],
  [               0,               0,               0,       5.8310267,      -7.6298622,       3.3234128 ],
  [               0,               0,               0,       26.176938,      -41.236609,       13.885624 ],
 ],
 [
  [               0,               0,               0,               0,     0.006731605,    0.0036016389 ],
  [               0,               0,               0,               0,       3.9625292,      0.40293469 ],
  [               0,               0,               0,               0,       3.6541364,      -4.1766713 ],
  [               0,               0,               0,               0,      -3.1222598,       2.4524561 ],
  [               0,               0,               0,               0,    0.0030464611,   -0.0023832679 ],
  [               0,               0,               0,               0,     -0.24765903,       2.1581299 ],
  [               0,               0,               0,               0,      -3.4230265,     -0.29012935 ],
 ],
 [
  [               0,               0,               0,   -0.0017436348,   0.00099280884,     0.010355097 ],
  [               0,               0,               0,       1.9524309,      -3.9904871,       5.3492345 ],
  [               0,               0,               0,    -0.094723932,  -0.00075965173,      -1.1855283 ],
  [               0,               0,               0,      -0.1056783,      0.40475448,     -0.28580461 ],
  [               0,               0,               0,    0.0015187632,   -0.0037253927,    0.0023979269 ],
  [               0,               0,               0,      -4.1298301,       6.7770439,      -3.8956354 ],
  [               0,               0,               0,      -2.3382249,       6.1566536,      -5.0876837 ],
 ],
 [
  [               0,   -0.0021849967,    -0.011027889,    -0.012645301,    0.0048547498,     0.063718399 ],
  [               0,       211.50746,      -146.88374,      -401.56671,      -274.58621,       765.24466 ],
  [               0,       -37.21859,        30.17857,       68.357433,       36.683869,      -126.34446 ],
  [               0,      -18.840945,       11.946699,       33.150836,       23.022871,      -58.004034 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.12497298,    0.0088555681,     -0.24049519 ],
  [               0,               0,               0,       5.8384567,       8.3500772,       15.944708 ],
  [               0,               0,               0,      -40.612549,       85.219525,      -41.997748 ],
  [               0,               0,               0,       0.1156363,      0.27381534,      0.03930639 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_55_e = [
 [
  [     0.074478555,   1.9224184e-05,    -0.099602528,     -0.21190402,     -0.30141925,      0.71046392 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,    0.0067962999,   -0.0077823612,     -0.01876617,   -0.0093392612,     0.058172371 ],
  [               0,      -13.067093,       12.468311,       39.480285,       30.812748,      -107.84958 ],
  [               0,       189.50606,      -145.74086,      -412.56406,        -291.162,       896.72121 ],
  [               0,      -112.13099,         82.4913,        244.9356,       181.64776,      -534.68002 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.001537233,    0.0069176441 ],
  [               0,               0,               0,               0,       32.580317,      -46.640647 ],
  [               0,               0,               0,               0,      -2.8990553,       3.1984742 ],
  [               0,               0,               0,               0,       -1.598509,        2.453905 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -2.7463331e-05,    0.0010412124,    0.0074072715 ],
  [               0,               0,               0,       4.3093467,      -15.544152,       17.068199 ],
  [               0,               0,               0,     -0.57645796,       1.9460757,      -2.9135593 ],
  [               0,               0,               0,        1.558939,      -6.6695194,       7.1087724 ],
  [               0,               0,               0,   -0.0034670992,  -0.00036636058,      0.01340035 ],
  [               0,               0,               0,       4.2761429,      -14.204588,       11.824272 ],
  [               0,               0,               0,      -4.1987601,       16.471186,      -18.349858 ],
 ],
 [
  [               0,               0,               0,    0.0011303122,   -0.0011707373,    0.0078446445 ],
  [               0,               0,               0,       1.2963549,      -7.2796503,       12.060078 ],
  [               0,               0,               0,      0.80003437,      -2.9799016,        1.489224 ],
  [               0,               0,               0,     -0.49259769,      -1.0391958,       5.0639057 ],
  [               0,               0,               0,    -0.007357495,    0.0023694174,     0.034402677 ],
  [               0,               0,               0,     -0.92435141,       3.8819585,      -3.9782941 ],
  [               0,               0,               0,       6.6268956,      -26.523383,       23.471681 ],
 ],
 [
  [               0,               0,               0,   -0.0020057204,    0.0010736062,     0.020112183 ],
  [               0,               0,               0,       19.331755,      -91.759163,       111.08429 ],
  [               0,               0,               0,      -2.8817271,       13.447959,      -16.848504 ],
  [               0,               0,               0,     -0.65597722,       19.275412,      -33.483936 ],
  [               0,               0,               0,     0.012277983,     0.012482745,  -5.1693795e-05 ],
  [               0,               0,               0,       1.3633894,      -6.3234735,       7.3335521 ],
  [               0,               0,               0,      -4.3080963,       18.154961,      -21.770935 ],
 ],
 [
  [               0,               0,               0,               0,    0.0033432136,    0.0042830185 ],
  [               0,               0,               0,               0,      -45.396925,       113.96878 ],
  [               0,               0,               0,               0,       8.5839513,      -22.422004 ],
  [               0,               0,               0,               0,       1.7729671,      -4.1478181 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00055624414,     0.011193183,    -0.026459955,    0.0053223736 ],
  [               0,               0,       421.76539,       757.10919,      -4471.8153,      -4.6071971 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       1.3696216,       33.625644,      -18.383038,        7.754018 ],
  [               0,               0,      0.99022859,        17.90151,      -2.1014931,      -1.3774737 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,  -0.00037107244,   0.00021558307,    0.0079405134 ],
  [               0,               0,               0,       2.5748809,      -1.4738571,       4.9804964 ],
  [               0,               0,               0,     -0.63519997,      0.37921356,      -1.3919136 ],
  [               0,               0,               0,       340.00545,      -176.94857,       17.646688 ],
  [               0,               0,               0,      -2.7075573,       1.4089161,      -0.1405233 ],
  [               0,               0,               0,       2.7935094,      -1.0854286,     0.071730665 ],
  [               0,               0,               0,      -4.3105336,       1.4094175,    -0.049276701 ],
 ],
 [
  [               0,               0,               0,   -0.0050520732,    0.0056413125,    0.0055607035 ],
  [               0,               0,               0,       3.1280133,      -4.9266669,       6.8956124 ],
  [               0,               0,               0,      -1.0740984,      0.83424861,      -1.4711702 ],
  [               0,               0,               0,      0.52939022,     -0.30118839,    -0.023595554 ],
  [               0,               0,               0,    0.0010068449,  -0.00057753684,   0.00044882017 ],
  [               0,               0,               0,      -1.4611416,       1.8304859,       0.2572356 ],
  [               0,               0,               0,       9.8456492,      -14.779586,       3.6850351 ],
 ],
 [
  [               0,               0,               0,               0,     0.051656119,    -0.032930625 ],
  [               0,               0,               0,               0,      -12.718584,       14.581795 ],
  [               0,               0,               0,               0,        8.642956,      -8.2271295 ],
  [               0,               0,               0,               0,      -40.002267,       31.340265 ],
  [               0,               0,               0,               0,      0.33639687,     -0.26345285 ],
  [               0,               0,               0,               0,      -5.3881377,       4.4395891 ],
  [               0,               0,               0,               0,       2.0810011,      -1.8005837 ],
 ],
 [
  [               0,               0,               0,   -0.0056744668,     0.013577336,     0.001816236 ],
  [               0,               0,               0,       2.3764784,      -5.9234185,       7.1195553 ],
  [               0,               0,               0,     0.011321168,    -0.096647674,      -1.2202444 ],
  [               0,               0,               0,     -0.12075723,      0.37360435,     -0.24259834 ],
  [               0,               0,               0,    0.0004474453,   -0.0010344237,   0.00048043496 ],
  [               0,               0,               0,      -4.6005064,       10.509137,      -2.7033427 ],
  [               0,               0,               0,        4.735125,      -9.3993498,     -0.18759322 ],
 ],
 [
  [               0,        2.061259,       -1.321432,      -3.8023962,      -2.7272901,       7.0164401 ],
  [               0,      -1686.0524,       1192.3746,       3164.1079,       2097.7502,       -5929.376 ],
  [               0,      -1651.4058,         1211.66,       3050.7043,       1904.5713,      -5636.6158 ],
  [               0,      -728.28758,       514.97645,        1371.708,       912.19429,      -2581.6566 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.18710925,    0.0076671562,     -0.37442889 ],
  [               0,               0,               0,       245.99114,       4.6256337,      -485.36462 ],
  [               0,               0,               0,      -12.914039,      0.44642286,       22.047904 ],
  [               0,               0,               0,     0.074575175,      0.12838459,    -0.070920141 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_35_e = [
 [
  [       88.965257,      -179.59376,      -142.91002,       210.97821,       513.67481,       -529.0769 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     -0.18701963,      0.14802077,      0.42224205,      0.30163092,     -0.94509035 ],
  [               0,       376.96551,      -281.65135,      -813.47514,      -613.32235,       1796.6251 ],
  [               0,       18.034332,      -2.9430063,      -49.520568,       -66.55844,       128.70431 ],
  [               0,      -47.824314,       36.013265,       102.38869,       72.678778,      -218.99863 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.11850349,       0.1850787 ],
  [               0,               0,               0,               0,       137.38829,      -204.22619 ],
  [               0,               0,               0,               0,      -67.880195,        101.4708 ],
  [               0,               0,               0,               0,       60.248745,      -91.046632 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00065860627,   -0.0008330196,    0.0069912662 ],
  [               0,               0,               0,       5.3920656,      -19.753571,        22.05002 ],
  [               0,               0,               0,      -3.6727446,       12.318154,       -11.65234 ],
  [               0,               0,               0,       2.3018373,      -7.6630564,       6.4059592 ],
  [               0,               0,               0,    0.0050914477,     -0.01694589,     0.014647538 ],
  [               0,               0,               0,       1.1987067,      -4.3443362,       4.6998286 ],
  [               0,               0,               0,      -23.619784,       79.356236,      -68.093057 ],
 ],
 [
  [               0,               0,               0,   -0.0043892267,     0.021886911,    -0.017043281 ],
  [               0,               0,               0,       2.9089585,      -14.515786,       20.731306 ],
  [               0,               0,               0,      0.68001007,      -2.5239541,      0.99679593 ],
  [               0,               0,               0,     -0.27304128,      -3.7888534,       10.338537 ],
  [               0,               0,               0,    -0.010749248,    0.0043976808,     0.049236528 ],
  [               0,               0,               0,     -0.54690548,       2.3433623,      -2.4323614 ],
  [               0,               0,               0,       7.2532021,       -29.20053,       26.387488 ],
 ],
 [
  [               0,               0,               0,   -0.0011538232,   -0.0046679801,      0.02767622 ],
  [               0,               0,               0,        21.46626,      -101.18941,       122.04029 ],
  [               0,               0,               0,        4.194155,      -19.578414,       21.557863 ],
  [               0,               0,               0,      -4.1074507,        31.73881,      -40.029596 ],
  [               0,               0,               0,  -0.00071091659,       0.0127187,     0.096372195 ],
  [               0,               0,               0,     -0.87309252,       3.9147414,      -4.3451449 ],
  [               0,               0,               0,       8.1088513,      -37.526124,       40.299773 ],
 ],
 [
  [               0,               0,               0,               0,    -0.047738367,      0.12144633 ],
  [               0,               0,               0,               0,       8.1505893,       -16.36861 ],
  [               0,               0,               0,               0,      -174.43048,       428.22382 ],
  [               0,               0,               0,               0,       176.13352,       -433.5571 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0027380358,    0.0021888884,   -0.0057269525,    -0.040775433 ],
  [               0,               0,      -1.1770952,      -1.3957978,      -3.4643562,      -4.5985207 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       521.04278,       435.87202,      -1054.0983,      -7928.1754 ],
  [               0,               0,       0.6997973,       1.4039604,       5.3845228,       15.652893 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,   -0.0082014881,    0.0039327878,    0.0075970426 ],
  [               0,               0,               0,       5.9393269,      -3.0848374,       5.1363536 ],
  [               0,               0,               0,      0.97420607,      -0.4773789,      -1.2868884 ],
  [               0,               0,               0,       48.636311,      -20.130257,      0.49662523 ],
  [               0,               0,               0,     -0.39468966,      0.16388615,   -0.0043308446 ],
  [               0,               0,               0,       8.6118806,      -3.8857466,      0.38649547 ],
  [               0,               0,               0,      -10.381367,        3.923205,     -0.31887996 ],
 ],
 [
  [               0,               0,               0,   -0.0034810469,    0.0045993967,    0.0055841874 ],
  [               0,               0,               0,         4.09733,      -6.4900635,       7.4804882 ],
  [               0,               0,               0,      -1.6721864,       1.6513568,      -1.7347121 ],
  [               0,               0,               0,      0.92084264,     -0.81999469,      0.13984912 ],
  [               0,               0,               0,   -0.0010100366,    0.0016322425,  -0.00011934313 ],
  [               0,               0,               0,       -4.527268,       5.3989404,     -0.74274151 ],
  [               0,               0,               0,       2.0859988,       -4.671648,      0.50462499 ],
 ],
 [
  [               0,               0,               0,               0,    0.0097772769,   0.00056159049 ],
  [               0,               0,               0,               0,      -1.3852254,       5.5005263 ],
  [               0,               0,               0,               0,       8.2856462,      -7.9065729 ],
  [               0,               0,               0,               0,      -6.1230476,       4.8085223 ],
  [               0,               0,               0,               0,     0.014884613,    -0.011374749 ],
  [               0,               0,               0,               0,       6.8780752,      -5.1371776 ],
  [               0,               0,               0,               0,       17.085933,      -14.148605 ],
 ],
 [
  [               0,               0,               0,   -0.0018816864,    0.0020754442,     0.007202865 ],
  [               0,               0,               0,       1.0216555,      -2.1137603,       5.5032241 ],
  [               0,               0,               0,     -0.83580955,       1.4853694,      -2.0027818 ],
  [               0,               0,               0,      0.45057862,     -0.74134423,      0.30149086 ],
  [               0,               0,               0,   0.00057947744,  -0.00040401682,   0.00029965303 ],
  [               0,               0,               0,      0.84154272,      -1.6534482,       1.6232849 ],
  [               0,               0,               0,       -4.794047,       10.013582,      -6.7823802 ],
 ],
 [
  [               0,       2.1155637,      -1.3557112,      -3.9018398,      -2.7991975,       7.1987117 ],
  [               0,      -1959.1206,       1384.1401,       3670.7376,       2432.2005,      -6869.8937 ],
  [               0,      -1559.8828,        1150.619,       2877.5872,       1781.9505,      -5309.0743 ],
  [               0,      -669.05975,       473.95883,       1259.5634,       835.61008,      -2369.4947 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,        19.99639,      -56.577043,       40.031321 ],
  [               0,               0,               0,       338.33699,       29.640994,      -708.57015 ],
  [               0,               0,               0,      -67.166642,       62.604313,       42.480076 ],
  [               0,               0,               0,      -1.7456624,      0.55376847,       3.2026908 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
ele1000_layer2_etaoffv4_1_37_e = [
 [
  [       79.624236,      -160.86423,      -127.96036,       189.20288,       460.49629,      -474.68229 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.037895904,    -0.015365876,    -0.063567331,    -0.059279696,      0.10364355 ],
  [               0,       435.57653,      -321.23678,      -953.29856,      -706.87413,       2088.4465 ],
  [               0,       20.535747,      -5.3268176,      -54.329389,      -68.498697,       137.82097 ],
  [               0,      -51.914428,       39.053247,       111.16656,       79.015266,        -237.833 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.1147595,      0.17891935 ],
  [               0,               0,               0,               0,       156.09843,      -232.41219 ],
  [               0,               0,               0,               0,      -86.345802,       129.41327 ],
  [               0,               0,               0,               0,        78.03257,      -117.93854 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00039636237,  -4.7593836e-05,    0.0064331337 ],
  [               0,               0,               0,       5.4269861,      -19.825626,       22.069671 ],
  [               0,               0,               0,      -3.6804465,       12.344343,      -11.672451 ],
  [               0,               0,               0,       2.3157279,       -7.712216,       6.4476062 ],
  [               0,               0,               0,    0.0045839218,    -0.015204208,     0.013170229 ],
  [               0,               0,               0,       1.4308092,      -5.1157428,       5.3313724 ],
  [               0,               0,               0,      -23.640838,       79.487834,      -68.239429 ],
 ],
 [
  [               0,               0,               0,   -0.0044702058,      0.02233033,    -0.017577686 ],
  [               0,               0,               0,       2.9121433,      -14.525619,       20.738983 ],
  [               0,               0,               0,      0.65928425,      -2.4400301,      0.91292878 ],
  [               0,               0,               0,     -0.23650054,      -4.2887955,       11.322842 ],
  [               0,               0,               0,     -0.01146327,    0.0045737337,     0.053014696 ],
  [               0,               0,               0,     -0.47972427,       2.0559054,       -2.130718 ],
  [               0,               0,               0,       7.1202817,      -28.642639,       25.809256 ],
 ],
 [
  [               0,               0,               0,   -0.0013518314,   -0.0040340219,     0.027565927 ],
  [               0,               0,               0,       23.584609,      -111.36029,       134.20303 ],
  [               0,               0,               0,       1.2313544,       -5.505472,       4.8589855 ],
  [               0,               0,               0,      -6.7827133,       40.845108,      -50.725969 ],
  [               0,               0,               0,     0.001644788,    0.0081109327,      0.05436372 ],
  [               0,               0,               0,     -0.79184967,       3.4793402,       -3.764623 ],
  [               0,               0,               0,       8.5132067,      -39.152376,       41.896842 ],
 ],
 [
  [               0,               0,               0,               0,    0.0013353494,    0.0073843819 ],
  [               0,               0,               0,               0,      -4.0920806,       12.857943 ],
  [               0,               0,               0,               0,       4.6798089,      -12.803494 ],
  [               0,               0,               0,               0,      0.93068276,      -2.2239693 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0010834358,   0.00040509331,   -0.0024785333,    -0.012824255 ],
  [               0,               0,      -1.1696393,      -1.3995112,      -3.4720997,      -4.6140437 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       1353.9761,       1136.4776,      -2730.0866,      -20554.105 ],
  [               0,               0,       1.7419048,       9.0442228,       21.121344,       43.717264 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 5 GeV photon clusters
#


gam5_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       8185.3434,       1972.3534 ],
  [               0,               0,               0,               0,    -0.015964966,     0.013852977 ],
  [               0,               0,               0,               0,     0.015963014,    -0.013852035 ],
  [               0,               0,               0,               0,     0.015965103,    -0.013852527 ],
  [               0,               0,               0,               0,   1.4999006e-07,   -6.141437e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0020815652,   -0.0024506011,   0.00029318101,    0.0003853144 ],
  [               0,               0,       108.32395,       -4.697073,      -16.120685,       1.6774807 ],
  [               0,               0,       103.56512,      -40.788956,      -35.987032,       15.715642 ],
  [               0,               0,      -4.3002419,      -3.0762598,       6.7918318,       6.1327616 ],
  [               0,               0,       18.830918,      -13.401945,      -9.1726738,       5.3311964 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -12059.019,       10058.556 ],
  [               0,               0,               0,               0,    0.0098478951,    0.0074772465 ],
  [               0,               0,               0,               0,    -0.009847772,   -0.0074770114 ],
  [               0,               0,               0,               0,   -0.0098464878,    -0.007477265 ],
  [               0,               0,               0,               0,   1.5031519e-08,  -1.7667773e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      -15017.911,       40719.334,       -35597.53,       10236.413 ],
  [               0,               0,       3.3130805,      -5.8492742,       3.4635986,     -0.66587219 ],
  [               0,               0,      -3.3124484,       5.8484105,      -3.4632039,      0.66581251 ],
  [               0,               0,       -3.312621,       5.8485577,      -3.4632068,      0.66580141 ],
  [               0,               0,     0.000168715,  -0.00030093855,   0.00017073916,  -3.1106701e-05 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -12853.182,       9191.7578 ],
  [               0,               0,               0,               0,    0.0093629699,    0.0079835336 ],
  [               0,               0,               0,               0,   -0.0093626796,   -0.0079832809 ],
  [               0,               0,               0,               0,   -0.0093613428,   -0.0079835645 ],
  [               0,               0,               0,               0,  -2.2860221e-09,  -1.8735072e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       189.86318,      -356.99858,       215.65419,       -40.84248 ],
  [               0,               0,       13.665468,      -21.351442,       11.332191,      -1.8963455 ],
  [               0,               0,      -13.343555,       20.944801,      -11.171359,       1.8776728 ],
  [               0,               0,      -13.409397,       21.039047,      -11.221069,       1.8884718 ],
  [               0,               0,     0.021002471,     -0.03746355,     0.020060931,   -0.0032582001 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.067706036,     0.047273933,   0.00053441174 ],
  [               0,               0,               0,       31.150033,      -18.353017,       4.1189512 ],
  [               0,               0,               0,      -6.4721111,       3.0844053,      -1.3455466 ],
  [               0,               0,               0,      0.39460963,      0.11460341,    -0.015879013 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.071306137,    -0.061197291,     0.015426875 ],
  [               0,               0,               0,       31.902572,      -50.342307,       20.797309 ],
  [               0,               0,               0,       2.5988776,      -1.3418118,      -1.2451303 ],
  [               0,               0,               0,     -0.79585359,       0.9933763,     -0.17535733 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.044387811,    -0.034245118 ],
  [               0,               0,               0,               0,      -153.89008,        131.2782 ],
  [               0,               0,               0,               0,       47.582981,      -40.523209 ],
  [               0,               0,               0,               0,      -33.713847,       27.589379 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     -0.15927667,     0.077028722,    0.0043890174 ],
  [               0,               0,               0,       26.638563,      -12.698124,       3.3526107 ],
  [               0,               0,               0,      -1.2766356,      0.57106545,      -1.1801731 ],
  [               0,               0,               0,      -1.1609975,      0.54488147,    -0.035268086 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.068015961,     0.083795867,    -0.012993783 ],
  [               0,               0,               0,         1.30326,      -3.1112897,        3.320834 ],
  [               0,               0,               0,     -0.87442358,       1.2578004,      -1.5619559 ],
  [               0,               0,               0,       1.1312214,      -1.1541563,      0.32466394 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.34409511,     -0.26753402 ],
  [               0,               0,               0,               0,      -117.42108,       97.417485 ],
  [               0,               0,               0,               0,       45.263187,      -37.739828 ],
  [               0,               0,               0,               0,      -7.7455738,       6.2810633 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,     -0.16737374,     0.079850764,    0.0043155947 ],
  [               0,               0,               0,        25.86713,      -11.914037,       3.1289029 ],
  [               0,               0,               0,      -3.5001461,       1.3626788,      -1.2005977 ],
  [               0,               0,               0,     -0.34516098,      0.21251069,    -0.010524237 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.10223553,      0.12514935,    -0.027358545 ],
  [               0,               0,               0,       15.039347,      -20.299139,       8.7684883 ],
  [               0,               0,               0,      -2.3849109,       3.1475705,      -2.1841276 ],
  [               0,               0,               0,        1.217134,       -1.196299,      0.33202994 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.049157268,    -0.028838589 ],
  [               0,               0,               0,               0,       8.1572284,      -4.5610985 ],
  [               0,               0,               0,               0,      -3.1513669,       1.3583457 ],
  [               0,               0,               0,               0,      -2.3783331,        1.933431 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam5_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 10 GeV photon clusters
#


gam10_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.0197278,    0.0083681822 ],
  [               0,               0,               0,               0,       2.1774105,      0.72563116 ],
  [               0,               0,               0,               0,      -405.76935,       51.035603 ],
  [               0,               0,               0,               0,      -405.63506,       51.029634 ],
  [               0,               0,               0,               0,       98.978137,      -23.892262 ],
  [               0,               0,               0,               0,      -10.430296,       4.4293081 ],
  [               0,               0,               0,               0,     -0.24340297,     0.031078136 ],
  [               0,               0,               0,               0,      0.18711833,       1.5256492 ],
 ],
 [
  [               0,               0,               0,    -0.011250866,     0.012739859,   -0.0023968405 ],
  [               0,               0,               0,       25.993315,      -42.496753,       13.643185 ],
  [               0,               0,               0,      -49.426428,       54.338336,      -13.422029 ],
  [               0,               0,               0,      -17.307891,       25.024529,      -7.7182331 ],
  [               0,               0,               0,       8.9339957,      -8.9000548,       1.9071007 ],
  [               0,               0,               0,   0.00061055518,   -0.0040696368,    0.0013647245 ],
  [               0,               0,               0,       2.4927022,      -3.2189211,       2.0287186 ],
  [               0,               0,               0,      -1.7270065,     -0.40876391,       2.5518905 ],
 ],
 [
  [               0,               0,               0,               0,    -0.014247026,     0.011746744 ],
  [               0,               0,               0,               0,       22.391231,       21.914731 ],
  [               0,               0,               0,               0,       939.01681,      -750.68365 ],
  [               0,               0,               0,               0,       378.48893,      -300.09448 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0039686087,   -0.0081120345,    0.0044906006 ],
  [               0,               0,               0,      -57.586881,       151.23554,      -94.012243 ],
  [               0,               0,               0,       16.800175,      -48.149855,       33.664298 ],
  [               0,               0,               0,      -3.0782061,       1.0913736,       3.2707933 ],
  [               0,               0,               0,     0.070563555,       1.7939041,      -2.7792607 ],
  [               0,               0,               0,   -0.0008869245,   0.00017588957,    0.0010697055 ],
  [               0,               0,               0,     -0.82662409,        -1.35472,       3.2994341 ],
  [               0,               0,               0,      -12.224228,       18.872579,      -8.3502557 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.27793611,     -0.38821087,      -0.5600543,      0.78623338 ],
  [               0,               0,      -4783.3297,       6851.0618,       9551.9352,      -13674.994 ],
  [               0,               0,       3684.0511,      -5246.0304,      -7704.7887,       10971.223 ],
  [               0,               0,        513.7159,      -736.75993,      -1104.1033,       1580.4299 ],
  [               0,               0,       1209.8178,      -1720.5527,       -2436.997,       3464.4274 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.010027087 ],
  [               0,               0,               0,               0,               0,       1.0687201 ],
  [               0,               0,               0,               0,               0,     -0.87412923 ],
  [               0,               0,               0,               0,               0,     -0.80984571 ],
  [               0,               0,               0,               0,               0,   -0.0013184969 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00071647474,   0.00042714579 ],
  [               0,               0,               0,               0,       17.960259,      -2.5591922 ],
  [               0,               0,               0,               0,      -24.525288,       3.2999329 ],
  [               0,               0,               0,               0,      -23.298839,       4.0968642 ],
  [               0,               0,               0,               0,      0.51219483,     0.067822787 ],
  [               0,               0,               0,               0,    -0.001934078,  -0.00016224204 ],
  [               0,               0,               0,               0,      -1.6484489,       1.2259258 ],
  [               0,               0,               0,               0,     -0.31529047,       1.5191536 ],
 ],
 [
  [               0,               0,               0,      0.03470599,    -0.025168859,    0.0045754919 ],
  [               0,               0,               0,       4.3267707,      -9.4698388,       6.1139566 ],
  [               0,               0,               0,       15.711939,       -20.96085,       5.8666832 ],
  [               0,               0,               0,       3.3114889,      -4.7703222,       0.6214082 ],
  [               0,               0,               0,      -3.9754622,       5.5819234,      -1.9993985 ],
  [               0,               0,               0,    0.0086127571,    -0.011495986,    0.0037095128 ],
  [               0,               0,               0,       15.384969,      -10.448869,       2.5333752 ],
  [               0,               0,               0,      -15.253498,       16.735322,       0.2058426 ],
 ],
 [
  [               0,               0,               0,               0,    -0.031489434,     0.025462044 ],
  [               0,               0,               0,               0,       22.472082,       21.979978 ],
  [               0,               0,               0,               0,       1015.9579,      -811.85643 ],
  [               0,               0,               0,               0,       459.29154,      -364.34799 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.072989852,     -0.12721739,     0.058139329 ],
  [               0,               0,               0,       7.8871462,      -19.526132,       12.937387 ],
  [               0,               0,               0,      -4.7356803,       11.041925,       -7.097241 ],
  [               0,               0,               0,      -5.2149281,       12.416456,      -8.0891108 ],
  [               0,               0,               0,     -0.28382931,      0.87792434,     -0.68101836 ],
  [               0,               0,               0,   0.00049347566,   -0.0018329599,    0.0013396206 ],
  [               0,               0,               0,      -2.1446923,      0.53398431,       3.2506837 ],
  [               0,               0,               0,      -17.688591,       27.631772,      -10.065396 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.15577492,       -0.196965,     -0.29830014,       0.3815932 ],
  [               0,               0,       4043.1987,      -5838.9751,      -8131.1279,       11745.214 ],
  [               0,               0,      -95.848202,        209.1855,      -201.51412,       140.02497 ],
  [               0,               0,       1646.8847,       -2305.325,      -3463.5065,       4849.1687 ],
  [               0,               0,       2101.6114,      -2991.3068,       -4253.493,        6052.709 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       898.45347 ],
  [               0,               0,               0,               0,               0,     0.018648919 ],
  [               0,               0,               0,               0,               0,    -0.018647229 ],
  [               0,               0,               0,               0,               0,     -0.01864661 ],
  [               0,               0,               0,               0,               0,   3.0754033e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0042564534,    0.0020755529 ],
  [               0,               0,               0,               0,       8.9802985,      -1.4663218 ],
  [               0,               0,               0,               0,      -9.3737703,        1.455916 ],
  [               0,               0,               0,               0,      -7.9038396,       1.9592154 ],
  [               0,               0,               0,               0,      0.41443685,     0.068631227 ],
  [               0,               0,               0,               0,  -0.00095994309,  -0.00018089356 ],
  [               0,               0,               0,               0,      -1.9999429,        1.511752 ],
  [               0,               0,               0,               0,       0.6904829,      0.93611556 ],
 ],
 [
  [               0,               0,               0,       9.3689665,      0.85979074,       4.2039874 ],
  [               0,               0,               0,      0.22534895,     -0.23147609,      0.14362727 ],
  [               0,               0,               0,     -0.22326435,      0.22895922,     -0.14269754 ],
  [               0,               0,               0,      -0.2247272,      0.23075822,      -0.1432512 ],
  [               0,               0,               0,  -0.00016589338,   0.00018684966,  -7.0611648e-05 ],
  [               0,               0,               0,   -0.0022567237,     0.002419986,  -0.00064797358 ],
  [               0,               0,               0,      -34.267993,       35.497044,      -5.4436381 ],
  [               0,               0,               0,       38.450048,      -34.475626,       8.9516115 ],
 ],
 [
  [               0,               0,               0,               0,     0.089191291,    -0.048351406 ],
  [               0,               0,               0,               0,      -40.457587,       32.862523 ],
  [               0,               0,               0,               0,       30.870086,      -25.167563 ],
  [               0,               0,               0,               0,       10.179728,      -8.0849517 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.017534272,    -0.046872324,     0.031617028 ],
  [               0,               0,               0,       27.920621,      -40.313132,       12.101467 ],
  [               0,               0,               0,      -1.8355571,       2.7631119,      -1.1272549 ],
  [               0,               0,               0,      -5.7893133,       7.9772851,      -2.3028506 ],
  [               0,               0,               0,      -5.5709349,       9.4839583,       -3.954307 ],
  [               0,               0,               0,     0.025126347,    -0.064743796,       0.0411326 ],
  [               0,               0,               0,      -7.1870976,       12.149772,      -4.3904061 ],
  [               0,               0,               0,      0.22682246,      -0.6721769,      -1.2987261 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       0.1896357,     -0.25587943,     -0.37440708,      0.51377967 ],
  [               0,               0,       2794.6824,      -4038.5673,      -5638.3593,       8149.9308 ],
  [               0,               0,      -736.46961,       1151.4679,       1047.4888,       -1699.456 ],
  [               0,               0,       1608.3566,      -2253.5975,      -3405.3254,       4772.5918 ],
  [               0,               0,       2406.3441,      -3428.4644,      -4872.4112,       6940.5683 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       271.93193 ],
  [               0,               0,               0,               0,               0,     0.029257647 ],
  [               0,               0,               0,               0,               0,    -0.029251908 ],
  [               0,               0,               0,               0,               0,    -0.029249791 ],
  [               0,               0,               0,               0,               0,   1.2226363e-07 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,   -0.0037165791,     0.012125494,    0.0054641118 ],
  [               0,               0,               0,       4.3569251,      -5.4870419,       3.7832014 ],
  [               0,               0,               0,      -2.1189776,       1.1854584,      -1.3895926 ],
  [               0,               0,               0,      0.75719448,     -0.23800752,     0.040244019 ],
  [               0,               0,               0,   -0.0016571502,  -0.00072366203,   0.00034218776 ],
  [               0,               0,               0,      -17.401771,      -5.1353954,       3.1251717 ],
  [               0,               0,               0,      -1.1878479,      -1.9575725,      -4.2961774 ],
 ],
 [
  [               0,               0,               0,    0.0092770862,     0.010980044,     0.013260617 ],
  [               0,               0,               0,      0.49514373,      -2.2168748,       2.8871217 ],
  [               0,               0,               0,      -2.7825075,       3.6345355,      -2.1232164 ],
  [               0,               0,               0,       1.8755834,      -2.1364797,      0.57017188 ],
  [               0,               0,               0,   -0.0023197984,    0.0026383647,     0.001487394 ],
  [               0,               0,               0,       2.5715482,      -3.5048873,       2.0852545 ],
  [               0,               0,               0,       16.618877,      -19.949219,      0.69253048 ],
 ],
 [
  [               0,               0,               0,               0,      0.44415574,     -0.34482254 ],
  [               0,               0,               0,               0,      -43.932671,       37.279035 ],
  [               0,               0,               0,               0,       26.461326,      -22.503744 ],
  [               0,               0,               0,               0,      -7.7338588,       6.2593947 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0027829137,   -0.0010453468,     0.012505062 ],
  [               0,               0,               0,      -0.5960838,       1.6318134,       1.1955701 ],
  [               0,               0,               0,      -3.6188374,       7.5045331,      -4.9598335 ],
  [               0,               0,               0,       1.5169461,      -2.5020311,       1.0235308 ],
  [               0,               0,               0,   -0.0079505763,     0.015371476,   -0.0065411961 ],
  [               0,               0,               0,      -1.1740658,       5.9996459,      -4.0139017 ],
  [               0,               0,               0,      -1.0164184,       2.9431253,      -1.4627992 ],
 ],
 [
  [               0,               0,               0,   -0.0083491043,    0.0010734027,     0.014566792 ],
  [               0,               0,               0,       1.9316648,       9.4571033,       22.971304 ],
  [               0,               0,               0,       -159.7553,      -7.3334132,       304.59673 ],
  [               0,               0,               0,       272.86232,      -5.1352686,      -493.72603 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,       3.6651875,      -10.343832,       7.2994735 ],
  [               0,               0,               0,      -508.40376,       281.33019,       643.14316 ],
  [               0,               0,               0,      -323.34673,       171.63358,       388.58994 ],
  [               0,               0,               0,       -4.917295,      0.73602766,       9.8973246 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,        125.6538 ],
  [               0,               0,               0,               0,               0,     0.046506871 ],
  [               0,               0,               0,               0,               0,    -0.046472307 ],
  [               0,               0,               0,               0,               0,  -7.2948685e-07 ],
  [               0,               0,               0,               0,               0,   0.00062595776 ],
  [               0,               0,               0,               0,               0,       1.6375718 ],
  [               0,               0,               0,               0,               0,      -7.9198434 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     0.026200218,   -0.0051087692,    0.0062344074 ],
  [               0,               0,               0,      -27.817545,       10.617659,        3.089842 ],
  [               0,               0,               0,        2.534125,     -0.66675713,       -1.251898 ],
  [               0,               0,               0,      0.81852467,     -0.48599583,     0.017532496 ],
  [               0,               0,               0,    0.0046487963,   -0.0027813142,    0.0010617161 ],
  [               0,               0,               0,       20.034779,       -8.906829,       1.4801986 ],
  [               0,               0,               0,       -1.184336,       3.0365805,      -1.5389951 ],
 ],
 [
  [               0,               0,               0,    -0.020755824,     0.023550057,   -0.0026236049 ],
  [               0,               0,               0,      -23.806997,       28.430647,      -3.6165785 ],
  [               0,               0,               0,      -7.1579283,       9.9714901,      -4.5909316 ],
  [               0,               0,               0,       6.2210139,      -7.7340392,       2.3757416 ],
  [               0,               0,               0,    -0.012782516,     0.015678023,   -0.0060852714 ],
  [               0,               0,               0,     -0.51861073,      0.27648684,      0.99946344 ],
  [               0,               0,               0,      -8.1995629,       12.062581,       -8.648324 ],
 ],
 [
  [               0,               0,               0,               0,      0.47678043,     -0.36811932 ],
  [               0,               0,               0,               0,      -47.649288,       40.436365 ],
  [               0,               0,               0,               0,       18.493415,       -16.05656 ],
  [               0,               0,               0,               0,      -3.3108336,       2.6713508 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.032549642,    -0.099349727,     0.083643033 ],
  [               0,               0,               0,       4.1329618,      -8.7930726,       6.7802116 ],
  [               0,               0,               0,     -0.46000239,      0.81058352,      -1.4512657 ],
  [               0,               0,               0,      0.44403698,     -0.61079731,      0.17697279 ],
  [               0,               0,               0,     0.019312424,    -0.047823356,     0.029932405 ],
  [               0,               0,               0,       4.1630997,      -8.2528933,       5.0471245 ],
  [               0,               0,               0,      -7.4965142,       12.365994,      -3.7568746 ],
 ],
 [
  [               0,               0,               0,      0.12552755,    -0.021955819,     -0.20256251 ],
  [               0,               0,               0,      -86.808809,      -14.138991,       187.76211 ],
  [               0,               0,               0,      -2960.5767,       8199.8296,      -5680.2865 ],
  [               0,               0,               0,       2213.1897,      -6147.5708,       4270.1574 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.013477939,   -0.0015189983,     0.034523005 ],
  [               0,               0,               0,      -433.42295,       4.9053012,       885.77323 ],
  [               0,               0,               0,        15.30444,      -91.800505,        99.75123 ],
  [               0,               0,               0,     -0.41071343,     -0.85448953,      -1.1769636 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0042787575 ],
  [               0,               0,               0,               0,               0,       3.0275215 ],
  [               0,               0,               0,               0,               0,      -1.5149996 ],
  [               0,               0,               0,               0,               0,       0.1529299 ],
  [               0,               0,               0,               0,               0,    0.0014747727 ],
  [               0,               0,               0,               0,               0,       1.0470709 ],
  [               0,               0,               0,               0,               0,      -8.2669513 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,     0.037907475,    -0.010307715,    0.0067545146 ],
  [               0,               0,               0,      -31.377265,       12.798052,        2.824666 ],
  [               0,               0,               0,       1.5947772,      -1.0104142,      -1.1681043 ],
  [               0,               0,               0,       2.8989846,       -0.701634,   -0.0024755419 ],
  [               0,               0,               0,     0.011588367,   -0.0041253696,    0.0011356911 ],
  [               0,               0,               0,       10.230957,      -7.1921996,       1.4190823 ],
  [               0,               0,               0,      -21.585662,       5.6975945,       -1.502178 ],
 ],
 [
  [               0,               0,               0,    -0.010117623,     0.011304028,    0.0062813408 ],
  [               0,               0,               0,      -10.628644,       11.691468,        -0.14486 ],
  [               0,               0,               0,        2.124624,      -1.8935017,     -0.91971468 ],
  [               0,               0,               0,     0.039612783,     -0.13627579,       0.1180336 ],
  [               0,               0,               0,     -0.01300812,      0.01444796,   -0.0041714977 ],
  [               0,               0,               0,      -10.152601,       11.313425,      -1.8152013 ],
  [               0,               0,               0,      -15.706057,       8.8327006,     -0.94341844 ],
 ],
 [
  [               0,               0,               0,               0,        0.449605,     -0.34649773 ],
  [               0,               0,               0,               0,       -36.71725,       31.586555 ],
  [               0,               0,               0,               0,       15.704469,      -13.806894 ],
  [               0,               0,               0,               0,      -3.7470404,       3.0179719 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0040555648,    -0.007523746,     0.012599093 ],
  [               0,               0,               0,     -0.49660399,     -0.30920343,       3.4276511 ],
  [               0,               0,               0,       0.7345772,      -1.1172645,     -0.84336265 ],
  [               0,               0,               0,     -0.77543951,       1.9634463,      -1.1510737 ],
  [               0,               0,               0,    0.0068052629,    -0.014541025,    0.0082110307 ],
  [               0,               0,               0,      -8.7550814,       22.058051,        -12.6109 ],
  [               0,               0,               0,      -1.0229109,       1.2138254,     -0.25252089 ],
 ],
 [
  [               0,               0,               0,       805.14822,      -2210.3518,       1517.2402 ],
  [               0,               0,               0,      -27.186877,       75.701669,      -52.494556 ],
  [               0,               0,               0,       41.430557,      -117.61672,       83.172681 ],
  [               0,               0,               0,      -89.731581,       247.57502,      -170.73794 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,       44.503822,      -125.37474,       88.308146 ],
  [               0,               0,               0,      -3341.0864,       9201.8187,      -6320.2893 ],
  [               0,               0,               0,       7985.6031,      -22499.681,       15859.155 ],
  [               0,               0,               0,       308.74388,      -528.39611,       121.10285 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam10_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0040854559 ],
  [               0,               0,               0,               0,               0,       2.6691158 ],
  [               0,               0,               0,               0,               0,      -1.3512232 ],
  [               0,               0,               0,               0,               0,     0.090101532 ],
  [               0,               0,               0,               0,               0,    0.0016662073 ],
  [               0,               0,               0,               0,               0,        1.079532 ],
  [               0,               0,               0,               0,               0,      -8.0580609 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 25 GeV photon clusters
#


gam25_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -3.5569007e-08,    0.0015243048 ],
  [               0,               0,               0,               0,    -0.068221265,       2.0204286 ],
  [               0,               0,               0,               0,     -0.73274038,      -0.7323691 ],
  [               0,               0,               0,               0,       1.0705237,      -1.2181219 ],
  [               0,               0,               0,               0,       9403.1329,      -2400.4796 ],
  [               0,               0,               0,               0,      -14.333141,       3.6589867 ],
  [               0,               0,               0,               0,    -0.035160752,    0.0074252254 ],
  [               0,               0,               0,               0,    0.0012579219,  -0.00095639001 ],
 ],
 [
  [               0,               0,               0,     0.002973595,   -0.0015466901,    0.0025112056 ],
  [               0,               0,               0,      -9.0094612,       9.9655926,      -1.0346027 ],
  [               0,               0,               0,       5.1322249,      -5.8614719,      0.71466393 ],
  [               0,               0,               0,       5.2289616,       -5.880868,      0.57073838 ],
  [               0,               0,               0,    -0.034777367,    -0.014531318,    -0.039633026 ],
  [               0,               0,               0,     0.002172884,   -0.0025238534,    0.0007033136 ],
  [               0,               0,               0,       2.7872266,       15.638639,      -4.4805577 ],
  [               0,               0,               0,       35.244625,       -35.01963,       11.058572 ],
 ],
 [
  [               0,               0,               0,               0,        672.6493,       184.27721 ],
  [               0,               0,               0,               0,    -0.024151388,     0.037056913 ],
  [               0,               0,               0,               0,     0.023774046,    -0.036755258 ],
  [               0,               0,               0,               0,    0.0005199778,   -0.0004133056 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.056543403,    -0.090942073,     0.038439334 ],
  [               0,               0,               0,     -0.24569464,      -1.7172315,        3.543224 ],
  [               0,               0,               0,       2.4028967,      -4.7261463,       1.3296869 ],
  [               0,               0,               0,       2.4573691,      -4.4479526,      0.83801786 ],
  [               0,               0,               0,     -0.34042262,      0.96391357,     -0.69445725 ],
  [               0,               0,               0,    0.0014731347,   -0.0033204636,    0.0017525508 ],
  [               0,               0,               0,       39.190833,      -87.614737,       50.837727 ],
  [               0,               0,               0,       4.7518092,      -10.482996,       9.8143666 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.038990621,     0.026888195,    -0.018886615,     -0.13098065 ],
  [               0,               0,       183.59171,      -279.81815,      -378.38672,       580.21791 ],
  [               0,               0,       862.49404,      -1245.0351,      -1785.8759,       2576.4343 ],
  [               0,               0,       -791.6866,       1116.2115,       1560.7321,      -2201.4438 ],
  [               0,               0,       77.576611,      -104.87866,       -153.4899,       206.70294 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,    -0.056394112,     0.081913874 ],
  [               0,               0,               0,               0,       1574.9971,      -2249.8562 ],
  [               0,               0,               0,               0,       72.873099,      -109.82068 ],
  [               0,               0,               0,               0,       84.282187,      -112.90147 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       645318.21 ],
  [               0,               0,               0,               0,               0,    0.0036410196 ],
  [               0,               0,               0,               0,               0,   -0.0036410091 ],
  [               0,               0,               0,               0,               0,   -0.0036410109 ],
  [               0,               0,               0,               0,               0,  -1.6175229e-09 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -3368.4131,       5159.2224 ],
  [               0,               0,               0,               0,     -0.94933742,        1.468901 ],
  [               0,               0,               0,               0,      0.94836178,      -1.4673556 ],
  [               0,               0,               0,               0,   -0.0012852267,    0.0019129583 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.003579837,   2.5168747e-05,     0.012384742 ],
  [               0,               0,               0,       9.0733141,      -25.481162,       19.296074 ],
  [               0,               0,               0,       17.640284,      -56.578478,       44.303185 ],
  [               0,               0,               0,       2.9185509,      -10.361307,       8.0716724 ],
  [               0,               0,               0,      -7.2902047,       23.363173,      -18.742418 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   4.8266406e-06,    0.0011127183 ],
  [               0,               0,               0,               0,       2.6910103,      -2.2794643 ],
  [               0,               0,               0,               0,      -2.6717093,       4.3439161 ],
  [               0,               0,               0,               0,     -0.58349531,      0.54864202 ],
  [               0,               0,               0,               0,     0.025811026,     -0.15709331 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0008077124,  -0.00055635517,    0.0015366738,      0.01038503 ],
  [               0,               0,       3.5559016,       -6.273643,      -16.680879,       31.776054 ],
  [               0,               0,         3.40869,      -7.2961783,      -14.949732,       30.843062 ],
  [               0,               0,      0.10927988,      0.25790475,      -2.6776877,       2.4875135 ],
  [               0,               0,     -0.65523119,       1.1236382,       4.1324057,      -7.5759707 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       -1.903091,       4.4147424 ],
  [               0,               0,               0,               0,      -23.817821,        55.39252 ],
  [               0,               0,               0,               0,       18.968642,      -44.185712 ],
  [               0,               0,               0,               0,     -0.12823842,      0.32475386 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00056671135,    0.0021559689 ],
  [               0,               0,               0,               0,      -4.5638493,        12.84472 ],
  [               0,               0,               0,               0,       2.0751337,      -5.9341329 ],
  [               0,               0,               0,               0,       1.7089836,      -5.1302389 ],
  [               0,               0,               0,               0,     -0.58731151,       1.3513897 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0095763589,     0.040472919,    -0.037255701 ],
  [               0,               0,               0,      -34.792915,       123.41312,      -86.447626 ],
  [               0,               0,               0,        -75.9504,       339.79512,      -379.03318 ],
  [               0,               0,               0,      -7.9445045,       58.071992,      -94.067313 ],
  [               0,               0,               0,   -0.0031278623,   -0.0028536144,     0.023475182 ],
  [               0,               0,               0,      -4.9896009,       24.385475,      -28.748885 ],
  [               0,               0,               0,      -20.334406,        77.81024,      -86.424121 ],
 ],
 [
  [               0,               0,               0,               0,       2.5637878,      -6.3490107 ],
  [               0,               0,               0,               0,      -1833.6072,        4560.765 ],
  [               0,               0,               0,               0,       2.7347044,       7.7005376 ],
  [               0,               0,               0,               0,      -413.32908,       1022.9038 ],
  [               0,               0,               0,               0,        1.531837,      -4.0363803 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    9.437185e-08,   0.00092616761 ],
  [               0,               0,               0,               0,     -0.45409712,       2.8016774 ],
  [               0,               0,               0,               0,    -0.080719115,     -0.76512503 ],
  [               0,               0,               0,               0,       1.4267238,      -1.2340311 ],
  [               0,               0,               0,               0,       6678.4894,      -1812.3343 ],
  [               0,               0,               0,               0,      -6.1856351,       1.6785295 ],
  [               0,               0,               0,               0,     -0.03738233,    0.0080706741 ],
  [               0,               0,               0,               0,     0.021289824,   -0.0066880264 ],
 ],
 [
  [               0,               0,               0,      0.01371897,    -0.014984253,    0.0064480501 ],
  [               0,               0,               0,      -11.821234,        13.51878,      -2.0588495 ],
  [               0,               0,               0,       4.7750533,      -5.5135029,      0.62303279 ],
  [               0,               0,               0,       3.7848569,      -4.1770741,     0.085944049 ],
  [               0,               0,               0,    -0.046765455,     0.002033303,    -0.044845738 ],
  [               0,               0,               0,    0.0017478881,    -0.002232888,   0.00066577832 ],
  [               0,               0,               0,       130.59373,      -145.50332,       40.827806 ],
  [               0,               0,               0,      -28.494089,       44.943395,      -11.689136 ],
 ],
 [
  [               0,               0,               0,               0,        672.6493,       184.27721 ],
  [               0,               0,               0,               0,    -0.024151388,     0.037056913 ],
  [               0,               0,               0,               0,     0.023774046,    -0.036755258 ],
  [               0,               0,               0,               0,    0.0005199778,   -0.0004133056 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.030263984,     0.016547181,    -0.021306947 ],
  [               0,               0,               0,     0.022744366,     -0.52689309,       1.1730453 ],
  [               0,               0,               0,      0.21941485,      -0.1759568,     -0.63181214 ],
  [               0,               0,               0,     0.036399847,      0.31992095,     -0.97207159 ],
  [               0,               0,               0,    -0.093527253,      0.24449084,     -0.16411989 ],
  [               0,               0,               0,   0.00033771114,  -0.00037030036,   -5.992088e-05 ],
  [               0,               0,               0,       14.091363,      -9.6857646,      -2.9018438 ],
  [               0,               0,               0,     -0.21379285,      -2.1028639,       5.7288536 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.29981077,     -0.39096911,     -0.58870178,      0.77141931 ],
  [               0,               0,       977.58347,      -1425.6045,      -1958.3016,       2861.2753 ],
  [               0,               0,       1222.5547,      -1740.7836,      -2719.6655,       3872.7502 ],
  [               0,               0,      -399.38668,       636.97165,       481.22153,      -826.06525 ],
  [               0,               0,       1698.4392,       -2410.386,      -3447.2303,       4890.7198 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,               0,   0.00086267129 ],
  [               0,               0,               0,               0,               0,      -39.998799 ],
  [               0,               0,               0,               0,               0,      -2.7581853 ],
  [               0,               0,               0,               0,               0,       2.6961222 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.022905746 ],
  [               0,               0,               0,               0,               0,       1.1798116 ],
  [               0,               0,               0,               0,               0,     -0.98352206 ],
  [               0,               0,               0,               0,               0,     -0.97243586 ],
  [               0,               0,               0,               0,               0,    -0.021496203 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.50493351,     -0.74469198 ],
  [               0,               0,               0,               0,         317.948,      -469.93237 ],
  [               0,               0,               0,               0,       101.15629,      -152.37667 ],
  [               0,               0,               0,               0,      -55.847842,        83.16719 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0034671656,    0.0010665744,     0.010262583 ],
  [               0,               0,               0,       3.9990083,      -9.3524399,        6.568177 ],
  [               0,               0,               0,       18.023977,       -57.89054,       45.397594 ],
  [               0,               0,               0,      -2.5241119,       7.8115657,      -7.0866242 ],
  [               0,               0,               0,      -7.3738521,        23.70673,      -19.080573 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -6.7404782e-05,    0.0012481767 ],
  [               0,               0,               0,               0,       2.8990629,      -2.6723638 ],
  [               0,               0,               0,               0,      -2.7218753,         4.43505 ],
  [               0,               0,               0,               0,     -0.54700382,      0.48435105 ],
  [               0,               0,               0,               0,     0.021043803,     -0.14685549 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00081911288,  -0.00057440166,    0.0015593351,     0.010636579 ],
  [               0,               0,       3.4129205,      -6.1108777,       -15.71724,       30.321664 ],
  [               0,               0,       3.0916188,      -6.7825905,      -12.887048,       27.198804 ],
  [               0,               0,    -0.011114086,      0.53960208,      -2.2706622,       1.5069462 ],
  [               0,               0,     -0.61942576,       1.0160029,       4.1042792,      -7.3739329 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.01478073,     0.048705324 ],
  [               0,               0,               0,               0,      -22.391444,       51.958728 ],
  [               0,               0,               0,               0,       19.749546,       -45.85731 ],
  [               0,               0,               0,               0,     -0.76148662,       1.7648872 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00064775812,    0.0023492971 ],
  [               0,               0,               0,               0,      -3.7385016,       10.772601 ],
  [               0,               0,               0,               0,       1.6140174,      -4.8467152 ],
  [               0,               0,               0,               0,       1.4147974,      -4.4130137 ],
  [               0,               0,               0,               0,     -0.52388136,       1.2064938 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0061753006,     0.026800311,    -0.023600417 ],
  [               0,               0,               0,      -31.670347,       109.11737,      -70.823504 ],
  [               0,               0,               0,      -75.218372,       340.90229,      -385.91126 ],
  [               0,               0,               0,      -9.5363324,        63.51309,       -97.93892 ],
  [               0,               0,               0,    0.0034429655,   0.00084304014,    -0.020688695 ],
  [               0,               0,               0,      -3.3401561,        16.23086,      -18.673215 ],
  [               0,               0,               0,      -31.646154,       129.41401,       -160.6608 ],
 ],
 [
  [               0,               0,               0,               0,     0.012368533,    -0.024043752 ],
  [               0,               0,               0,               0,      -3273.9321,       8139.3064 ],
  [               0,               0,               0,               0,      -5.4962825,      -14.130002 ],
  [               0,               0,               0,               0,      -77.830287,       191.12151 ],
  [               0,               0,               0,               0,      0.10172391,     -0.37307993 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   1.1309283e-07,    0.0010604456 ],
  [               0,               0,               0,               0,     -0.48867394,        2.559489 ],
  [               0,               0,               0,               0,     0.020290751,     -0.82445606 ],
  [               0,               0,               0,               0,       1.2448581,      -1.2079468 ],
  [               0,               0,               0,               0,       3387.8242,      -781.54046 ],
  [               0,               0,               0,               0,      -3.5928383,      0.82877837 ],
  [               0,               0,               0,               0,    -0.047879451,     0.010693238 ],
  [               0,               0,               0,               0,     0.034401594,    -0.010188899 ],
 ],
 [
  [               0,               0,               0,   0.00040119574,   -0.0017195366,    0.0020269026 ],
  [               0,               0,               0,      -32.416693,       38.024526,      -7.6874856 ],
  [               0,               0,               0,       13.333474,      -14.459018,       2.7264715 ],
  [               0,               0,               0,       9.0563193,      -9.3282616,       1.0610936 ],
  [               0,               0,               0,      -1.2793801,       1.1529347,     -0.35366804 ],
  [               0,               0,               0,  -0.00061060719,    0.0005915475,  -5.7991453e-05 ],
  [               0,               0,               0,       28.850681,      -31.845795,       10.782581 ],
  [               0,               0,               0,       7.5219446,      -7.0693369,       3.3688788 ],
 ],
 [
  [               0,               0,               0,       2.4639006,      0.48193676,     -0.79515099 ],
  [               0,               0,               0,       27.874816,      -49.935637,       22.244986 ],
  [               0,               0,               0,       9.6429639,      -10.237131,       1.8815508 ],
  [               0,               0,               0,      0.74583218,     -0.86656049,      0.21760195 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0017187376,    -0.004929595,    0.0037894806 ],
  [               0,               0,               0,       31.084949,      -47.720498,       21.847997 ],
  [               0,               0,               0,       4.5825721,      -4.8130939,     -0.12454251 ],
  [               0,               0,               0,       9.8806304,      -17.483604,       6.3816006 ],
  [               0,               0,               0,     -0.45651041,     -0.31277313,      0.30987872 ],
  [               0,               0,               0,   0.00024807766,  -0.00062663224,   0.00028920826 ],
  [               0,               0,               0,       6.2778433,      -12.719928,        9.526172 ],
  [               0,               0,               0,       4.8024017,      -10.568261,       10.154935 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.34645593,      -0.4593888,     -0.68603473,      0.91385607 ],
  [               0,               0,       1547.8963,      -2243.6326,      -3103.6141,       4503.9714 ],
  [               0,               0,       1285.2725,      -1829.3473,      -2867.2677,       4081.5289 ],
  [               0,               0,      -272.68675,       461.16797,       208.53091,      -447.19469 ],
  [               0,               0,       1870.7578,      -2655.6975,      -3797.7772,       5389.6608 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,    -0.078361806,      0.11376357 ],
  [               0,               0,               0,               0,      -16.663832,      -15.770801 ],
  [               0,               0,               0,               0,      -182.25476,        262.5548 ],
  [               0,               0,               0,               0,      -64.166845,       92.857004 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00095481912 ],
  [               0,               0,               0,               0,               0,       7.4477939 ],
  [               0,               0,               0,               0,               0,      -2.0485543 ],
  [               0,               0,               0,               0,               0,      0.21983901 ],
  [               0,               0,               0,               0,               0,     -0.48745656 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.083816218,      0.12620904 ],
  [               0,               0,               0,               0,       1310.4102,      -1932.4912 ],
  [               0,               0,               0,               0,       215.63473,      -321.10248 ],
  [               0,               0,               0,               0,       49.865125,      -73.910501 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      -2284.1458,       7398.7327,      -5892.2312 ],
  [               0,               0,               0,       1.1615116,      -3.8443595,       3.2226588 ],
  [               0,               0,               0,      -1.1599794,       3.8393717,      -3.2185796 ],
  [               0,               0,               0,      -1.1605859,       3.8412971,       -3.220106 ],
  [               0,               0,               0,  -0.00028647444,   0.00093011907,  -0.00075564448 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00050670088,    0.0016585207 ],
  [               0,               0,               0,               0,       15.402162,      -25.097999 ],
  [               0,               0,               0,               0,      -6.5734659,       12.085576 ],
  [               0,               0,               0,               0,     -0.37449898,      0.63359107 ],
  [               0,               0,               0,               0,     0.088582706,     -0.36690162 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   9.6254906e-05,  -0.00012220074,  -0.00020879951,    0.0022015413 ],
  [               0,               0,       -1.027862,       2.2568386,       3.0509423,      -4.0659179 ],
  [               0,               0,       1.1998562,      -2.2906076,      -5.9470526,       10.390811 ],
  [               0,               0,     -0.71451849,       1.6447341,       2.0374841,      -5.7952041 ],
  [               0,               0,     -0.42681556,       0.8369016,       2.3598675,      -4.7387271 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -1.4826849,       3.4627489 ],
  [               0,               0,               0,               0,      -19.831147,       46.158796 ],
  [               0,               0,               0,               0,       17.825692,       -41.50282 ],
  [               0,               0,               0,               0,     -0.74728309,       1.7391353 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.02315972,     0.058087441 ],
  [               0,               0,               0,               0,       6.1064058,      -12.333002 ],
  [               0,               0,               0,               0,      0.11702086,      -1.2963769 ],
  [               0,               0,               0,               0,      -1.8410165,       3.2551014 ],
  [               0,               0,               0,               0,     -0.79113219,       1.8290622 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.006443924,     0.005655814,    -0.043745305 ],
  [               0,               0,               0,      -12.585885,       30.502715,       5.6458549 ],
  [               0,               0,               0,      -68.595327,        332.6651,      -404.70791 ],
  [               0,               0,               0,      -13.245052,       67.295753,      -85.325106 ],
  [               0,               0,               0,    0.0022343511,   -0.0021209967,   -0.0076213988 ],
  [               0,               0,               0,      0.61598771,      -5.8993763,       12.439615 ],
  [               0,               0,               0,       7.2743524,       24.514326,      -109.79317 ],
 ],
 [
  [               0,               0,               0,               0,       5.6499658,      -13.988217 ],
  [               0,               0,               0,               0,      -119.71312,         300.339 ],
  [               0,               0,               0,               0,       24286.261,       -60430.05 ],
  [               0,               0,               0,               0,      -107.83871,       265.99352 ],
  [               0,               0,               0,               0,      -4.8152427,       12.051754 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.048720115,     0.026263261,    0.0057598704 ],
  [               0,               0,               0,       13.402096,      -7.0748805,       3.8641296 ],
  [               0,               0,               0,      0.48827639,    -0.043405832,       -1.279164 ],
  [               0,               0,               0,     -0.46629033,      0.23139569,    -0.017288939 ],
  [               0,               0,               0,   -0.0006597088,   -0.0018635981,   0.00052545749 ],
  [               0,               0,               0,      -6.8289239,      -4.3412998,       1.2728997 ],
  [               0,               0,               0,      -29.982732,       9.5011756,      -1.9631902 ],
 ],
 [
  [               0,               0,               0,     0.010169739,   -0.0046861423,    0.0052623104 ],
  [               0,               0,               0,      -8.9717468,       8.7796331,       1.5999979 ],
  [               0,               0,               0,       5.6156505,      -7.2165848,      0.72904015 ],
  [               0,               0,               0,     -0.69383298,       1.5595479,     -0.49555099 ],
  [               0,               0,               0,    0.0092639611,    -0.011718982,     0.004478591 ],
  [               0,               0,               0,      -4.9552711,       4.6223485,    -0.081999916 ],
  [               0,               0,               0,      -2.2192657,      -2.0889296,      -0.2486647 ],
 ],
 [
  [               0,               0,               0,               0,      0.15835342,     -0.12294232 ],
  [               0,               0,               0,               0,       29.622391,      -20.454348 ],
  [               0,               0,               0,               0,       29.998006,      -26.237233 ],
  [               0,               0,               0,               0,       -19.20228,       15.829161 ],
  [               0,               0,               0,               0,     -0.02205076,     0.019632426 ],
  [               0,               0,               0,               0,      -4.0628392,       4.2726706 ],
  [               0,               0,               0,               0,      -23.785011,       4.3461081 ],
 ],
 [
  [               0,               0,               0,     0.014287826,    -0.031310832,     0.019235776 ],
  [               0,               0,               0,      -20.380139,       42.057487,      -16.627238 ],
  [               0,               0,               0,      -19.103817,       42.474488,      -24.471276 ],
  [               0,               0,               0,       7.8158673,      -16.542235,       8.8948234 ],
  [               0,               0,               0,   -0.0031323167,    0.0064674431,   -0.0015308706 ],
  [               0,               0,               0,      0.57361427,       -1.152745,       1.5753966 ],
  [               0,               0,               0,      -12.037179,       27.277094,       -16.48992 ],
 ],
 [
  [               0,               0,     -0.91372275,       1.2340869,       1.7224713,      -2.3245331 ],
  [               0,               0,      -8.0926156,       4.6438736,       6.2479808,        12.40912 ],
  [               0,               0,       5295.0225,      -7252.6025,        -9900.99,       13558.061 ],
  [               0,               0,      -4708.7406,       6393.4755,       8973.2751,      -12183.249 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.26339871,     -0.37947803,     0.025588083 ],
  [               0,               0,               0,      -361.31679,        162.0878,       521.34321 ],
  [               0,               0,               0,      -34.909328,       82.968725,      -50.058148 ],
  [               0,               0,               0,      -1.5281244,       1.4128866,       1.4795354 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,     0.058102078,    -0.052265387,     -0.27087057,      0.32865112 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       63.093205,      -93.752619,      -140.51862,       208.81913 ],
  [               0,               0,      -153.79612,       237.66935,       270.45188,      -422.29874 ],
  [               0,               0,       155.99847,      -219.75875,      -340.74313,       479.93199 ],
  [               0,               0,      -4.2603261,       6.7010291,       5.6761046,      -9.3055761 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.065700278,    -0.027038391 ],
  [               0,               0,               0,               0,        8.202292,      -10.963114 ],
  [               0,               0,               0,               0,       38.272572,      -58.861909 ],
  [               0,               0,               0,               0,      -39.677424,       60.005694 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.013547538,     -0.03641437,     0.028851212 ],
  [               0,               0,               0,      0.67419697,      -5.5998599,        10.64001 ],
  [               0,               0,               0,      -8.2020202,       29.623777,      -27.892967 ],
  [               0,               0,               0,       7.1752092,      -25.404067,       22.409732 ],
  [               0,               0,               0,     0.029313116,     -0.10679768,     0.096818515 ],
  [               0,               0,               0,       -1.897902,       6.6463358,       -5.125948 ],
  [               0,               0,               0,      -28.218758,       90.432402,      -87.902501 ],
 ],
 [
  [               0,               0,               0,      0.23512358,     -0.98651218,       1.0403662 ],
  [               0,               0,               0,      -37.599823,       156.68709,         -160.37 ],
  [               0,               0,               0,       8.2572292,      -34.336677,       34.402566 ],
  [               0,               0,               0,      -1.5697307,        6.575459,      -6.8436884 ],
  [               0,               0,               0,    -0.028482641,      0.11932434,     -0.12444184 ],
  [               0,               0,               0,      -36.903563,       152.77415,      -157.27197 ],
  [               0,               0,               0,       18.245195,      -73.664348,       65.957612 ],
 ],
 [
  [               0,               0,               0,   -0.0044666661,   0.00059886304,     0.035934469 ],
  [               0,               0,               0,       19.641723,      -91.180446,       107.59075 ],
  [               0,               0,               0,        9.104725,      -43.972751,        52.00234 ],
  [               0,               0,               0,      -10.266519,       49.338506,      -59.254732 ],
  [               0,               0,               0,  -0.00025286536,  -0.00047586648,    0.0026963139 ],
  [               0,               0,               0,       0.7196101,      0.16118093,     -0.98713172 ],
  [               0,               0,               0,     -0.26030881,     -0.92350253,      -3.4724252 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.043924634 ],
  [               0,               0,               0,               0,               0,      0.91123132 ],
  [               0,               0,               0,               0,               0,     -0.77701567 ],
  [               0,               0,               0,               0,               0,     0.051501102 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0013696945,   -0.0033774729 ],
  [               0,               0,               0,               0,      -10.138132,      -14.795376 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,       20028.848,      -49176.515 ],
  [               0,               0,               0,               0,       68.776073,       269.51247 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.052165202,     0.031327409,    0.0040138502 ],
  [               0,               0,               0,       27.057869,      -15.321508,       5.7571541 ],
  [               0,               0,               0,       2.1411813,     -0.32791137,      -1.3315496 ],
  [               0,               0,               0,      0.35026716,     -0.25034686,     -0.01008447 ],
  [               0,               0,               0,    0.0064313255,   -0.0041526283,    0.0012181853 ],
  [               0,               0,               0,       28.709335,      -10.699591,        1.287862 ],
  [               0,               0,               0,      -3.9545163,       5.7531444,      -1.8042706 ],
 ],
 [
  [               0,               0,               0,    0.0070674991,   -0.0049009122,    0.0061277959 ],
  [               0,               0,               0,      -9.6506607,       9.5148529,       2.3595101 ],
  [               0,               0,               0,       1.0423283,      -1.8128525,     -0.89703766 ],
  [               0,               0,               0,        1.381886,     -0.96866385,      0.24854791 ],
  [               0,               0,               0,    0.0059747785,   -0.0074945218,     0.003351981 ],
  [               0,               0,               0,      -2.5786617,        2.386352,      0.36713756 ],
  [               0,               0,               0,      -5.4111917,       3.8011415,      -8.7059798 ],
 ],
 [
  [               0,               0,               0,               0,      0.08272226,    -0.062736687 ],
  [               0,               0,               0,               0,       -45.32756,       42.521595 ],
  [               0,               0,               0,               0,       27.517179,      -24.171469 ],
  [               0,               0,               0,               0,      -15.844227,       13.026173 ],
  [               0,               0,               0,               0,    -0.014229445,     0.013210844 ],
  [               0,               0,               0,               0,      -4.9456657,       4.9127576 ],
  [               0,               0,               0,               0,      -5.6945801,       2.5548559 ],
 ],
 [
  [               0,               0,               0,   -0.0096855036,     0.024507577,    -0.010816382 ],
  [               0,               0,               0,       14.458901,      -37.054899,       27.488271 ],
  [               0,               0,               0,      -2.1163544,       4.5336985,      -3.9943742 ],
  [               0,               0,               0,      0.86857088,      -1.4757899,      0.81468753 ],
  [               0,               0,               0,    0.0047556723,    -0.011381881,    0.0080436638 ],
  [               0,               0,               0,     -0.22912283,      0.62655009,      0.54733458 ],
  [               0,               0,               0,      -1.2747009,       2.6751178,      -3.0258559 ],
 ],
 [
  [               0,               0,      0.10557306,    -0.019522461,     -0.11170648,    -0.073313711 ],
  [               0,               0,      -3404.5798,       4784.9736,       6008.9115,      -8453.8737 ],
  [               0,               0,      -1104.3756,       1533.4985,       2074.9873,      -2883.0443 ],
  [               0,               0,       542.39084,      -742.85056,      -1018.1366,       1394.6997 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.051293002,    0.0019124339,     -0.10232698 ],
  [               0,               0,               0,        401.3008,       3.0437624,      -792.97206 ],
  [               0,               0,               0,      -110.12543,       54.283608,        136.9392 ],
  [               0,               0,               0,     -0.18395792,     -0.14057648,     -0.18021905 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,     -0.50409004,       0.7822157,      0.91365742,      -1.4310912 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.002040154,  -7.7900723e-05,    0.0025814566,    0.0034669894 ],
  [               0,               0,         10.1194,      0.24446163,       0.3639929,        4.628699 ],
  [               0,               0,      -24.553896,       2.9679044,       16.293511,       45.010818 ],
  [               0,               0,       -9.102332,         5.35237,       3.7499778,       13.096535 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       2.2905163,      -3.4308974 ],
  [               0,               0,               0,               0,       68.346643,      -101.07964 ],
  [               0,               0,               0,               0,        81.46403,      -124.47351 ],
  [               0,               0,               0,               0,      -78.942323,       119.38787 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.11245967,       0.3910545,     -0.33046648 ],
  [               0,               0,               0,       42.376499,       -144.3214,       125.81074 ],
  [               0,               0,               0,      -5.1321982,       18.939227,      -18.763269 ],
  [               0,               0,               0,       2.6194749,      -9.7471066,       9.0567983 ],
  [               0,               0,               0,     0.017762186,    -0.064177052,     0.058401958 ],
  [               0,               0,               0,       4.7228079,      -12.176705,       8.2881114 ],
  [               0,               0,               0,     -0.57581073,      -4.7061637,      -6.5876446 ],
 ],
 [
  [               0,               0,               0,      0.12696679,     -0.51720833,      0.53568087 ],
  [               0,               0,               0,      -19.727159,        78.59217,       -75.20589 ],
  [               0,               0,               0,       3.6486499,      -14.271975,       12.652717 ],
  [               0,               0,               0,     -0.25449474,      0.92486944,     -0.81453808 ],
  [               0,               0,               0,     0.017864743,    -0.074600481,     0.078001062 ],
  [               0,               0,               0,      -18.483585,        75.96822,      -76.284133 ],
  [               0,               0,               0,      -57.935864,       234.00354,      -240.25629 ],
 ],
 [
  [               0,               0,               0,     0.045964434,     -0.26538477,      0.37512965 ],
  [               0,               0,               0,       198.10977,      -925.00555,       1082.6182 ],
  [               0,               0,               0,       97.836059,       -460.4945,       540.38097 ],
  [               0,               0,               0,      -61.336077,       289.67316,      -341.86823 ],
  [               0,               0,               0,     0.024527611,      -0.1035742,      0.10937825 ],
  [               0,               0,               0,      -18.593421,       88.479039,      -104.25868 ],
  [               0,               0,               0,      -42.748315,       209.75196,      -258.75122 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.020342399 ],
  [               0,               0,               0,               0,               0,       1.4974837 ],
  [               0,               0,               0,               0,               0,     -0.99165612 ],
  [               0,               0,               0,               0,               0,     0.024136857 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -2.7770825,       6.9645932 ],
  [               0,               0,               0,               0,       6.0681976,       24.913661 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,       8.8365178,      -21.908981 ],
  [               0,               0,               0,               0,      0.36308962,      0.62894408 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.001856136,   0.00083006557,    0.0077275837 ],
  [               0,               0,               0,       5.6290392,      -2.2935958,       4.1380669 ],
  [               0,               0,               0,      0.85064568,     -0.48854854,      -1.2551375 ],
  [               0,               0,               0,       525.00714,      -221.37897,       4.3366394 ],
  [               0,               0,               0,      -4.1147386,       1.7362443,    -0.034240038 ],
  [               0,               0,               0,       0.5322979,     -0.24942346,     0.046481084 ],
  [               0,               0,               0,      -2.9562268,       1.2238254,     -0.18609298 ],
 ],
 [
  [               0,               0,               0,     0.003532136,    0.0022672784,    0.0029625684 ],
  [               0,               0,               0,      -4.0377129,      0.87345164,       5.5391346 ],
  [               0,               0,               0,       2.0942979,      -2.8847632,     -0.62492159 ],
  [               0,               0,               0,       0.8079937,     -0.35678604,     0.082308051 ],
  [               0,               0,               0,     0.006314145,   -0.0089727552,    0.0041553091 ],
  [               0,               0,               0,      -3.4022111,       3.2255278,       0.1766635 ],
  [               0,               0,               0,      -4.7469218,        2.943535,      -8.3561806 ],
 ],
 [
  [               0,               0,               0,               0,     0.068357304,    -0.051145389 ],
  [               0,               0,               0,               0,      -19.646295,       21.593409 ],
  [               0,               0,               0,               0,      -5.4963639,       2.8173977 ],
  [               0,               0,               0,               0,       9.6703731,      -7.6578642 ],
  [               0,               0,               0,               0,     0.019098528,    -0.013353625 ],
  [               0,               0,               0,               0,      -21.359467,       18.035628 ],
  [               0,               0,               0,               0,      -35.137242,       26.468578 ],
 ],
 [
  [               0,               0,               0,    -0.025706285,      0.05931964,    -0.028659985 ],
  [               0,               0,               0,       17.705602,      -43.163488,       29.814211 ],
  [               0,               0,               0,       1.9949163,      -3.7753403,      0.14017028 ],
  [               0,               0,               0,      -1.0825407,       2.4153656,      -1.0957581 ],
  [               0,               0,               0,      0.01062741,    -0.023980275,     0.014494762 ],
  [               0,               0,               0,      0.13715176,     -0.19473854,        0.974288 ],
  [               0,               0,               0,       7.9282588,       -16.22763,       6.4171519 ],
 ],
 [
  [               0,               0,     0.068574368,     0.030372183,     -0.04197781,     -0.16657124 ],
  [               0,               0,      -231.44847,       270.14252,        366.6489,      -412.13919 ],
  [               0,               0,      -18.261648,       37.570691,        42.58939,      -83.424815 ],
  [               0,               0,       8.2239828,      -6.9664023,      -12.299073,       9.2077357 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.047698738,    0.0057180543,    -0.082087758 ],
  [               0,               0,               0,       18.087721,     -0.14260515,      -32.950403 ],
  [               0,               0,               0,      -7.7892966,       6.4106768,       5.4629788 ],
  [               0,               0,               0,    -0.031106963,    -0.028324207,     -0.32542171 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam25_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -2.2379264,        3.231846,       4.4904577,      -6.4825262 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.013272054,    -0.015054846,    -0.026067872,     0.029278665 ],
  [               0,               0,       2388.2157,      -3217.0379,      -4783.9511,       6394.0599 ],
  [               0,               0,       1848.9537,      -2749.1312,      -4137.6343,       6148.3189 ],
  [               0,               0,       3.4431853,      -19.745332,      -31.329346,       79.534569 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.066019478,    -0.098425406 ],
  [               0,               0,               0,               0,       1895.2036,      -2849.6748 ],
  [               0,               0,               0,               0,      -250.33373,       378.08194 ],
  [               0,               0,               0,               0,       205.03481,      -310.55555 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.09300003,     -0.30313087,      0.25504683 ],
  [               0,               0,               0,      -8.2831015,       25.425456,      -16.327818 ],
  [               0,               0,               0,       1.0582053,      -3.2576024,       1.2538836 ],
  [               0,               0,               0,     -0.11804919,      0.49265731,     -0.51689507 ],
  [               0,               0,               0,     0.014049133,    -0.049184868,     0.042878701 ],
  [               0,               0,               0,      -54.638874,       193.46584,       -166.4478 ],
  [               0,               0,               0,      -65.174133,       139.77078,      -146.84876 ],
 ],
 [
  [               0,               0,               0,      0.11117638,     -0.45088468,      0.46627358 ],
  [               0,               0,               0,      -17.420795,       68.875318,      -65.010563 ],
  [               0,               0,               0,       2.9297327,      -11.308395,       9.6143805 ],
  [               0,               0,               0,    -0.080717086,      0.21904739,     -0.10452376 ],
  [               0,               0,               0,     0.016228814,    -0.067661977,     0.070657304 ],
  [               0,               0,               0,      -18.981025,       77.742915,      -77.789113 ],
  [               0,               0,               0,      -52.982318,       213.02277,      -218.24424 ],
 ],
 [
  [               0,               0,               0,   -0.0025026638,  -3.5797206e-05,     0.027950244 ],
  [               0,               0,               0,       20.392085,      -96.544397,         116.382 ],
  [               0,               0,               0,       0.1875327,     -0.81624323,     -0.23913502 ],
  [               0,               0,               0,      -5.3286336,       25.459955,      -30.402103 ],
  [               0,               0,               0,    0.0010738027,    0.0017733369,    -0.010416478 ],
  [               0,               0,               0,       1.9121856,      0.19272353,      -9.8143023 ],
  [               0,               0,               0,      -8.4999309,       3.6881795,       36.347402 ],
 ],
 [
  [               0,               0,               0,               0,               0,      0.02190873 ],
  [               0,               0,               0,               0,               0,        1.441393 ],
  [               0,               0,               0,               0,               0,     -0.97695147 ],
  [               0,               0,               0,               0,               0,      0.02302287 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.010648106,    -0.017124478 ],
  [               0,               0,               0,               0,       -1169.884,       2907.8187 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,      -78.777199,       193.54981 ],
  [               0,               0,               0,               0,      0.16094293,     -0.19123279 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 50 GeV photon clusters (large statistics)
#


gam50large_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,       -15620.48,       280.51966,       -1.730025,     0.002695836 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0016817524,    0.0011785287 ],
  [               0,               0,               0,               0,       3.2740767,       2.5833842 ],
  [               0,               0,               0,               0,      -164.87066,       20.880858 ],
  [               0,               0,               0,               0,      -165.32922,       20.840412 ],
  [               0,               0,               0,               0,       16.190072,      -5.6049638 ],
  [               0,               0,               0,               0,      -0.1710264,       0.1220403 ],
  [               0,               0,               0,               0,      -0.5080734,     0.067201116 ],
  [               0,               0,               0,               0,      0.16008088,       1.5167142 ],
 ],
 [
  [               0,               0,               0,    0.0047851221,   -0.0026907337,   0.00054220218 ],
  [               0,               0,               0,       23.746381,      -37.844377,       17.276412 ],
  [               0,               0,               0,       13.842796,      -21.359348,       6.6066657 ],
  [               0,               0,               0,        10.86083,      -14.487869,       2.8227565 ],
  [               0,               0,               0,      -0.9317748,       2.2521777,      -1.3670099 ],
  [               0,               0,               0,    0.0072292301,   -0.0065320331,    0.0017486471 ],
  [               0,               0,               0,      -1.3517176,    -0.058728405,       1.2515142 ],
  [               0,               0,               0,      -2.1947196,       2.2092214,      -1.9436823 ],
 ],
 [
  [               0,               0,               0,               0,     0.021118004,     0.007465804 ],
  [               0,               0,               0,               0,       4.3921364,      -2.9418716 ],
  [               0,               0,               0,               0,      -17.528739,        13.42854 ],
  [               0,               0,               0,               0,       16.745547,      -13.316095 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00094124444,   0.00049326997,    0.0031163283 ],
  [               0,               0,               0,      -1.6962375,       4.3755829,     -0.69407415 ],
  [               0,               0,               0,       1.1357265,      -2.4949579,      0.32016194 ],
  [               0,               0,               0,       1.4059857,      -3.2072901,      0.52679968 ],
  [               0,               0,               0,      -0.1105806,      0.16726313,     -0.15500853 ],
  [               0,               0,               0,   0.00013719912,  -0.00028004696,   8.2295246e-05 ],
  [               0,               0,               0,      0.62158662,       1.1745041,      0.19384352 ],
  [               0,               0,               0,      -1.4603169,         3.95104,       8.2186713 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00016437926 ],
  [               0,               0,               0,               0,               0,       39.959029 ],
  [               0,               0,               0,               0,               0,      -6.2180952 ],
  [               0,               0,               0,               0,               0,       -1.663498 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0072509106,    0.0033332675,   -0.0048216789,    -0.019454355 ],
  [               0,               0,       2.1894389,       4.9800786,       6.8637945,       13.450607 ],
  [               0,               0,      -7.8619753,       -9.327039,      -10.731881,      -12.204019 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,        4.627896,       6.5466458,        9.249519,        12.75278 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.28132288,     -0.47085978,      -0.6266731,       1.0453062 ],
  [               0,               0,       412.78687,      -597.49797,      -789.97346,       1148.2735 ],
  [               0,               0,      -644.71378,       996.14503,       965.94615,        -1533.58 ],
  [               0,               0,       158.97971,      -145.41803,      -716.00248,       852.87291 ],
  [               0,               0,       2390.9589,      -3407.1638,      -4858.3253,       6921.7113 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,    -0.060061324,     0.086818726 ],
  [               0,               0,               0,               0,       6.3927785,       30.649965 ],
  [               0,               0,               0,               0,       222.79359,      -322.75436 ],
  [               0,               0,               0,               0,       131.07691,      -178.97741 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.034563471,    -0.050007973 ],
  [               0,               0,               0,               0,      -91.104588,       148.41288 ],
  [               0,               0,               0,               0,      -766.37255,       1130.0579 ],
  [               0,               0,               0,               0,       807.49892,       -1157.812 ],
  [               0,               0,               0,               0,     -0.29584111,      -1.2294643 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.091452843,      0.14048228 ],
  [               0,               0,               0,               0,      -182.51435,       276.06907 ],
  [               0,               0,               0,               0,       8.0833316,      -13.177392 ],
  [               0,               0,               0,               0,       63.052859,      -93.573662 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00033558711 ],
  [               0,               0,               0,               0,               0,      -39.913614 ],
  [               0,               0,               0,               0,               0,       10.502371 ],
  [               0,               0,               0,               0,               0,     -0.10598986 ],
  [               0,               0,               0,               0,               0,        8.575104 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0077779662,   -0.0039960486,     0.010995359,     0.052759504 ],
  [               0,               0,      -5.0050802,       29.007598,      -54.414232,       33.997566 ],
  [               0,               0,      -3.7143834,       15.242259,      -20.113081,       7.7465523 ],
  [               0,               0,      0.78354717,      -7.3182435,       17.533901,       -13.15268 ],
  [               0,               0,      0.67857144,      -3.5425347,        6.125327,      -3.5112322 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.005577901,   -0.0083098267 ],
  [               0,               0,               0,               0,      -3.1367386,       8.1032529 ],
  [               0,               0,               0,               0,     -0.17464727,      -0.7356532 ],
  [               0,               0,               0,               0,    -0.025817612,     -0.95030641 ],
  [               0,               0,               0,               0,    -0.023976478,     0.021369594 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0065291788,    -0.015705624,    -0.035077593,      0.08555773 ],
  [               0,               0,       1.3763365,       6.5112943,      -39.203692,        43.22219 ],
  [               0,               0,       22.406008,      -133.39254,       263.76823,      -174.30656 ],
  [               0,               0,        -17.0973,       113.91485,       -251.2453,        182.5387 ],
  [               0,               0,       10.832976,      -71.224811,       155.87221,      -113.54314 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00081970377,    0.0034255395 ],
  [               0,               0,               0,               0,      -95.949952,       222.24394 ],
  [               0,               0,               0,               0,       49.980569,      -116.12508 ],
  [               0,               0,               0,               0,      -2.6662284,       6.3711173 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00036447074,    0.0014762282 ],
  [               0,               0,               0,               0,      -1.4145291,       6.1449311 ],
  [               0,               0,               0,               0,     -0.21625035,     -0.63900853 ],
  [               0,               0,               0,               0,      0.48582685,      -2.1534802 ],
  [               0,               0,               0,               0,      0.24076641,      -0.5404972 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0020638623,    0.0050477607 ],
  [               0,               0,               0,               0,      -75.858802,       221.17501 ],
  [               0,               0,               0,               0,      -371.13668,         886.234 ],
  [               0,               0,               0,               0,       3862.7308,      -9258.9566 ],
  [               0,               0,               0,               0,      -2.5398971,      -6.1725409 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0088939123,    0.0088223177,    -0.065382398 ],
  [               0,               0,               0,      -31.645763,       101.76556,      -54.756672 ],
  [               0,               0,               0,       -88.50089,       434.93888,      -535.80159 ],
  [               0,               0,               0,      -20.871935,       103.42137,      -128.02085 ],
  [               0,               0,               0,    0.0047172517,   -0.0013167345,    -0.025182842 ],
  [               0,               0,               0,      -8.9796289,       1.4842954,       48.892232 ],
  [               0,               0,               0,      -18.380631,       1.0859838,       103.17047 ],
 ],
 [
  [               0,               0,               0,               0,       1.2023834,      -2.9766914 ],
  [               0,               0,               0,               0,      -3487.2639,       8671.2411 ],
  [               0,               0,               0,               0,      -6.8067581,       20.513911 ],
  [               0,               0,               0,               0,      -330.02012,       817.20108 ],
  [               0,               0,               0,               0,      0.99901466,      -2.3937458 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,      -6774.3552,       183.93531,      -1.2923946,    0.0017258211 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00063621237,   0.00082360443 ],
  [               0,               0,               0,               0,      -1.0799343,       3.2385915 ],
  [               0,               0,               0,               0,          -4.445,       0.4941054 ],
  [               0,               0,               0,               0,      -4.3087738,      0.33012356 ],
  [               0,               0,               0,               0,       3.9271428,      -2.3676451 ],
  [               0,               0,               0,               0,      0.00137273,    0.0031718332 ],
  [               0,               0,               0,               0,     -0.41316061,      0.13077356 ],
  [               0,               0,               0,               0,       1.1257891,      0.91229291 ],
 ],
 [
  [               0,               0,               0,    0.0030495704,   -0.0014477614,   0.00028432925 ],
  [               0,               0,               0,       29.306076,      -45.701662,       20.296315 ],
  [               0,               0,               0,       18.385184,      -29.450237,       9.6347539 ],
  [               0,               0,               0,       17.858659,      -25.391058,       6.4488613 ],
  [               0,               0,               0,     -0.36867074,       1.5352923,      -1.2348165 ],
  [               0,               0,               0,     0.008995407,   -0.0077841158,    0.0019541953 ],
  [               0,               0,               0,     -0.23805977,      -1.4536877,       1.6870566 ],
  [               0,               0,               0,      -1.5985976,       1.6112979,      -1.8258788 ],
 ],
 [
  [               0,               0,               0,               0,     0.021741608,    0.0077317284 ],
  [               0,               0,               0,               0,       4.3572787,      -2.9206638 ],
  [               0,               0,               0,               0,      -17.048001,       13.051384 ],
  [               0,               0,               0,               0,       16.223403,      -12.900854 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00069117757,   0.00022598902,    0.0026538634 ],
  [               0,               0,               0,      -1.8185678,       4.9356867,     -0.92228503 ],
  [               0,               0,               0,       1.0670184,      -2.2883774,      0.17265981 ],
  [               0,               0,               0,       1.5776731,      -3.6844386,      0.79665141 ],
  [               0,               0,               0,    -0.091255969,     0.088338216,     -0.11323504 ],
  [               0,               0,               0,   0.00021725325,  -0.00051144805,   0.00023939313 ],
  [               0,               0,               0,     -0.60233915,       3.3374279,     -0.56128888 ],
  [               0,               0,               0,    -0.099820726,      0.48436541,       10.417792 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00016751297 ],
  [               0,               0,               0,               0,               0,       37.690162 ],
  [               0,               0,               0,               0,               0,      -6.0710166 ],
  [               0,               0,               0,               0,               0,      -1.6370482 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.15104214,     0.072011383,    -0.096794025,     -0.41444392 ],
  [               0,               0,      -536.43774,      -250.16234,       347.28055,       1464.9273 ],
  [               0,               0,      -201.00998,       -92.97745,       135.02491,        561.4995 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,       1.7225996,     -0.30631518,      -4.6742165,       -13.32161 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.21184412,     -0.37690564,     -0.48789419,      0.85817149 ],
  [               0,               0,       324.11535,      -471.65958,      -598.73453,       876.64534 ],
  [               0,               0,      -541.96898,       853.94832,       731.50321,      -1208.0131 ],
  [               0,               0,       161.88015,      -154.36259,      -712.36412,       857.41875 ],
  [               0,               0,       2466.6265,       -3514.126,      -5012.3999,       7139.4599 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,     0.038993708,    -0.055007068 ],
  [               0,               0,               0,               0,      -20.103705,      -10.479384 ],
  [               0,               0,               0,               0,      -188.52965,       276.49167 ],
  [               0,               0,               0,               0,      -209.77212,       301.21584 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.028476096,    -0.041166368 ],
  [               0,               0,               0,               0,        236.3314,      -326.66242 ],
  [               0,               0,               0,               0,      -798.33008,       1176.2558 ],
  [               0,               0,               0,               0,       602.42329,      -861.37489 ],
  [               0,               0,               0,               0,    -0.071675475,      -1.2456123 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.14883195,      0.22553535 ],
  [               0,               0,               0,               0,      -22.882339,       39.306797 ],
  [               0,               0,               0,               0,      -11.426502,       15.685946 ],
  [               0,               0,               0,               0,       59.928268,       -88.89823 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00022905539 ],
  [               0,               0,               0,               0,               0,       -32.13559 ],
  [               0,               0,               0,               0,               0,       10.483803 ],
  [               0,               0,               0,               0,               0,       2.1461143 ],
  [               0,               0,               0,               0,               0,       11.383141 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       1.3876507,      -2.6251958,      -4.0767725,       7.8963737 ],
  [               0,               0,      -39.639527,       198.73503,      -331.23441,       183.85032 ],
  [               0,               0,       37.720494,      -189.17066,       315.40081,      -175.13588 ],
  [               0,               0,       38.734576,      -194.19962,       323.70107,      -179.69531 ],
  [               0,               0,      0.13236921,     -0.66897615,       1.1230261,     -0.62644185 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0054801382,   -0.0082074032 ],
  [               0,               0,               0,               0,      -3.3378717,        8.526986 ],
  [               0,               0,               0,               0,     -0.18565029,     -0.71248795 ],
  [               0,               0,               0,               0,    -0.073076509,     -0.85421443 ],
  [               0,               0,               0,               0,    -0.028023791,     0.028612426 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0072579679,    -0.017464886,     -0.03833686,     0.093487462 ],
  [               0,               0,       2.1158624,       2.9371487,      -33.823746,        40.80252 ],
  [               0,               0,       21.388052,      -126.55654,       248.48412,      -162.92764 ],
  [               0,               0,      -15.587897,       104.62454,       -232.2408,       169.62008 ],
  [               0,               0,       10.137265,      -66.756765,       146.31462,      -106.73374 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.31954247,     -0.72822654 ],
  [               0,               0,               0,               0,       -110.9871,       256.67944 ],
  [               0,               0,               0,               0,       67.477036,       -156.2819 ],
  [               0,               0,               0,               0,       -11.35028,       26.297742 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00036887369,    0.0014857146 ],
  [               0,               0,               0,               0,      -1.2837124,       5.8217048 ],
  [               0,               0,               0,               0,     -0.26202289,     -0.53167389 ],
  [               0,               0,               0,               0,      0.43859249,      -2.0418045 ],
  [               0,               0,               0,               0,      0.22509915,     -0.50346051 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0019043521,    0.0046706463 ],
  [               0,               0,               0,               0,       9.8453102,       15.581886 ],
  [               0,               0,               0,               0,      -363.59495,       868.46106 ],
  [               0,               0,               0,               0,       3614.7928,      -8665.0049 ],
  [               0,               0,               0,               0,      -2.4163696,       -5.898061 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0021829054,     0.013648151,    -0.037484224 ],
  [               0,               0,               0,      -40.538195,       166.58638,      -159.88316 ],
  [               0,               0,               0,      -97.000359,       470.25645,      -571.42903 ],
  [               0,               0,               0,     -0.13334371,       7.4727892,      -17.285543 ],
  [               0,               0,               0,    0.0094383659,   -0.0049795589,    -0.043124994 ],
  [               0,               0,               0,       9.9380905,      -32.472139,       20.015667 ],
  [               0,               0,               0,       6.6079566,       2.1463114,      -42.529997 ],
 ],
 [
  [               0,               0,               0,               0,       1.1411695,      -2.8251365 ],
  [               0,               0,               0,               0,      -3492.6899,       8684.6739 ],
  [               0,               0,               0,               0,       800.40162,      -1989.2968 ],
  [               0,               0,               0,               0,      -316.61159,       784.05564 ],
  [               0,               0,               0,               0,     -0.51050211,       1.3794348 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,       14566.054,      -41.263616,     -0.63411495,    0.0013245686 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00057124446,   0.00083790862 ],
  [               0,               0,               0,               0,     -0.97077291,       3.2030342 ],
  [               0,               0,               0,               0,      -1.4971344,     -0.39618058 ],
  [               0,               0,               0,               0,      -1.3356728,     -0.57225355 ],
  [               0,               0,               0,               0,       1.3801768,     -0.99388469 ],
  [               0,               0,               0,               0,  -0.00011440057,   0.00094596245 ],
  [               0,               0,               0,               0,     -0.60290064,      0.20645935 ],
  [               0,               0,               0,               0,       1.5416365,       0.5178573 ],
 ],
 [
  [               0,               0,               0,    0.0027126074,  -0.00092723149,   0.00017138875 ],
  [               0,               0,               0,       21.836068,      -34.774724,       16.246597 ],
  [               0,               0,               0,       15.363747,      -24.351042,       7.4744599 ],
  [               0,               0,               0,       17.074124,      -24.303197,       6.0778059 ],
  [               0,               0,               0,      0.13011773,      0.50659862,     -0.72992847 ],
  [               0,               0,               0,     0.010055944,   -0.0084871772,    0.0020583132 ],
  [               0,               0,               0,      0.43736553,      -2.2500573,       1.8594395 ],
  [               0,               0,               0,      -1.3606085,       1.4438697,      -1.8340684 ],
 ],
 [
  [               0,               0,               0,               0,     0.021837786,    0.0076580381 ],
  [               0,               0,               0,               0,        3.351838,      -2.1173168 ],
  [               0,               0,               0,               0,      -16.413015,        12.54336 ],
  [               0,               0,               0,               0,       16.281788,       -12.94713 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00056543077,   0.00012868794,    0.0025091584 ],
  [               0,               0,               0,      -0.7185173,       1.9536645,        1.083141 ],
  [               0,               0,               0,      0.39069072,     -0.52328847,     -0.93938844 ],
  [               0,               0,               0,       1.3445125,       -2.935197,      0.24306889 ],
  [               0,               0,               0,    -0.035385694,    -0.041286968,    -0.047299747 ],
  [               0,               0,               0,  -4.8792601e-05,   0.00016469127,   -0.0001809755 ],
  [               0,               0,               0,     -0.15188684,       2.5427218,     -0.19215756 ],
  [               0,               0,               0,       2.2660009,      -5.0453816,       13.625848 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00016337938 ],
  [               0,               0,               0,               0,               0,       39.978923 ],
  [               0,               0,               0,               0,               0,      -6.4216436 ],
  [               0,               0,               0,               0,               0,      -1.6588229 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0040761942,    0.0018723655,    -0.002733099,    -0.010904117 ],
  [               0,               0,      -5.1736963,      -5.0593078,       -6.864564,      -5.3276425 ],
  [               0,               0,      -10.065826,      -18.388311,      -25.000923,      -31.201544 ],
  [               0,               0,       11.523514,        26726.99,       29077.964,       12631.938 ],
  [               0,               0,       11.303145,       5.4436929,       31.251063,       20.391718 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       2.4546491,      -3.3776304,      -4.8360559,       6.6572236 ],
  [               0,               0,       2434.2321,      -3499.4807,      -4978.8746,         7160.76 ],
  [               0,               0,       1690.5423,      -2442.0225,      -3503.9428,       5059.3965 ],
  [               0,               0,      -485.61817,       695.14247,       929.89224,      -1333.3995 ],
  [               0,               0,       174.94158,      -237.62219,      -349.94213,        474.2249 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,     0.046149316,     -0.06490508 ],
  [               0,               0,               0,               0,      -20.050638,       -10.51803 ],
  [               0,               0,               0,               0,      -181.22749,       265.24549 ],
  [               0,               0,               0,               0,      -123.10938,       176.30469 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.036389001,    -0.012066459 ],
  [               0,               0,               0,               0,       42.616783,      -62.468158 ],
  [               0,               0,               0,               0,      -61.374841,       90.118191 ],
  [               0,               0,               0,               0,      -15.493445,       23.086543 ],
  [               0,               0,               0,               0,      0.31798959,     -0.48384522 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.11082764,      0.16739967 ],
  [               0,               0,               0,               0,      -166.48115,       254.73881 ],
  [               0,               0,               0,               0,      -114.76955,       169.55777 ],
  [               0,               0,               0,               0,       198.57299,      -294.96319 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       7.5726049 ],
  [               0,               0,               0,               0,               0,    -0.049331913 ],
  [               0,               0,               0,               0,               0,     0.049582939 ],
  [               0,               0,               0,               0,               0,     0.049308952 ],
  [               0,               0,               0,               0,               0,   0.00034797548 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.035995693,    -0.017919829,      0.04784826,      0.22095655 ],
  [               0,               0,      -29.165156,       149.20617,      -253.51222,       143.54094 ],
  [               0,               0,       25.948344,      -132.80959,       225.78639,      -127.96141 ],
  [               0,               0,       28.479578,      -145.39644,       246.62075,      -139.44087 ],
  [               0,               0,      0.22969423,      -1.1910997,       2.0469233,       -1.166816 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0053794182,   -0.0078382681 ],
  [               0,               0,               0,               0,       -2.672018,       7.1470482 ],
  [               0,               0,               0,               0,     -0.21933905,     -0.64229569 ],
  [               0,               0,               0,               0,    -0.044876092,     -0.91198693 ],
  [               0,               0,               0,               0,   -0.0041276138,    -0.020949433 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0085346046,     -0.02058452,    -0.044146913,      0.10784656 ],
  [               0,               0,       2.6965379,      0.81165553,      -32.480227,       41.864387 ],
  [               0,               0,       22.435586,      -133.45154,       263.35757,      -173.44657 ],
  [               0,               0,      -16.134781,       108.50965,      -241.52876,       177.05518 ],
  [               0,               0,        10.51897,      -69.122197,        151.1745,      -110.04815 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00074921735,    0.0033647102 ],
  [               0,               0,               0,               0,      -102.48047,       237.29463 ],
  [               0,               0,               0,               0,       54.299348,      -126.06096 ],
  [               0,               0,               0,               0,      -3.2166883,       7.6428401 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.009221364,     0.024527172 ],
  [               0,               0,               0,               0,       4.9683924,      -9.1955507 ],
  [               0,               0,               0,               0,     -0.34887001,     -0.33139703 ],
  [               0,               0,               0,               0,      0.15396485,      -1.4120629 ],
  [               0,               0,               0,               0,      0.16031808,     -0.36072892 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       0.1416392,      0.32983373 ],
  [               0,               0,               0,               0,      -42.347444,        101.6162 ],
  [               0,               0,               0,               0,       41.843726,      -100.40777 ],
  [               0,               0,               0,               0,       40.978775,      -98.337889 ],
  [               0,               0,               0,               0,     0.056317128,     -0.13701033 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0027815981,   -0.0012482205,   -0.0051065325 ],
  [               0,               0,               0,      -2.9955225,      -8.0751459,        42.92266 ],
  [               0,               0,               0,      -51.872006,       243.58867,      -287.17349 ],
  [               0,               0,               0,      -16.273633,       86.903823,      -114.97936 ],
  [               0,               0,               0,    0.0087566867,    0.0039735464,     -0.06117113 ],
  [               0,               0,               0,       -4.012377,       4.0730911,       13.357308 ],
  [               0,               0,               0,       5.4078908,       13.320905,      -62.240747 ],
 ],
 [
  [               0,               0,               0,               0,     -0.18046379,      0.45791769 ],
  [               0,               0,               0,               0,      -3481.1045,        8655.865 ],
  [               0,               0,               0,               0,      -1270.2579,       3161.5714 ],
  [               0,               0,               0,               0,       -55.70811,       136.74463 ],
  [               0,               0,               0,               0,      -1.2427331,       3.4301754 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,   8.8426489e-05,  -3.4930186e-05,      0.00871588 ],
  [               0,               0,               0,     -0.67627883,     0.089858161,       3.4073184 ],
  [               0,               0,               0,     -0.50474252,      0.19941211,      -1.3038083 ],
  [               0,               0,               0,       5611.2247,      -2369.3955,       217.55027 ],
  [               0,               0,               0,       48.905314,      -20.651024,       1.8960662 ],
  [               0,               0,               0,      0.25783255,     -0.10662619,    0.0092456306 ],
  [               0,               0,               0,      -1.0121595,      0.42387879,      -3.1524147 ],
 ],
 [
  [               0,               0,               0,    0.0098347803,    -0.013229628,    0.0099960632 ],
  [               0,               0,               0,      -5.5187069,       5.8237961,       2.5551569 ],
  [               0,               0,               0,       2.0060861,      -2.9703071,     -0.41621293 ],
  [               0,               0,               0,     -0.22198152,      0.80823996,     -0.29632274 ],
  [               0,               0,               0,    0.0075713764,   -0.0091444609,    0.0018041488 ],
  [               0,               0,               0,      -2.7310743,       3.2200686,     0.034055837 ],
  [               0,               0,               0,       7.6715027,      -9.9412868,      -1.7210508 ],
 ],
 [
  [               0,               0,               0,               0,     0.052122769,    -0.035836045 ],
  [               0,               0,               0,               0,       14.120732,      -7.6617329 ],
  [               0,               0,               0,               0,      -4.0364407,       1.9071959 ],
  [               0,               0,               0,               0,       4.1391866,      -3.2967827 ],
  [               0,               0,               0,               0,     0.016622371,    -0.012264244 ],
  [               0,               0,               0,               0,      -23.293697,       19.624735 ],
  [               0,               0,               0,               0,       -50.19499,       26.065802 ],
 ],
 [
  [               0,               0,               0,    0.0069208797,    -0.019760417,     0.018401047 ],
  [               0,               0,               0,      0.11972715,     0.056512129,       3.3812104 ],
  [               0,               0,               0,      0.62138699,      -1.8506669,     -0.23532716 ],
  [               0,               0,               0,     -0.23570035,       1.1669094,     -0.76035525 ],
  [               0,               0,               0,   -0.0020391129,    0.0054268781,   -0.0025571324 ],
  [               0,               0,               0,      -0.8030563,       1.9356407,     -0.17138124 ],
  [               0,               0,               0,    -0.079751354,      0.44973974,      -2.0020474 ],
 ],
 [
  [               0,               0,    -0.013191954,    -0.023344937,   -0.0018343404,      0.09128696 ],
  [               0,               0,       368.84328,      -509.61074,      -707.40159,       979.38062 ],
  [               0,               0,      -137.72403,       195.72525,       250.45343,      -357.56485 ],
  [               0,               0,      -30.954816,        39.33302,       54.487359,      -68.705083 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.12086788,      0.20247072 ],
  [               0,               0,               0,               0,      -598.12421,        877.3846 ],
  [               0,               0,               0,               0,     -0.25461192,        -1.32934 ],
  [               0,               0,               0,               0,       11.478436,      -15.985462 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      0.40678917,     -0.54167662,     -0.98726384,       1.3306898 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,      -2.0971195,      -0.5439595,       1.5195431,       3.4953565,       3.7405581 ],
  [               0,       5.0136702,      -17.596026,       5.0259707,       42.965349,       -41.33685 ],
  [               0,       4.7843028,      -7.8243461,      -7.7836993,       5.7050758,       10.749928 ],
  [               0,     0.092425378,     -0.70825297,      0.42240442,       2.2626541,      -2.4229504 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0012102164,    0.0035688128,    0.0084484721 ],
  [               0,               0,               0,       6.5880625,       1.7151692,       -15.79115 ],
  [               0,               0,               0,       10.484637,     -0.60396459,      -24.143698 ],
  [               0,               0,               0,      -11.454634,     -0.11094768,       26.368343 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.054396719,      0.18100668,     -0.14064882 ],
  [               0,               0,               0,       20.447419,      -68.665093,       59.948937 ],
  [               0,               0,               0,      -3.2077211,       10.479616,      -9.6813778 ],
  [               0,               0,               0,      0.84845475,      -2.6477409,        2.038013 ],
  [               0,               0,               0,  -0.00032820155,    0.0028238818,   -0.0037659015 ],
  [               0,               0,               0,      0.41740609,        -3.80256,       6.3739634 ],
  [               0,               0,               0,       64.818505,      -240.03777,       194.99082 ],
 ],
 [
  [               0,               0,               0,      0.15812247,     -0.66223273,      0.69960659 ],
  [               0,               0,               0,      -18.010778,       74.282355,      -73.909961 ],
  [               0,               0,               0,       2.8462396,      -11.599046,       10.598875 ],
  [               0,               0,               0,     -0.88901859,       3.6200354,      -3.6510923 ],
  [               0,               0,               0,    -0.023378022,     0.098252123,     -0.10251265 ],
  [               0,               0,               0,      -10.927299,       45.202887,      -45.848066 ],
  [               0,               0,               0,       11.540888,       -46.92947,       46.112725 ],
 ],
 [
  [               0,               0,               0,    -0.012660323,   -0.0046017576,     0.093655559 ],
  [               0,               0,               0,       63.589749,      -295.51846,       345.13565 ],
  [               0,               0,               0,     -0.56744344,       1.3466286,      -1.1172863 ],
  [               0,               0,               0,      -23.046671,       109.61886,      -130.27075 ],
  [               0,               0,               0,    0.0019720361,   -0.0011565917,   -0.0077247601 ],
  [               0,               0,               0,       9.3656958,      -46.036589,       56.736224 ],
  [               0,               0,               0,      -8.8467045,       65.828808,      -109.08377 ],
 ],
 [
  [               0,               0,               0,               0,    0.0048022149,   0.00027169802 ],
  [               0,               0,               0,               0,       -49.56684,       124.08434 ],
  [               0,               0,               0,               0,       11.879498,      -30.473695 ],
  [               0,               0,               0,               0,       1.9612699,      -4.6055036 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.047203371,    0.0013144141 ],
  [               0,               0,               0,               0,      -3.7006382,       -13.06146 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,       36.529372,       -90.05827 ],
  [               0,               0,               0,               0,   9.1316236e-05,      -1.1523874 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     -0.04709001,     0.020973266,    0.0053197614 ],
  [               0,               0,               0,       15.280534,      -7.0747256,       5.1632891 ],
  [               0,               0,               0,      -1.0715823,      0.47076301,      -1.3809387 ],
  [               0,               0,               0,       1.0073694,      -0.3872582,     0.010237232 ],
  [               0,               0,               0,    0.0083184845,   -0.0038658435,   0.00095859702 ],
  [               0,               0,               0,       19.012589,      -8.1661092,       1.2034319 ],
  [               0,               0,               0,      -15.858409,       6.6879095,      -1.8715344 ],
 ],
 [
  [               0,               0,               0,     0.010407148,    -0.013881527,     0.010582366 ],
  [               0,               0,               0,      -8.6549164,        9.525474,       2.2120001 ],
  [               0,               0,               0,       1.0155115,      -1.7265262,     -0.80027909 ],
  [               0,               0,               0,    -0.001006896,      0.45292266,     -0.19602588 ],
  [               0,               0,               0,   -0.0046816908,    0.0060469838,    -0.001018764 ],
  [               0,               0,               0,      -4.1679511,       5.0104197,     -0.62689742 ],
  [               0,               0,               0,       6.3451445,      -8.6094287,      -11.476395 ],
 ],
 [
  [               0,               0,               0,               0,     0.017443109,   -0.0060258038 ],
  [               0,               0,               0,               0,       15.167402,      -8.2396965 ],
  [               0,               0,               0,               0,       5.5758992,      -5.9145484 ],
  [               0,               0,               0,               0,        8.257901,       10.351408 ],
  [               0,               0,               0,               0,      0.39164785,     -0.17892894 ],
  [               0,               0,               0,               0,     -0.35309488,      0.31972888 ],
  [               0,               0,               0,               0,      0.11825342,      -9.4899372 ],
 ],
 [
  [               0,               0,               0,   0.00024566177,    0.0016515099,    0.0049452939 ],
  [               0,               0,               0,       1.2959576,      -4.4126731,       7.0738583 ],
  [               0,               0,               0,      0.43582604,      -1.1688055,     -0.73132794 ],
  [               0,               0,               0,     -0.23009778,      0.85179171,     -0.49955499 ],
  [               0,               0,               0,    0.0014158438,    -0.003377069,     0.002597972 ],
  [               0,               0,               0,     -0.53059948,      0.90642345,       0.3916557 ],
  [               0,               0,               0,      -1.4307468,       1.7077554,      -2.1754508 ],
 ],
 [
  [               0,       0.4406674,      -0.2432949,     -0.73947563,     -0.55476406,       1.2311739 ],
  [               0,       -1830.711,       1135.5479,       3920.5074,       1719.9193,      -6180.0736 ],
  [               0,      -456.47882,       321.32224,       866.94115,       583.98379,      -1647.3231 ],
  [               0,       181.37062,      -124.29988,       -339.7449,      -231.85097,       636.23936 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.026803772,    0.0030007604,    -0.041972386 ],
  [               0,               0,               0,      -102.74338,      0.64110767,       218.50745 ],
  [               0,               0,               0,       50.163993,      -142.20621,        99.58294 ],
  [               0,               0,               0,     0.066185871,     -0.19062324,     -0.36705475 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,       -1.278493,       1.8505851,       2.4833767,      -3.5945672 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,      0.77045896,     -0.53396396,       -1.570469,      -1.1537574,       3.1974711 ],
  [               0,       38.869612,      -41.442149,       -50.93231,       5.4467293,       52.996895 ],
  [               0,       415.68343,      -314.47712,      -902.17151,      -645.97619,       1955.7877 ],
  [               0,      -340.76078,        250.3707,       741.39292,       548.52899,      -1612.1176 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0029010625,    0.0015038456,   -0.0016106391 ],
  [               0,               0,               0,       32.834238,   -0.0052808485,      -72.482416 ],
  [               0,               0,               0,      -1.7979022,      0.17742544,        2.616599 ],
  [               0,               0,               0,       0.3323198,      0.12356143,      -0.9154063 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.060380999,       0.1950264,     -0.14908227 ],
  [               0,               0,               0,       30.940861,      -102.06312,       87.245638 ],
  [               0,               0,               0,      -3.5379957,       11.388266,      -10.398715 ],
  [               0,               0,               0,       1.8389387,      -6.0645303,       5.0148665 ],
  [               0,               0,               0,     0.027335382,    -0.091363491,     0.076760298 ],
  [               0,               0,               0,      -4.9356613,       19.286908,       -17.84568 ],
  [               0,               0,               0,      -0.5550092,       2.8543688,      -11.396227 ],
 ],
 [
  [               0,               0,               0,    0.0061417339,    -0.018869345,     0.021618402 ],
  [               0,               0,               0,      0.71788955,      -5.8548218,       11.943551 ],
  [               0,               0,               0,      0.20626138,     -0.31994305,      -1.4749731 ],
  [               0,               0,               0,      -4.5684133,       18.006859,      -17.557876 ],
  [               0,               0,               0,    0.0016113966,   -0.0033614362,    0.0019359226 ],
  [               0,               0,               0,       3.0929799,      -13.410605,       14.629406 ],
  [               0,               0,               0,       15.739505,      -60.257234,       55.134336 ],
 ],
 [
  [               0,               0,               0,    -0.022592004,     0.084945962,     -0.06482184 ],
  [               0,               0,               0,       55.006492,      -258.43582,       306.11081 ],
  [               0,               0,               0,       5.5543105,      -26.768218,       31.034075 ],
  [               0,               0,               0,       -6.473645,       53.876632,      -93.489322 ],
  [               0,               0,               0,    -0.053088361,    0.0055687869,      0.30882097 ],
  [               0,               0,               0,      -4.2337318,       19.763155,      -23.076779 ],
  [               0,               0,               0,      -24.254335,        120.1125,      -147.94247 ],
 ],
 [
  [               0,               0,               0,               0,    0.0014950667,    0.0076696262 ],
  [               0,               0,               0,               0,      -17.559329,       45.676493 ],
  [               0,               0,               0,               0,       6.8597942,      -18.073802 ],
  [               0,               0,               0,               0,     0.087398431,     -0.16536917 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0033066182,   -0.0081944925 ],
  [               0,               0,               0,               0,      -10.880459,      -12.919068 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,      -2452.5781,       6641.6529 ],
  [               0,               0,               0,               0,       194.41728,      -64.411317 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.046490851,     0.020811656,    0.0053169711 ],
  [               0,               0,               0,       15.124502,      -7.0654329,       5.1691274 ],
  [               0,               0,               0,      -1.1001337,      0.48782941,      -1.3831812 ],
  [               0,               0,               0,         1.05701,     -0.41372494,     0.012444292 ],
  [               0,               0,               0,    0.0076632226,   -0.0035688792,   0.00096082883 ],
  [               0,               0,               0,       18.588194,      -8.0429954,       1.1915961 ],
  [               0,               0,               0,      -15.366152,       6.5374326,      -1.8701515 ],
 ],
 [
  [               0,               0,               0,     0.010056735,     -0.01324985,     0.010392275 ],
  [               0,               0,               0,      -8.6784157,       9.3870471,        2.267418 ],
  [               0,               0,               0,      0.94190305,      -1.6134243,     -0.83761404 ],
  [               0,               0,               0,     0.028076324,      0.41123466,      -0.1821423 ],
  [               0,               0,               0,    -0.004226417,    0.0054667548,   -0.0008400038 ],
  [               0,               0,               0,      -4.2594798,          5.1142,     -0.66836566 ],
  [               0,               0,               0,       6.2644483,      -8.4675718,      -5.2430468 ],
 ],
 [
  [               0,               0,               0,               0,    0.0092553887,   0.00054441642 ],
  [               0,               0,               0,               0,       19.087534,       -11.38882 ],
  [               0,               0,               0,               0,        5.558207,       -5.903022 ],
  [               0,               0,               0,               0,        -6.96484,         13.5469 ],
  [               0,               0,               0,               0,     0.051255875,     0.022309476 ],
  [               0,               0,               0,               0,     -0.53781497,      0.48529683 ],
  [               0,               0,               0,               0,      0.90551383,      -10.102719 ],
 ],
 [
  [               0,               0,               0,   0.00081861262,   1.8657737e-05,    0.0059771896 ],
  [               0,               0,               0,       1.3168924,      -4.3646353,       6.9995055 ],
  [               0,               0,               0,      0.49584921,       -1.319951,       -0.643905 ],
  [               0,               0,               0,     -0.22310723,      0.84887523,     -0.50138686 ],
  [               0,               0,               0,    0.0010081141,    -0.002452907,    0.0021130534 ],
  [               0,               0,               0,      -1.0765891,       2.1505095,      -0.2997771 ],
  [               0,               0,               0,      -1.4880133,       2.0261654,      -2.4169029 ],
 ],
 [
  [               0,               0,      0.12642201,    0.0098389684,     -0.11051569,     -0.18030418 ],
  [               0,               0,      -4153.0348,       5809.5194,       7497.3167,      -10491.489 ],
  [               0,               0,       -723.1198,       1013.6808,       1362.7385,      -1911.9397 ],
  [               0,               0,       544.77094,      -746.17956,      -1023.7895,       1402.4454 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.049726581,     0.002563325,    -0.087981936 ],
  [               0,               0,               0,      -88.013829,      0.61491203,       187.16782 ],
  [               0,               0,               0,       56.522144,      -162.31435,       115.22852 ],
  [               0,               0,               0,     0.082616377,      -0.2137264,     -0.36554335 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam50large_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,       -1.495451,       2.1631591,       2.9326264,      -4.2418552 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     -0.11409421,      0.10627458,      0.28181725,      0.18560584,     -0.67792267 ],
  [               0,        142.0299,      -148.30038,      -247.46159,      -42.353296,       405.66615 ],
  [               0,       365.51746,      -282.63321,      -789.57808,      -549.34847,       1703.6558 ],
  [               0,      -287.65905,       210.33125,       627.20133,       467.22999,       -1366.705 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      -24.087029,       72.933294,      -55.201254 ],
  [               0,               0,               0,       58.973387,      -305.38829,       329.61105 ],
  [               0,               0,               0,      -5234.1019,       15884.346,      -12052.153 ],
  [               0,               0,               0,       2589.6428,      -7837.8608,       5930.5764 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.11087294,     -0.37551738,      0.32192912 ],
  [               0,               0,               0,      -36.587421,       124.80081,      -101.94707 ],
  [               0,               0,               0,       5.6471281,      -19.437325,        15.29203 ],
  [               0,               0,               0,      -2.1570282,        7.490687,       -6.430282 ],
  [               0,               0,               0,    -0.027365962,     0.092781278,    -0.077131242 ],
  [               0,               0,               0,     -0.30216515,      0.90459108,      0.26364886 ],
  [               0,               0,               0,       2.9361567,      -9.9489965,      -5.8816593 ],
 ],
 [
  [               0,               0,               0,     0.014014218,    -0.043123132,     0.037731327 ],
  [               0,               0,               0,       2.5247735,      -14.914243,       23.017353 ],
  [               0,               0,               0,      0.11937375,       0.3578543,      -2.5308929 ],
  [               0,               0,               0,      -3.7206802,        14.99578,      -14.999931 ],
  [               0,               0,               0,    0.0030916048,    -0.010611456,    0.0097754561 ],
  [               0,               0,               0,       8.6498688,       -38.13461,          42.165 ],
  [               0,               0,               0,       30.626888,      -124.14885,       123.58071 ],
 ],
 [
  [               0,               0,               0,   -0.0061936493,       0.0010262,     0.042115346 ],
  [               0,               0,               0,       59.212571,      -277.33283,       327.24753 ],
  [               0,               0,               0,       8.0453035,      -38.648106,       45.199072 ],
  [               0,               0,               0,      -7.1751652,       46.857525,       -71.89914 ],
  [               0,               0,               0,    -0.030330349,    0.0036362252,      0.17507139 ],
  [               0,               0,               0,      -8.8696593,       41.481294,      -48.521965 ],
  [               0,               0,               0,       -35.06374,       172.36057,      -211.00934 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0031947298,     0.021677422 ],
  [               0,               0,               0,               0,      -12.452187,       33.038591 ],
  [               0,               0,               0,               0,       21.470286,      -54.048747 ],
  [               0,               0,               0,               0,      -14.729422,       36.314136 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00068644712,   -0.0017008157 ],
  [               0,               0,               0,               0,      -10.745849,      -12.972537 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,       497.92564,       1286.9994 ],
  [               0,               0,               0,               0,       336.13467,       1139.5172 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 75 GeV photon clusters
#


gam75_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0011878447,   0.00086113985 ],
  [               0,               0,               0,               0,       4.5601122,       2.9852082 ],
  [               0,               0,               0,               0,       -4.320228,      -1.8976288 ],
  [               0,               0,               0,               0,       -4.101525,      -2.1410993 ],
  [               0,               0,               0,               0,      0.60235053,     -0.26577085 ],
  [               0,               0,               0,               0,  -0.00093389044,    0.0014301158 ],
  [               0,               0,               0,               0,      0.73017986,      0.22889146 ],
  [               0,               0,               0,               0,     -0.49715324,      -1.4423191 ],
 ],
 [
  [               0,               0,               0,   -0.0065028969,    0.0081983418,   -0.0016152477 ],
  [               0,               0,               0,      -18.108302,       17.510703,     -0.56704834 ],
  [               0,               0,               0,       28.983421,      -33.132447,       8.5002852 ],
  [               0,               0,               0,      -14.474285,       17.290789,      -6.4136696 ],
  [               0,               0,               0,      -8.7433526,       10.083999,      -3.0954175 ],
  [               0,               0,               0,    0.0058511519,   -0.0068295729,    0.0020319282 ],
  [               0,               0,               0,       18.814255,       -21.80767,       6.9872715 ],
  [               0,               0,               0,     -0.96567928,      0.99062865,     -0.95670183 ],
 ],
 [
  [               0,               0,               0,               0,     0.025207318,      0.01013597 ],
  [               0,               0,               0,               0,      -1.2789616,       1.5149875 ],
  [               0,               0,               0,               0,      -13.818848,       10.527107 ],
  [               0,               0,               0,               0,       14.250457,      -11.332802 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0012419428,   5.3604118e-05,    0.0013456596 ],
  [               0,               0,               0,       3.3880642,      -6.9803412,       5.4506504 ],
  [               0,               0,               0,      -2.1663491,       4.2238145,      -3.0402311 ],
  [               0,               0,               0,      -1.0751109,       1.9114421,      -2.0343766 ],
  [               0,               0,               0,     0.078232419,     -0.12226202,    -0.048100805 ],
  [               0,               0,               0,    0.0011413546,   -0.0021604913,    0.0010345555 ],
  [               0,               0,               0,       4.3525772,      -10.857718,       8.7993157 ],
  [               0,               0,               0,       9.2655182,      -21.772969,       14.190287 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.054549428,     -0.13746578,     -0.15440869,      0.35119202 ],
  [               0,               0,       634.35254,      -908.56646,      -1239.6015,       1779.6271 ],
  [               0,               0,      -389.14237,       647.01575,       397.49885,      -754.22421 ],
  [               0,               0,       280.32945,      -346.39546,      -857.45179,        1111.434 ],
  [               0,               0,       2142.1465,      -3051.7779,       -4353.447,       6200.6638 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,               0,   9.7841701e-07 ],
  [               0,               0,               0,               0,               0,      -24.961077 ],
  [               0,               0,               0,               0,               0,      -4912.9886 ],
  [               0,               0,               0,               0,               0,       5612.6666 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0036212179 ],
  [               0,               0,               0,               0,               0,       4.2611655 ],
  [               0,               0,               0,               0,               0,     -0.46472228 ],
  [               0,               0,               0,               0,               0,     -0.82338286 ],
  [               0,               0,               0,               0,               0,     -0.32053956 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       0.3344367,     -0.49329285 ],
  [               0,               0,               0,               0,       456.58637,      -674.99204 ],
  [               0,               0,               0,               0,       36.017914,      -56.493521 ],
  [               0,               0,               0,               0,       77.980547,      -115.68344 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00077276937,  -0.00072258857,   0.00076100706,    0.0068773633 ],
  [               0,               0,       10.810557,      -14.697971,      -38.180264,       55.995764 ],
  [               0,               0,      -6.4327753,       11.417315,        15.50099,       -28.62101 ],
  [               0,               0,      -5.3051645,       9.3662309,       13.076014,      -24.101206 ],
  [               0,               0,     -0.40055151,      0.67391957,      0.98466431,      -1.6571041 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.001608707,    0.0032628999 ],
  [               0,               0,               0,               0,     -0.35438671,         2.00252 ],
  [               0,               0,               0,               0,      0.16847028,      -1.2683079 ],
  [               0,               0,               0,               0,      0.30920125,      -1.5269019 ],
  [               0,               0,               0,               0,    0.0010074727,   -0.0094695694 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00051023616,  -0.00027474183,    0.0011774509,    0.0068618584 ],
  [               0,               0,    0.0075420167,      0.62003945,     -0.25791557,      0.15743033 ],
  [               0,               0,       2.0976172,      -4.4236059,      -9.3983111,       18.615106 ],
  [               0,               0,      0.69266454,      -1.1656264,       -3.823847,        5.619112 ],
  [               0,               0,     -0.34073097,      0.62726911,       2.0490022,      -3.9030281 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0012439376,    0.0025731029 ],
  [               0,               0,               0,               0,      -99.745512,       231.00168 ],
  [               0,               0,               0,               0,       43.628202,      -101.47237 ],
  [               0,               0,               0,               0,       4.5650122,      -10.263346 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.013320195,     0.033780987 ],
  [               0,               0,               0,               0,       12.354093,      -26.012054 ],
  [               0,               0,               0,               0,     0.015904426,      -1.2441137 ],
  [               0,               0,               0,               0,      0.37989788,      -1.9897096 ],
  [               0,               0,               0,               0,      0.17225387,     -0.37985812 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.018676371,     0.084277057,    -0.086386588 ],
  [               0,               0,               0,      -107.97687,       518.33647,      -615.88144 ],
  [               0,               0,               0,      -241.55206,       1159.7269,       -1393.408 ],
  [               0,               0,               0,        -18.0004,       100.24682,      -137.11952 ],
  [               0,               0,               0,     0.012883834,    0.0043675747,    -0.085819325 ],
  [               0,               0,               0,       -4.289747,       34.188958,      -57.306121 ],
  [               0,               0,               0,      -7.2021528,       16.592801,      -0.4282909 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0080114083,     0.029285545 ],
  [               0,               0,               0,               0,      -3346.4532,       8324.1117 ],
  [               0,               0,               0,               0,      -20.546053,      -52.712475 ],
  [               0,               0,               0,               0,      -8.5257878,       20.226586 ],
  [               0,               0,               0,               0,      0.18630282,      0.14991269 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0011011405,   0.00082628519 ],
  [               0,               0,               0,               0,        4.557841,       3.0624264 ],
  [               0,               0,               0,               0,      -8.2831543,        -2.86563 ],
  [               0,               0,               0,               0,      -7.9318914,      -3.1437248 ],
  [               0,               0,               0,               0,       1.5346928,     -0.51305976 ],
  [               0,               0,               0,               0,   -0.0034128379,    0.0034861546 ],
  [               0,               0,               0,               0,      0.59847586,      0.16551408 ],
  [               0,               0,               0,               0,     -0.41507488,      -1.4623273 ],
 ],
 [
  [               0,               0,               0,   -0.0063507242,    0.0080650178,    -0.001609013 ],
  [               0,               0,               0,      -17.886681,       17.027326,     -0.31925423 ],
  [               0,               0,               0,       29.335036,      -33.605135,       8.6718387 ],
  [               0,               0,               0,      -15.004165,       17.890313,      -6.5794909 ],
  [               0,               0,               0,      -9.0499691,       10.452188,      -3.2078454 ],
  [               0,               0,               0,    0.0060023243,   -0.0069952637,    0.0020746547 ],
  [               0,               0,               0,       16.816869,       -19.43943,       6.3408442 ],
  [               0,               0,               0,      -1.0057587,       1.0824227,     -0.99376628 ],
 ],
 [
  [               0,               0,               0,               0,      -0.1377975,      0.13835769 ],
  [               0,               0,               0,               0,      -1.3887858,        1.612425 ],
  [               0,               0,               0,               0,      -15.033296,       11.485136 ],
  [               0,               0,               0,               0,       15.244101,       -12.12306 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0010817682,  -2.1955582e-05,    0.0019716723 ],
  [               0,               0,               0,        2.124768,      -4.2607873,       3.8723768 ],
  [               0,               0,               0,      -1.6448197,       3.2093311,      -2.5428388 ],
  [               0,               0,               0,     -0.50583159,      0.76948204,      -1.4359911 ],
  [               0,               0,               0,      0.10744953,     -0.20704961,     0.018628899 ],
  [               0,               0,               0,   0.00089519134,   -0.0016914965,   0.00081145408 ],
  [               0,               0,               0,       4.1031689,      -10.544486,       8.7902823 ],
  [               0,               0,               0,       11.750529,      -27.529422,       17.479167 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.058789461,     -0.13922975,     -0.16006774,      0.35013469 ],
  [               0,               0,       447.95391,      -643.50924,       -853.7426,         1231.04 ],
  [               0,               0,      -284.89895,       496.81326,       164.01936,      -418.02667 ],
  [               0,               0,       353.48407,      -447.75282,      -1025.4266,       1344.7716 ],
  [               0,               0,       2392.0454,      -3406.7252,      -4861.2739,       6921.7987 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,     -0.01984352,     0.030009689 ],
  [               0,               0,               0,               0,      -18.148478,      -13.616952 ],
  [               0,               0,               0,               0,      -172.58193,       252.17424 ],
  [               0,               0,               0,               0,       -101.3869,       144.49198 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.045362631 ],
  [               0,               0,               0,               0,               0,       1.0887602 ],
  [               0,               0,               0,               0,               0,     -0.84470704 ],
  [               0,               0,               0,               0,               0,     -0.88147514 ],
  [               0,               0,               0,               0,               0,     -0.02197577 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.37356019,     -0.55092456 ],
  [               0,               0,               0,               0,       430.36927,      -635.96386 ],
  [               0,               0,               0,               0,       46.217364,      -71.480872 ],
  [               0,               0,               0,               0,       43.725899,      -64.768703 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00071149785,  -0.00061353934,   0.00074364085,    0.0061613644 ],
  [               0,               0,       11.136815,      -15.443624,      -38.446296,       57.045592 ],
  [               0,               0,       -6.561181,       11.664261,       15.792294,      -29.194653 ],
  [               0,               0,      -5.4536996,       9.5017623,       13.923717,      -25.197751 ],
  [               0,               0,     -0.42787403,      0.74110996,      0.98592962,      -1.7192765 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0016725612,    0.0033637403 ],
  [               0,               0,               0,               0,     -0.34046515,       1.9552405 ],
  [               0,               0,               0,               0,       0.1647306,      -1.2543288 ],
  [               0,               0,               0,               0,      0.29801789,      -1.4988325 ],
  [               0,               0,               0,               0,   0.00040278766,   -0.0079871064 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00055169409,  -0.00027722667,    0.0012979458,    0.0072910919 ],
  [               0,               0,     -0.14343044,      0.99383757,     0.031717232,      -0.8122645 ],
  [               0,               0,       1.9691459,      -4.1537312,       -8.905296,         17.5842 ],
  [               0,               0,      0.67304966,      -1.1154468,      -3.9336944,       5.8087199 ],
  [               0,               0,     -0.31006245,      0.53728191,       2.0064732,      -3.7017226 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0010736962,    0.0029556145 ],
  [               0,               0,               0,               0,      -100.05281,       231.70973 ],
  [               0,               0,               0,               0,       43.692896,      -101.62185 ],
  [               0,               0,               0,               0,       4.5944357,      -10.330523 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.01322314,     0.033674185 ],
  [               0,               0,               0,               0,       11.195856,      -23.389855 ],
  [               0,               0,               0,               0,       -0.119932,     -0.92698833 ],
  [               0,               0,               0,               0,      0.45264427,      -2.1633427 ],
  [               0,               0,               0,               0,      0.16052399,     -0.35355123 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0088904319,     0.029031919,    -0.010512085 ],
  [               0,               0,               0,       -42.92074,       200.67486,      -227.84583 ],
  [               0,               0,               0,      -136.61212,       648.00855,      -769.74323 ],
  [               0,               0,               0,      -21.616635,       116.46153,      -155.13508 ],
  [               0,               0,               0,    0.0092660282,   -0.0003810953,    -0.052739404 ],
  [               0,               0,               0,      -9.8193872,       52.480418,      -68.707995 ],
  [               0,               0,               0,      -4.4996295,       8.6552799,       2.9791203 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19663273,       0.4976718 ],
  [               0,               0,               0,               0,      -3235.0425,       8048.3183 ],
  [               0,               0,               0,               0,      -31.571312,      -46.969976 ],
  [               0,               0,               0,               0,       29.092553,      -72.843805 ],
  [               0,               0,               0,               0,      0.20570367,      0.23380081 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00068508746,   0.00077527666 ],
  [               0,               0,               0,               0,       1.8606799,       3.1942659 ],
  [               0,               0,               0,               0,     -0.32334621,     -0.72088871 ],
  [               0,               0,               0,               0,      0.91897715,     -0.90574885 ],
  [               0,               0,               0,               0,      0.03948257,    -0.050608097 ],
  [               0,               0,               0,               0,   0.00034832811,  -5.6842819e-06 ],
  [               0,               0,               0,               0,      -1.9791398,       1.8617094 ],
  [               0,               0,               0,               0,     -0.20893265,      -1.8989399 ],
 ],
 [
  [               0,               0,               0,   -0.0050598002,    0.0048996562,  -9.8307912e-05 ],
  [               0,               0,               0,      -10.340025,       11.249287,      0.18240685 ],
  [               0,               0,               0,       24.349797,      -26.138028,       5.9740993 ],
  [               0,               0,               0,      -4.2893207,       4.1818681,      -2.2880064 ],
  [               0,               0,               0,      -20.477853,       28.129806,      -10.077574 ],
  [               0,               0,               0,     0.018656611,    -0.027049317,    0.0099559708 ],
  [               0,               0,               0,       18.614292,      -18.757408,       4.7919426 ],
  [               0,               0,               0,       -3.914371,       2.1872456,     -0.18517838 ],
 ],
 [
  [               0,               0,               0,               0,      0.02274797,    0.0084147193 ],
  [               0,               0,               0,               0,      -10.285955,       8.7055463 ],
  [               0,               0,               0,               0,       -9.395413,       6.9880724 ],
  [               0,               0,               0,               0,       16.041823,      -12.757199 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0061452913,    0.0079639142,     0.010799568 ],
  [               0,               0,               0,       4.0874687,      -8.4189637,       4.7730495 ],
  [               0,               0,               0,      -3.3746036,       6.9153694,      -3.9421655 ],
  [               0,               0,               0,      -3.3692911,       6.9784763,      -4.0440097 ],
  [               0,               0,               0,     -0.01084745,     0.041287013,    -0.042575227 ],
  [               0,               0,               0,   -0.0054850244,     0.011020406,   -0.0051758667 ],
  [               0,               0,               0,       2.9334775,      -5.7892182,       3.9854838 ],
  [               0,               0,               0,       4.4698784,      -10.303762,       10.665826 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.070214318,     -0.15395692,     -0.18246668,      0.37879199 ],
  [               0,               0,       567.54674,      -814.71939,      -1095.2898,       1576.8234 ],
  [               0,               0,      -274.80683,       482.31834,       142.65001,      -387.33919 ],
  [               0,               0,       366.02037,      -465.35132,      -1050.0916,       1379.3782 ],
  [               0,               0,       2392.1345,      -3406.8573,      -4861.3732,       6921.9479 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,       23.356789,      -33.201807 ],
  [               0,               0,               0,               0,       18.677869,      -27.134953 ],
  [               0,               0,               0,               0,      -18.886805,       27.442748 ],
  [               0,               0,               0,               0,     -0.16097137,      0.22485815 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       20251.976 ],
  [               0,               0,               0,               0,               0,     0.011981333 ],
  [               0,               0,               0,               0,               0,     -0.01198087 ],
  [               0,               0,               0,               0,               0,    -0.011980877 ],
  [               0,               0,               0,               0,               0,  -4.6099074e-08 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,       0.4633075,     -0.68328101 ],
  [               0,               0,               0,               0,       416.98548,      -616.01999 ],
  [               0,               0,               0,               0,       89.729081,      -136.03049 ],
  [               0,               0,               0,               0,      -8.4112376,       12.745728 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00067701197,  -0.00055669813,   0.00072418329,    0.0057688978 ],
  [               0,               0,       11.866829,      -16.855112,      -39.798052,       59.887637 ],
  [               0,               0,      -6.9132856,       12.151921,       17.128809,      -31.147904 ],
  [               0,               0,      -5.8273227,        10.15784,       14.915171,      -26.942025 ],
  [               0,               0,     -0.44971108,      0.75180079,       1.1182267,      -1.8695922 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0016490178,    0.0035165301 ],
  [               0,               0,               0,               0,     -0.22581381,       1.7248648 ],
  [               0,               0,               0,               0,     0.069197747,      -1.0677645 ],
  [               0,               0,               0,               0,      0.18686665,      -1.2864483 ],
  [               0,               0,               0,               0,  -0.00073164726,   -0.0072429521 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00035759639,   -0.0002068762,   0.00088186426,    0.0056266804 ],
  [               0,               0,      -1.0950637,       2.7801838,       4.2764232,      -8.6128915 ],
  [               0,               0,       1.9444546,      -4.0028347,       -8.976548,        17.29034 ],
  [               0,               0,       0.6294403,      -1.0283121,      -3.5141773,       4.9398901 ],
  [               0,               0,     -0.36420305,      0.63466893,       2.3255894,      -4.3019755 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0011434798,    0.0029040247 ],
  [               0,               0,               0,               0,       -96.99638,        224.6623 ],
  [               0,               0,               0,               0,       44.231289,      -102.83681 ],
  [               0,               0,               0,               0,       4.4263888,      -9.9539039 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00043923473,    0.0015007353 ],
  [               0,               0,               0,               0,       1.0656263,      0.38810317 ],
  [               0,               0,               0,               0,     -0.84369331,      0.78075044 ],
  [               0,               0,               0,               0,     -0.41223237,     -0.13642643 ],
  [               0,               0,               0,               0,     -0.04285759,      0.12060546 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.010156373,    0.0030762881,    -0.059232395 ],
  [               0,               0,               0,      -14.037245,       19.548719,       41.351313 ],
  [               0,               0,               0,       4.7596725,      -26.973792,       35.777402 ],
  [               0,               0,               0,      -1.0790862,      -3.9001266,       18.508329 ],
  [               0,               0,               0,    0.0049988178,    0.0029073943,    -0.015428588 ],
  [               0,               0,               0,       6.0699261,      -27.999509,       32.325337 ],
  [               0,               0,               0,      0.24469927,      -3.4363078,       3.7999299 ],
 ],
 [
  [               0,               0,               0,               0,    -0.016145108,     0.047512008 ],
  [               0,               0,               0,               0,       -3448.794,       8578.3897 ],
  [               0,               0,               0,               0,      -5.3715352,      -13.297379 ],
  [               0,               0,               0,               0,       62.427853,      -155.34437 ],
  [               0,               0,               0,               0,      0.18946047,      0.26153748 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,     -0.23573032,     0.097433704,    -0.011726044,    0.0081553372 ],
  [               0,               0,       127.59239,      -65.420427,       9.5551471,       3.5012133 ],
  [               0,               0,      0.91357454,     -0.71253612,      0.13499358,      -1.3275235 ],
  [               0,               0,      -10.417456,       8.4457012,      -2.2877954,      0.21348169 ],
  [               0,               0,     0.052407698,    0.0027278086,   -0.0097369966,    0.0016698669 ],
  [               0,               0,       -96.18653,       62.890681,      -8.6699626,      0.42104902 ],
  [               0,               0,       65.361919,      -51.563127,       14.777494,      -3.2164479 ],
 ],
 [
  [               0,               0,               0,    -0.011958286,     0.016066559,     0.003175416 ],
  [               0,               0,               0,      -1.1361123,     -0.25839349,       4.0360044 ],
  [               0,               0,               0,    -0.077554599,       0.1223012,      -1.3788321 ],
  [               0,               0,               0,      -2.8394684,      0.86717349,       1.2688963 ],
  [               0,               0,               0,    -0.014338746,    0.0012997954,     0.011230972 ],
  [               0,               0,               0,      0.54995021,     -0.70437291,      0.24427069 ],
  [               0,               0,               0,       11.363866,      -11.138366,      -6.4105075 ],
 ],
 [
  [               0,               0,               0,               0,     0.022365902,   -0.0092131954 ],
  [               0,               0,               0,               0,       2.1330667,       1.6699267 ],
  [               0,               0,               0,               0,         8.46019,      -8.0740279 ],
  [               0,               0,               0,               0,      -5.0560817,       4.0414675 ],
  [               0,               0,               0,               0,    -0.014878376,     0.012132519 ],
  [               0,               0,               0,               0,      -41.990975,       31.658342 ],
  [               0,               0,               0,               0,       56.533782,      -46.447723 ],
 ],
 [
  [               0,               0,               0,    -0.035347878,     0.077762316,    -0.032711813 ],
  [               0,               0,               0,        9.559453,      -21.745487,       14.987262 ],
  [               0,               0,               0,       -1.142197,        2.336969,      -2.4892374 ],
  [               0,               0,               0,       7.2273069,      -12.011823,       5.3691709 ],
  [               0,               0,               0,   -0.0029573158,     0.017984673,    -0.002317964 ],
  [               0,               0,               0,      0.22027608,     -0.47450285,      0.25526341 ],
  [               0,               0,               0,      -4.5709353,       7.4324875,       -4.833119 ],
 ],
 [
  [               0,               0,     0.011837253,    -0.027636517,    -0.028604532,      0.07020508 ],
  [               0,               0,       618.77392,      -857.91219,      -1176.7488,       1633.6646 ],
  [               0,               0,      -115.69108,       169.62147,       201.80498,      -298.64709 ],
  [               0,               0,      -52.786691,       69.067928,       93.818244,      -122.24076 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.68498299,     -0.95695969 ],
  [               0,               0,               0,               0,      -457.99038,       681.18567 ],
  [               0,               0,               0,               0,      -13.296778,       15.553463 ],
  [               0,               0,               0,               0,       4.4977431,      -5.8450001 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,     -0.46099495,      0.69174453,      0.79421854,      -1.2011466 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,       0.1613699,     -0.53522294,      0.44010073 ],
  [               0,               0,               0,       9213.0551,      -27028.914,       19825.599 ],
  [               0,               0,               0,       550.09655,      -1633.2368,       1211.0275 ],
  [               0,               0,               0,       143.05764,      -398.93271,       277.07965 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.28169457,     -0.36419341 ],
  [               0,               0,               0,               0,      -1.5546695,       3.7938251 ],
  [               0,               0,               0,               0,       34.080144,      -52.504908 ],
  [               0,               0,               0,               0,      -33.966027,       51.376359 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0028720379,    0.0082356425,    0.0034034024 ],
  [               0,               0,               0,       4.5549541,       -15.25315,       15.392244 ],
  [               0,               0,               0,      -3.1896285,       10.646082,      -10.079868 ],
  [               0,               0,               0,       1.5321751,      -5.0149203,       4.1115365 ],
  [               0,               0,               0,  -0.00063322797,  -0.00041687396,    0.0024420109 ],
  [               0,               0,               0,       16.699484,      -63.017302,       58.620317 ],
  [               0,               0,               0,      -5.8043712,       27.081988,      -30.790042 ],
 ],
 [
  [               0,               0,               0,     0.060473643,     -0.25217291,      0.26451202 ],
  [               0,               0,               0,      -35.661013,       146.18062,      -145.01129 ],
  [               0,               0,               0,      -1.5076362,       6.9734982,       -9.114789 ],
  [               0,               0,               0,      -1.6753201,       6.5924057,      -6.4297285 ],
  [               0,               0,               0,    -0.021373188,     0.088344681,    -0.089175029 ],
  [               0,               0,               0,       -1.063127,       4.2121536,      -3.1702511 ],
  [               0,               0,               0,       1.2533759,       -4.713589,       2.9079799 ],
 ],
 [
  [               0,               0,               0,     -0.62454138,       2.9656638,      -3.5076057 ],
  [               0,               0,               0,       74.655376,      -354.72191,       423.22777 ],
  [               0,               0,               0,      -26.419131,       124.55467,       -147.8824 ],
  [               0,               0,               0,       4.1987997,      -19.126205,       21.780218 ],
  [               0,               0,               0,    -0.065285668,      0.32029431,      -0.3921797 ],
  [               0,               0,               0,       13.154458,      -51.743156,       48.842879 ],
  [               0,               0,               0,       200.08623,      -967.69434,       1166.9826 ],
 ],
 [
  [               0,               0,               0,               0,    0.0036175418,     0.015516898 ],
  [               0,               0,               0,               0,      -20.662999,       52.781398 ],
  [               0,               0,               0,               0,       37.000002,      -92.268309 ],
  [               0,               0,               0,               0,      -30.327477,       74.795419 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0029480565,    0.0023723058,   -0.0060788018,    -0.044246854 ],
  [               0,               0,       -1.155768,      -1.4164712,      -3.5080746,      -4.6881848 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       10.200978,        9.256931,      -16.823091,      -141.64049 ],
  [               0,               0,       0.4053228,       0.9884946,       2.4693554,       6.0157148 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.023378192,    0.0093834995,    0.0066133331 ],
  [               0,               0,               0,       7.5444159,      -3.4518772,       4.9610384 ],
  [               0,               0,               0,      -1.9463404,       0.9233581,      -1.4657193 ],
  [               0,               0,               0,       2.4931675,      -1.4127811,      0.21508152 ],
  [               0,               0,               0,     0.012944445,   -0.0073983996,    0.0016199801 ],
  [               0,               0,               0,       10.592939,      -3.5492826,       0.5094774 ],
  [               0,               0,               0,      -11.781745,       7.2947136,      -2.9183309 ],
 ],
 [
  [               0,               0,               0,     0.021411181,    -0.023366505,      0.01237016 ],
  [               0,               0,               0,        -12.4353,       12.583128,       1.8909711 ],
  [               0,               0,               0,       1.5431531,      -2.2931338,     -0.64392317 ],
  [               0,               0,               0,       0.2004884,      0.27165788,      -0.1758907 ],
  [               0,               0,               0,   -0.0048550083,    0.0055482285,  -0.00067898086 ],
  [               0,               0,               0,      -5.5253035,       6.2746898,     -0.90358248 ],
  [               0,               0,               0,     -0.21668589,      -2.3133408,     -0.29964404 ],
 ],
 [
  [               0,               0,               0,               0,      0.17135122,     -0.12972312 ],
  [               0,               0,               0,               0,      -43.229943,       38.956628 ],
  [               0,               0,               0,               0,       21.348073,      -18.680878 ],
  [               0,               0,               0,               0,      -20.997952,        18.24303 ],
  [               0,               0,               0,               0,      0.12029449,    -0.086124771 ],
  [               0,               0,               0,               0,      -10.243635,       8.3562168 ],
  [               0,               0,               0,               0,      -10.767852,       5.5715102 ],
 ],
 [
  [               0,               0,               0,   -0.0045681152,     0.012763176,  -9.3268967e-05 ],
  [               0,               0,               0,       3.5698235,      -9.4651443,       9.5911359 ],
  [               0,               0,               0,     -0.41014054,      0.60182265,      -1.5750284 ],
  [               0,               0,               0,      0.78547058,     -0.96287551,        0.274783 ],
  [               0,               0,               0,     0.009446938,     -0.01709205,    0.0082016036 ],
  [               0,               0,               0,       1.9408023,      -4.6941514,       3.1588981 ],
  [               0,               0,               0,       4.2615382,        -12.7296,       6.3366388 ],
 ],
 [
  [               0,               0,      0.17594327,     0.011594014,      -0.1551086,     -0.24405486 ],
  [               0,               0,       123.33005,      -155.36026,        -229.716,       292.26707 ],
  [               0,               0,       19.027825,      -10.508967,      -40.231204,       24.305123 ],
  [               0,               0,       463.00306,      -632.45563,      -871.39054,        1190.405 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.090359363,      0.14323683 ],
  [               0,               0,               0,               0,      -1096.1895,       1581.8277 ],
  [               0,               0,               0,               0,       6.4252024,      -9.8727772 ],
  [               0,               0,               0,               0,     -0.16026671,     -0.54723802 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -1.1552433,       1.6871108,       2.2329638,      -3.2630711 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.97739097,      -3.0323464,       2.3555482 ],
  [               0,               0,               0,      -690.09955,       2057.1364,      -1531.1913 ],
  [               0,               0,               0,      -8.2385138,       9.5913768,       2.6555355 ],
  [               0,               0,               0,      -1.8708935,       7.8200412,       -7.375009 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.010873714,     0.012118925 ],
  [               0,               0,               0,               0,       20.436046,      -29.107598 ],
  [               0,               0,               0,               0,       31.576945,      -48.849409 ],
  [               0,               0,               0,               0,      -34.569999,       52.304126 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0025124333,   -0.0076334268,     0.013808074 ],
  [               0,               0,               0,       8.6321379,      -29.692715,       28.833679 ],
  [               0,               0,               0,        4.600673,      -15.734683,        12.03832 ],
  [               0,               0,               0,      -1.4013213,       4.9362926,      -4.2120493 ],
  [               0,               0,               0,    0.0025170237,   -0.0076328814,    0.0065503584 ],
  [               0,               0,               0,      -15.227717,       51.596575,      -43.000523 ],
  [               0,               0,               0,       9.6503163,      -32.754285,       25.130152 ],
 ],
 [
  [               0,               0,               0,     0.098532265,     -0.41358476,      0.43701235 ],
  [               0,               0,               0,      -35.113625,       145.35805,      -145.79901 ],
  [               0,               0,               0,        1.549112,      -6.0812512,       4.6433076 ],
  [               0,               0,               0,       -1.474728,       5.9312902,      -5.9098665 ],
  [               0,               0,               0,    -0.034092091,      0.14300018,     -0.14791609 ],
  [               0,               0,               0,      -3.9614306,       16.247514,       -15.68227 ],
  [               0,               0,               0,       1.5869053,      -6.0867879,       4.2720066 ],
 ],
 [
  [               0,               0,               0,    -0.013090306,     0.032701778,    0.0061866952 ],
  [               0,               0,               0,       59.760108,      -279.36936,       328.96645 ],
  [               0,               0,               0,       8.7488617,      -42.354457,       50.029436 ],
  [               0,               0,               0,       12.938045,       22.516357,      -122.40434 ],
  [               0,               0,               0,   0.00015027685,    -0.054170174,      0.23368617 ],
  [               0,               0,               0,     0.077555466,     -0.39245788,      0.49234118 ],
  [               0,               0,               0,      -9.0417381,       50.554629,      -67.104914 ],
 ],
 [
  [               0,               0,               0,               0,    0.0010454914,    0.0099635054 ],
  [               0,               0,               0,               0,      -20.545234,       52.924694 ],
  [               0,               0,               0,               0,       4.8708961,       -13.16449 ],
  [               0,               0,               0,               0,      0.14786221,     -0.31109693 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00048587781,   -0.0033515244,     0.023801608,    -0.045797356 ],
  [               0,               0,      -1.1631982,       -1.412963,      -3.5017764,      -4.6794342 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       155.95781,       161.44362,      -240.85741,      -2615.1026 ],
  [               0,               0,      0.35148929,       14.230604,      -13.017861,       51.638189 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.020726048,    0.0082503476,    0.0066941493 ],
  [               0,               0,               0,       5.6008485,      -2.5998069,       4.8901528 ],
  [               0,               0,               0,      -1.8594508,      0.88812985,      -1.4618237 ],
  [               0,               0,               0,       2.5253378,      -1.4133484,      0.20984959 ],
  [               0,               0,               0,     0.012042022,   -0.0069302017,    0.0015810539 ],
  [               0,               0,               0,       10.288079,      -3.4549488,      0.50972682 ],
  [               0,               0,               0,      -13.285385,        7.894565,      -2.9217899 ],
 ],
 [
  [               0,               0,               0,     0.023884729,      -0.0268811,     0.013621832 ],
  [               0,               0,               0,      -14.211769,       14.998746,       1.0711562 ],
  [               0,               0,               0,       1.3313657,      -2.0650059,      -0.7001794 ],
  [               0,               0,               0,      0.34797093,      0.11472356,     -0.13621982 ],
  [               0,               0,               0,   -0.0049370424,    0.0058291797,  -0.00081977569 ],
  [               0,               0,               0,      -5.7534851,       6.6026396,      -1.0329475 ],
  [               0,               0,               0,     -0.22966924,      -2.3823328,     -0.24024947 ],
 ],
 [
  [               0,               0,               0,               0,      0.16373529,     -0.12366162 ],
  [               0,               0,               0,               0,      -41.149027,       37.285178 ],
  [               0,               0,               0,               0,       20.094256,      -17.658272 ],
  [               0,               0,               0,               0,      -18.702814,       16.240738 ],
  [               0,               0,               0,               0,     0.097791701,    -0.069259489 ],
  [               0,               0,               0,               0,      -10.571648,       8.6266644 ],
  [               0,               0,               0,               0,      -11.910625,       6.5032519 ],
 ],
 [
  [               0,               0,               0,   -0.0040313953,     0.011511951,   0.00062694565 ],
  [               0,               0,               0,       3.4175522,      -9.1165732,       9.3942969 ],
  [               0,               0,               0,     -0.39770711,      0.58118379,      -1.5674722 ],
  [               0,               0,               0,      0.89759256,      -1.1434191,      0.34793072 ],
  [               0,               0,               0,     0.010093735,    -0.018020709,    0.0085427174 ],
  [               0,               0,               0,       1.9683111,      -4.7584808,       3.1680278 ],
  [               0,               0,               0,       3.9194422,      -11.928088,       5.8852424 ],
 ],
 [
  [               0,               0,      0.17447706,     0.017021448,      -0.1499035,     -0.25753046 ],
  [               0,               0,      -104.16178,       158.13243,       181.91772,      -275.10816 ],
  [               0,               0,      -50.600582,        85.83649,       90.896207,      -157.11375 ],
  [               0,               0,        488.1672,      -667.45899,      -918.82604,       1256.3736 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0061197176,    0.0041609324 ],
  [               0,               0,               0,               0,      -1151.9626,       1660.2547 ],
  [               0,               0,               0,               0,      -5.5447122,       6.8893099 ],
  [               0,               0,               0,               0,     -0.19935324,     -0.56480059 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam75_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -1.2465643,       1.8193196,       2.4235966,      -3.5392381 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.035588735,    0.0021556551,     0.079228571 ],
  [               0,               0,               0,      -2032.2032,       6046.8713,      -4494.9173 ],
  [               0,               0,               0,       55.356639,      -195.70989,       166.73406 ],
  [               0,               0,               0,      -27.817585,       90.153651,      -72.274441 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.51626896,     -0.76345848 ],
  [               0,               0,               0,               0,       23.274674,      -33.223056 ],
  [               0,               0,               0,               0,       26.209642,      -40.828085 ],
  [               0,               0,               0,               0,      -25.882685,       39.176299 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00080814461,  -0.00070331592,    0.0073533845 ],
  [               0,               0,               0,       7.1707812,      -24.958877,       24.947446 ],
  [               0,               0,               0,      0.46885654,      -1.7690339,      0.31954694 ],
  [               0,               0,               0,     -0.34853647,       1.8620919,      -1.8357093 ],
  [               0,               0,               0,    0.0016833051,  -0.00045351194,   -0.0014515029 ],
  [               0,               0,               0,     -0.15832263,      0.55459327,     -0.24673547 ],
  [               0,               0,               0,      -1.4961757,       4.4958845,       -6.156918 ],
 ],
 [
  [               0,               0,               0,    -0.024197213,      0.10238318,    -0.098597774 ],
  [               0,               0,               0,       5.4858127,      -24.716447,        30.48394 ],
  [               0,               0,               0,      0.25172304,     -0.84227073,     -0.57588287 ],
  [               0,               0,               0,       -24.26837,       100.00463,      -103.12576 ],
  [               0,               0,               0,     0.014943667,    -0.058377034,     0.069455288 ],
  [               0,               0,               0,      0.27940955,      -1.1382496,       1.1716182 ],
  [               0,               0,               0,       17.471903,      -71.987267,       72.678955 ],
 ],
 [
  [               0,               0,               0,   -0.0047787369,     0.009785886,     0.013611816 ],
  [               0,               0,               0,       45.303487,       -213.9656,       255.20871 ],
  [               0,               0,               0,       11.672156,      -55.867087,       65.634406 ],
  [               0,               0,               0,      -37.125075,        182.3023,      -223.80606 ],
  [               0,               0,               0,   -0.0066023133,    -0.020063285,     0.089440313 ],
  [               0,               0,               0,      -15.325969,       72.184779,      -85.053467 ],
  [               0,               0,               0,      -62.190752,       304.74875,      -372.04453 ],
 ],
 [
  [               0,               0,               0,               0,    0.0010487277,     0.010376706 ],
  [               0,               0,               0,               0,      -22.357899,       57.345807 ],
  [               0,               0,               0,               0,        5.207727,      -13.984505 ],
  [               0,               0,               0,               0,      0.14773815,     -0.31102354 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0022532773,    0.0017209261,   -0.0047416308,    -0.033105289 ],
  [               0,               0,      -1.1707957,      -1.4049143,      -3.4811215,      -4.6255632 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,     -0.66682941,      0.83601463,       8.4763533,       36.419709 ],
  [               0,               0,      0.63468754,       1.5657253,       3.8700628,       9.2844155 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 100 GeV photon clusters (large stats)
#


gam100large_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,       6730.3295,       63.117332,      -1.0471101,    0.0018154977 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00024915266,   0.00053315291 ],
  [               0,               0,               0,               0,      0.30714614,        4.771605 ],
  [               0,               0,               0,               0,      -89.865854,      -7.4294656 ],
  [               0,               0,               0,               0,      -91.133822,      -7.4130716 ],
  [               0,               0,               0,               0,     -0.80979467,    -0.040667923 ],
  [               0,               0,               0,               0,     0.086237596,    0.0065491357 ],
  [               0,               0,               0,               0,    -0.084104388,      0.18151068 ],
  [               0,               0,               0,               0,    -0.002191649,      -1.5686035 ],
 ],
 [
  [               0,               0,               0,    0.0037026245,   -0.0037881203,    0.0019352554 ],
  [               0,               0,               0,      -5.5815108,       5.5784569,        2.060227 ],
  [               0,               0,               0,      -2.1995641,       2.6269838,      -1.6987617 ],
  [               0,               0,               0,     -0.78833971,      0.82747288,      -1.4012747 ],
  [               0,               0,               0,      0.34041029,     -0.57077574,     0.071157394 ],
  [               0,               0,               0,  -0.00020752001,   0.00017265227,  -4.9887858e-05 ],
  [               0,               0,               0,      0.36124406,      -0.9378244,       2.9086789 ],
  [               0,               0,               0,      -8.7990626,       8.7072089,       -2.797217 ],
 ],
 [
  [               0,               0,               0,               0,     0.022662779,    0.0068094346 ],
  [               0,               0,               0,               0,       4.1865565,      -2.7840441 ],
  [               0,               0,               0,               0,      -24.152568,       18.706176 ],
  [               0,               0,               0,               0,       18.026036,      -14.335295 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00035176304,    0.0019116366,    0.0031901995 ],
  [               0,               0,               0,       2.3293741,      -4.5604573,       3.4948343 ],
  [               0,               0,               0,      -1.5660927,       3.1059493,      -2.3767321 ],
  [               0,               0,               0,      -1.5275793,       2.8886472,      -2.3082749 ],
  [               0,               0,               0,    -0.029701744,     0.048740593,    -0.070588423 ],
  [               0,               0,               0,  -0.00070729919,    0.0013075142,  -0.00056472597 ],
  [               0,               0,               0,       3.7470198,       -8.858525,       7.2044286 ],
  [               0,               0,               0,      0.79942809,      -1.2781057,       4.6969319 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.003314085 ],
  [               0,               0,               0,               0,               0,       1.6135209 ],
  [               0,               0,               0,               0,               0,     -0.98715919 ],
  [               0,               0,               0,               0,               0,    -0.080047829 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.012415513,     -0.00610776,    0.0067814654,     0.031173536 ],
  [               0,               0,       1.8226569,       4.6661352,       7.3799184,       15.484303 ],
  [               0,               0,      0.22204686,      0.27987005,      0.34687012,     0.020586306 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,     -0.21091695,     -0.30253221,     -0.43275699,     -0.91745086 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.17412098,     -0.29432195,      -0.3902291,      0.65895561 ],
  [               0,               0,       749.58112,       -1077.904,      -1474.2922,       2124.8148 ],
  [               0,               0,      -396.72683,         675.958,       323.99137,      -686.02684 ],
  [               0,               0,       358.77953,      -430.89328,      -1124.8393,       1436.7458 ],
  [               0,               0,       2691.0068,      -3833.1521,      -5468.7331,       7788.0986 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,       2.5460625,       -3.622302 ],
  [               0,               0,               0,               0,      -49.101807,       70.502084 ],
  [               0,               0,               0,               0,       19.387965,      -27.844815 ],
  [               0,               0,               0,               0,       2.3562878,      -3.2396668 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.037007788,    -0.053563213 ],
  [               0,               0,               0,               0,       96.592493,      -102.41636 ],
  [               0,               0,               0,               0,      -1020.3285,       1504.2388 ],
  [               0,               0,               0,               0,       623.34557,      -887.47139 ],
  [               0,               0,               0,               0,      0.10455783,     -0.98150062 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.13567215,       0.2079875 ],
  [               0,               0,               0,               0,      -187.06514,       282.89214 ],
  [               0,               0,               0,               0,       17.713136,      -27.651745 ],
  [               0,               0,               0,               0,       27.003052,      -39.936037 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,       271.28003 ],
  [               0,               0,               0,               0,               0,    -0.019796374 ],
  [               0,               0,               0,               0,               0,      0.01981043 ],
  [               0,               0,               0,               0,               0,     0.019792916 ],
  [               0,               0,               0,               0,               0,   7.7123521e-06 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0016916685,   -0.0015470743,    0.0024012751,     0.018311024 ],
  [               0,               0,      -10.798929,       56.528034,      -97.971035,       57.264386 ],
  [               0,               0,      -4.2850798,       19.504239,      -29.336315,       13.737982 ],
  [               0,               0,      -2.0615906,       8.3215138,      -10.651537,       3.3742538 ],
  [               0,               0,     -0.55733296,       2.4536933,      -3.5600053,       1.6996178 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0042902211,   -0.0061684481 ],
  [               0,               0,               0,               0,      -4.1797161,       10.597569 ],
  [               0,               0,               0,               0,    -0.077444096,      -1.0459486 ],
  [               0,               0,               0,               0,    -0.029785469,      -1.0273308 ],
  [               0,               0,               0,               0,     -0.05921775,      0.10871843 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.011881077,    -0.025441999,    -0.055493245,      0.12135087 ],
  [               0,               0,       -5.856786,       11.352519,       32.045047,      -60.314714 ],
  [               0,               0,       24.160338,      -153.37609,       324.24583,      -229.45424 ],
  [               0,               0,       -18.62542,       113.87113,      -230.92966,       154.20795 ],
  [               0,               0,        10.98128,      -72.909445,       161.11021,      -118.47054 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.16370924,      0.38066522 ],
  [               0,               0,               0,               0,      -108.92709,       252.28043 ],
  [               0,               0,               0,               0,       28.198549,      -66.204991 ],
  [               0,               0,               0,               0,       14.608655,      -33.240329 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.010329296,     0.027234806 ],
  [               0,               0,               0,               0,       5.6756611,      -10.625386 ],
  [               0,               0,               0,               0,      0.67038542,      -2.7862211 ],
  [               0,               0,               0,               0,      0.79630286,      -2.9919314 ],
  [               0,               0,               0,               0,      0.14174194,     -0.30700307 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00079727332,     0.003177713 ],
  [               0,               0,               0,               0,      -1.5426615,        5.846378 ],
  [               0,               0,               0,               0,      -1.7769527,       3.2427099 ],
  [               0,               0,               0,               0,      0.39359377,      -2.9262864 ],
  [               0,               0,               0,               0,     -0.16972311,     -0.70780922 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.28049806,       1.3665012,      -1.6556561 ],
  [               0,               0,               0,      -684.68907,       3314.9083,       -4005.912 ],
  [               0,               0,               0,       -76.06866,       366.58253,       -443.2001 ],
  [               0,               0,               0,      -5.8991932,       35.436065,      -51.162135 ],
  [               0,               0,               0,     0.012109854,    -0.013077019,    -0.039719746 ],
  [               0,               0,               0,      -384.34108,       1888.7806,      -2319.9634 ],
  [               0,               0,               0,      -177.42063,       905.60423,      -1150.7831 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.024983666 ],
  [               0,               0,               0,               0,               0,       3.7428226 ],
  [               0,               0,               0,               0,               0,      -5.4848134 ],
  [               0,               0,               0,               0,               0,     -0.97800749 ],
  [               0,               0,               0,               0,               0,      0.39162964 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,       6680.9351,       62.684273,     -0.89895897,    0.0010592673 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00012169513,    0.0004931249 ],
  [               0,               0,               0,               0,     -0.14972746,       4.9277988 ],
  [               0,               0,               0,               0,      -46.656616,      -13.372407 ],
  [               0,               0,               0,               0,      -47.810387,      -13.375474 ],
  [               0,               0,               0,               0,      -0.5294023,    -0.034822693 ],
  [               0,               0,               0,               0,     0.040167151,     0.011036054 ],
  [               0,               0,               0,               0,    -0.043721681,      0.18749784 ],
  [               0,               0,               0,               0,   -0.0014739547,      -1.5699708 ],
 ],
 [
  [               0,               0,               0,    0.0033457864,   -0.0033732686,      0.00181817 ],
  [               0,               0,               0,      -4.0132386,       3.8568446,       2.5266289 ],
  [               0,               0,               0,      -2.4751432,       2.8942068,      -1.7572722 ],
  [               0,               0,               0,     -0.90154794,      0.92840824,      -1.4286158 ],
  [               0,               0,               0,       0.3284579,     -0.55422761,     0.064499096 ],
  [               0,               0,               0,  -0.00032452156,    0.0002984227,  -8.2481596e-05 ],
  [               0,               0,               0,     0.035813381,     -0.85649052,       3.0189724 ],
  [               0,               0,               0,      -7.4850561,       7.1742464,      -2.4207459 ],
 ],
 [
  [               0,               0,               0,               0,     0.023480635,    0.0066752968 ],
  [               0,               0,               0,               0,     -0.12640821,      0.64666088 ],
  [               0,               0,               0,               0,      -21.964388,       16.966262 ],
  [               0,               0,               0,               0,       17.888315,      -14.225873 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0027062393,    0.0055725703,    0.0074404903 ],
  [               0,               0,               0,      0.87680616,      -2.2963348,       2.1521362 ],
  [               0,               0,               0,     -0.89282401,       2.1836509,      -1.9049824 ],
  [               0,               0,               0,     -0.72373902,       1.8458346,      -1.7712895 ],
  [               0,               0,               0,       0.0237424,    -0.039402029,  -0.00030224223 ],
  [               0,               0,               0,   0.00060706308,  -0.00097104104,   0.00042455592 ],
  [               0,               0,               0,       3.7616926,      -11.577844,       9.8427139 ],
  [               0,               0,               0,      -3.5205495,       8.0537176,    -0.029426817 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0030436683 ],
  [               0,               0,               0,               0,               0,       1.6919061 ],
  [               0,               0,               0,               0,               0,      -1.0098629 ],
  [               0,               0,               0,               0,               0,    -0.087199536 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       -360.4662,       671.98022,       235.56493,      -657.85317 ],
  [               0,               0,      -188.57072,       101.58864,       190.79175,       45.178126 ],
  [               0,               0,      -79.013896,       111.73391,       45.057962,      -63.581663 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,        1.444075,      -1.3140559,      -1.2181313,     0.070167756 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.10157433,     -0.20627775,     -0.25227351,      0.49437835 ],
  [               0,               0,        700.0463,      -1008.1818,      -1356.1422,        1957.875 ],
  [               0,               0,      -377.08152,       649.63679,       271.93798,      -615.21437 ],
  [               0,               0,        348.8675,      -415.51103,      -1115.4298,       1420.7538 ],
  [               0,               0,       2784.6507,       -3966.077,      -5660.0403,       8059.6231 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    -0.005332115,    0.0091759962 ],
  [               0,               0,               0,               0,      -20.205111,      -10.525733 ],
  [               0,               0,               0,               0,      -128.89952,       187.75875 ],
  [               0,               0,               0,               0,       -68.42482,       98.254489 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.03320739,    -0.047989138 ],
  [               0,               0,               0,               0,       401.46598,      -552.38081 ],
  [               0,               0,               0,               0,      -1078.3824,       1589.0221 ],
  [               0,               0,               0,               0,       1150.0727,      -1660.0664 ],
  [               0,               0,               0,               0,      0.17852486,     -0.92965393 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19276646,      0.29425425 ],
  [               0,               0,               0,               0,      -31.411786,       51.004907 ],
  [               0,               0,               0,               0,       19.594001,      -30.568751 ],
  [               0,               0,               0,               0,       0.9196989,      -1.1759566 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00051298872 ],
  [               0,               0,               0,               0,               0,   -0.0093449284 ],
  [               0,               0,               0,               0,               0,        2.523684 ],
  [               0,               0,               0,               0,               0,      0.33090376 ],
  [               0,               0,               0,               0,               0,       4.2235135 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0035881853,    -0.001892138,    0.0055517014,      0.02756077 ],
  [               0,               0,      -6.8361075,       39.026857,      -72.676899,       45.226271 ],
  [               0,               0,      -1.6877232,       4.5770889,      -1.3093967,       -3.448336 ],
  [               0,               0,     -0.40559978,      -2.0232791,       9.9711513,      -9.8440503 ],
  [               0,               0,     -0.46929139,       2.0969901,      -3.0960347,       1.5090713 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0041150044,   -0.0058959316 ],
  [               0,               0,               0,               0,      -4.2597486,       10.795333 ],
  [               0,               0,               0,               0,    -0.092836163,      -1.0121244 ],
  [               0,               0,               0,               0,     -0.12210291,     -0.84288693 ],
  [               0,               0,               0,               0,    -0.069338111,      0.12801265 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.011777959,     -0.02519469,    -0.055084144,      0.12034293 ],
  [               0,               0,      -6.0724672,       12.645066,       29.478616,      -58.617914 ],
  [               0,               0,       24.193061,      -153.64281,         324.932,      -230.02152 ],
  [               0,               0,      -18.671032,        114.1822,      -231.63955,       154.74907 ],
  [               0,               0,       11.052943,      -73.372125,       162.10513,      -119.18316 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.16487534,      0.38339061 ],
  [               0,               0,               0,               0,      -108.61877,        251.5553 ],
  [               0,               0,               0,               0,        28.16126,      -66.111395 ],
  [               0,               0,               0,               0,       14.615557,       -33.25987 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.009169237,     0.024389962 ],
  [               0,               0,               0,               0,       5.6214467,      -10.431251 ],
  [               0,               0,               0,               0,      0.67176736,      -2.7893972 ],
  [               0,               0,               0,               0,      0.76694709,      -2.9150364 ],
  [               0,               0,               0,               0,      0.13702883,     -0.29447526 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.019315593,    -0.045498799 ],
  [               0,               0,               0,               0,      -53.046081,        130.1016 ],
  [               0,               0,               0,               0,      -85.725138,       204.89126 ],
  [               0,               0,               0,               0,      -278.50549,       664.59622 ],
  [               0,               0,               0,               0,     -0.63413252,     -0.29468526 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0087190979,      0.05088781,     -0.16686685 ],
  [               0,               0,               0,       -88.85117,       375.38904,      -381.10388 ],
  [               0,               0,               0,       384.93469,      -1873.0005,       2276.5864 ],
  [               0,               0,               0,       -29.51849,       144.85853,      -177.60507 ],
  [               0,               0,               0,    0.0050070062,    -0.035019701,     0.053128379 ],
  [               0,               0,               0,       23.581004,      -95.553634,       92.257326 ],
  [               0,               0,               0,       59.439706,      -168.97479,       54.130734 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.023916858 ],
  [               0,               0,               0,               0,               0,       4.0694469 ],
  [               0,               0,               0,               0,               0,      -5.3084444 ],
  [               0,               0,               0,               0,               0,     -0.95811083 ],
  [               0,               0,               0,               0,               0,      0.42927888 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,       10639.024,       7.3677345,     -0.68171317,   0.00088410566 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   9.0275452e-05,   0.00051476424 ],
  [               0,               0,               0,               0,      -1.5596765,       4.8635751 ],
  [               0,               0,               0,               0,      -9.5566595,      -13.407219 ],
  [               0,               0,               0,               0,      -10.524965,      -13.405107 ],
  [               0,               0,               0,               0,     -0.40993845,   -0.0026319341 ],
  [               0,               0,               0,               0,    0.0067539384,     0.012717077 ],
  [               0,               0,               0,               0,     0.069263508,      0.16868667 ],
  [               0,               0,               0,               0,  -0.00064494174,      -1.5713932 ],
 ],
 [
  [               0,               0,               0,    0.0028616519,   -0.0027348448,    0.0016102934 ],
  [               0,               0,               0,      -3.4537536,        3.058172,       2.8205447 ],
  [               0,               0,               0,      -2.0628598,       2.4070956,      -1.6174683 ],
  [               0,               0,               0,     -0.69110364,      0.60114105,      -1.3185884 ],
  [               0,               0,               0,      0.26418406,     -0.48446822,     0.044628591 ],
  [               0,               0,               0,  -0.00031354274,   0.00028878379,  -8.1702992e-05 ],
  [               0,               0,               0,    -0.035565672,     -0.46003362,       2.8271033 ],
  [               0,               0,               0,      -4.9055548,       3.6647505,       -1.317253 ],
 ],
 [
  [               0,               0,               0,               0,     0.023935168,     0.006788341 ],
  [               0,               0,               0,               0,       1.0573194,     -0.30025116 ],
  [               0,               0,               0,               0,      -22.215695,       17.170144 ],
  [               0,               0,               0,               0,       17.468025,      -13.891485 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00060671082,  -0.00092120384,    0.0016909068 ],
  [               0,               0,               0,      -4.3388836,       9.2008056,      -1.7728471 ],
  [               0,               0,               0,      0.45680979,      -0.9769679,     -0.50203754 ],
  [               0,               0,               0,       1.4980123,       -3.424243,      0.48689169 ],
  [               0,               0,               0,     0.087488477,     -0.20264121,    -0.068954369 ],
  [               0,               0,               0,  -0.00039743829,   0.00092357162,  -0.00046313024 ],
  [               0,               0,               0,        3.463376,      -6.9478453,       5.8655102 ],
  [               0,               0,               0,      0.27659422,      -1.6083455,       3.0704762 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00080070091 ],
  [               0,               0,               0,               0,               0,       3.6800453 ],
  [               0,               0,               0,               0,               0,      -1.8314857 ],
  [               0,               0,               0,               0,               0,     -0.30209605 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.024697079,    -0.011581686,     0.015877139,     0.067514527 ],
  [               0,               0,       6.1306979,       6.6096579,        4.486128,       3.7913218 ],
  [               0,               0,       5.8454837,       5.2716365,       3.3421874,      -1.3198854 ],
  [               0,               0,               0,               0,               0,              -1 ],
  [               0,               0,      -2.3772443,       -3.357711,      -4.7205294,      -6.9063231 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.10634757,     -0.21276043,     -0.26183733,      0.50734111 ],
  [               0,               0,       714.98699,      -1029.4341,      -1387.0259,       2001.8082 ],
  [               0,               0,      -384.09572,       659.14455,       288.93013,      -638.44352 ],
  [               0,               0,       347.80562,      -414.61495,      -1110.7321,       1415.3191 ],
  [               0,               0,       2772.5867,      -3948.9657,      -5635.4481,       8024.7486 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,     0.033405291,    -0.046043491 ],
  [               0,               0,               0,               0,       -19.52843,      -11.567275 ],
  [               0,               0,               0,               0,      -90.416145,       132.64962 ],
  [               0,               0,               0,               0,      -71.937458,       103.28906 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.025769351,    -0.036855333 ],
  [               0,               0,               0,               0,       743.67093,      -1078.5432 ],
  [               0,               0,               0,               0,      -840.07968,       1237.6291 ],
  [               0,               0,               0,               0,       1370.2405,      -1988.4252 ],
  [               0,               0,               0,               0,     0.031376246,     -0.62386087 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.018903555,     0.038394862 ],
  [               0,               0,               0,               0,      -66.295081,       102.29011 ],
  [               0,               0,               0,               0,       48.625733,      -73.772898 ],
  [               0,               0,               0,               0,      -29.866315,       44.577629 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,   0.00056813512 ],
  [               0,               0,               0,               0,               0,   -0.0018875288 ],
  [               0,               0,               0,               0,               0,       2.4104893 ],
  [               0,               0,               0,               0,               0,      0.22023743 ],
  [               0,               0,               0,               0,               0,       3.8828733 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0031867273,   -0.0017155887,    0.0048841464,     0.024630175 ],
  [               0,               0,      -7.1803461,       40.868364,      -75.944702,       47.192913 ],
  [               0,               0,      -3.0264672,       11.205239,      -12.201847,       2.4680809 ],
  [               0,               0,     -0.50800711,      -1.4862259,       9.0624218,      -9.3719186 ],
  [               0,               0,     -0.37424182,       1.6033051,      -2.2472205,       1.0255043 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0039590007,   -0.0055096171 ],
  [               0,               0,               0,               0,      -3.6067003,       9.4743113 ],
  [               0,               0,               0,               0,     -0.13362314,     -0.93675777 ],
  [               0,               0,               0,               0,    -0.082593789,     -0.93596308 ],
  [               0,               0,               0,               0,    -0.044801587,     0.077234906 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.012225722,    -0.026159696,    -0.057058963,      0.12465161 ],
  [               0,               0,      -5.7626982,       10.404578,       34.783342,      -62.795931 ],
  [               0,               0,       24.816399,      -157.32509,       332.14408,      -234.69575 ],
  [               0,               0,       -17.60654,       107.76314,      -218.82014,       146.27487 ],
  [               0,               0,       10.555527,      -70.136267,       155.09138,      -114.12144 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.17013953,      0.39508877 ],
  [               0,               0,               0,               0,      -109.82123,       254.42038 ],
  [               0,               0,               0,               0,       24.549715,      -57.873496 ],
  [               0,               0,               0,               0,       18.631115,      -42.455675 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0086616882,     0.023165919 ],
  [               0,               0,               0,               0,       5.6163106,      -10.412229 ],
  [               0,               0,               0,               0,      0.60243733,      -2.6289444 ],
  [               0,               0,               0,               0,      0.61001861,      -2.5435247 ],
  [               0,               0,               0,               0,     0.096749072,     -0.20094869 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.93328328,       2.2418801 ],
  [               0,               0,               0,               0,       -880.2116,       2112.3124 ],
  [               0,               0,               0,               0,       604.64453,      -1451.2411 ],
  [               0,               0,               0,               0,        -658.224,       1574.1944 ],
  [               0,               0,               0,               0,     -0.29465724,      0.38855892 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     -0.11428884,       0.5562767,     -0.66811903 ],
  [               0,               0,               0,      -657.82088,       3184.1097,      -3846.9007 ],
  [               0,               0,               0,      -91.221996,       439.67264,      -531.31156 ],
  [               0,               0,               0,      -17.437606,       92.007051,      -120.49783 ],
  [               0,               0,               0,    0.0089416755,    0.0018265263,    -0.057155153 ],
  [               0,               0,               0,      -347.23698,       1711.0442,      -2107.2428 ],
  [               0,               0,               0,      -120.70424,       624.54658,      -802.72365 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.024257171 ],
  [               0,               0,               0,               0,               0,       4.0726203 ],
  [               0,               0,               0,               0,               0,      -5.3193587 ],
  [               0,               0,               0,               0,               0,      -0.9525428 ],
  [               0,               0,               0,               0,               0,      0.43607159 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.022724052,     0.013380843,    0.0063578906 ],
  [               0,               0,               0,       3.2335019,      -2.8041183,       4.2285497 ],
  [               0,               0,               0,       1.5868595,     -0.59857124,      -1.2510895 ],
  [               0,               0,               0,      0.47787482,     -0.16645986,     -0.02900068 ],
  [               0,               0,               0,   -0.0010474953,  -0.00024651866,   0.00050573837 ],
  [               0,               0,               0,        27.15736,      -11.538009,        1.380496 ],
  [               0,               0,               0,      -18.135669,       9.3862867,      -1.7003709 ],
 ],
 [
  [               0,               0,               0,    -0.015812153,     0.025015621,   -0.0023764703 ],
  [               0,               0,               0,       1.3400992,      -4.6088049,       6.2546218 ],
  [               0,               0,               0,      -1.5318412,       1.4984663,      -1.7185566 ],
  [               0,               0,               0,       1.5412192,        -1.42846,      0.34063854 ],
  [               0,               0,               0,    0.0077723585,    -0.010826823,    0.0042020671 ],
  [               0,               0,               0,     -0.10805786,     -0.40069342,        1.171368 ],
  [               0,               0,               0,      -10.259588,       9.7657115,      -10.139817 ],
 ],
 [
  [               0,               0,               0,               0,      0.10511418,    -0.076406701 ],
  [               0,               0,               0,               0,      -5.0352866,       7.5424472 ],
  [               0,               0,               0,               0,       15.144332,      -13.555598 ],
  [               0,               0,               0,               0,      -9.2735884,       7.4922155 ],
  [               0,               0,               0,               0,    -0.034912254,     0.028474435 ],
  [               0,               0,               0,               0,       2.4098944,      -1.0584363 ],
  [               0,               0,               0,               0,       6.8866208,      -7.9704135 ],
 ],
 [
  [               0,               0,               0,   -0.0010937046,    0.0026816324,    0.0070546041 ],
  [               0,               0,               0,       1.1951916,       -3.354506,       5.2820198 ],
  [               0,               0,               0,      0.51941832,      -1.3931187,     -0.49339503 ],
  [               0,               0,               0,      -3.9862684,      0.70362301,       8.0538433 ],
  [               0,               0,               0,     -0.03754243,     0.011358888,      0.06674794 ],
  [               0,               0,               0,     -0.12710122,      0.40394154,     -0.23565068 ],
  [               0,               0,               0,      0.65142088,      -1.4983817,      -2.2562421 ],
 ],
 [
  [               0,               0,    -0.007981213,    -0.031612817,    -0.012637851,      0.10735561 ],
  [               0,               0,       493.38844,      -680.51703,      -938.57768,       1296.9444 ],
  [               0,               0,      -130.41953,       186.91569,       232.83435,      -335.73091 ],
  [               0,               0,      -48.575714,       63.425783,       86.966087,      -113.12228 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.70561457,      -0.9846016 ],
  [               0,               0,               0,               0,      -206.34148,       328.88156 ],
  [               0,               0,               0,               0,      -29.843366,       38.783592 ],
  [               0,               0,               0,               0,       1.1348132,     -0.86790041 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,      0.46349064,     -0.60940429,      -1.0900771,         1.44994 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0036516759,    0.0043865451,    0.0077096807,   -0.0061677997 ],
  [               0,               0,       51.049802,      -50.788732,      -94.198218,       88.098911 ],
  [               0,               0,       96.476718,      -152.81202,      -217.25431,        342.4466 ],
  [               0,               0,       33.481346,       -45.14462,       -69.79828,       93.504768 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0021061149,   -0.0022582976 ],
  [               0,               0,               0,               0,       381.87668,      -572.37913 ],
  [               0,               0,               0,               0,      -630.40679,       952.48488 ],
  [               0,               0,               0,               0,       594.62024,      -899.13737 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.094506282,      0.31641263,     -0.25511552 ],
  [               0,               0,               0,       23.737015,      -80.307494,       70.499069 ],
  [               0,               0,               0,      -1.7426178,       5.9103212,       -6.189331 ],
  [               0,               0,               0,      0.08734886,     -0.23259212,      0.14522939 ],
  [               0,               0,               0,     0.012938402,    -0.040931418,     0.032214924 ],
  [               0,               0,               0,      -22.120124,       81.330907,      -73.773403 ],
  [               0,               0,               0,      -12.050114,       42.592669,      -38.901035 ],
 ],
 [
  [               0,               0,               0,   -0.0009682277,    0.0040410557,    0.0053252671 ],
  [               0,               0,               0,    -0.078725618,      -0.9937355,       4.8668767 ],
  [               0,               0,               0,      0.32365111,      -1.0346057,     -0.46078045 ],
  [               0,               0,               0,      -369.37861,       1552.2548,      -1613.6228 ],
  [               0,               0,               0,      -3.5149909,       14.774515,      -15.361883 ],
  [               0,               0,               0,      0.38385968,      -1.6767303,       1.8404769 ],
  [               0,               0,               0,       2.0843237,      -8.3645968,        5.279687 ],
 ],
 [
  [               0,               0,               0,   -0.0029636075,  -0.00083368948,     0.029596013 ],
  [               0,               0,               0,       28.547816,      -134.73602,       161.07359 ],
  [               0,               0,               0,      -2.9928981,       13.644134,      -16.659891 ],
  [               0,               0,               0,      -12.499643,       61.543978,      -75.611885 ],
  [               0,               0,               0,   0.00026717414,    -0.005975968,     0.013897746 ],
  [               0,               0,               0,       16.986524,      -78.297782,       90.235795 ],
  [               0,               0,               0,       4.1374807,       -40.88809,       72.031452 ],
 ],
 [
  [               0,               0,               0,               0,    0.0031891978,    0.0033425531 ],
  [               0,               0,               0,               0,      -87.953102,       218.69403 ],
  [               0,               0,               0,               0,       14.895978,      -37.870022 ],
  [               0,               0,               0,               0,       7.3008125,      -17.734863 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0017209203,   0.00054865985,   -0.0021268045,    -0.024286473 ],
  [               0,               0,      -1.1952418,      -1.1735251,      -3.3528777,        -4.47889 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      0.21357406,        4.596239,       14.106929,       28.011273 ],
  [               0,               0,       1.7679912,       1.7802348,       13.561592,       4.4965443 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     -0.02556197,     0.015707367,    0.0053918582 ],
  [               0,               0,               0,       9.9629382,      -6.6365618,       5.5691158 ],
  [               0,               0,               0,     -0.40750738,       0.0347901,      -1.3224671 ],
  [               0,               0,               0,       1.8919554,     -0.54880074,    -0.016825551 ],
  [               0,               0,               0,    0.0038590386,   -0.0023979108,   0.00092463834 ],
  [               0,               0,               0,        11.87264,      -6.5832467,       1.1680304 ],
  [               0,               0,               0,      -30.811943,       10.707715,       -1.851077 ],
 ],
 [
  [               0,               0,               0,    0.0022898437,   -0.0020037836,    0.0070529023 ],
  [               0,               0,               0,      -5.4320049,       4.9722535,       3.7964715 ],
  [               0,               0,               0,     -0.90128565,      0.70967363,      -1.5067335 ],
  [               0,               0,               0,       1.1899513,      -1.0611216,      0.23822209 ],
  [               0,               0,               0,  -0.00069226264,   0.00087847931,   0.00039394453 ],
  [               0,               0,               0,      -1.1451478,       1.1840871,      0.48841085 ],
  [               0,               0,               0,      -3.1227947,       2.3176793,      -1.9683474 ],
 ],
 [
  [               0,               0,               0,               0,     0.016484374,   -0.0038935833 ],
  [               0,               0,               0,               0,       9.9600002,      -4.1417746 ],
  [               0,               0,               0,               0,       11.292993,      -10.378553 ],
  [               0,               0,               0,               0,      -9.0217427,       7.2493802 ],
  [               0,               0,               0,               0,    -0.023475046,     0.019399311 ],
  [               0,               0,               0,               0,       21.816288,      -16.147763 ],
  [               0,               0,               0,               0,      -105.52842,         54.4471 ],
 ],
 [
  [               0,               0,               0,   -0.0097353975,     0.019649806,   -0.0019016532 ],
  [               0,               0,               0,       4.4884354,      -10.155429,       9.5170344 ],
  [               0,               0,               0,    -0.082188184,     -0.17500023,      -1.1562215 ],
  [               0,               0,               0,     0.084420952,    -0.060164505,      0.12120093 ],
  [               0,               0,               0,    0.0040141363,   -0.0099907598,    0.0068323046 ],
  [               0,               0,               0,      0.17252994,       0.7715033,     -0.57983456 ],
  [               0,               0,               0,          5.4742,      -11.389996,         3.37524 ],
 ],
 [
  [               0,               0,      0.17023926,    -0.056462463,     -0.19849785,    -0.048539242 ],
  [               0,               0,      -3974.3985,       5522.3509,       7314.7225,      -10161.022 ],
  [               0,               0,      -998.88614,       1391.5664,       1879.1385,      -2619.6374 ],
  [               0,               0,       534.23188,      -729.61332,      -1003.1312,       1370.1672 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.29886309,     -0.41691748 ],
  [               0,               0,               0,               0,       11.354784,       23.659732 ],
  [               0,               0,               0,               0,      -143.82883,       198.65459 ],
  [               0,               0,               0,               0,      -4.6037674,       6.7340476 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,       -1.749531,       2.5334575,       3.4322374,      -4.9705709 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.29816333,     -0.31613697,     -0.55469341,       0.5607074 ],
  [               0,               0,       226.02362,      -341.66904,      -405.42069,       617.87182 ],
  [               0,               0,       997.09154,      -1460.4523,      -2207.8057,       3232.5501 ],
  [               0,               0,      -966.60434,         1428.56,       2092.2291,      -3091.6672 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.12848111,     -0.19118002 ],
  [               0,               0,               0,               0,       60.506356,      -87.385647 ],
  [               0,               0,               0,               0,       -91.50999,       136.66315 ],
  [               0,               0,               0,               0,       112.21107,      -169.53263 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0004192349,    0.0026491239,    0.0051330121 ],
  [               0,               0,               0,       6.7139129,      -23.285559,       23.566347 ],
  [               0,               0,               0,       1.0148484,       -3.524931,       1.7362999 ],
  [               0,               0,               0,      -2.4434803,       1.2704997,       7.0397256 ],
  [               0,               0,               0,    -0.020081466,     0.011269969,     0.056785478 ],
  [               0,               0,               0,      -1.3308475,       4.6077289,      -3.9034032 ],
  [               0,               0,               0,       5.1255864,      -16.926557,       10.916935 ],
 ],
 [
  [               0,               0,               0,    0.0060125487,    -0.021432279,     0.027339223 ],
  [               0,               0,               0,      -1.0848766,       2.1676975,       3.2593756 ],
  [               0,               0,               0,      0.30695309,     -0.87382504,     -0.78673789 ],
  [               0,               0,               0,      -4.3766687,       16.759369,       -15.68257 ],
  [               0,               0,               0,    0.0018519089,   -0.0076301545,     0.010756993 ],
  [               0,               0,               0,       1.7739195,      -7.6891954,       8.4098853 ],
  [               0,               0,               0,       12.011455,      -45.031413,       39.447949 ],
 ],
 [
  [               0,               0,               0,   -0.0034079228,     -0.01107927,     0.055474476 ],
  [               0,               0,               0,       41.661456,      -194.09152,       228.60886 ],
  [               0,               0,               0,       8.5160749,       -40.81617,       47.684866 ],
  [               0,               0,               0,      -10.574436,       26.144282,      -6.6865455 ],
  [               0,               0,               0,     0.017937469,    -0.054455128,     0.074413955 ],
  [               0,               0,               0,       1.3158794,      -6.0038403,       6.8486809 ],
  [               0,               0,               0,      -11.737643,       60.760264,      -78.504671 ],
 ],
 [
  [               0,               0,               0,               0,    -0.046733909,      0.12170718 ],
  [               0,               0,               0,               0,       -49.52351,       124.94115 ],
  [               0,               0,               0,               0,      -43.260188,       105.31139 ],
  [               0,               0,               0,               0,       53.397204,      -131.39193 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0011049243,    0.0030022669,    -0.010919592,   -0.0081690589 ],
  [               0,               0,      -1.2028134,      -1.3772989,      -3.4183146,      -4.4863476 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       3.8680668,       10.314758,       4.7295467,      -2.1683183 ],
  [               0,               0,       3.5001994,       3.2417376,       0.8131988,       4.1089841 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.024648438,     0.015197551,    0.0054387173 ],
  [               0,               0,               0,       9.2874917,       -6.271395,       5.5335904 ],
  [               0,               0,               0,      -0.4566194,     0.052945354,      -1.3232693 ],
  [               0,               0,               0,       1.9341143,     -0.57326184,    -0.015075682 ],
  [               0,               0,               0,    0.0032819216,   -0.0021367896,   0.00091767176 ],
  [               0,               0,               0,       11.651721,      -6.4484605,       1.1576555 ],
  [               0,               0,               0,      -30.055603,       10.528681,       -1.850445 ],
 ],
 [
  [               0,               0,               0,    0.0050813392,   -0.0061475883,    0.0086303768 ],
  [               0,               0,               0,      -6.8787448,       7.0413894,       3.0384224 ],
  [               0,               0,               0,      -0.6455881,      0.37844881,      -1.4019887 ],
  [               0,               0,               0,       1.0407795,     -0.86772928,      0.17916794 ],
  [               0,               0,               0,   -0.0014021007,    0.0020082634,  -3.9506169e-05 ],
  [               0,               0,               0,      -1.6684745,       1.9788639,      0.16741584 ],
  [               0,               0,               0,      -1.3016709,     0.032843316,      -1.2810865 ],
 ],
 [
  [               0,               0,               0,               0,     0.063350399,    -0.041920084 ],
  [               0,               0,               0,               0,     -0.47641846,       4.2774365 ],
  [               0,               0,               0,               0,       9.1299323,      -8.7153211 ],
  [               0,               0,               0,               0,      0.90608417,     -0.66676888 ],
  [               0,               0,               0,               0,    -0.011135932,     0.010884735 ],
  [               0,               0,               0,               0,       12.921403,      -9.7913007 ],
  [               0,               0,               0,               0,      -102.80418,       29.161598 ],
 ],
 [
  [               0,               0,               0,    -0.010252603,     0.020653325,   -0.0023730453 ],
  [               0,               0,               0,       4.5403369,      -10.254619,        9.559338 ],
  [               0,               0,               0,     -0.11525126,     -0.10943155,       -1.189423 ],
  [               0,               0,               0,      0.13262832,     -0.17089043,      0.18696549 ],
  [               0,               0,               0,    0.0044284507,    -0.010972821,    0.0074327317 ],
  [               0,               0,               0,      0.30297012,      0.49971718,     -0.44342198 ],
  [               0,               0,               0,       5.4938371,       -11.32359,       3.3041499 ],
 ],
 [
  [               0,               0,      0.17121053,    -0.028367919,     -0.17944694,     -0.12795474 ],
  [               0,               0,      -3850.8351,       5377.4857,       7007.2768,      -9786.1636 ],
  [               0,               0,      -624.53792,       879.98988,       1174.3153,      -1656.3626 ],
  [               0,               0,       553.16636,      -756.69417,      -1040.3109,       1423.1847 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.40242273,     -0.56336109 ],
  [               0,               0,               0,               0,      -1023.9048,        1477.559 ],
  [               0,               0,               0,               0,       28.429748,      -43.699015 ],
  [               0,               0,               0,               0,    -0.080033762,     -0.48999392 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam100large_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -1.9285483,       2.7902531,       3.7997259,      -5.4977872 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.24480713,     -0.22089934,      -0.4259959,      0.33694629 ],
  [               0,               0,        200.5661,      -309.21073,      -335.27359,       525.27957 ],
  [               0,               0,       955.64133,      -1406.5281,      -2101.8018,        3091.995 ],
  [               0,               0,      -887.90464,       1310.9295,       1926.6544,      -2844.1286 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.22408098,     -0.33366397 ],
  [               0,               0,               0,               0,       33.075056,      -46.772207 ],
  [               0,               0,               0,               0,       -16.07166,       22.746825 ],
  [               0,               0,               0,               0,        32.09206,      -48.425328 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.087429016,      0.29839806,     -0.24168253 ],
  [               0,               0,               0,       21.189181,      -72.080832,       63.702863 ],
  [               0,               0,               0,      -2.4259467,       8.1781003,      -8.0581881 ],
  [               0,               0,               0,      0.15831521,     -0.47394485,       0.3569228 ],
  [               0,               0,               0,    0.0025832852,   -0.0086707383,    0.0070919546 ],
  [               0,               0,               0,       8.9744525,      -29.875551,       27.749874 ],
  [               0,               0,               0,     -0.71070814,        2.514457,      -6.8975043 ],
 ],
 [
  [               0,               0,               0,    0.0057467604,    -0.020185697,     0.025954859 ],
  [               0,               0,               0,     -0.97382303,       1.7157931,       3.7114746 ],
  [               0,               0,               0,      0.30282766,     -0.85704581,     -0.80301567 ],
  [               0,               0,               0,      -4.5215447,       17.298491,      -16.164215 ],
  [               0,               0,               0,     0.001619324,   -0.0067459529,     0.010113248 ],
  [               0,               0,               0,        1.711572,      -7.4217799,       8.1224237 ],
  [               0,               0,               0,       11.853296,      -44.430886,       38.876289 ],
 ],
 [
  [               0,               0,               0,    -0.027031015,     0.096546227,     -0.06682502 ],
  [               0,               0,               0,       48.502902,      -225.59031,       264.83155 ],
  [               0,               0,               0,       5.2863606,      -25.798594,       30.233071 ],
  [               0,               0,               0,       13.444311,       7.0224624,      -90.855613 ],
  [               0,               0,               0,    -0.012673683,    -0.081152421,      0.35525901 ],
  [               0,               0,               0,     -0.06328051,        0.334618,      -0.4352398 ],
  [               0,               0,               0,       7.4359078,      -42.964051,       58.345674 ],
 ],
 [
  [               0,               0,               0,               0,     -0.06494554,      0.16710018 ],
  [               0,               0,               0,               0,      -40.594521,       102.89765 ],
  [               0,               0,               0,               0,      -33.741451,       81.880148 ],
  [               0,               0,               0,               0,       43.518672,      -107.07184 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0015138904,    0.0012314791,    -0.003365565,      -0.0222665 ],
  [               0,               0,      -1.1977042,      -1.3760584,      -3.4142097,      -4.4613899 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       261.71153,       222.71955,      -520.95218,      -3936.0802 ],
  [               0,               0,       1.2795661,       1.2707448,       24.875473,    -0.019121766 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 200 GeV photon clusters
#


gam200_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00095515073,     0.004921058 ],
  [               0,               0,               0,               0,     -0.30584384,       2.3411867 ],
  [               0,               0,               0,               0,    -0.053323914,        -1.18536 ],
  [               0,               0,               0,               0,      0.12269854,      -1.3563506 ],
  [               0,               0,               0,               0,    0.0067115618,    -0.033970682 ],
  [               0,               0,               0,               0,  -0.00017854866,   0.00074433018 ],
  [               0,               0,               0,               0,    -0.057634182,       1.3037569 ],
  [               0,               0,               0,               0,      0.25000048,       1.3980603 ],
 ],
 [
  [               0,               0,               0,   -0.0016959973,    0.0006913986,    0.0013410603 ],
  [               0,               0,               0,       1.3629906,       3.6133075,      0.25529019 ],
  [               0,               0,               0,       3.7851054,      -4.5693253,      0.48308481 ],
  [               0,               0,               0,       3.3520666,        -5.10452,      0.71202444 ],
  [               0,               0,               0,      -1.0162989,      0.67769909,     -0.17443767 ],
  [               0,               0,               0,   0.00054197728,  -0.00087107688,   0.00036669696 ],
  [               0,               0,               0,     -0.32880291,       1.8489236,        1.344393 ],
  [               0,               0,               0,      -13.375685,       8.8288041,      -3.0510535 ],
 ],
 [
  [               0,               0,               0,               0,     0.022589012,    0.0067423952 ],
  [               0,               0,               0,               0,        4.817619,      -3.2736353 ],
  [               0,               0,               0,               0,       -24.61263,       19.061813 ],
  [               0,               0,               0,               0,       18.587937,      -14.782572 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00065974014,   2.0637191e-05,    0.0009755148 ],
  [               0,               0,               0,       -1.937121,       5.3630129,     -0.80702381 ],
  [               0,               0,               0,      0.55046523,      -1.8514173,      0.24207764 ],
  [               0,               0,               0,       1.1502949,       -3.765043,        1.338133 ],
  [               0,               0,               0,    -0.080563059,      0.15200288,      -0.2105512 ],
  [               0,               0,               0,   0.00058797362,   -0.0011272408,   0.00060545893 ],
  [               0,               0,               0,      -1.8212022,       4.6849334,     -0.68756554 ],
  [               0,               0,               0,        8.915336,       -23.97578,       17.685088 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.21180422,     -0.40029081,      -0.5045801,      0.93169615 ],
  [               0,               0,      -456.77174,         644.451,        988.5185,      -1391.3039 ],
  [               0,               0,      -929.67368,       1458.5488,       1372.2979,      -2226.8847 ],
  [               0,               0,      -54.116524,       196.92217,      -448.19825,       392.47885 ],
  [               0,               0,       2845.9885,      -4058.5034,      -5777.7045,       8237.5309 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,     0.058534204,    -0.083215182 ],
  [               0,               0,               0,               0,       1400.2305,      -1995.7786 ],
  [               0,               0,               0,               0,       120.00177,      -176.19159 ],
  [               0,               0,               0,               0,      -510.91935,       741.43687 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.078127955,       0.1152477 ],
  [               0,               0,               0,               0,       1994.4175,       -2898.585 ],
  [               0,               0,               0,               0,      -211.82724,       322.11311 ],
  [               0,               0,               0,               0,       486.13243,        -704.593 ],
  [               0,               0,               0,               0,      -1.4698147,       1.9867463 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.021821418,     0.048934526 ],
  [               0,               0,               0,               0,      -169.70673,       255.50449 ],
  [               0,               0,               0,               0,       77.111314,      -115.85978 ],
  [               0,               0,               0,               0,      -42.780795,        63.74309 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.002539972,     0.026934545,    -0.080625511,     0.073994957 ],
  [               0,               0,      0.26395957,       18.233196,      -58.167531,       46.903323 ],
  [               0,               0,      -15.481738,       73.793703,      -117.25599,       61.075867 ],
  [               0,               0,       8.0346076,      -41.829866,       72.123773,       -42.22353 ],
  [               0,               0,      -2.7875206,       13.087803,      -20.419815,       10.593183 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0038338525,   -0.0052867931 ],
  [               0,               0,               0,               0,      -4.1179922,       10.575842 ],
  [               0,               0,               0,               0,     -0.41166299,     -0.38947834 ],
  [               0,               0,               0,               0,     -0.12897039,     -0.86798275 ],
  [               0,               0,               0,               0,    0.0083842775,     -0.01888742 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00026623971,  -0.00022869626,   0.00059402361,    0.0050132596 ],
  [               0,               0,       1.9479183,      -3.5573076,      -9.4769725,       20.451627 ],
  [               0,               0,      0.54343649,      -1.4511859,     -0.77250393,        1.772406 ],
  [               0,               0,      -0.6868245,       2.0946161,       1.3171374,      -6.7520175 ],
  [               0,               0,     -0.29154317,      0.32229183,       2.7067594,       -4.382724 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.064639396 ],
  [               0,               0,               0,               0,               0,      0.53271978 ],
  [               0,               0,               0,               0,               0,      -0.5084044 ],
  [               0,               0,               0,               0,               0,     0.023081839 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00071685321,    0.0019837212 ],
  [               0,               0,               0,               0,      0.95935692,      0.21853334 ],
  [               0,               0,               0,               0,     -0.66865108,      0.35019638 ],
  [               0,               0,               0,               0,      0.54993568,      -2.4436663 ],
  [               0,               0,               0,               0,       0.1475591,      -0.3240187 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.021195947,    -0.043192655 ],
  [               0,               0,               0,               0,      -59.264087,       151.20011 ],
  [               0,               0,               0,               0,       2.1451556,      -6.8249281 ],
  [               0,               0,               0,               0,      0.22553271,     -0.44022149 ],
  [               0,               0,               0,               0,    -0.028556982,     0.069984118 ],
  [               0,               0,               0,               0,       64.271577,      -154.70387 ],
  [               0,               0,               0,               0,       141.49327,      -345.15423 ],
 ],
 [
  [               0,               0,               0,               0,       1.0350837,      -2.5618933 ],
  [               0,               0,               0,               0,       -3089.972,       7688.7878 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       9.9274386,      -24.428037 ],
  [               0,               0,               0,               0,     -0.35742871,       1.7789461 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00029553656,   0.00077455413 ],
  [               0,               0,               0,               0,      0.85040082,       3.7438413 ],
  [               0,               0,               0,               0,     -0.21944512,     -0.89150338 ],
  [               0,               0,               0,               0,      -0.8160034,     -0.92712642 ],
  [               0,               0,               0,               0,     -0.28113102,    -0.012488453 ],
  [               0,               0,               0,               0,  -6.7518377e-06,   1.6571267e-05 ],
  [               0,               0,               0,               0,      -6.9877999,       4.5226614 ],
  [               0,               0,               0,               0,      -8.5363216,      -1.7034494 ],
 ],
 [
  [               0,               0,               0,   0.00096278424,   9.9263387e-05,   0.00061652689 ],
  [               0,               0,               0,      -12.398483,       14.889921,        -1.34993 ],
  [               0,               0,               0,       1.4506624,      -2.2875001,    0.0043281971 ],
  [               0,               0,               0,       4.9211329,      -7.3549833,       1.5304123 ],
  [               0,               0,               0,      0.34192698,      -0.6733671,       0.1340887 ],
  [               0,               0,               0,  -0.00046670229,    0.0008117165,  -0.00035925464 ],
  [               0,               0,               0,       1.9525447,      0.37926918,        1.353511 ],
  [               0,               0,               0,       1.8858349,      -5.7783992,      -2.8453163 ],
 ],
 [
  [               0,               0,               0,               0,     0.023147893,    0.0072619794 ],
  [               0,               0,               0,               0,       4.7430891,      -3.2223978 ],
  [               0,               0,               0,               0,      -23.787665,       18.412016 ],
  [               0,               0,               0,               0,       17.885331,      -14.223801 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.054937616,      0.10186547,      0.39578899 ],
  [               0,               0,               0,     -0.64878747,       1.6912734,     -0.82848445 ],
  [               0,               0,               0,      0.63482968,      -1.6539091,      0.80892163 ],
  [               0,               0,               0,      0.64338356,      -1.6773772,      0.82308008 ],
  [               0,               0,               0,   0.00073066014,   -0.0024480509,    0.0012112498 ],
  [               0,               0,               0,    0.0011008614,   -0.0028877823,    0.0019009852 ],
  [               0,               0,               0,      0.99388885,       1.6204553,     -0.38800885 ],
  [               0,               0,               0,       2.9760891,       -6.224365,       7.4055751 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.065511291,     -0.25893996,     -0.25303091,      0.71098365 ],
  [               0,               0,      -1182.0464,       1682.2753,       2476.7886,       -3520.574 ],
  [               0,               0,      -351.78097,       658.39097,       161.64371,      -548.90663 ],
  [               0,               0,      -3.7883334,       121.52411,      -580.22361,       587.94376 ],
  [               0,               0,       3075.5321,       -4387.504,      -6245.5877,       8907.9762 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,   -0.0019417757,     0.002192227 ],
  [               0,               0,               0,               0,       -1818.499,       2631.5358 ],
  [               0,               0,               0,               0,      -309.74011,       446.26209 ],
  [               0,               0,               0,               0,       685.10728,      -988.75296 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.039778293,     0.058925003 ],
  [               0,               0,               0,               0,       1620.0436,      -2349.0519 ],
  [               0,               0,               0,               0,      -1606.8457,       2370.8594 ],
  [               0,               0,               0,               0,        940.8238,      -1364.6035 ],
  [               0,               0,               0,               0,      0.45601434,     -0.95250264 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.013732721,     0.035674763 ],
  [               0,               0,               0,               0,      -121.64433,       184.01911 ],
  [               0,               0,               0,               0,       52.885099,      -79.940804 ],
  [               0,               0,               0,               0,      -32.444195,       48.380896 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.002585231,     0.026974083,    -0.080344229,     0.073625352 ],
  [               0,               0,       0.2678981,       18.215687,      -58.153449,       46.910383 ],
  [               0,               0,      -15.474821,       73.744824,      -117.14921,       61.001589 ],
  [               0,               0,       8.0193613,      -41.775703,       72.071646,      -42.216512 ],
  [               0,               0,      -2.7831864,       13.066308,      -20.384413,       10.573833 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0038393169,   -0.0052961387 ],
  [               0,               0,               0,               0,      -4.1195728,       10.578166 ],
  [               0,               0,               0,               0,      -0.4080204,     -0.39672819 ],
  [               0,               0,               0,               0,     -0.13164132,     -0.86275674 ],
  [               0,               0,               0,               0,    0.0073158192,    -0.016761926 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00026663848,  -0.00022841829,   0.00059541982,    0.0050141479 ],
  [               0,               0,       1.9398476,      -3.5401322,      -9.4380487,       20.368212 ],
  [               0,               0,      0.55127496,      -1.4675669,     -0.81167117,       1.8539501 ],
  [               0,               0,     -0.68407691,       2.0872828,       1.3099746,       -6.730236 ],
  [               0,               0,     -0.29208436,      0.32426326,       2.7059538,      -4.3847078 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.063527373 ],
  [               0,               0,               0,               0,               0,      0.53629295 ],
  [               0,               0,               0,               0,               0,     -0.51151633 ],
  [               0,               0,               0,               0,               0,     0.023490595 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00076127776,    0.0021106675 ],
  [               0,               0,               0,               0,      0.80589437,      0.49559715 ],
  [               0,               0,               0,               0,      -0.7322957,      0.50845472 ],
  [               0,               0,               0,               0,      0.53828013,      -2.4176632 ],
  [               0,               0,               0,               0,      0.13676976,     -0.29991648 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.037548355,    -0.082660587 ],
  [               0,               0,               0,               0,      -64.592448,          163.78 ],
  [               0,               0,               0,               0,     -0.34725253,     -0.71212701 ],
  [               0,               0,               0,               0,       2.4024348,      -5.7764853 ],
  [               0,               0,               0,               0,   -0.0024370794,    0.0061154733 ],
  [               0,               0,               0,               0,       10.091702,      -8.6551364 ],
  [               0,               0,               0,               0,      -27.992985,       67.869425 ],
 ],
 [
  [               0,               0,               0,               0,       1.0350837,      -2.5618933 ],
  [               0,               0,               0,               0,       -3089.972,       7688.7878 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       9.9274386,      -24.428037 ],
  [               0,               0,               0,               0,     -0.35742871,       1.7789461 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00095940059,   0.00063021431 ],
  [               0,               0,               0,               0,      -2.4333241,       4.3618179 ],
  [               0,               0,               0,               0,     -0.90177296,     -0.75503314 ],
  [               0,               0,               0,               0,      -1.7459748,     -0.78860758 ],
  [               0,               0,               0,               0,      -1.1320377,     0.057325804 ],
  [               0,               0,               0,               0,   -0.0011852028,   0.00011461671 ],
  [               0,               0,               0,               0,      -1.1640674,      0.11939293 ],
  [               0,               0,               0,               0,      -1.5934265,      -1.8823494 ],
 ],
 [
  [               0,               0,               0,   0.00043481574,  -0.00030326764,   0.00099643657 ],
  [               0,               0,               0,      -3.3042186,       6.4036363,       1.1096053 ],
  [               0,               0,               0,      -3.0066997,       3.0196939,      -1.6626339 ],
  [               0,               0,               0,       4.8422631,      -6.5776415,      0.89987153 ],
  [               0,               0,               0,      0.73250479,      -1.1610934,      0.26556223 ],
  [               0,               0,               0,    0.0010911384,   -0.0011528384,   0.00032627385 ],
  [               0,               0,               0,      -23.083742,       28.125568,      -5.9770095 ],
  [               0,               0,               0,       22.260254,       -33.21463,       8.1179985 ],
 ],
 [
  [               0,               0,               0,               0,     0.021958125,    0.0060422847 ],
  [               0,               0,               0,               0,       4.9117814,      -3.3377132 ],
  [               0,               0,               0,               0,      -25.749319,       19.957436 ],
  [               0,               0,               0,               0,       19.544171,      -15.543014 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.018779158,      0.03160274,     0.027014082 ],
  [               0,               0,               0,      0.19631075,      -1.1799547,        1.923248 ],
  [               0,               0,               0,      -0.2754566,     -0.56967812,      0.79896898 ],
  [               0,               0,               0,     -0.26956771,     -0.56692841,       0.7609955 ],
  [               0,               0,               0,    0.0002961083,     0.034548959,    -0.059596785 ],
  [               0,               0,               0,     0.034363804,   -0.0089200356,    -0.057127105 ],
  [               0,               0,               0,    -0.085436258,    -0.011246782,       0.4273218 ],
  [               0,               0,               0,     0.011868683,    -0.015116956,       4.6938905 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.060939933,     -0.25383618,     -0.24459016,      0.70195649 ],
  [               0,               0,      -1234.8215,       1756.9047,       2582.3662,      -3669.8574 ],
  [               0,               0,      -410.32454,       742.74286,       278.21275,      -716.89598 ],
  [               0,               0,       17.430719,       92.419473,      -626.62969,       651.76746 ],
  [               0,               0,       3093.4471,      -4413.0033,      -6281.9459,       8959.7241 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,    -0.067861945,      0.09872836 ],
  [               0,               0,               0,               0,      -43.745998,       24.168669 ],
  [               0,               0,               0,               0,      -79.163547,       113.44703 ],
  [               0,               0,               0,               0,      -121.95818,       177.04634 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.081150308,      0.11974538 ],
  [               0,               0,               0,               0,       1411.5787,      -2048.7408 ],
  [               0,               0,               0,               0,      -528.87121,       786.24086 ],
  [               0,               0,               0,               0,       508.88892,      -738.50452 ],
  [               0,               0,               0,               0,       0.2162111,     -0.56845728 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.012769757,     0.036138219 ],
  [               0,               0,               0,               0,      -97.523126,       148.01992 ],
  [               0,               0,               0,               0,       55.030481,      -83.137133 ],
  [               0,               0,               0,               0,       -37.79158,       56.325805 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0012287506,  -0.00071374932,     0.001505239,    0.0083845217 ],
  [               0,               0,       4.3363957,       -3.757312,      -20.301834,       25.862425 ],
  [               0,               0,      -1.2910652,        3.329517,      -0.7388106,      -3.0735234 ],
  [               0,               0,      -1.2320583,       2.2312301,       2.4657677,      -5.5998005 ],
  [               0,               0,     -0.20220593,      0.41155396,       0.2559561,     -0.62621434 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0032736437,    -0.004372581 ],
  [               0,               0,               0,               0,      -4.0386929,       10.551325 ],
  [               0,               0,               0,               0,     -0.62695319,     0.031368986 ],
  [               0,               0,               0,               0,     -0.17713266,      -0.7647282 ],
  [               0,               0,               0,               0,     0.042917315,    -0.086709457 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00035307464,  -0.00031573565,   0.00072708805,    0.0061531608 ],
  [               0,               0,       2.6357916,      -4.9000118,      -12.338632,       25.843577 ],
  [               0,               0,      0.67784829,      -1.6720127,      -1.6955262,       3.4413084 ],
  [               0,               0,     -0.60327345,       1.7870059,       1.4509571,      -6.4537542 ],
  [               0,               0,     -0.34064334,      0.48198565,       2.6839639,      -4.5853142 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.061618286 ],
  [               0,               0,               0,               0,               0,      0.54262294 ],
  [               0,               0,               0,               0,               0,     -0.51716927 ],
  [               0,               0,               0,               0,               0,     0.024273215 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0025845532,    0.0095689284 ],
  [               0,               0,               0,               0,       2.5964366,      -3.5285063 ],
  [               0,               0,               0,               0,      -1.2323967,       1.6669535 ],
  [               0,               0,               0,               0,      0.70831317,      -2.8193167 ],
  [               0,               0,               0,               0,      0.24283562,     -0.54521323 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.12029969,     -0.28490513 ],
  [               0,               0,               0,               0,      -117.88184,         294.611 ],
  [               0,               0,               0,               0,       25.090247,      -63.405094 ],
  [               0,               0,               0,               0,      -11.966084,       29.640569 ],
  [               0,               0,               0,               0,    -0.066119056,      0.16357839 ],
  [               0,               0,               0,               0,      -1.7366506,       3.2989037 ],
  [               0,               0,               0,               0,       17.673872,      -46.181586 ],
 ],
 [
  [               0,               0,               0,               0,    0.0086704824,      0.01586402 ],
  [               0,               0,               0,               0,       7.1783775,       21.816312 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,       39.126361,      -97.000124 ],
  [               0,               0,               0,               0,      0.33641048,      0.59558655 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.026217776,      0.01053875,    0.0074731004 ],
  [               0,               0,               0,       11.493248,      -4.8742479,       4.2811223 ],
  [               0,               0,               0,      0.60063303,     -0.18263107,      -1.2915658 ],
  [               0,               0,               0,      -2.4404553,      0.75133898,    -0.035282927 ],
  [               0,               0,               0,    -0.018236593,    0.0048803066,  -1.1043109e-05 ],
  [               0,               0,               0,      -2.3357217,      -1.0003474,      0.30371221 ],
  [               0,               0,               0,       2.8795756,      -2.9155542,      -2.0304682 ],
 ],
 [
  [               0,               0,               0,     0.003924697,   -0.0007643801,     0.005965893 ],
  [               0,               0,               0,      -6.0546938,       5.5330338,       3.0854307 ],
  [               0,               0,               0,      -1.4554914,       1.6234375,      -1.7980887 ],
  [               0,               0,               0,       1.4869177,      -1.5606569,      0.40877122 ],
  [               0,               0,               0,    0.0016551618,   -0.0032562401,    0.0018197626 ],
  [               0,               0,               0,      -3.8573334,       3.7069365,     0.083302903 ],
  [               0,               0,               0,      -11.551525,       12.853584,      -11.372228 ],
 ],
 [
  [               0,               0,               0,               0,     0.026464246,    -0.016950703 ],
  [               0,               0,               0,               0,      -23.694712,       23.912496 ],
  [               0,               0,               0,               0,       26.987619,      -23.095028 ],
  [               0,               0,               0,               0,      -14.113157,       11.327412 ],
  [               0,               0,               0,               0,   0.00065973738,    0.0010520657 ],
  [               0,               0,               0,               0,       2.7114642,      -1.1347819 ],
  [               0,               0,               0,               0,        19.19305,      -17.095395 ],
 ],
 [
  [               0,               0,               0,     0.015205195,    -0.037941199,     0.029677442 ],
  [               0,               0,               0,     -0.88275917,       2.4602422,       2.0398634 ],
  [               0,               0,               0,       2.1751271,      -5.0717399,       1.4361401 ],
  [               0,               0,               0,      -1.1311314,       2.7771156,       -1.527509 ],
  [               0,               0,               0,   -0.0062476733,      0.01468875,   -0.0078001192 ],
  [               0,               0,               0,      -4.2776999,       9.8419654,      -4.6818087 ],
  [               0,               0,               0,       3.7818503,      -7.0995422,       1.2397178 ],
 ],
 [
  [               0,               0,  -0.00097959836,     -0.03078897,    -0.018498231,     0.096501536 ],
  [               0,               0,       541.06595,      -749.10168,      -1023.9722,       1420.0973 ],
  [               0,               0,      -114.94985,       166.63686,        202.4145,      -295.78101 ],
  [               0,               0,      -35.289334,       45.289745,        61.79015,      -78.781397 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.37795015,     -0.50863294 ],
  [               0,               0,               0,               0,       11.892884,       22.903734 ],
  [               0,               0,               0,               0,       -21.71761,       27.856653 ],
  [               0,               0,               0,               0,      -3.5833295,       5.9273537 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_55_e = [
 [
  [               0,               0,       0.6906468,     -0.92612476,      -1.5396326,       2.0760895 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.024000692,  -0.00032079847,      0.06098262 ],
  [               0,               0,               0,      -399.47939,       1214.1112,      -920.02183 ],
  [               0,               0,               0,       231.46069,      -701.35078,       530.03271 ],
  [               0,               0,               0,       71.736676,      -207.65412,       150.17622 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.54384311,     -0.81174169 ],
  [               0,               0,               0,               0,      -127.43266,       195.03972 ],
  [               0,               0,               0,               0,       39.655807,      -60.923392 ],
  [               0,               0,               0,               0,      -13.772128,       20.836868 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.002692434,    0.0093719998,  -0.00076620147 ],
  [               0,               0,               0,       4.0565478,      -14.283298,       15.697931 ],
  [               0,               0,               0,      -8.0833851,       27.615999,      -24.714838 ],
  [               0,               0,               0,       4.3086352,      -14.656159,        12.42138 ],
  [               0,               0,               0,   0.00021158485,  -0.00034949679,   0.00060570539 ],
  [               0,               0,               0,       2.0003286,      -7.0627105,       7.1484495 ],
  [               0,               0,               0,      -42.671853,       146.23187,      -132.42811 ],
 ],
 [
  [               0,               0,               0,    -0.028952727,      0.12151384,     -0.11726531 ],
  [               0,               0,               0,       6.1731348,       -27.24768,       32.396376 ],
  [               0,               0,               0,     -0.46676453,       2.1575271,      -3.6645633 ],
  [               0,               0,               0,      -28.975885,       119.18625,      -122.78201 ],
  [               0,               0,               0,  -0.00010920611,    0.0055706946,      0.00819889 ],
  [               0,               0,               0,     0.082788732,     -0.33599698,      0.34510686 ],
  [               0,               0,               0,       14.701239,       -60.48274,        60.77633 ],
 ],
 [
  [               0,               0,               0,    0.0012878366,    -0.011954959,     0.031772402 ],
  [               0,               0,               0,       30.520821,      -145.01563,       174.44912 ],
  [               0,               0,               0,       1.4337597,      -7.1794252,       7.7961181 ],
  [               0,               0,               0,       16.906291,      -33.380702,      -8.4269357 ],
  [               0,               0,               0,     0.086368954,     0.011089516,     -0.43104643 ],
  [               0,               0,               0,       1.2029482,      -5.5559703,       6.4187625 ],
  [               0,               0,               0,      0.65937129,      -5.8821029,       7.2955741 ],
 ],
 [
  [               0,               0,               0,               0,     -0.25562824,      0.64171196 ],
  [               0,               0,               0,               0,      -6.0088126,       17.046779 ],
  [               0,               0,               0,               0,       4.3927686,      -12.103415 ],
  [               0,               0,               0,               0,       1.1946054,      -2.7189057 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   0.00058717888,   0.00032146943,   -0.0031416307,   -0.0031257929 ],
  [               0,               0,      -1.1682959,      -1.4052189,      -3.4806911,      -4.6224364 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       736.21403,        691.2054,      -1495.4396,      -11313.905 ],
  [               0,               0,      -4.3996878,       41.231492,       35.552241,       42.757939 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,     0.016599774,   -0.0082266499,    0.0077099385 ],
  [               0,               0,               0,      -3.8483715,       2.0098721,       4.8273679 ],
  [               0,               0,               0,       3.4200465,      -1.6163489,      -1.2102171 ],
  [               0,               0,               0,      -2.2454856,       1.0600969,     -0.10651101 ],
  [               0,               0,               0,   -0.0047279623,    0.0022400267,   0.00030615762 ],
  [               0,               0,               0,      -6.2373344,         3.14198,      0.50476118 ],
  [               0,               0,               0,       28.987699,      -13.901458,     -0.21203556 ],
 ],
 [
  [               0,               0,               0,     0.016017161,    -0.018137392,      0.01374119 ],
  [               0,               0,               0,      -11.166409,       12.362875,      0.86650118 ],
  [               0,               0,               0,      0.37271816,     -0.48621298,      -1.1869519 ],
  [               0,               0,               0,      0.42927041,     -0.40312172,     0.087931457 ],
  [               0,               0,               0,   -0.0010109685,    0.0016674958,  -0.00059728657 ],
  [               0,               0,               0,       1.6529482,     -0.65580081,       4.8159173 ],
  [               0,               0,               0,       35.136163,      -52.039316,      -7.4336439 ],
 ],
 [
  [               0,               0,               0,               0,     0.037208613,      -0.0220868 ],
  [               0,               0,               0,               0,      -15.056008,       16.486968 ],
  [               0,               0,               0,               0,       8.6538225,      -8.2977626 ],
  [               0,               0,               0,               0,      -4.3945982,       3.5227664 ],
  [               0,               0,               0,               0,    -0.012829252,     0.010721117 ],
  [               0,               0,               0,               0,       33.620086,      -26.111713 ],
  [               0,               0,               0,               0,        6.633192,      -7.0234537 ],
 ],
 [
  [               0,               0,               0,   -0.0026508162,    0.0010159589,    0.0090489881 ],
  [               0,               0,               0,       4.0222607,      -7.8350504,        7.962712 ],
  [               0,               0,               0,      0.27910217,      -1.0177676,     -0.69266265 ],
  [               0,               0,               0,    -0.085336546,      0.45577678,     -0.29179643 ],
  [               0,               0,               0,   0.00061982791,  -0.00015167346,   0.00010833429 ],
  [               0,               0,               0,      -4.0099067,       9.6442108,      -4.8950272 ],
  [               0,               0,               0,       1.1678108,      -1.6818461,      -1.5404867 ],
 ],
 [
  [               0,               0,       0.1694574,  -0.00042828803,     -0.15772617,     -0.20330725 ],
  [               0,               0,       372.45691,      -499.31029,      -692.15605,       931.48179 ],
  [               0,               0,      -444.48874,       628.89496,       826.97671,      -1172.0672 ],
  [               0,               0,       573.92626,      -783.84867,      -1079.7864,       1474.8279 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.039507161,    -0.040953247 ],
  [               0,               0,               0,               0,      -1039.8379,       1503.1332 ],
  [               0,               0,               0,               0,     -0.64624591,     -0.23592755 ],
  [               0,               0,               0,               0,     -0.10080063,     -0.44351595 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_35_e = [
 [
  [               0,               0,      -3.0181242,       4.3665772,       6.0014943,      -8.6825728 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,     0.083849051,     0.032095332,     -0.22290761 ],
  [               0,               0,               0,       217.32259,      -776.81904,       677.00637 ],
  [               0,               0,               0,       123.63039,      -331.82498,       218.94449 ],
  [               0,               0,               0,       -53.44626,       146.34989,      -99.364978 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0040710372,    0.0020912547 ],
  [               0,               0,               0,               0,       63.837199,      -93.495149 ],
  [               0,               0,               0,               0,     -0.57254967,      -0.3656141 ],
  [               0,               0,               0,               0,       -2.448913,       3.7477476 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.027964566,      0.09548291,     -0.07269883 ],
  [               0,               0,               0,       14.844294,      -51.083948,        47.40305 ],
  [               0,               0,               0,      -1.1689558,       3.9353464,      -4.5953468 ],
  [               0,               0,               0,       12.983318,      -72.123383,       90.095464 ],
  [               0,               0,               0,    -0.026527733,     0.010125451,      0.13043327 ],
  [               0,               0,               0,     0.075610621,     -0.26061079,      0.22682982 ],
  [               0,               0,               0,     -0.47688416,       5.1280889,       -9.576575 ],
 ],
 [
  [               0,               0,               0,     0.004381277,    -0.016632524,     0.024654238 ],
  [               0,               0,               0,     -0.08862355,       -1.537979,       6.6996413 ],
  [               0,               0,               0,      0.18269889,     -0.43377373,      -1.1696395 ],
  [               0,               0,               0,      -3.5640413,        13.73726,      -12.977562 ],
  [               0,               0,               0,   0.00096258165,   -0.0031139069,    0.0045200777 ],
  [               0,               0,               0,       1.9486853,      -8.4276643,       9.2059096 ],
  [               0,               0,               0,       9.9533112,      -36.214239,       30.084546 ],
 ],
 [
  [               0,               0,               0,  -0.00072376382,   -0.0041663622,     0.023672595 ],
  [               0,               0,               0,         32.7505,      -155.21375,       186.66961 ],
  [               0,               0,               0,       4.8170892,      -23.416595,       27.190561 ],
  [               0,               0,               0,      -12.479785,        66.45984,      -83.000403 ],
  [               0,               0,               0,   -0.0005726758,  -0.00044714196,     0.049045977 ],
  [               0,               0,               0,       2.8723092,      -13.329499,       15.472694 ],
  [               0,               0,               0,      -8.3803122,       36.790326,      -42.944888 ],
 ],
 [
  [               0,               0,               0,               0,    0.0010473737,     0.008888898 ],
  [               0,               0,               0,               0,      -21.273643,       54.954224 ],
  [               0,               0,               0,               0,       4.7285425,      -12.850689 ],
  [               0,               0,               0,               0,       0.6732529,      -1.5996823 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0028165447,    0.0022006118,    -0.005780961,    -0.042040746 ],
  [               0,               0,      -1.1801797,      -1.3942115,      -3.4593647,      -4.5834931 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,        305.1088,       256.24726,      -613.14421,      -4622.8898 ],
  [               0,               0,      0.73070898,       2.0179138,       4.4271548,       14.612667 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,   -0.0048711986,    0.0023068832,    0.0074432202 ],
  [               0,               0,               0,       5.2923368,      -2.4483464,        4.973216 ],
  [               0,               0,               0,      0.38679841,     -0.21737363,      -1.3147069 ],
  [               0,               0,               0,      -17.972461,       8.8306978,      -1.4466664 ],
  [               0,               0,               0,      0.16739492,    -0.082002065,     0.013230672 ],
  [               0,               0,               0,      -1.1799446,      0.55398626,   -0.0025779958 ],
  [               0,               0,               0,      -2.5253377,            1.14,     -0.66883667 ],
 ],
 [
  [               0,               0,               0,    0.0043844395,   -0.0053518845,    0.0091360801 ],
  [               0,               0,               0,      -7.6107545,       8.4452441,       2.4655776 ],
  [               0,               0,               0,     -0.13765972,     0.035033836,      -1.3422202 ],
  [               0,               0,               0,     -0.93098661,      0.68521877,      0.41221004 ],
  [               0,               0,               0,     0.018745232,   -0.0039491162,    -0.012986283 ],
  [               0,               0,               0,         0.24896,     -0.22375779,     0.065907189 ],
  [               0,               0,               0,     -0.95373222,      0.62947646,      -24.035294 ],
 ],
 [
  [               0,               0,               0,               0,     0.035675725,    -0.020764934 ],
  [               0,               0,               0,               0,       -13.23189,       14.988232 ],
  [               0,               0,               0,               0,       8.7165532,      -8.3431143 ],
  [               0,               0,               0,               0,      -4.5433856,       3.6404341 ],
  [               0,               0,               0,               0,    -0.012686687,     0.010599327 ],
  [               0,               0,               0,               0,       33.684292,      -26.185951 ],
  [               0,               0,               0,               0,       10.329892,      -9.9744972 ],
 ],
 [
  [               0,               0,               0,   -0.0021216121,   0.00030634451,    0.0092194527 ],
  [               0,               0,               0,       3.8511522,       -7.658597,       7.9615683 ],
  [               0,               0,               0,      0.34961879,      -1.1690364,     -0.61810451 ],
  [               0,               0,               0,     -0.13942748,      0.58110463,      -0.3595249 ],
  [               0,               0,               0,   4.0172947e-05,    0.0010267457,  -0.00046637333 ],
  [               0,               0,               0,      -3.9258792,         9.36531,      -4.6918597 ],
  [               0,               0,               0,       1.6719906,      -2.9030633,     -0.83047658 ],
 ],
 [
  [               0,               0,      0.17465733,    0.0094944718,     -0.15549665,     -0.23710571 ],
  [               0,               0,        373.3404,      -503.22356,      -696.41554,       942.14189 ],
  [               0,               0,      -378.02658,       537.81035,       701.60654,      -1000.2014 ],
  [               0,               0,       581.06557,       -794.1707,      -1093.5723,       1494.7107 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.24967109,       0.3744115 ],
  [               0,               0,               0,               0,      -1037.7254,       1500.2239 ],
  [               0,               0,               0,               0,       28.249302,        -40.6315 ],
  [               0,               0,               0,               0,    -0.083085414,     -0.48424014 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam200_layer2_etaoffv4_1_37_e = [
 [
  [               0,               0,      -2.9905395,       4.3313577,       5.9457146,      -8.6116373 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,      0.08477926,    0.0025945758,      -0.1813993 ],
  [               0,               0,               0,      -44.592045,       23.295216,       66.423326 ],
  [               0,               0,               0,      -5.0734375,       41.741118,      -52.023747 ],
  [               0,               0,               0,      -6.8644657,       10.433017,     -0.25633742 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.51675599,     -0.77122941 ],
  [               0,               0,               0,               0,      -41.788423,       66.034486 ],
  [               0,               0,               0,               0,       28.779706,      -44.823344 ],
  [               0,               0,               0,               0,      -18.442933,       27.940832 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    -0.027762757,     0.096254585,    -0.074570921 ],
  [               0,               0,               0,       14.540503,      -50.567039,       47.404602 ],
  [               0,               0,               0,       -1.731677,       5.8957432,      -6.3019674 ],
  [               0,               0,               0,       24.659087,      -83.314179,        75.62402 ],
  [               0,               0,               0,    0.0019907743,    0.0088451346,     0.027175733 ],
  [               0,               0,               0,      0.72659868,      -2.4787233,       2.1232224 ],
  [               0,               0,               0,      -5.3274546,       18.210277,      -18.428159 ],
 ],
 [
  [               0,               0,               0,    0.0042542055,     -0.01549255,     0.022798246 ],
  [               0,               0,               0,      0.29126384,       -3.298313,       8.7375272 ],
  [               0,               0,               0,      0.15679747,     -0.30372539,      -1.3311462 ],
  [               0,               0,               0,      -3.6761439,       14.193494,      -13.438887 ],
  [               0,               0,               0,   0.00060783754,   -0.0014108751,    0.0025581265 ],
  [               0,               0,               0,       1.9606625,      -8.5039708,       9.3177576 ],
  [               0,               0,               0,       10.131843,      -37.037703,       31.029047 ],
 ],
 [
  [               0,               0,               0,   -0.0047616196,    -0.004393641,     0.047187348 ],
  [               0,               0,               0,       55.107398,      -257.49131,       303.42272 ],
  [               0,               0,               0,       5.8853226,      -28.629423,       33.555305 ],
  [               0,               0,               0,      -1.1251496,       40.052785,      -93.961402 ],
  [               0,               0,               0,    -0.031571629,     0.018396809,      0.27243055 ],
  [               0,               0,               0,     -0.16693082,        0.807424,     -0.97409456 ],
  [               0,               0,               0,       5.9156513,       -31.51006,       40.281309 ],
 ],
 [
  [               0,               0,               0,               0,   0.00086082668,    0.0094708902 ],
  [               0,               0,               0,               0,      -22.488074,       57.939699 ],
  [               0,               0,               0,               0,        4.869468,      -13.197237 ],
  [               0,               0,               0,               0,      0.75767286,      -1.8058677 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   1.6780277e-05,    0.0001217638,     0.013227108,    -0.033795483 ],
  [               0,               0,      -1.1810687,      -1.4012501,      -3.4731332,      -4.6103903 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      -1.9595978,       21.703084,        39.50175,       11.269155 ],
  [               0,               0,     -0.26032945,       12.507259,       32.404443,       9.2648864 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 500 GeV photon clusters
#


gam500_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   1.6953925e-05,    0.0010177373 ],
  [               0,               0,               0,               0,      -1.1537675,       4.1415009 ],
  [               0,               0,               0,               0,       1.4684268,      -1.5439816 ],
  [               0,               0,               0,               0,      -1.4487606,     -0.74425394 ],
  [               0,               0,               0,               0,     -0.36976777,       0.1304426 ],
  [               0,               0,               0,               0,   1.8148357e-05,  -0.00011676895 ],
  [               0,               0,               0,               0,      -3.9425127,       1.5997085 ],
  [               0,               0,               0,               0,      0.96541168,     -0.57258153 ],
 ],
 [
  [               0,               0,               0,   -0.0014388757,    0.0026542993,   0.00056193709 ],
  [               0,               0,               0,      -9.4247939,        9.275151,      0.93098605 ],
  [               0,               0,               0,      -5.6656027,       7.5379553,      -3.1496694 ],
  [               0,               0,               0,       5.0864158,      -5.5828632,      0.18607015 ],
  [               0,               0,               0,       1.5009849,      -2.2218713,      0.55104858 ],
  [               0,               0,               0,    0.0020351806,   -0.0033292152,   0.00091015214 ],
  [               0,               0,               0,       -7.893428,       8.2898054,       -1.318301 ],
  [               0,               0,               0,       4.0695295,      -5.4341715,     -0.31937406 ],
 ],
 [
  [               0,               0,               0,               0,     0.024698992,     0.011779898 ],
  [               0,               0,               0,               0,       14.522226,      -11.005646 ],
  [               0,               0,               0,               0,       -28.61648,       22.258477 ],
  [               0,               0,               0,               0,       14.727662,      -11.713467 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0038810554,    0.0099833361,   -0.0043264545 ],
  [               0,               0,               0,       3.8728071,      -9.0106446,       7.6689301 ],
  [               0,               0,               0,     -0.25139937,     0.028902441,     -0.71497157 ],
  [               0,               0,               0,        1.295008,      -2.9290835,      0.27074297 ],
  [               0,               0,               0,     -0.14295312,      0.42547431,      -0.4500142 ],
  [               0,               0,               0,   0.00092570227,   -0.0019106867,    0.0010056794 ],
  [               0,               0,               0,      -4.1298773,       7.8073263,      -1.1491572 ],
  [               0,               0,               0,       5.4614994,      -15.189321,       6.4598482 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.14469409,     -0.30283556,     -0.36605591,      0.73041319 ],
  [               0,               0,       572.33146,      -823.27814,      -1087.4356,       1569.2087 ],
  [               0,               0,      -688.91331,       1114.2542,       856.75215,      -1489.5857 ],
  [               0,               0,       6.4343392,       72.507755,      -472.92838,       504.94288 ],
  [               0,               0,        3064.819,      -4367.1835,      -6233.4535,       8880.3873 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,       -1.157848,       1.6960537 ],
  [               0,               0,               0,               0,       15.549008,      -20.503791 ],
  [               0,               0,               0,               0,       45.416913,      -66.506826 ],
  [               0,               0,               0,               0,      -47.294366,       68.192806 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.012670903,     0.019491901 ],
  [               0,               0,               0,               0,      -1350.6672,       2004.2638 ],
  [               0,               0,               0,               0,      -1209.8053,       1781.4958 ],
  [               0,               0,               0,               0,       958.61576,      -1390.4704 ],
  [               0,               0,               0,               0,      0.38564714,     -0.15207896 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.022473562,     0.046379836 ],
  [               0,               0,               0,               0,      -147.62195,       223.28455 ],
  [               0,               0,               0,               0,       47.504743,      -71.990938 ],
  [               0,               0,               0,               0,      -28.445656,       42.446051 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00086872114,  -0.00067676708,   0.00093519405,    0.0071828478 ],
  [               0,               0,       2.8649816,        0.288376,      -21.318414,       23.290889 ],
  [               0,               0,      -1.4020049,        3.893141,      -1.8688027,      -2.2948172 ],
  [               0,               0,     -0.76811541,       2.0550152,     -0.77660634,      -1.8763578 ],
  [               0,               0,     -0.21177014,        0.450926,      0.23033259,     -0.64156001 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0051996487,   -0.0077155741 ],
  [               0,               0,               0,               0,      -5.7310174,       13.749557 ],
  [               0,               0,               0,               0,      0.34666222,      -1.9161704 ],
  [               0,               0,               0,               0,    -0.032787705,      -1.0761637 ],
  [               0,               0,               0,               0,     -0.16586139,      0.32756075 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00024510832,  -0.00015452631,   0.00069114404,    0.0046557592 ],
  [               0,               0,       1.5673246,       -3.041187,       -7.259577,       16.978056 ],
  [               0,               0,      0.61651643,      -1.5278601,       -1.704139,        3.329211 ],
  [               0,               0,      -0.7238239,       1.9130884,       2.2058528,      -7.4960792 ],
  [               0,               0,     -0.34420186,      0.56175189,       2.3676964,      -4.2289674 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.58519275,       -1.339697 ],
  [               0,               0,               0,               0,      -198.19729,       457.44825 ],
  [               0,               0,               0,               0,       137.73491,      -318.05172 ],
  [               0,               0,               0,               0,      -29.207167,       67.475222 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0081088672,    -0.016734224 ],
  [               0,               0,               0,               0,      -11.002115,       29.552703 ],
  [               0,               0,               0,               0,     -0.65868422,      0.28512685 ],
  [               0,               0,               0,               0,      -0.6519445,      0.43474188 ],
  [               0,               0,               0,               0,     -0.23492596,      0.60739641 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.001245876,    0.0031423399 ],
  [               0,               0,               0,               0,        1.540173,      -2.3154495 ],
  [               0,               0,               0,               0,       -3.320472,       7.0502665 ],
  [               0,               0,               0,               0,       -44.94323,       106.34271 ],
  [               0,               0,               0,               0,     0.010877295,     -0.28064251 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0028336694,     0.041543599,    -0.075613863 ],
  [               0,               0,               0,      -37.118561,       129.42749,      -88.101945 ],
  [               0,               0,               0,       161.71661,      -790.74482,       964.91017 ],
  [               0,               0,               0,      -117.28198,       575.23239,      -705.16039 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.026295486 ],
  [               0,               0,               0,               0,               0,       4.8688857 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,      -0.9550842 ],
  [               0,               0,               0,               0,               0,      0.49613376 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -1.1126569e-05,    0.0010308955 ],
  [               0,               0,               0,               0,     -0.94116329,       4.0541555 ],
  [               0,               0,               0,               0,       1.2022104,      -1.4428639 ],
  [               0,               0,               0,               0,      -1.1680161,     -0.85792447 ],
  [               0,               0,               0,               0,     -0.34542246,     0.097266931 ],
  [               0,               0,               0,               0,   7.8079505e-06,  -9.0949112e-05 ],
  [               0,               0,               0,               0,      -3.9948426,       1.5738342 ],
  [               0,               0,               0,               0,       1.0813279,     -0.65582114 ],
 ],
 [
  [               0,               0,               0,     0.016299151,    -0.011265453,    0.0029896543 ],
  [               0,               0,               0,       12.029606,      -17.166932,       8.7428893 ],
  [               0,               0,               0,      -5.1587737,        4.164539,      -1.5498567 ],
  [               0,               0,               0,    -0.067095334,     -0.65638771,      -0.9904244 ],
  [               0,               0,               0,     -0.44129678,       1.1693089,      -0.7008877 ],
  [               0,               0,               0,    0.0025928367,   -0.0025944489,   0.00088595137 ],
  [               0,               0,               0,       -4.186585,       1.7649013,     -0.56775529 ],
  [               0,               0,               0,      -2.5772282,      0.75831265,      -0.6164869 ],
 ],
 [
  [               0,               0,               0,               0,     0.024698992,     0.011779898 ],
  [               0,               0,               0,               0,       14.522226,      -11.005646 ],
  [               0,               0,               0,               0,       -28.61648,       22.258477 ],
  [               0,               0,               0,               0,       14.727662,      -11.713467 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0043134389,     0.010708761,   -0.0044968249 ],
  [               0,               0,               0,       3.1806286,      -7.5807748,       6.8082917 ],
  [               0,               0,               0,       -0.377219,      0.45494111,       -1.017261 ],
  [               0,               0,               0,       1.4846877,      -3.1459112,      0.31663726 ],
  [               0,               0,               0,    -0.056860388,      0.22720168,     -0.32814922 ],
  [               0,               0,               0,   0.00082407222,   -0.0017082072,   0.00090420322 ],
  [               0,               0,               0,      -5.3985712,       10.299908,      -2.3086508 ],
  [               0,               0,               0,       2.9019598,      -9.0738605,       3.1649029 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.045821882,     -0.16198473,     -0.16317513,      0.44166952 ],
  [               0,               0,      -196.22413,       274.20948,       473.43462,      -659.47373 ],
  [               0,               0,      -601.80592,       986.37995,       688.85439,      -1242.7698 ],
  [               0,               0,      -41.485887,       136.89916,      -368.06537,       363.43518 ],
  [               0,               0,       3012.8998,      -4292.4058,      -6128.3586,       8729.0343 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,               0,    -0.043875759,     0.063298083 ],
  [               0,               0,               0,               0,       11.430651,       23.375143 ],
  [               0,               0,               0,               0,      -37.321995,       41.278833 ],
  [               0,               0,               0,               0,      -7.4942175,       23.587655 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.019241136,     0.029017826 ],
  [               0,               0,               0,               0,       42.498969,      -30.619719 ],
  [               0,               0,               0,               0,      -1411.1585,       2078.4948 ],
  [               0,               0,               0,               0,       1166.3223,      -1692.7558 ],
  [               0,               0,               0,               0,       0.4330443,      -0.2079186 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.024900723,     0.051988145 ],
  [               0,               0,               0,               0,        -113.049,       171.68084 ],
  [               0,               0,               0,               0,       54.072405,       -81.76228 ],
  [               0,               0,               0,               0,      -38.977424,       58.094597 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00086812953,  -0.00067673924,   0.00093405089,    0.0071804234 ],
  [               0,               0,       2.8666252,       0.2897444,      -21.335067,       23.307639 ],
  [               0,               0,      -1.4022552,       3.8947561,      -1.8717657,      -2.2932319 ],
  [               0,               0,     -0.76811668,       2.0564251,     -0.78079577,      -1.8732659 ],
  [               0,               0,     -0.21183663,      0.45100208,      0.23063331,      -0.6419568 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0051942054,   -0.0077024737 ],
  [               0,               0,               0,               0,      -5.7219703,       13.730863 ],
  [               0,               0,               0,               0,      0.35637418,      -1.9356821 ],
  [               0,               0,               0,               0,    -0.028716067,      -1.0843076 ],
  [               0,               0,               0,               0,     -0.16600615,      0.32787562 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00026258951,  -0.00016652249,   0.00071564506,    0.0048173133 ],
  [               0,               0,       1.5667945,      -2.9851532,      -7.2628003,       16.745436 ],
  [               0,               0,      0.64991254,      -1.5903962,      -1.9122891,        3.732264 ],
  [               0,               0,     -0.66509084,       1.7954943,       1.9237229,      -6.9296453 ],
  [               0,               0,     -0.34845736,      0.57658413,       2.3690124,      -4.2575384 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.58519275,       -1.339697 ],
  [               0,               0,               0,               0,      -198.19729,       457.44825 ],
  [               0,               0,               0,               0,       137.73491,      -318.05172 ],
  [               0,               0,               0,               0,      -29.207167,       67.475222 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0064844601,      -0.0128103 ],
  [               0,               0,               0,               0,       -7.547965,        21.28567 ],
  [               0,               0,               0,               0,     -0.54657215,     0.013953184 ],
  [               0,               0,               0,               0,     -0.20139527,     -0.63164461 ],
  [               0,               0,               0,               0,     -0.15171061,      0.41169351 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.001245876,    0.0031423399 ],
  [               0,               0,               0,               0,        1.540173,      -2.3154495 ],
  [               0,               0,               0,               0,       -3.320472,       7.0502665 ],
  [               0,               0,               0,               0,       -44.94323,       106.34271 ],
  [               0,               0,               0,               0,     0.010877295,     -0.28064251 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0060474549,     0.061040388,     -0.10400544 ],
  [               0,               0,               0,      -54.127693,       209.93703,      -183.28453 ],
  [               0,               0,               0,       154.48433,      -755.30338,       921.51343 ],
  [               0,               0,               0,      -124.13147,       608.61283,      -745.82869 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.025864161 ],
  [               0,               0,               0,               0,               0,       4.9339793 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.95280072 ],
  [               0,               0,               0,               0,               0,      0.49948955 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00018740731,   0.00094683723 ],
  [               0,               0,               0,               0,     -0.37779798,       4.2123433 ],
  [               0,               0,               0,               0,      -16.990313,       2.3225922 ],
  [               0,               0,               0,               0,      -17.924916,       2.3448469 ],
  [               0,               0,               0,               0,    -0.042573923,     -0.07311668 ],
  [               0,               0,               0,               0,   0.00042677659,   -0.0041896886 ],
  [               0,               0,               0,               0,      -1.2741591,      0.25462454 ],
  [               0,               0,               0,               0,     0.074833576,      -1.5887666 ],
 ],
 [
  [               0,               0,               0,   -0.0015630823,    0.0017326567,   0.00026559035 ],
  [               0,               0,               0,      -5.7626226,       4.4207625,       3.5938264 ],
  [               0,               0,               0,      -3.9363072,       4.5069412,      -1.7102657 ],
  [               0,               0,               0,     -0.47820642,     -0.15355683,      -1.5999513 ],
  [               0,               0,               0,      -1.2185053,      0.95174742,     -0.55669377 ],
  [               0,               0,               0,    0.0021664778,   -0.0021960592,   0.00070265001 ],
  [               0,               0,               0,      -4.6570382,        5.133886,     -0.25133657 ],
  [               0,               0,               0,      -2.7103718,       1.7843574,     -0.97405922 ],
 ],
 [
  [               0,               0,               0,               0,     0.023879858,      0.01144006 ],
  [               0,               0,               0,               0,       14.719224,      -11.155549 ],
  [               0,               0,               0,               0,      -29.305489,       22.801171 ],
  [               0,               0,               0,               0,       15.230969,      -12.113774 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0011282589,      0.00643445,   -0.0029716372 ],
  [               0,               0,               0,      -3.2846753,       5.2963882,     0.037304583 ],
  [               0,               0,               0,       2.6541361,      -5.6709781,       2.0759184 ],
  [               0,               0,               0,       4.1998753,       -8.613071,       3.1751535 ],
  [               0,               0,               0,  -0.00052074456,      0.15771121,     -0.28173795 ],
  [               0,               0,               0,     0.001169391,   -0.0024782923,    0.0013416232 ],
  [               0,               0,               0,      -3.8649365,       7.4973519,      -1.2591028 ],
  [               0,               0,               0,       2.6565772,      -6.1388216,       1.1092249 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.051091706,     -0.16889647,     -0.17356914,      0.45521358 ],
  [               0,               0,      -160.63496,       223.25946,       401.56814,      -556.59138 ],
  [               0,               0,      -601.92999,       986.79417,       687.29094,       -1241.016 ],
  [               0,               0,      -35.552521,       128.66912,      -380.99899,       381.40083 ],
  [               0,               0,       3027.1201,        -4312.67,      -6157.2371,       8770.1786 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,               0,    -0.027764317,     0.040053238 ],
  [               0,               0,               0,               0,      -18.884255,      -12.344659 ],
  [               0,               0,               0,               0,      -439.10908,       616.33849 ],
  [               0,               0,               0,               0,       293.21058,      -402.02534 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.016339903,     0.024735676 ],
  [               0,               0,               0,               0,       48.245863,      -35.818829 ],
  [               0,               0,               0,               0,      -1510.0576,       2224.0869 ],
  [               0,               0,               0,               0,       1216.5928,      -1765.3252 ],
  [               0,               0,               0,               0,      0.46177451,     -0.25050697 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.32272224,      0.49258667 ],
  [               0,               0,               0,               0,      -60.570338,       93.892677 ],
  [               0,               0,               0,               0,       37.006627,      -56.403475 ],
  [               0,               0,               0,               0,      -25.734161,       38.421203 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00085491756,  -0.00069586769,   0.00089079651,    0.0072071939 ],
  [               0,               0,       3.0073596,      0.14232866,      -21.854041,       23.951357 ],
  [               0,               0,      -1.3460749,       3.7861491,        -1.98719,      -2.0654117 ],
  [               0,               0,     -0.81488259,       2.1849198,     -0.84627083,      -1.9080501 ],
  [               0,               0,     -0.23453265,      0.49017744,      0.28013762,      -0.7280921 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0056955478,   -0.0084761333 ],
  [               0,               0,               0,               0,      -5.1115702,       12.445395 ],
  [               0,               0,               0,               0,      0.33669435,      -1.9072085 ],
  [               0,               0,               0,               0,    -0.048162697,      -1.0586712 ],
  [               0,               0,               0,               0,     -0.16773776,      0.32967063 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00020494226,  -0.00012616912,   0.00061946238,    0.0042009079 ],
  [               0,               0,       1.3273257,      -2.6149539,      -6.1526285,       15.069379 ],
  [               0,               0,      0.66624731,      -1.6292759,      -1.9509125,       3.8330402 ],
  [               0,               0,     -0.75887045,       1.9971228,       2.3539254,       -7.851437 ],
  [               0,               0,     -0.36457074,      0.60374571,       2.4740726,      -4.4491008 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,    0.0059481936 ],
  [               0,               0,               0,               0,               0,       1.5798893 ],
  [               0,               0,               0,               0,               0,      -1.2564032 ],
  [               0,               0,               0,               0,               0,        0.278231 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00042245579,    0.0015604661 ],
  [               0,               0,               0,               0,     -0.68543889,       4.9983214 ],
  [               0,               0,               0,               0,     -0.35026244,     -0.44956852 ],
  [               0,               0,               0,               0,      0.45761836,      -2.1953722 ],
  [               0,               0,               0,               0,     0.020989972,    0.0019052831 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0018011526,    0.0046924829 ],
  [               0,               0,               0,               0,       1.2466675,      -1.8184313 ],
  [               0,               0,               0,               0,     -0.12289478,     -0.56429045 ],
  [               0,               0,               0,               0,      -31.190481,       73.581478 ],
  [               0,               0,               0,               0,     0.027018734,     -0.23867332 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   -0.0067625222,     0.063581606,     -0.10590315 ],
  [               0,               0,               0,       -57.84565,        231.2246,      -213.16035 ],
  [               0,               0,               0,        137.6125,      -673.52969,       822.45708 ],
  [               0,               0,               0,      -130.02464,        637.5308,      -781.30288 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.026191309 ],
  [               0,               0,               0,               0,               0,       5.0531751 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.93769187 ],
  [               0,               0,               0,               0,               0,      0.51885379 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,    -0.024143519,     0.012650608,    0.0064298632 ],
  [               0,               0,               0,       3.5736365,      -2.5772102,       4.6173405 ],
  [               0,               0,               0,       1.9823856,     -0.68365064,      -1.2618922 ],
  [               0,               0,               0,     0.075008504,     -0.09617026,    -0.027369543 ],
  [               0,               0,               0,    0.0041282773,   -0.0021515856,   0.00053469224 ],
  [               0,               0,               0,        29.17942,      -12.443723,       1.5585294 ],
  [               0,               0,               0,      -27.436811,       13.198635,      -1.7885643 ],
 ],
 [
  [               0,               0,               0,   -0.0043676707,     0.010319551,    0.0027765587 ],
  [               0,               0,               0,       5.3550201,      -9.1208692,        7.765457 ],
  [               0,               0,               0,      -1.8150273,       2.4007798,      -2.1040369 ],
  [               0,               0,               0,       1.0356184,      -1.2692805,      0.37113372 ],
  [               0,               0,               0,   0.00086189852,   -0.0020767738,    0.0013051083 ],
  [               0,               0,               0,      -7.0748973,       7.1905446,     -0.85998562 ],
  [               0,               0,               0,       2.2769328,      0.44443311,      -2.5541967 ],
 ],
 [
  [               0,               0,               0,               0,     0.010412503,    -0.003879848 ],
  [               0,               0,               0,               0,       3.5713159,       2.3246131 ],
  [               0,               0,               0,               0,       12.661752,       -11.48641 ],
  [               0,               0,               0,               0,      -9.2646965,       7.3576495 ],
  [               0,               0,               0,               0,   0.00092308359,   0.00075610779 ],
  [               0,               0,               0,               0,      0.72313002,       0.4503621 ],
  [               0,               0,               0,               0,       4.8703001,      -5.5650149 ],
 ],
 [
  [               0,               0,               0,     0.012214125,    -0.025232173,     0.022084822 ],
  [               0,               0,               0,      -1.9660461,       3.2954248,       2.0879699 ],
  [               0,               0,               0,     0.044881618,    -0.058575781,      -1.2890437 ],
  [               0,               0,               0,    -0.042312294,      0.21168822,     -0.15154213 ],
  [               0,               0,               0,    0.0029288515,   -0.0064666132,     0.003624935 ],
  [               0,               0,               0,       6.5306817,      -14.031101,       5.9147531 ],
  [               0,               0,               0,      -12.212485,       28.835549,      -17.772106 ],
 ],
 [
  [               0,               0,   -0.0042523779,   -0.0022325622,    0.0041230329,     0.019138678 ],
  [               0,               0,      -19.155506,       16.589822,       28.496092,      -17.732756 ],
  [               0,               0,      -55.148036,       80.037244,       106.86154,       -156.1804 ],
  [               0,               0,       -12.02035,       12.384202,       19.838335,      -19.403211 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.33392479,     -0.44892993 ],
  [               0,               0,               0,               0,       12.143538,        22.54858 ],
  [               0,               0,               0,               0,      -30.213443,       40.048425 ],
  [               0,               0,               0,               0,      0.31775683,      0.22608558 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_55_e = [
 [
  [     0.097825768,   -0.0060088107,     -0.13745933,     -0.26746748,     -0.32298612,      0.85109871 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    -0.015120671,   -0.0061115535,     0.014256186,     0.054908934 ],
  [               0,               0,      -13.746044,        26.75678,       35.281687,       -64.71787 ],
  [               0,               0,       22.899682,      -37.218811,      -51.591511,       82.722397 ],
  [               0,               0,       6.9794439,      -9.9128894,      -14.743931,       20.906228 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0010808836,    0.0070275045 ],
  [               0,               0,               0,               0,       32.479693,      -46.387078 ],
  [               0,               0,               0,               0,      -1.5895424,       1.2036631 ],
  [               0,               0,               0,               0,       -1.671856,       2.5636355 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,  -0.00044732527,    0.0043104294,    0.0030579219 ],
  [               0,               0,               0,       5.1649243,      -18.917075,       20.312244 ],
  [               0,               0,               0,       1.0746096,      -3.6627641,       1.8049466 ],
  [               0,               0,               0,     -0.36471348,      -7.2742441,       15.312764 ],
  [               0,               0,               0,     0.023529424,   -0.0069497264,    -0.073063382 ],
  [               0,               0,               0,       1.2778247,      -4.5578572,       3.9701239 ],
  [               0,               0,               0,      -4.4459314,       14.895435,      -31.365855 ],
 ],
 [
  [               0,               0,               0,    0.0002343547,     0.001627832,    0.0054563884 ],
  [               0,               0,               0,       1.4418488,      -7.8690334,       12.737132 ],
  [               0,               0,               0,      0.45459573,      -1.6092089,      0.13166216 ],
  [               0,               0,               0,     -0.70838964,      -1.6271923,       7.6583285 ],
  [               0,               0,               0,    -0.010343369,    0.0020107712,     0.052103217 ],
  [               0,               0,               0,     -0.46276476,       1.9721092,      -2.0327216 ],
  [               0,               0,               0,       5.9262366,      -23.765912,       20.767997 ],
 ],
 [
  [               0,               0,               0,   -0.0040873304,    -0.006679513,     0.050418098 ],
  [               0,               0,               0,       23.426981,      -109.13335,       129.23417 ],
  [               0,               0,               0,       1.8884304,      -9.4165698,       10.545803 ],
  [               0,               0,               0,      -7.8543509,       43.210847,      -57.855085 ],
  [               0,               0,               0,     0.016878936,   -0.0069019097,    -0.073583102 ],
  [               0,               0,               0,       0.4972046,      -3.0106473,       4.3986198 ],
  [               0,               0,               0,      -1.0925298,       11.100742,      -22.994306 ],
 ],
 [
  [               0,               0,               0,               0,    -0.077473503,      0.19530374 ],
  [               0,               0,               0,               0,      -37.099462,       94.193982 ],
  [               0,               0,               0,               0,      -122.86185,       300.91704 ],
  [               0,               0,               0,               0,       133.26597,      -327.56524 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0012674704,  -0.00052477599,   -0.0034187873,   -0.0076250096 ],
  [               0,               0,      -1.2132805,      -1.3735761,       -3.405374,      -4.4354317 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       6.9584998,        8.708218,        21.84027,       26.084413 ],
  [               0,               0,       1.7927082,       4.0849073,       33.136005,       65.296966 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.022095674,     0.012359088,    0.0059826813 ],
  [               0,               0,               0,       8.0416971,      -4.8601182,       5.7156135 ],
  [               0,               0,               0,      -1.3753409,      0.78763862,      -1.4532959 ],
  [               0,               0,               0,     -0.13558964,     -0.29713095,     0.031697366 ],
  [               0,               0,               0,      0.01940576,   -0.0066812156,   0.00086150134 ],
  [               0,               0,               0,       8.3482455,      -6.2123043,       1.3058235 ],
  [               0,               0,               0,      -17.799708,       10.251901,      -2.2736109 ],
 ],
 [
  [               0,               0,               0,    -0.013446947,     0.017010238,    0.0023907641 ],
  [               0,               0,               0,       5.9889571,      -8.6849261,       7.8906045 ],
  [               0,               0,               0,      0.15334353,     -0.37448377,      -1.1714592 ],
  [               0,               0,               0,      0.57941489,     -0.37161137,    -0.042903383 ],
  [               0,               0,               0,   -0.0017832985,    0.0014854398,   0.00033888163 ],
  [               0,               0,               0,       6.4995838,      -7.7108552,       2.5567767 ],
  [               0,               0,               0,      -5.0521077,       3.2693898,      -1.1139796 ],
 ],
 [
  [               0,               0,               0,               0,      0.02987012,    -0.015637574 ],
  [               0,               0,               0,               0,      -5.6289475,       8.9412871 ],
  [               0,               0,               0,               0,       7.4077954,      -7.2760233 ],
  [               0,               0,               0,               0,      -5.0411373,       4.4512459 ],
  [               0,               0,               0,               0,    0.0045955791,   3.8247013e-05 ],
  [               0,               0,               0,               0,      -5.6302213,       4.6196861 ],
  [               0,               0,               0,               0,      -3.1772469,       -0.382713 ],
 ],
 [
  [               0,               0,               0,  -0.00038671698,    0.0061339158,    0.0010749698 ],
  [               0,               0,               0,      0.71757855,        -4.21874,       8.1518045 ],
  [               0,               0,               0,      0.10207071,      -0.3802362,      -1.0937175 ],
  [               0,               0,               0,      0.12079028,     0.016081404,     -0.12578087 ],
  [               0,               0,               0,   -0.0037728059,    0.0092323576,   -0.0060815089 ],
  [               0,               0,               0,      -1.1157088,       1.5303912,      0.44945481 ],
  [               0,               0,               0,      -6.1201728,       11.021793,      -9.5990571 ],
 ],
 [
  [               0,               0,      0.19380696,    -0.054861851,     -0.21916294,    -0.080887901 ],
  [               0,               0,       264.01124,      -336.28317,      -479.60019,       614.03126 ],
  [               0,               0,      -510.15667,       720.20614,       946.09203,      -1337.8401 ],
  [               0,               0,       673.89868,      -918.86242,      -1268.5109,       1729.6993 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.012330575,    0.0032569127 ],
  [               0,               0,               0,               0,       3.8313499,        9.575625 ],
  [               0,               0,               0,               0,     -0.81033899,     -0.23690818 ],
  [               0,               0,               0,               0,     0.031159937,     -0.25534901 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_35_e = [
 [
  [       88.825478,      -178.98268,       -143.0272,       209.92001,       513.18053,      -527.58346 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.042015213,    -0.032888963,    -0.068541243,     0.047602828 ],
  [               0,               0,        750.8712,      -1107.0217,      -1646.1061,       2429.0981 ],
  [               0,               0,       3.1274032,        2.416718,      -29.249078,       26.628419 ],
  [               0,               0,      -64.824603,       95.571702,       137.45203,       -202.5165 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0018107125,    0.0033345279 ],
  [               0,               0,               0,               0,       56.327098,      -81.534795 ],
  [               0,               0,               0,               0,      -42.454665,       62.904524 ],
  [               0,               0,               0,               0,       40.139941,      -60.630661 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0012289324,   0.00068887528,    0.0030721564 ],
  [               0,               0,               0,       7.6373215,      -28.379679,       30.103319 ],
  [               0,               0,               0,       2.9857741,      -10.308142,       7.4896672 ],
  [               0,               0,               0,       -1.352402,        4.915687,      -4.3574942 ],
  [               0,               0,               0,    0.0048806571,    -0.016016683,      0.01371318 ],
  [               0,               0,               0,       -9.496377,       30.826278,       -24.18549 ],
  [               0,               0,               0,       16.776364,      -60.196852,       51.632034 ],
 ],
 [
  [               0,               0,               0,    0.0057226898,    -0.021308731,     0.028678094 ],
  [               0,               0,               0,    -0.042788014,      -1.9960497,       7.5654476 ],
  [               0,               0,               0,      0.71193367,      -2.6753044,       1.1573703 ],
  [               0,               0,               0,      -10.568593,       44.294578,      -45.768678 ],
  [               0,               0,               0,    -0.084953875,      0.35888585,     -0.37334706 ],
  [               0,               0,               0,     0.097345307,     -0.65053198,       1.0619998 ],
  [               0,               0,               0,       8.1930557,      -33.572622,       31.468942 ],
 ],
 [
  [               0,               0,               0,   -0.0013491114,   -0.0034827753,     0.026080895 ],
  [               0,               0,               0,       25.626343,      -120.87043,       145.28748 ],
  [               0,               0,               0,       4.1707807,      -20.101842,       22.969006 ],
  [               0,               0,               0,      -15.209499,       83.497553,      -108.14419 ],
  [               0,               0,               0,    0.0031183337,    0.0026423099,     0.022959695 ],
  [               0,               0,               0,       1.9377152,      -8.9726675,       10.392366 ],
  [               0,               0,               0,      -4.1233813,        15.87488,      -17.249288 ],
 ],
 [
  [               0,               0,               0,               0,    -0.069059296,       0.1771257 ],
  [               0,               0,               0,               0,       2.9343565,      -4.1713549 ],
  [               0,               0,               0,               0,      -46.021708,       112.05544 ],
  [               0,               0,               0,               0,       50.204622,      -123.53221 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.050599391,    -0.073146453,     -0.10182288,    -0.067874151 ],
  [               0,               0,       -1.189447,      -1.4010448,      -3.4569853,      -4.4230106 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      0.19720829,      0.38456664,      0.95323542,       1.0746222 ],
  [               0,               0,    0.0058796346,      0.10152893,      0.49737395,       1.1139203 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.019312292,     0.011333769,    0.0060336085 ],
  [               0,               0,               0,       6.0300337,      -4.0962738,       5.6712077 ],
  [               0,               0,               0,      -1.2732743,      0.74554563,      -1.4477555 ],
  [               0,               0,               0,     -0.22895196,     -0.30855163,     0.031231777 ],
  [               0,               0,               0,     0.019372954,   -0.0063399255,   0.00084941212 ],
  [               0,               0,               0,       8.0251414,      -6.0917703,        1.299487 ],
  [               0,               0,               0,      -18.149201,       10.563988,      -2.2715665 ],
 ],
 [
  [               0,               0,               0,   -0.0096383127,     0.012459062,    0.0037711223 ],
  [               0,               0,               0,       4.0391002,      -6.2992046,       7.1618918 ],
  [               0,               0,               0,      0.33504649,     -0.57120286,      -1.1220363 ],
  [               0,               0,               0,      0.92302888,      0.23940281,     -0.85868778 ],
  [               0,               0,               0,   -0.0089807613,    0.0035109861,    0.0043842719 ],
  [               0,               0,               0,       1.7151697,      -2.1331493,      0.74412697 ],
  [               0,               0,               0,      -2.5882472,       1.1107017,     -0.22609026 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0050994654,    0.0089366792 ],
  [               0,               0,               0,               0,       9.5458476,      -1.5972657 ],
  [               0,               0,               0,               0,       20.022623,       -17.48208 ],
  [               0,               0,               0,               0,      -12.354386,       9.8914341 ],
  [               0,               0,               0,               0,    0.0024257227,  -0.00041528586 ],
  [               0,               0,               0,               0,       1.0110978,      0.20849157 ],
  [               0,               0,               0,               0,       14.180527,      -13.077819 ],
 ],
 [
  [               0,               0,               0,    0.0011292063,   -0.0013066157,    0.0089085465 ],
  [               0,               0,               0,      -1.0482635,       1.1089868,       4.1604982 ],
  [               0,               0,               0,    -0.049340409,     0.081058092,      -1.3766263 ],
  [               0,               0,               0,     0.038499546,     0.034618091,    -0.064559137 ],
  [               0,               0,               0,  -0.00053532645,    0.0010562682,  -0.00031465862 ],
  [               0,               0,               0,       2.1805912,      -4.7516401,       4.3343232 ],
  [               0,               0,               0,      -1.1432213,       1.7710243,      -5.3738794 ],
 ],
 [
  [               0,               0,      0.19818875,    -0.052644792,     -0.22161077,    -0.092217622 ],
  [               0,               0,       394.13109,      -514.80641,      -722.66786,       947.35874 ],
  [               0,               0,      -440.72546,       625.05985,       814.56766,      -1157.5791 ],
  [               0,               0,       711.61217,      -970.75476,      -1339.6184,        1827.524 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.50647169,     -0.70951705 ],
  [               0,               0,               0,               0,      -909.02604,       1319.4129 ],
  [               0,               0,               0,               0,         26.1506,       15.866556 ],
  [               0,               0,               0,               0,       1502.6702,      -2160.3644 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam500_layer2_etaoffv4_1_37_e = [
 [
  [       81.691307,      -164.68206,      -131.58801,       193.28567,        472.4921,      -485.97351 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.049276184,    0.0059028506,     -0.04713743,    -0.088984606 ],
  [               0,               0,        797.3402,      -1177.4004,      -1721.6608,       2544.2304 ],
  [               0,               0,       429.40708,      -628.75662,      -952.33946,       1393.2775 ],
  [               0,               0,      -450.16386,       663.62295,       976.83735,      -1439.7482 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00097855573,    0.0072266254 ],
  [               0,               0,               0,               0,       64.994376,      -94.570834 ],
  [               0,               0,               0,               0,      -53.129304,       79.047082 ],
  [               0,               0,               0,               0,       50.354469,      -76.077614 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0013297788,   0.00035031956,     0.003383104 ],
  [               0,               0,               0,       7.4795673,      -27.820068,       29.602702 ],
  [               0,               0,               0,       2.8805647,       -9.950604,       7.1872786 ],
  [               0,               0,               0,      -1.3520625,       4.9117551,      -4.3518831 ],
  [               0,               0,               0,    0.0045513325,    -0.014906804,     0.012791053 ],
  [               0,               0,               0,        -9.07581,       29.437797,      -23.047929 ],
  [               0,               0,               0,       16.431731,       -58.95491,       50.523752 ],
 ],
 [
  [               0,               0,               0,    0.0072981678,    -0.027594549,     0.034958531 ],
  [               0,               0,               0,     -0.24183035,      -1.2020298,       6.7779342 ],
  [               0,               0,               0,       0.8040169,      -3.0549513,       1.5466482 ],
  [               0,               0,               0,      -8.0897022,       33.820858,      -34.865907 ],
  [               0,               0,               0,    -0.063266605,      0.26694901,     -0.27736643 ],
  [               0,               0,               0,      0.11424182,     -0.74625752,       1.2116356 ],
  [               0,               0,               0,       9.2443968,      -37.887768,       35.943623 ],
 ],
 [
  [               0,               0,               0,   -0.0010036312,   -0.0033094151,     0.023774694 ],
  [               0,               0,               0,       21.506345,      -101.61266,       122.79822 ],
  [               0,               0,               0,       4.4358393,      -21.329278,       24.387149 ],
  [               0,               0,               0,      -10.896467,       66.200265,      -87.661041 ],
  [               0,               0,               0,    0.0049409297,    0.0089855687,     0.035961272 ],
  [               0,               0,               0,       1.5607212,      -7.2395869,       8.3999904 ],
  [               0,               0,               0,      -3.8508867,       15.676527,       -18.40923 ],
 ],
 [
  [               0,               0,               0,               0,    -0.069219148,      0.17752519 ],
  [               0,               0,               0,               0,     -0.54356966,       4.3974937 ],
  [               0,               0,               0,               0,      -46.691693,       113.70444 ],
  [               0,               0,               0,               0,       51.374829,       -126.4101 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,     0.001616661,   -0.0025409131,   -0.0054054086,     0.004365657 ],
  [               0,               0,      -1.2217127,      -1.3557897,      -3.3658817,      -4.3565805 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       23.078721,       51.610458,      -8.8212472,      -293.30477 ],
  [               0,               0,       3.8433111,       21.984564,       34.220675,       21.826392 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Tables for v4_1 sshape correction, 1000 GeV photon clusters
#


gam1000_layer1_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00018060473,   0.00077601766 ],
  [               0,               0,               0,               0,      -2.9657719,       5.4532545 ],
  [               0,               0,               0,               0,      0.34138231,      -1.1290424 ],
  [               0,               0,               0,               0,      0.33399678,      -1.2405273 ],
  [               0,               0,               0,               0,     -0.18487028,    -0.025166963 ],
  [               0,               0,               0,               0,   0.00020508829,  -2.7851627e-05 ],
  [               0,               0,               0,               0,       8.0173184,     -0.65559793 ],
  [               0,               0,               0,               0,      -1.0399634,       -2.039384 ],
 ],
 [
  [               0,               0,               0,   -0.0016310582,    0.0036196132,  -0.00047206192 ],
  [               0,               0,               0,       8.2517569,      -15.235554,       9.9373298 ],
  [               0,               0,               0,       1.4437127,       -1.981494,     -0.30819303 ],
  [               0,               0,               0,       4.0137897,       -5.233843,      0.29472319 ],
  [               0,               0,               0,      -0.3786088,      0.36877175,     -0.25537444 ],
  [               0,               0,               0,   0.00012751194,  -0.00025795646,   0.00011091603 ],
  [               0,               0,               0,      -22.378351,       10.575983,      -7.5550836 ],
  [               0,               0,               0,      -1.8106632,      -1.5840892,     -0.78671141 ],
 ],
 [
  [               0,               0,               0,               0,     0.023336578,    0.0077932666 ],
  [               0,               0,               0,               0,       34.850335,      -27.171084 ],
  [               0,               0,               0,               0,      -46.189222,       36.235168 ],
  [               0,               0,               0,               0,       16.183702,      -12.872961 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0011029376,   0.00089328119,  -0.00027412184 ],
  [               0,               0,               0,       2.0161407,      -6.4740272,       6.9992076 ],
  [               0,               0,               0,      0.63269271,      -1.6186813,     0.062207653 ],
  [               0,               0,               0,      0.68370147,      -1.5176577,      -0.4924779 ],
  [               0,               0,               0,     -0.18531799,      0.55303209,     -0.53635838 ],
  [               0,               0,               0,   0.00027524356,  -0.00067559251,   0.00044174086 ],
  [               0,               0,               0,      0.31193266,     -0.80302571,       2.8916319 ],
  [               0,               0,               0,     -0.62903468,     -0.57923769,      -1.5271086 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,      0.25299607,     -0.51077436,     -0.62949708,       1.2150468 ],
  [               0,               0,      -1124.5674,       1599.8301,       2386.2273,      -3389.5481 ],
  [               0,               0,      -818.23316,       1357.5587,       981.39522,      -1786.5379 ],
  [               0,               0,      -390.23425,        815.4939,      -280.75927,      -130.12193 ],
  [               0,               0,       3922.0466,      -5590.2915,      -7981.6788,       11374.281 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer1_etaoffv4_1_55_e = [
 [
  [               0,               0,               0,               0,     0.017183679,    -0.010158961 ],
  [               0,               0,               0,               0,       21.828357,      -29.329642 ],
  [               0,               0,               0,               0,       39.025246,      -57.519195 ],
  [               0,               0,               0,               0,      -48.271872,       69.805546 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.082938648,      0.12545759 ],
  [               0,               0,               0,               0,      -59.769647,       93.387892 ],
  [               0,               0,               0,               0,      -167.20265,       245.18067 ],
  [               0,               0,               0,               0,       25.956487,      -38.445654 ],
  [               0,               0,               0,               0,      0.16696515,     -0.15245848 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.17163977,      0.26159882 ],
  [               0,               0,               0,               0,      -267.35435,        403.9756 ],
  [               0,               0,               0,               0,      -24.404183,       34.792097 ],
  [               0,               0,               0,               0,       48.826264,       -72.32893 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0006760918,  -0.00040536965,   0.00087403586,    0.0052379489 ],
  [               0,               0,       3.8427514,      -4.3418836,      -13.912615,       19.504188 ],
  [               0,               0,     -0.43081891,       1.0823683,     -0.54243787,      -1.2596383 ],
  [               0,               0,      -1.1388053,       2.1948378,       1.7835062,       -4.845446 ],
  [               0,               0,     -0.39888288,       0.6226135,       1.1734116,       -1.816908 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     0.004916339,   -0.0072560818 ],
  [               0,               0,               0,               0,      -5.0070218,       12.593867 ],
  [               0,               0,               0,               0,      0.12954287,      -1.5551962 ],
  [               0,               0,               0,               0,     -0.80190464,      0.38162159 ],
  [               0,               0,               0,               0,     -0.31874204,      0.62559697 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00019417364,  -6.8828292e-05,   0.00068541369,    0.0039271717 ],
  [               0,               0,      0.46084378,     -0.75463219,      -2.0731667,       6.3257611 ],
  [               0,               0,      0.48513415,      -1.1166121,      -1.8901106,       3.1036333 ],
  [               0,               0,     -0.34758878,      0.86346141,       1.3187079,      -4.5184633 ],
  [               0,               0,      -0.2965714,      0.57877596,       1.6549317,      -3.2512192 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0031635677,     0.010405226 ],
  [               0,               0,               0,               0,      -331.27934,       764.32348 ],
  [               0,               0,               0,               0,       110.54663,      -255.79766 ],
  [               0,               0,               0,               0,       4.3463564,      -9.4201814 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00024668737,    0.0035142771 ],
  [               0,               0,               0,               0,       0.5542884,       1.9745823 ],
  [               0,               0,               0,               0,     -0.28667314,     -0.65429077 ],
  [               0,               0,               0,               0,      0.23691952,      -1.7115751 ],
  [               0,               0,               0,               0,      0.11315057,     -0.21530497 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.14965539,      0.35945028 ],
  [               0,               0,               0,               0,       -93.53903,       229.17943 ],
  [               0,               0,               0,               0,      -108.29107,       257.47734 ],
  [               0,               0,               0,               0,       429.35158,      -1030.8791 ],
  [               0,               0,               0,               0,     -0.22010118,     -0.84284725 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.10596963,      0.26395847 ],
  [               0,               0,               0,               0,       145.62998,      -343.80074 ],
  [               0,               0,               0,               0,      -34.928148,       82.571421 ],
  [               0,               0,               0,               0,       19.147362,      -46.127572 ],
  [               0,               0,               0,               0,     0.059299206,     -0.14239539 ],
  [               0,               0,               0,               0,      -2.6117805,       7.5288191 ],
  [               0,               0,               0,               0,       13.601171,      -35.442525 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.026531123 ],
  [               0,               0,               0,               0,               0,       10.063021 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.67038534 ],
  [               0,               0,               0,               0,               0,      0.85227305 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer1_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   0.00015556271,     0.001279524 ],
  [               0,               0,               0,               0,      -1.5826685,       4.3140753 ],
  [               0,               0,               0,               0,       1.0338676,      -1.2660621 ],
  [               0,               0,               0,               0,     -0.37215105,      -1.3925945 ],
  [               0,               0,               0,               0,     -0.42463463,    -0.018916647 ],
  [               0,               0,               0,               0,   6.0248815e-05,   0.00013695528 ],
  [               0,               0,               0,               0,     -0.84471257,       1.4507723 ],
  [               0,               0,               0,               0,      -1.5817169,       1.4428617 ],
 ],
 [
  [               0,               0,               0,               0,   0.00012842991,   5.5803964e-05 ],
  [               0,               0,               0,               0,      -1.1684575,       7.6898259 ],
  [               0,               0,               0,               0,      -6.9157237,       9.1283494 ],
  [               0,               0,               0,               0,      -12.406793,       11.753438 ],
  [               0,               0,               0,               0,      -1.9017471,     0.095329391 ],
  [               0,               0,               0,               0,   9.1666229e-06,   0.00022051543 ],
  [               0,               0,               0,               0,     -0.99094938,       2.4300281 ],
  [               0,               0,               0,               0,     0.061941487,      -1.8241852 ],
 ],
 [
  [               0,               0,               0,               0,     0.023336578,    0.0077932666 ],
  [               0,               0,               0,               0,       34.850335,      -27.171084 ],
  [               0,               0,               0,               0,      -46.189222,       36.235168 ],
  [               0,               0,               0,               0,       16.183702,      -12.872961 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00097592526,   0.00090454839,  -0.00017832623 ],
  [               0,               0,               0,       2.1984333,      -6.9810443,       7.3642854 ],
  [               0,               0,               0,      0.34472673,     -0.88971181,     -0.37266177 ],
  [               0,               0,               0,      0.59364651,      -1.2638627,     -0.67474965 ],
  [               0,               0,               0,     -0.14260688,      0.44412752,     -0.47474618 ],
  [               0,               0,               0,   0.00024066628,  -0.00059663559,   0.00039277476 ],
  [               0,               0,               0,      0.28869401,     -0.48688635,       2.6994204 ],
  [               0,               0,               0,     -0.32507505,     -0.81087489,      -1.6315601 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.042010089,     -0.10293254,     -0.19240973,    -0.062568588,      0.62304736 ],
  [               0,      -851.94276,       588.40087,       1750.1026,        1297.489,      -3587.9129 ],
  [               0,      -663.18313,       595.32376,       1205.5452,        504.3336,      -2154.0088 ],
  [               0,       54.301159,        93.18475,      -305.72535,      -609.80274,       992.96386 ],
  [               0,       1489.4879,      -1059.1849,       -3030.444,      -2165.7897,       6160.9151 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer1_etaoffv4_1_35_e = [
 [
  [               0,               0,               0,   -0.0064731619,    -0.018931779,     0.040881363 ],
  [               0,               0,               0,      -9.5405097,      -9.3974401,      -5.9010679 ],
  [               0,               0,               0,        72.75425,      -2.3689521,      -148.00589 ],
  [               0,               0,               0,      -76.476006,       1.2623773,       161.99071 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      -0.0866696,      0.13123168 ],
  [               0,               0,               0,               0,       -120.6894,       182.26412 ],
  [               0,               0,               0,               0,      -146.31897,       214.43004 ],
  [               0,               0,               0,               0,       10.271588,      -15.790654 ],
  [               0,               0,               0,               0,      0.16220346,     -0.15356836 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19088029,      0.29062434 ],
  [               0,               0,               0,               0,      -211.58621,       320.81683 ],
  [               0,               0,               0,               0,      -14.382267,       19.882875 ],
  [               0,               0,               0,               0,       35.768343,      -52.923948 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00067568646,  -0.00040381005,    0.0008763267,    0.0052382298 ],
  [               0,               0,       4.1036152,      -4.7147852,      -14.833249,       20.850695 ],
  [               0,               0,     -0.66128888,       1.4797557,     0.023261761,      -2.2308716 ],
  [               0,               0,      -1.2549432,       2.3885459,       2.0889819,       -5.351274 ],
  [               0,               0,     -0.37186667,      0.58338086,       1.0832805,       -1.683949 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0049147674,   -0.0072545894 ],
  [               0,               0,               0,               0,      -5.0128234,       12.606657 ],
  [               0,               0,               0,               0,      0.12858143,      -1.5534007 ],
  [               0,               0,               0,               0,     -0.80363951,      0.38496181 ],
  [               0,               0,               0,               0,     -0.31898546,      0.62605473 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00021773323,  -0.00013388599,   0.00066731451,    0.0044937794 ],
  [               0,               0,      0.98246066,      -1.7747148,      -4.5463501,       11.160539 ],
  [               0,               0,      0.39967911,      -1.1087584,     -0.73350623,       1.4371379 ],
  [               0,               0,     -0.57034534,       1.5292959,       1.6297312,      -6.0389955 ],
  [               0,               0,     -0.28308146,      0.46317748,       1.9745867,      -3.5378024 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0031635677,     0.010405226 ],
  [               0,               0,               0,               0,      -331.27934,       764.32348 ],
  [               0,               0,               0,               0,       110.54663,      -255.79766 ],
  [               0,               0,               0,               0,       4.3463564,      -9.4201814 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00032705478,    0.0037140431 ],
  [               0,               0,               0,               0,      0.78642243,       1.4182051 ],
  [               0,               0,               0,               0,      -0.4252094,     -0.32523259 ],
  [               0,               0,               0,               0,      0.27139465,      -1.7952576 ],
  [               0,               0,               0,               0,      0.14901187,     -0.30096859 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.14965539,      0.35945028 ],
  [               0,               0,               0,               0,       -93.53903,       229.17943 ],
  [               0,               0,               0,               0,      -108.29107,       257.47734 ],
  [               0,               0,               0,               0,       429.35158,      -1030.8791 ],
  [               0,               0,               0,               0,     -0.22010118,     -0.84284725 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.10303424,      0.25699433 ],
  [               0,               0,               0,               0,       126.94773,        -298.802 ],
  [               0,               0,               0,               0,      -29.121967,       68.664428 ],
  [               0,               0,               0,               0,       16.445861,      -39.651795 ],
  [               0,               0,               0,               0,     0.059677431,     -0.14355947 ],
  [               0,               0,               0,               0,       1.9212687,      -5.8358348 ],
  [               0,               0,               0,               0,       -9.032065,       17.840252 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.026531123 ],
  [               0,               0,               0,               0,               0,       10.063021 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.67038534 ],
  [               0,               0,               0,               0,               0,      0.85227305 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer1_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -8.8160192e-05,   0.00037484961 ],
  [               0,               0,               0,               0,       -4.042399,       7.6138512 ],
  [               0,               0,               0,               0,      0.48283606,      0.25987821 ],
  [               0,               0,               0,               0,        1.566192,     -0.57069456 ],
  [               0,               0,               0,               0,     -0.36541308,     -0.12456099 ],
  [               0,               0,               0,               0,   0.00010579903,    0.0001065987 ],
  [               0,               0,               0,               0,    0.0085299947,       1.7979467 ],
  [               0,               0,               0,               0,      -1.3372811,      -1.2243212 ],
 ],
 [
  [               0,               0,               0,               0,    0.0016154056,   0.00046392944 ],
  [               0,               0,               0,               0,     -0.37668859,       2.8301345 ],
  [               0,               0,               0,               0,     -0.42884303,     -0.66774819 ],
  [               0,               0,               0,               0,     -0.95092522,     -0.56569723 ],
  [               0,               0,               0,               0,     -0.17553649,    -0.014083319 ],
  [               0,               0,               0,               0,  -0.00020604626,   0.00019154979 ],
  [               0,               0,               0,               0,       1.0560005,       1.9008453 ],
  [               0,               0,               0,               0,      -1.2434229,      -1.3607853 ],
 ],
 [
  [               0,               0,               0,               0,     0.022019668,    0.0074393094 ],
  [               0,               0,               0,               0,       34.453119,      -26.842569 ],
  [               0,               0,               0,               0,      -46.180245,       36.217908 ],
  [               0,               0,               0,               0,       16.884288,      -13.430077 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0043632989,     0.005480168,  -0.00017831283 ],
  [               0,               0,               0,     -0.44153553,       0.1309537,       1.4159509 ],
  [               0,               0,               0,      0.50181548,     -0.64348205,     -0.69295853 ],
  [               0,               0,               0,       0.6654263,      -0.8984902,     -0.69535322 ],
  [               0,               0,               0,   -0.0030604152,     0.036224677,    -0.065990507 ],
  [               0,               0,               0,   0.00080570005,    -0.001727925,   0.00096243113 ],
  [               0,               0,               0,       -6.016226,        11.51195,      -2.4546841 ],
  [               0,               0,               0,    -0.063611223,     -0.20148451,      -1.7826319 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       1.0849866,      -1.7524363,      -2.3197366,       3.7408945 ],
  [               0,               0,       -1908.881,       2547.3962,       4405.8083,      -5922.5844 ],
  [               0,               0,      -1738.3548,       2680.1572,       2818.5345,      -4427.7253 ],
  [               0,               0,       198.78578,      -93.422426,      -1256.1335,       1401.0516 ],
  [               0,               0,       4034.7286,      -5748.9735,      -8210.7305,       11696.927 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer1_etaoffv4_1_37_e = [
 [
  [               0,               0,               0,     0.077017476,    -0.082069811,    -0.039169024 ],
  [               0,               0,               0,       18.854957,      -6.5032323,      -69.029718 ],
  [               0,               0,               0,      -59.657872,       1.2629133,       125.40458 ],
  [               0,               0,               0,      -22.812007,      0.38749639,       47.013552 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.10197083,      0.15361054 ],
  [               0,               0,               0,               0,      -205.02962,       305.43422 ],
  [               0,               0,               0,               0,      -160.01059,       234.72317 ],
  [               0,               0,               0,               0,       13.403389,      -20.393287 ],
  [               0,               0,               0,               0,      0.16389858,      -0.1589875 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,     -0.19509098,      0.29701073 ],
  [               0,               0,               0,               0,      -215.91246,       327.15855 ],
  [               0,               0,               0,               0,      -11.976332,       16.303991 ],
  [               0,               0,               0,               0,       33.640435,      -49.755869 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,   -0.0006907651,  -0.00042462712,     0.000876839,    0.0053698156 ],
  [               0,               0,       3.9844628,      -4.3262469,      -14.999143,       20.601452 ],
  [               0,               0,     -0.57813301,       1.4156945,     -0.47313999,      -1.6119463 ],
  [               0,               0,      -1.1089199,       2.1627569,        1.602099,      -4.5909788 ],
  [               0,               0,      -0.3740969,      0.59701702,       1.0575993,      -1.6687292 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0049898964,   -0.0073980341 ],
  [               0,               0,               0,               0,      -5.1626557,       12.863453 ],
  [               0,               0,               0,               0,      0.20003107,      -1.6724924 ],
  [               0,               0,               0,               0,     -0.78640371,      0.36188106 ],
  [               0,               0,               0,               0,     -0.32564199,      0.63785719 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -0.00024162314,  -0.00014929456,   0.00070686496,    0.0047207635 ],
  [               0,               0,      0.77148331,      -1.2818983,       -3.525559,       8.7810183 ],
  [               0,               0,      0.52688515,      -1.4140512,       -1.139046,       2.4576397 ],
  [               0,               0,     -0.47276515,       1.3564178,       1.0173395,      -4.8915513 ],
  [               0,               0,     -0.28667568,       0.4507332,       2.0673192,       -3.643665 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,   0.00026012283,   0.00040780747,   0.00099754246 ],
  [               0,               0,               0,        -66.5656,     -0.37294091,       355.30527 ],
  [               0,               0,               0,       19.784659,     -0.24559965,      -105.69295 ],
  [               0,               0,               0,       1.2565886,      0.12534013,      -6.3730518 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0012264354,     0.005801811 ],
  [               0,               0,               0,               0,       2.2262153,      -1.9446866 ],
  [               0,               0,               0,               0,     -0.98580264,      0.99325113 ],
  [               0,               0,               0,               0,      0.34072343,      -1.9565966 ],
  [               0,               0,               0,               0,      0.24136294,     -0.51776672 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.026947529,     0.064960477 ],
  [               0,               0,               0,               0,        559.2199,      -1302.2083 ],
  [               0,               0,               0,               0,      -302.09928,       724.08125 ],
  [               0,               0,               0,               0,      -1090.0889,       2606.9925 ],
  [               0,               0,               0,               0,     -0.47265988,      -2.1497509 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00087868201,     0.010229722 ],
  [               0,               0,               0,               0,      -26.184632,       71.407094 ],
  [               0,               0,               0,               0,      -11.001714,       24.975095 ],
  [               0,               0,               0,               0,       1492.1027,       356.89912 ],
  [               0,               0,               0,               0,       8.5043613,       11.571725 ],
  [               0,               0,               0,               0,     -0.10898774,       0.2627356 ],
  [               0,               0,               0,               0,      0.14464334,      -3.4978708 ],
 ],
 [
  [               0,               0,               0,               0,               0,     0.025651166 ],
  [               0,               0,               0,               0,               0,       8.8614603 ],
  [               0,               0,               0,               0,               0,            -0.4 ],
  [               0,               0,               0,               0,               0,     -0.71534214 ],
  [               0,               0,               0,               0,               0,      0.79823148 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_55_b = [
 [
  [               0,               0,               0,   -0.0067389045,    0.0045109444,    0.0075407932 ],
  [               0,               0,               0,       2.9563998,      -2.4269586,       4.6052923 ],
  [               0,               0,               0,       2.2142386,     -0.88885642,      -1.2388354 ],
  [               0,               0,               0,     -0.71662633,      0.29900932,    -0.054626007 ],
  [               0,               0,               0,   -0.0041205745,   0.00074245878,    0.0001545945 ],
  [               0,               0,               0,       45.215724,      -18.717345,       1.8028424 ],
  [               0,               0,               0,      -94.367101,       29.441692,      -1.3451346 ],
 ],
 [
  [               0,               0,               0,  -0.00018327367,   0.00044918711,    0.0078053902 ],
  [               0,               0,               0,      -3.6927623,       3.7521054,       3.3330384 ],
  [               0,               0,               0,     -0.26885985,      0.13300279,      -1.2877814 ],
  [               0,               0,               0,       2576.9469,      -397.67895,       -2053.625 ],
  [               0,               0,               0,       -21.41787,        4.446439,       15.971412 ],
  [               0,               0,               0,      0.02330965,    -0.034010318,     0.012649888 ],
  [               0,               0,               0,     -0.31928234,      0.26925706,    -0.057465607 ],
 ],
 [
  [               0,               0,               0,               0,     0.036793787,    -0.020203352 ],
  [               0,               0,               0,               0,      -5.5075396,       8.0523706 ],
  [               0,               0,               0,               0,       5.2116674,       -5.304119 ],
  [               0,               0,               0,               0,      -7.5467843,       5.8117058 ],
  [               0,               0,               0,               0,     0.046732548,    -0.035913895 ],
  [               0,               0,               0,               0,      -5.1338903,       4.5381371 ],
  [               0,               0,               0,               0,       11.284432,      -9.0530833 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00035090139,    0.0089236597 ],
  [               0,               0,               0,               0,     -0.31946368,       3.9221121 ],
  [               0,               0,               0,               0,    -0.098438753,       -1.193743 ],
  [               0,               0,               0,               0,       3.9407184,      -2.8886329 ],
  [               0,               0,               0,               0,    0.0065707051,      0.02259923 ],
  [               0,               0,               0,               0,    0.0033842947,   -0.0016909444 ],
  [               0,               0,               0,               0,      -1.0882627,      -19.646767 ],
 ],
 [
  [               0,               0,   -0.0093964769,   -0.0045189418,    0.0073730001,      0.03212578 ],
  [               0,               0,      -14.545442,       11.568205,       20.948666,      -9.6555945 ],
  [               0,               0,      -60.539852,       87.020304,       116.68953,        -168.862 ],
  [               0,               0,      -12.946986,       13.791832,       21.659645,      -22.188205 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.47619189,     -0.65474474 ],
  [               0,               0,               0,               0,       11.052532,       24.109959 ],
  [               0,               0,               0,               0,      -33.843497,       45.013909 ],
  [               0,               0,               0,               0,     -0.56178375,       1.4394908 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_55_e = [
 [
  [       96.110871,      -194.75153,      -154.19379,       229.88964,       557.05852,      -576.13952 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.030816202,     0.032046176,      0.01464743,    -0.045921799,      -0.1961149 ],
  [               0,       35.918212,      -27.549264,      -70.348337,       -45.84671,       140.45657 ],
  [               0,       200.67676,      -152.13488,      -437.23676,      -313.68028,       951.23108 ],
  [               0,      -156.07359,        114.8188,       340.38704,       252.15679,      -742.03118 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0032058677,    0.0026799052 ],
  [               0,               0,               0,               0,        53.11374,      -76.952411 ],
  [               0,               0,               0,               0,      -2.3773373,       2.3171163 ],
  [               0,               0,               0,               0,     -0.60663908,      0.96457913 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0015543696,    -0.003810941,     0.010779492 ],
  [               0,               0,               0,        4.407499,      -16.416308,       18.455646 ],
  [               0,               0,               0,     -0.62946489,       2.2194801,      -3.2761001 ],
  [               0,               0,               0,   -0.0035336164,     -0.44140428,      0.92347492 ],
  [               0,               0,               0,   -0.0015241684,     0.001225141,    0.0034225492 ],
  [               0,               0,               0,        4.899388,      -15.968237,       13.210462 ],
  [               0,               0,               0,       3.1546375,      -9.1415453,       3.6821269 ],
 ],
 [
  [               0,               0,               0,    0.0015216509,   -0.0038555654,     0.011262046 ],
  [               0,               0,               0,      0.89943545,      -5.5418187,       10.327416 ],
  [               0,               0,               0,      0.28183621,     -0.86536637,       -0.689799 ],
  [               0,               0,               0,     -0.53367922,      -1.5214885,       6.4928783 ],
  [               0,               0,               0,   -0.0088969054,    0.0034369536,     0.040928285 ],
  [               0,               0,               0,     -0.12869266,      0.55503411,     -0.51942953 ],
  [               0,               0,               0,       3.8383377,      -15.130907,       11.855794 ],
 ],
 [
  [               0,               0,               0,   -0.0031672615,     0.007299526,     0.011485547 ],
  [               0,               0,               0,       10.917046,      -51.802241,       63.751428 ],
  [               0,               0,               0,       4.0876232,      -19.213357,       21.360727 ],
  [               0,               0,               0,      -5.4211956,       31.496425,      -38.041903 ],
  [               0,               0,               0,    0.0029261041,    0.0076730537,     0.032589389 ],
  [               0,               0,               0,        2.608095,      -12.130228,       14.065954 ],
  [               0,               0,               0,      -12.038655,       56.021499,      -55.770357 ],
 ],
 [
  [               0,               0,               0,               0,      -0.2097385,      0.52912036 ],
  [               0,               0,               0,               0,      -13.504455,       35.519054 ],
  [               0,               0,               0,               0,       4.5601201,      -12.531797 ],
  [               0,               0,               0,               0,      0.88971775,      -1.9754281 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,    0.0010157151,   0.00014741215,   -0.0037271556,   -0.0071095712 ],
  [               0,               0,      -1.1943656,      -1.3780013,      -3.4151413,       -4.564551 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,      -1.9215006,       -21.71689,       81.228961,       196.94842 ],
  [               0,               0,      -1.1695553,       8.2153989,       38.171335,       13.756231 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_35_b = [
 [
  [               0,               0,               0,    -0.026208703,     0.016300742,    0.0057364839 ],
  [               0,               0,               0,       13.444668,      -8.8766152,       6.2061044 ],
  [               0,               0,               0,       2.4609665,     -0.97536452,      -1.2660542 ],
  [               0,               0,               0,      -1.3406327,       0.5764124,    -0.072762742 ],
  [               0,               0,               0,     0.013818562,   -0.0069191238,   0.00084869196 ],
  [               0,               0,               0,       20.681637,      -8.2732228,       1.3359951 ],
  [               0,               0,               0,      -75.159324,       15.208643,       -1.770419 ],
 ],
 [
  [               0,               0,               0,     0.014117586,    -0.018626345,     0.012761847 ],
  [               0,               0,               0,      -9.3937756,       11.429971,       2.0193542 ],
  [               0,               0,               0,      -1.2049772,        1.317799,      -1.6678173 ],
  [               0,               0,               0,      0.68520742,     -0.72036684,      0.13383341 ],
  [               0,               0,               0,   -0.0057643386,    0.0075388787,   -0.0018411756 ],
  [               0,               0,               0,      -8.0681025,       10.069804,      -2.2119554 ],
  [               0,               0,               0,       7.6443295,      -10.231143,       2.0072887 ],
 ],
 [
  [               0,               0,               0,               0,    -0.011816829,     0.017239168 ],
  [               0,               0,               0,               0,      -3.6905053,       7.6824879 ],
  [               0,               0,               0,               0,       12.993935,      -11.560739 ],
  [               0,               0,               0,               0,      -7.9104126,       6.2013956 ],
  [               0,               0,               0,               0,     0.025283938,    -0.019543579 ],
  [               0,               0,               0,               0,       7.0128496,      -4.7846463 ],
  [               0,               0,               0,               0,       5.7104458,      -5.1575147 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00013654186,    0.0068435696 ],
  [               0,               0,               0,               0,     -0.45355485,       5.2842918 ],
  [               0,               0,               0,               0,    -0.068723434,      -1.3450297 ],
  [               0,               0,               0,               0,      0.11458501,    -0.089446965 ],
  [               0,               0,               0,               0,   -6.766451e-05,  -0.00062579785 ],
  [               0,               0,               0,               0,    -0.099091111,       1.0086338 ],
  [               0,               0,               0,               0,       0.1449883,      -30.263644 ],
 ],
 [
  [               0,               0,       0.1601009,    -0.036564829,       -0.174803,    -0.091931407 ],
  [               0,               0,        102.5415,      -99.426622,      -165.40052,       155.45108 ],
  [               0,               0,      -1046.8305,       1452.1078,       1963.6061,      -2725.7586 ],
  [               0,               0,       471.02499,       -638.1801,      -882.39006,       1195.6767 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,      0.47693312,     -0.66564185 ],
  [               0,               0,               0,               0,      -586.05106,       865.51278 ],
  [               0,               0,               0,               0,      -18.999738,       23.668831 ],
  [               0,               0,               0,               0,      -1.2271134,       1.7617549 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_35_e = [
 [
  [       101.02445,      -203.57368,      -162.55454,        238.7292,        583.1474,      -599.56185 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.020647895,     0.023442343,     0.013621637,    -0.029236678,     -0.14662658 ],
  [               0,       301.96304,      -232.63755,       -647.9899,      -449.08514,       1391.4917 ],
  [               0,       180.18289,      -132.55233,      -393.29734,      -291.01142,       856.03911 ],
  [               0,      -183.88204,       135.05464,       399.67525,       295.62489,      -868.31334 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.049114621,     0.080361831 ],
  [               0,               0,               0,               0,       115.54903,      -170.52978 ],
  [               0,               0,               0,               0,      -38.107383,       56.311272 ],
  [               0,               0,               0,               0,       34.814022,      -52.577084 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0094520707,    -0.024803168,     0.022155455 ],
  [               0,               0,               0,       2.6041238,      -12.633617,       18.260361 ],
  [               0,               0,               0,       1.1866053,      -4.0464698,       2.0308457 ],
  [               0,               0,               0,     -0.13982206,      0.67842882,     -0.66309134 ],
  [               0,               0,               0,    0.0068760087,    -0.024089122,     0.021700692 ],
  [               0,               0,               0,      -1.5815148,       3.9416439,      -1.3736015 ],
  [               0,               0,               0,      -4.4099798,       11.979747,      -9.5707664 ],
 ],
 [
  [               0,               0,               0,    -0.016577725,     0.070871644,    -0.066367261 ],
  [               0,               0,               0,       5.4781516,      -24.902525,       31.352715 ],
  [               0,               0,               0,     -0.63789894,       2.9667022,      -4.7318794 ],
  [               0,               0,               0,      -5.0392392,       20.774611,      -21.048786 ],
  [               0,               0,               0,     -0.04357735,      0.18197363,      -0.1868116 ],
  [               0,               0,               0,       2.3017232,      -9.8932392,       10.758417 ],
  [               0,               0,               0,       9.2382326,      -37.103795,       34.429794 ],
 ],
 [
  [               0,               0,               0,  -3.1982174e-05,    0.0029736353,    0.0031904765 ],
  [               0,               0,               0,       11.611093,      -56.837048,       72.353974 ],
  [               0,               0,               0,       5.9135791,      -27.461962,        30.57464 ],
  [               0,               0,               0,       -13.54253,       64.454472,      -51.986178 ],
  [               0,               0,               0,     0.008679631,     0.037813848,      0.10983627 ],
  [               0,               0,               0,     -0.79614389,       3.6273097,      -4.1059803 ],
  [               0,               0,               0,       5.3438087,      -24.838131,       25.746196 ],
 ],
 [
  [               0,               0,               0,               0,    0.0014850524,    0.0070022367 ],
  [               0,               0,               0,               0,      -11.440047,       30.976266 ],
  [               0,               0,               0,               0,       5.6440475,      -15.179321 ],
  [               0,               0,               0,               0,      0.48280551,      -1.1185741 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,  -8.1439659e-05,    0.0014868144,     0.011357817,    -0.036026824 ],
  [               0,               0,      -1.2072937,      -1.3677159,      -3.3963037,      -4.4344422 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       1349.3011,        1125.382,      -2668.7601,      -20564.344 ],
  [               0,               0,      0.77003314,       8.7151539,        17.52035,       64.803559 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_37_b = [
 [
  [               0,               0,               0,    -0.025050001,     0.015561007,    0.0058382628 ],
  [               0,               0,               0,       16.232598,       -9.392595,       6.1898057 ],
  [               0,               0,               0,       1.5299165,     -0.77022069,      -1.2646522 ],
  [               0,               0,               0,      -0.9082663,      0.47353711,    -0.073576817 ],
  [               0,               0,               0,    0.0071164328,   -0.0049608347,   0.00075086375 ],
  [               0,               0,               0,       30.072186,      -9.0306895,       1.3478593 ],
  [               0,               0,               0,      -25.947519,       10.032362,      -1.6510916 ],
 ],
 [
  [               0,               0,               0,     0.013838977,    -0.018289865,     0.012673799 ],
  [               0,               0,               0,      -9.1528599,       11.120082,       2.1062417 ],
  [               0,               0,               0,      -1.3183098,       1.4381326,      -1.6985798 ],
  [               0,               0,               0,      0.74590237,     -0.78137818,      0.14850206 ],
  [               0,               0,               0,   -0.0057359871,    0.0075130525,   -0.0018280687 ],
  [               0,               0,               0,      -8.0248716,       10.028679,      -2.2062925 ],
  [               0,               0,               0,       6.8888934,      -9.4077865,        1.783572 ],
 ],
 [
  [               0,               0,               0,               0,   -0.0098344715,      0.01570118 ],
  [               0,               0,               0,               0,      -4.0689931,       7.9652838 ],
  [               0,               0,               0,               0,       13.206246,       -11.71912 ],
  [               0,               0,               0,               0,      -8.0159308,       6.2780808 ],
  [               0,               0,               0,               0,     0.025959337,    -0.020049725 ],
  [               0,               0,               0,               0,       6.8193262,      -4.6344267 ],
  [               0,               0,               0,               0,       5.3769533,      -4.8524008 ],
 ],
 [
  [               0,               0,               0,               0,  -0.00018166625,    0.0069319208 ],
  [               0,               0,               0,               0,     -0.44389222,       5.2604329 ],
  [               0,               0,               0,               0,    -0.093314712,      -1.3113183 ],
  [               0,               0,               0,               0,      0.12857895,     -0.10787179 ],
  [               0,               0,               0,               0,   0.00011845504,   0.00056577654 ],
  [               0,               0,               0,               0,    -0.064359293,      0.96086223 ],
  [               0,               0,               0,               0,     0.043922326,      -20.701177 ],
 ],
 [
  [               0,               0,      0.16361331,     -0.03405466,     -0.17630831,     -0.10307984 ],
  [               0,               0,       72.350113,      -64.503855,      -113.80211,       96.542957 ],
  [               0,               0,      -947.31038,       1316.3258,       1775.7406,      -2469.3939 ],
  [               0,               0,       502.84577,      -682.43285,      -942.96621,       1279.8775 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    0.0086678495,    0.0081486834 ],
  [               0,               0,               0,               0,       -555.9625,       823.15386 ],
  [               0,               0,               0,               0,      -22.139608,       29.467039 ],
  [               0,               0,               0,               0,      -1.9971636,       2.9083105 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]
gam1000_layer2_etaoffv4_1_37_e = [
 [
  [       88.948079,      -179.43642,      -143.07825,       210.71524,       513.99279,      -529.05731 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,     0.025027473,     0.021490519,    0.0059296189,    -0.036277823,     -0.13357113 ],
  [               0,       296.75642,       -230.3743,      -632.07664,      -443.01253,       1365.2689 ],
  [               0,       180.14984,       -132.3239,      -393.39077,      -291.65068,       856.61054 ],
  [               0,      -186.28646,        136.8671,       404.92827,       299.42756,      -879.78811 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,               0,    -0.058724659,     0.094926943 ],
  [               0,               0,               0,               0,       119.73686,      -176.89571 ],
  [               0,               0,               0,               0,      -39.654816,       58.658857 ],
  [               0,               0,               0,               0,       36.351049,      -54.902822 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,               0,    0.0097949232,     -0.02598521,     0.023214755 ],
  [               0,               0,               0,        2.409647,      -11.922359,       17.600901 ],
  [               0,               0,               0,       1.1738535,      -3.9937406,       1.9786437 ],
  [               0,               0,               0,     -0.13467646,      0.65842993,     -0.64373279 ],
  [               0,               0,               0,    0.0066875726,    -0.023405463,     0.021090917 ],
  [               0,               0,               0,      -1.5086201,       3.6808316,      -1.1483672 ],
  [               0,               0,               0,      -4.2787451,       11.590841,      -9.2909439 ],
 ],
 [
  [               0,               0,               0,    0.0084131104,    -0.033675339,     0.042660023 ],
  [               0,               0,               0,      -1.2354342,       3.2815721,       1.8651106 ],
  [               0,               0,               0,      0.41019261,      -1.4513341,    -0.094832804 ],
  [               0,               0,               0,      -3.7274757,       15.492056,      -15.849071 ],
  [               0,               0,               0,    -0.030471073,      0.12815671,     -0.13268053 ],
  [               0,               0,               0,       2.3537263,      -10.152886,       11.137398 ],
  [               0,               0,               0,       5.7927346,      -23.047867,       20.221701 ],
 ],
 [
  [               0,               0,               0,   0.00012493025,    0.0019213577,    0.0049475173 ],
  [               0,               0,               0,       10.617341,       -52.04559,       66.567915 ],
  [               0,               0,               0,       6.9033756,      -32.172946,       36.177766 ],
  [               0,               0,               0,       -12.94881,       55.528991,      -47.542142 ],
  [               0,               0,               0,   -0.0071778926,    0.0086618847,      0.13507446 ],
  [               0,               0,               0,      -1.6051301,       7.4092999,      -8.5147414 ],
  [               0,               0,               0,       8.2474482,      -38.406049,       41.605746 ],
 ],
 [
  [               0,               0,               0,               0,    0.0014403936,    0.0072006645 ],
  [               0,               0,               0,               0,      -13.089084,       35.036898 ],
  [               0,               0,               0,               0,       6.1196543,       -16.35107 ],
  [               0,               0,               0,               0,      0.37782045,     -0.85823737 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
 [
  [               0,               0,       30.556079,      -129.20615,        77.84366,       135.54702 ],
  [               0,               0,       461.36742,      -788.04528,      -711.35481,      -446.97775 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,       83.954048,      -134.99136,      -138.12778,      -99.053091 ],
  [               0,               0,     -0.47992384,       0.7080246,       1.6186601,       2.4706535 ],
  [               0,               0,               0,               0,               0,               0 ],
  [               0,               0,               0,               0,               0,               0 ],
 ],
]


###############################################################################
# Now, organize these by type/size/barrel-endcap.
#


ele_layer1_etaoffv4_1_55_b = [
    ele5_layer1_etaoffv4_1_55_b,
    ele10_layer1_etaoffv4_1_55_b,
    ele25_layer1_etaoffv4_1_55_b,
    ele50large_layer1_etaoffv4_1_55_b,
    ele75_layer1_etaoffv4_1_55_b,
    ele100large_layer1_etaoffv4_1_55_b,
    ele200_layer1_etaoffv4_1_55_b,
    ele500_layer1_etaoffv4_1_55_b,
    ele1000_layer1_etaoffv4_1_55_b,
    ]

ele_layer1_etaoffv4_1_55_e = [
    ele5_layer1_etaoffv4_1_55_e,
    ele10_layer1_etaoffv4_1_55_e,
    ele25_layer1_etaoffv4_1_55_e,
    ele50large_layer1_etaoffv4_1_55_e,
    ele75_layer1_etaoffv4_1_55_e,
    ele100large_layer1_etaoffv4_1_55_e,
    ele200_layer1_etaoffv4_1_55_e,
    ele500_layer1_etaoffv4_1_55_e,
    ele1000_layer1_etaoffv4_1_55_e,
    ]



gam_layer1_etaoffv4_1_55_b = [
    gam5_layer1_etaoffv4_1_55_b,
    gam10_layer1_etaoffv4_1_55_b,
    gam25_layer1_etaoffv4_1_55_b,
    gam50large_layer1_etaoffv4_1_55_b,
    gam75_layer1_etaoffv4_1_55_b,
    gam100large_layer1_etaoffv4_1_55_b,
    gam200_layer1_etaoffv4_1_55_b,
    gam500_layer1_etaoffv4_1_55_b,
    gam1000_layer1_etaoffv4_1_55_b,
    ]

gam_layer1_etaoffv4_1_55_e = [
    gam5_layer1_etaoffv4_1_55_e,
    gam10_layer1_etaoffv4_1_55_e,
    gam25_layer1_etaoffv4_1_55_e,
    gam50large_layer1_etaoffv4_1_55_e,
    gam75_layer1_etaoffv4_1_55_e,
    gam100large_layer1_etaoffv4_1_55_e,
    gam200_layer1_etaoffv4_1_55_e,
    gam500_layer1_etaoffv4_1_55_e,
    gam1000_layer1_etaoffv4_1_55_e,
    ]



ele_layer1_etaoffv4_1_35_b = [
    ele5_layer1_etaoffv4_1_35_b,
    ele10_layer1_etaoffv4_1_35_b,
    ele25_layer1_etaoffv4_1_35_b,
    ele50large_layer1_etaoffv4_1_35_b,
    ele75_layer1_etaoffv4_1_35_b,
    ele100large_layer1_etaoffv4_1_35_b,
    ele200_layer1_etaoffv4_1_35_b,
    ele500_layer1_etaoffv4_1_35_b,
    ele1000_layer1_etaoffv4_1_35_b,
    ]

ele_layer1_etaoffv4_1_35_e = [
    ele5_layer1_etaoffv4_1_35_e,
    ele10_layer1_etaoffv4_1_35_e,
    ele25_layer1_etaoffv4_1_35_e,
    ele50large_layer1_etaoffv4_1_35_e,
    ele75_layer1_etaoffv4_1_35_e,
    ele100large_layer1_etaoffv4_1_35_e,
    ele200_layer1_etaoffv4_1_35_e,
    ele500_layer1_etaoffv4_1_35_e,
    ele1000_layer1_etaoffv4_1_35_e,
    ]



gam_layer1_etaoffv4_1_35_b = [
    gam5_layer1_etaoffv4_1_35_b,
    gam10_layer1_etaoffv4_1_35_b,
    gam25_layer1_etaoffv4_1_35_b,
    gam50large_layer1_etaoffv4_1_35_b,
    gam75_layer1_etaoffv4_1_35_b,
    gam100large_layer1_etaoffv4_1_35_b,
    gam200_layer1_etaoffv4_1_35_b,
    gam500_layer1_etaoffv4_1_35_b,
    gam1000_layer1_etaoffv4_1_35_b,
    ]

gam_layer1_etaoffv4_1_35_e = [
    gam5_layer1_etaoffv4_1_35_e,
    gam10_layer1_etaoffv4_1_35_e,
    gam25_layer1_etaoffv4_1_35_e,
    gam50large_layer1_etaoffv4_1_35_e,
    gam75_layer1_etaoffv4_1_35_e,
    gam100large_layer1_etaoffv4_1_35_e,
    gam200_layer1_etaoffv4_1_35_e,
    gam500_layer1_etaoffv4_1_35_e,
    gam1000_layer1_etaoffv4_1_35_e,
    ]



ele_layer1_etaoffv4_1_37_b = [
    ele5_layer1_etaoffv4_1_37_b,
    ele10_layer1_etaoffv4_1_37_b,
    ele25_layer1_etaoffv4_1_37_b,
    ele50large_layer1_etaoffv4_1_37_b,
    ele75_layer1_etaoffv4_1_37_b,
    ele100large_layer1_etaoffv4_1_37_b,
    ele200_layer1_etaoffv4_1_37_b,
    ele500_layer1_etaoffv4_1_37_b,
    ele1000_layer1_etaoffv4_1_37_b,
    ]

ele_layer1_etaoffv4_1_37_e = [
    ele5_layer1_etaoffv4_1_37_e,
    ele10_layer1_etaoffv4_1_37_e,
    ele25_layer1_etaoffv4_1_37_e,
    ele50large_layer1_etaoffv4_1_37_e,
    ele75_layer1_etaoffv4_1_37_e,
    ele100large_layer1_etaoffv4_1_37_e,
    ele200_layer1_etaoffv4_1_37_e,
    ele500_layer1_etaoffv4_1_37_e,
    ele1000_layer1_etaoffv4_1_37_e,
    ]



gam_layer1_etaoffv4_1_37_b = [
    gam5_layer1_etaoffv4_1_37_b,
    gam10_layer1_etaoffv4_1_37_b,
    gam25_layer1_etaoffv4_1_37_b,
    gam50large_layer1_etaoffv4_1_37_b,
    gam75_layer1_etaoffv4_1_37_b,
    gam100large_layer1_etaoffv4_1_37_b,
    gam200_layer1_etaoffv4_1_37_b,
    gam500_layer1_etaoffv4_1_37_b,
    gam1000_layer1_etaoffv4_1_37_b,
    ]

gam_layer1_etaoffv4_1_37_e = [
    gam5_layer1_etaoffv4_1_37_e,
    gam10_layer1_etaoffv4_1_37_e,
    gam25_layer1_etaoffv4_1_37_e,
    gam50large_layer1_etaoffv4_1_37_e,
    gam75_layer1_etaoffv4_1_37_e,
    gam100large_layer1_etaoffv4_1_37_e,
    gam200_layer1_etaoffv4_1_37_e,
    gam500_layer1_etaoffv4_1_37_e,
    gam1000_layer1_etaoffv4_1_37_e,
    ]



ele_layer2_etaoffv4_1_55_b = [
    ele5_layer2_etaoffv4_1_55_b,
    ele10_layer2_etaoffv4_1_55_b,
    ele25_layer2_etaoffv4_1_55_b,
    ele50large_layer2_etaoffv4_1_55_b,
    ele75_layer2_etaoffv4_1_55_b,
    ele100large_layer2_etaoffv4_1_55_b,
    ele200_layer2_etaoffv4_1_55_b,
    ele500_layer2_etaoffv4_1_55_b,
    ele1000_layer2_etaoffv4_1_55_b,
    ]

ele_layer2_etaoffv4_1_55_e = [
    ele5_layer2_etaoffv4_1_55_e,
    ele10_layer2_etaoffv4_1_55_e,
    ele25_layer2_etaoffv4_1_55_e,
    ele50large_layer2_etaoffv4_1_55_e,
    ele75_layer2_etaoffv4_1_55_e,
    ele100large_layer2_etaoffv4_1_55_e,
    ele200_layer2_etaoffv4_1_55_e,
    ele500_layer2_etaoffv4_1_55_e,
    ele1000_layer2_etaoffv4_1_55_e,
    ]


gam_layer2_etaoffv4_1_55_b = [
    gam5_layer2_etaoffv4_1_55_b,
    gam10_layer2_etaoffv4_1_55_b,
    gam25_layer2_etaoffv4_1_55_b,
    gam50large_layer2_etaoffv4_1_55_b,
    gam75_layer2_etaoffv4_1_55_b,
    gam100large_layer2_etaoffv4_1_55_b,
    gam200_layer2_etaoffv4_1_55_b,
    gam500_layer2_etaoffv4_1_55_b,
    gam1000_layer2_etaoffv4_1_55_b,
    ]

gam_layer2_etaoffv4_1_55_e = [
    gam5_layer2_etaoffv4_1_55_e,
    gam10_layer2_etaoffv4_1_55_e,
    gam25_layer2_etaoffv4_1_55_e,
    gam50large_layer2_etaoffv4_1_55_e,
    gam75_layer2_etaoffv4_1_55_e,
    gam100large_layer2_etaoffv4_1_55_e,
    gam200_layer2_etaoffv4_1_55_e,
    gam500_layer2_etaoffv4_1_55_e,
    gam1000_layer2_etaoffv4_1_55_e,
    ]


ele_layer2_etaoffv4_1_35_b = [
    ele5_layer2_etaoffv4_1_35_b,
    ele10_layer2_etaoffv4_1_35_b,
    ele25_layer2_etaoffv4_1_35_b,
    ele50large_layer2_etaoffv4_1_35_b,
    ele75_layer2_etaoffv4_1_35_b,
    ele100large_layer2_etaoffv4_1_35_b,
    ele200_layer2_etaoffv4_1_35_b,
    ele500_layer2_etaoffv4_1_35_b,
    ele1000_layer2_etaoffv4_1_35_b,
    ]

ele_layer2_etaoffv4_1_35_e = [
    ele5_layer2_etaoffv4_1_35_e,
    ele10_layer2_etaoffv4_1_35_e,
    ele25_layer2_etaoffv4_1_35_e,
    ele50large_layer2_etaoffv4_1_35_e,
    ele75_layer2_etaoffv4_1_35_e,
    ele100large_layer2_etaoffv4_1_35_e,
    ele200_layer2_etaoffv4_1_35_e,
    ele500_layer2_etaoffv4_1_35_e,
    ele1000_layer2_etaoffv4_1_35_e,
    ]


gam_layer2_etaoffv4_1_35_b = [
    gam5_layer2_etaoffv4_1_35_b,
    gam10_layer2_etaoffv4_1_35_b,
    gam25_layer2_etaoffv4_1_35_b,
    gam50large_layer2_etaoffv4_1_35_b,
    gam75_layer2_etaoffv4_1_35_b,
    gam100large_layer2_etaoffv4_1_35_b,
    gam200_layer2_etaoffv4_1_35_b,
    gam500_layer2_etaoffv4_1_35_b,
    gam1000_layer2_etaoffv4_1_35_b,
    ]

gam_layer2_etaoffv4_1_35_e = [
    gam5_layer2_etaoffv4_1_35_e,
    gam10_layer2_etaoffv4_1_35_e,
    gam25_layer2_etaoffv4_1_35_e,
    gam50large_layer2_etaoffv4_1_35_e,
    gam75_layer2_etaoffv4_1_35_e,
    gam100large_layer2_etaoffv4_1_35_e,
    gam200_layer2_etaoffv4_1_35_e,
    gam500_layer2_etaoffv4_1_35_e,
    gam1000_layer2_etaoffv4_1_35_e,
    ]


ele_layer2_etaoffv4_1_37_b = [
    ele5_layer2_etaoffv4_1_37_b,
    ele10_layer2_etaoffv4_1_37_b,
    ele25_layer2_etaoffv4_1_37_b,
    ele50large_layer2_etaoffv4_1_37_b,
    ele75_layer2_etaoffv4_1_37_b,
    ele100large_layer2_etaoffv4_1_37_b,
    ele200_layer2_etaoffv4_1_37_b,
    ele500_layer2_etaoffv4_1_37_b,
    ele1000_layer2_etaoffv4_1_37_b,
    ]

ele_layer2_etaoffv4_1_37_e = [
    ele5_layer2_etaoffv4_1_37_e,
    ele10_layer2_etaoffv4_1_37_e,
    ele25_layer2_etaoffv4_1_37_e,
    ele50large_layer2_etaoffv4_1_37_e,
    ele75_layer2_etaoffv4_1_37_e,
    ele100large_layer2_etaoffv4_1_37_e,
    ele200_layer2_etaoffv4_1_37_e,
    ele500_layer2_etaoffv4_1_37_e,
    ele1000_layer2_etaoffv4_1_37_e,
    ]


gam_layer2_etaoffv4_1_37_b = [
    gam5_layer2_etaoffv4_1_37_b,
    gam10_layer2_etaoffv4_1_37_b,
    gam25_layer2_etaoffv4_1_37_b,
    gam50large_layer2_etaoffv4_1_37_b,
    gam75_layer2_etaoffv4_1_37_b,
    gam100large_layer2_etaoffv4_1_37_b,
    gam200_layer2_etaoffv4_1_37_b,
    gam500_layer2_etaoffv4_1_37_b,
    gam1000_layer2_etaoffv4_1_37_b,
    ]

gam_layer2_etaoffv4_1_37_e = [
    gam5_layer2_etaoffv4_1_37_e,
    gam10_layer2_etaoffv4_1_37_e,
    gam25_layer2_etaoffv4_1_37_e,
    gam50large_layer2_etaoffv4_1_37_e,
    gam75_layer2_etaoffv4_1_37_e,
    gam100large_layer2_etaoffv4_1_37_e,
    gam200_layer2_etaoffv4_1_37_e,
    gam500_layer2_etaoffv4_1_37_e,
    gam1000_layer2_etaoffv4_1_37_e,
    ]


###############################################################################
# Region tables.
#

layer1_etaoffv4_1_regions_b = [
  [ 0.0, 0.00625, 5, 0.003125, 0 ],
  [ 0.00625, 0.4, 11, 0.003125, 0 ],
  [ 0.4, 0.79375, 11, 0.003125, 0 ],
  [ 0.79365, 0.8, 3, 0.003125, 0 ],
  [ 0.8, 1.39375, 11, 0.003125, 0 ],
  [ 1.39375, 1.3955, 3, 0.003125, 0 ],
  [ 1.3955, 1.4, 0, 0.025, 0 ],
  [ 1.4, 1.45, 0, 0.025, 0 ],
]
layer1_etaoffv4_1_regions_e = [
  [ 1.425, 1.454, 3, 0.025, 0 ],
  [ 1.454, 1.475, 0, 0.025, 0 ],
  [ 1.475, 1.5, 3, 0.025, 0 ],
  [ 1.5, 1.503125, 0, 0.003125, 0 ],
  [ 1.503125, 1.8, 0, 0.003125, 0 ],
  [ 1.8, 2.0, 0, 0.00416666666667, 0 ],
  [ 2.0, 2.29375, 0, 0.00625, 0 ],
  [ 2.29375, 2.30625, 3, 0.00625, 0 ],
  [ 2.30625, 2.39375, 0, 0.00625, 0 ],
  [ 2.39375, 2.4, 0, 0.00625, 0 ],
  [ 2.4, 2.475, 10, 0.025, 0 ],
  [ 2.475, 2.5, 0, 0.025, 0 ],
]

layer2_etaoffv4_1_regions_b = [
  [ 0.0, 0.4, 10, 0.025, 0 ],
  [ 0.4, 0.775, 10, 0.025, 0 ],
  [ 0.775, 0.825, 10, 0.025, 0 ],
  [ 0.825, 1.35, 10, 0.025, 0 ],
  [ 1.35, 1.4, 3, 0.025, 0 ],
  [ 1.4, 1.437, 3, 0.075, -0.025 ],
]
layer2_etaoffv4_1_regions_e = [
  [ 1.4, 1.45, 5, 0.025, 0 ],
  [ 1.45, 1.5, 3, 0.025, 0 ],
  [ 1.5, 1.525, 3, 0.025, 0 ],
  [ 1.525, 1.9, 10, 0.025, 0 ],
  [ 1.9, 2.3, 10, 0.025, 0 ],
  [ 2.3, 2.45, 10, 0.025, 0 ],
  [ 2.45, 2.475, 3, 0.025, 0 ],
  [ 2.475, 2.5, 0, 0.025, 0 ],
]


ele_etaoffv4_1_energies = [
    5*GeV, 10*GeV, 25*GeV, 50*GeV, 75*GeV, 100*GeV, 200*GeV, 500*GeV, 1000*GeV
    ]
gam_etaoffv4_1_energies = [
    5*GeV, 10*GeV, 25*GeV, 50*GeV, 75*GeV, 100*GeV, 200*GeV, 500*GeV, 1000*GeV
    ]
etaoffv4_1_energies = {'ele55' : ele_etaoffv4_1_energies,
                       'ele35' : ele_etaoffv4_1_energies,
                       'ele37' : ele_etaoffv4_1_energies,
                       'gam55' : gam_etaoffv4_1_energies,
                       'gam35' : gam_etaoffv4_1_energies,
                       'gam37' : gam_etaoffv4_1_energies}



class CaloSwEtaoff_v4_1_b1_parms:
    region = CALOCORR_EMB1
    regions = layer1_etaoffv4_1_regions_b
    energy_degree = 3
    forms = [[]]

    correction = {'ele55' : ele_layer1_etaoffv4_1_55_b,
                  'ele35' : ele_layer1_etaoffv4_1_35_b,
                  'ele37' : ele_layer1_etaoffv4_1_37_b,
                  'gam55' : gam_layer1_etaoffv4_1_55_b,
                  'gam35' : gam_layer1_etaoffv4_1_35_b,
                  'gam37' : gam_layer1_etaoffv4_1_37_b,

                  # Use 5x5 for other cluster sizes.
                  'ele33' : ele_layer1_etaoffv4_1_55_b,
                  'ele57' : ele_layer1_etaoffv4_1_55_b,
                  'ele77' : ele_layer1_etaoffv4_1_55_b,
                  'gam33' : gam_layer1_etaoffv4_1_55_b,
                  'gam57' : gam_layer1_etaoffv4_1_55_b,
                  'gam77' : gam_layer1_etaoffv4_1_55_b,
                  }

    energies = etaoffv4_1_energies


class CaloSwEtaoff_v4_1_b2_parms:
    region = CALOCORR_EMB2
    regions = layer2_etaoffv4_1_regions_b
    energy_degree = 3
    forms = [[]]

    correction = {'ele55' : ele_layer2_etaoffv4_1_55_b,
                  'ele35' : ele_layer2_etaoffv4_1_35_b,
                  'ele37' : ele_layer2_etaoffv4_1_37_b,
                  'gam55' : gam_layer2_etaoffv4_1_55_b,
                  'gam35' : gam_layer2_etaoffv4_1_35_b,
                  'gam37' : gam_layer2_etaoffv4_1_37_b,

                  # Use 5x5 for other cluster sizes.
                  'ele33' : ele_layer2_etaoffv4_1_55_b,
                  'ele57' : ele_layer2_etaoffv4_1_55_b,
                  'ele77' : ele_layer2_etaoffv4_1_55_b,
                  'gam33' : gam_layer2_etaoffv4_1_55_b,
                  'gam57' : gam_layer2_etaoffv4_1_55_b,
                  'gam77' : gam_layer2_etaoffv4_1_55_b,
                  }

    energies = etaoffv4_1_energies


class CaloSwEtaoff_v4_1_e1_parms:
    region = CALOCORR_EME1
    regions = layer1_etaoffv4_1_regions_e
    energy_degree = 3
    forms = [[]]

    correction = {'ele55' : ele_layer1_etaoffv4_1_55_e,
                  'ele35' : ele_layer1_etaoffv4_1_35_e,
                  'ele37' : ele_layer1_etaoffv4_1_37_e,
                  'gam55' : gam_layer1_etaoffv4_1_55_e,
                  'gam35' : gam_layer1_etaoffv4_1_35_e,
                  'gam37' : gam_layer1_etaoffv4_1_37_e,

                  # Use 5x5 for other cluster sizes.
                  'ele33' : ele_layer1_etaoffv4_1_55_e,
                  'ele57' : ele_layer1_etaoffv4_1_55_e,
                  'ele77' : ele_layer1_etaoffv4_1_55_e,
                  'gam33' : gam_layer1_etaoffv4_1_55_e,
                  'gam57' : gam_layer1_etaoffv4_1_55_e,
                  'gam77' : gam_layer1_etaoffv4_1_55_e,
                  }

    energies = etaoffv4_1_energies


class CaloSwEtaoff_v4_1_e2_parms:
    region = CALOCORR_EME2
    regions = layer2_etaoffv4_1_regions_e
    energy_degree = 3
    forms = [[]]

    correction = {'ele55' : ele_layer2_etaoffv4_1_55_e,
                  'ele35' : ele_layer2_etaoffv4_1_35_e,
                  'ele37' : ele_layer2_etaoffv4_1_37_e,
                  'gam55' : gam_layer2_etaoffv4_1_55_e,
                  'gam35' : gam_layer2_etaoffv4_1_35_e,
                  'gam37' : gam_layer2_etaoffv4_1_37_e,

                  # Use 5x5 for other cluster sizes.
                  'ele33' : ele_layer2_etaoffv4_1_55_e,
                  'ele57' : ele_layer2_etaoffv4_1_55_e,
                  'ele77' : ele_layer2_etaoffv4_1_55_e,
                  'gam33' : gam_layer2_etaoffv4_1_55_e,
                  'gam57' : gam_layer2_etaoffv4_1_55_e,
                  'gam77' : gam_layer2_etaoffv4_1_55_e,
                  }

    energies = etaoffv4_1_energies




CaloSwEtaoff_v4_1_parms = { EMB1 : CaloSwEtaoff_v4_1_b1_parms,
                            EMB2 : CaloSwEtaoff_v4_1_b2_parms,
                            EME1 : CaloSwEtaoff_v4_1_e1_parms,
                            EME2 : CaloSwEtaoff_v4_1_e2_parms,
                          }



#
# Notes on the derivation of the above tables.
# The following are the descriptions of the fits used.
# The D constructor has the signature:
#
#     def __init__ (self, sampling, cellsize, lo, hi, degree, form,
#                   ysize = 0.001, prange = (None, None),
#                   psize = None, cond = None,
#                   outlier_sigma = 5,
#                   phase = 0,
#                   initvals = [],
#                   cmds = [],
#                   minevents = MINEVENTS_DEFAULT,
#                   fitter_type = 1,
#                   ignorepars = []):
#
#
# layer1_regions_v41 = [
#     D(EMB1, 0.025/8, 0.0,     0.00625, 3, 5, 0.004),                      # 0
#     D(EMB1, 0.025/8, 0.00625, 0.4,     1, 11,                             # 1
#       prange=(0.1, 0.15), psize=0.0001,
#       ignorepars = {'ele5' : [(5,0),(6,0),(7,0)],
#                     'gam5' : [(5,0),(6,0),(7,0)]}),
#     D(EMB1, 0.025/8, 0.4,     0.79375, {'ele1000':1,                      # 2
#                                         'gam1000-35':1,
#                                         'gam1000-37':1,
#                                         'gam5' : 3,
#                                         None: 2},
#                                        11, prange=(0.6, 0.65), psize=0.0001,
#       minevents = {'ele5' : 1e9},
#       ignorepars = {'gam5' : [(5,0),(6,0),(7,0)],
#                     'ele10' : [(5,0),(6,0),(7,0)]}),
#     D(EMB1, 0.025/8, 0.79365, 0.8, {'gam25-37' : 2,                       # 3
#                                     None : 1}, 3),
#     D(EMB1, 0.025/8, 0.8,     1.39375, 2, 11,                             # 4
#       prange=(1.0, 1.05), psize=0.0001,
#       minevents = {'gam5' : 1e9}),
#     D(EMB1, 0.025/8, 1.39375, 1.3955,  0, 3, 0.002, psize=0.0005),        # 5
#     D(EMB1, 0.025,   1.3955,  1.4,     3, 0, 0.01),                       # 6
#     D(EMB1, 0.025,   1.4,     1.45,    {'gam1000-35' : 4,                 # 7
#                                         None: 3}, 0, 0.01),  
#     D(EME1, 0.025,   1.425,   1.454,   {'ele25-55' : 0,                   # 8
#                                         'gam25-35' : 0,
#                                         'ele75' : 0,
#                                         'gam75-55' : 0,
#                                         'gam1000-35' : 2,
#                                         'gam1000-37' : 2,
#                                         'ele1000-35' : 2,
#                                         'ele1000-37' : 2,
#                                         None : 1},
#                                        3, 0.015),
#     D(EME1, 0.025,   1.454,   1.475,   {'ele25' : 0,                      # 9
#                                         'gam25' : 0,
#                                         'ele50' : 0,
#                                         'gam50' : 0,
#                                         'ele75' : 0,
#                                         'gam75' : 0,
#                                         None :    1}, 0, 0.01),
#     D(EME1, 0.025,   1.475,   1.5,     {'ele25': 0,                       #10
#                                         None   : 1}, 3, 0.01), 
#     D(EME1, 0.025/8, 1.5,     1.503125,0, 0, 0.005),                      #11
#     D(EME1, 0.025/8, 1.503125, 1.8, {'ele10' : 0,                         #12
#                                      'gam10' : 0,
#                                      'gam25' : 2,
#                                      None:3}, 0, prange=(1.6, 1.65)),
#     D(EME1, 0.025/6, 1.8,     2.0,     1, 0, prange=(1.85, 1.9),          #13
#       minevents = {'gam10' : 1e9}),
#     D(EME1, 0.025/4, 2.0,     2.29375, 3, 0, prange=(2.1, 2.15)),         #14
#     D(EME1, 0.025/4, 2.29375, 2.30625, {'gam1000-37' : 2,                 #15
#                                         'gam200' : 0,
#                                         'gam500-37' : 0,
#                                         'ele25'  : 0,
#                                         None: 1}, 3, 0.003),
#     D(EME1, 0.025/4, 2.30625, 2.39375, {'ele25' : 0,                      #16
#                                         None :    1}, 0),
#     D(EME1, 0.025/4, 2.39375, 2.4,     {'ele200' : 0,                     #17
#                                         'gam200' : 0,
#                                         'ele75'  : 0,
#                                         'gam75'  : 0,
#                                         'ele50'  : 0,
#                                         'gam50'  : 0,
#                                         'gam100' : 0,
#                                         None: 1},
#       0, 0.004),
#     D(EME1, 0.025,   2.4,     2.475,   {'gam100' : 0,                     #18
#                                         'ele200' : 1,
#                                         'gam200' : 1,
#                                         'gam1000' : 1,
#                                         None:2},
#       10, 0.008,
#       ignorepars={'gam500' : [(4,0),(5,0),(6,0)]}),
#     D(EME1, 0.025,   2.475,   2.5,     {'gam50-37':2,                     #19
#                                         'gam100large': 0,
#                                         'ele50' : 2,
#                                         'ele75-55' : 2,
#                                         'ele200' : 0,
#                                         'gam500' : 0,
#                                         'gam1000' : 0,
#                                         None:1},
#       0, 0.008, outlier_sigma=3),
#     ]


# layer2_regions_v41 = [
#     D(EMB2, 0.025,   0.0,   0.4,       {'gam75-55' : 3,                   # 0
#                                         'ele100'   : 3,
#                                         None       : 2},
#       10, psize=0.0003,
#       fitter_type = {'gam75-55' : 2,
#                      'ele100'   : 2},
#       ignorepars={'ele5' : [(4,0),(5,0),(6,0)],
#                   'gam5' : [(4,0),(5,0),(6,0)]}),
#     D(EMB2, 0.025,   0.4,   0.775,     {'ele50' : 1,                      # 1
#                                         'ele100large-35' : 0,
#                                         'ele500' : 1,
#                                         None:2}, 10, psize=0.0003,
#       minevents={'ele5' : 1e9},
#       ignorepars={'gam5' : [(4,0),(5,0),(6,0)]}),
#     D(EMB2, 0.025,   0.775, 0.825,     {'ele10': 0,                       # 2
#                                         None: 1}, 10, 0.005, psize=0.001,
#       ignorepars={'gam5'  : [(4,0),(5,0),(6,0)],
#                   'gam10' : [(4,0),(5,0),(6,0)],
#                   'ele10' : [(4,0),(5,0),(6,0)],
#                   'ele25' : [(4,0),(5,0),(6,0)]}),
#     D(EMB2, 0.025,   0.825, 1.35,      {'ele200' : 1,                     # 3
#                                         'gam1000' : 1,
#                                         'ele50-37'   : 3,
#                                         None:2}, 10, 0.002, psize=0.0003,
#       ignorepars={'ele10' : [(4,0),(5,0),(6,0)]},
#       fitter_type = {'ele50-37' : 2},
#       minevents={'gam5' : 1e9}),
#     D(EMB2, 0.025,   1.35,  1.4,       {'ele25-55' : 2,                   # 4
#                                         'ele25-35' : 2,
#                                         'ele25-37' : 4,
#                                         'ele1000'  : 4,
#                                         'ele50large-55' : 1,
#                                         'gam50large-35' : 4,
#                                         'gam10'  : 2,
#                                         'gam100' : 2,
#                                         'ele100' : 2,
#                                         'ele75'  : 2,
#                                         None: 3}, 3, 0.005, psize=0.0004), 
#     D(EMB2, 0.075,   1.4,   1.437,     {'ele10' : 1,                      # 5
#                                         'gam10' : 2,
#                                         'ele25' : 2,
#                                         'gam25' : 2,
#                                         'ele50' : 2,
#                                         'gam50' : 2,
#                                         'ele50large' : 2,
#                                         'gam50large-55' : 1,
#                                         'gam50large-35' : 2,
#                                         'gam50large-37' : 2,
#                                         'ele75-35' : 2,
#                                         'ele75-37' : 2,
#                                         'ele75-55' : 1,
#                                         'gam75' : 1,
#                                         'ele100' : 2,
#                                         'gam100' : 1,
#                                         'ele100large' : 2,
#                                         'gam100large' : 1,
#                                         'ele200' : 2,
#                                         'gam200' : 1,
#                                         'ele500' : 1,
#                                         'gam500' : 1,
#                                         'ele1000' : 2,
#                                         'gam1000' : 1,
#                                         None: 1},
#       3, 0.008, psize=0.005, phase=-0.025),
#     D(EME2, 0.025,   1.4,  1.45,      {'ele500':5,                        # 6
#                                        'gam500':5,
#                                        'ele1000':5,
#                                        'gam1000':5,
#                                        None:3},
#       5, 0.02, psize=0.02, outlier_sigma=2),
#     D(EME2, 0.025,   1.45, 1.5,       {'ele25':1,                         # 7
#                                        'gam50large' : 4,
#                                        'gam75' : 2,
#                                        'ele100large' : 4,
#                                        'gam100' : 2,
#                                        'ele200' : 2,
#                                        'gam200' : 2,
#                                        'ele1000': 4,
#                                        'gam1000': 4,
#                                        None: 3},
#       3, 0.01, psize=0.005),
#     D(EME2, 0.025,   1.5,    1.525,   {'gam50large' : 2,                  # 8
#                                        None : 1},
#       3, 0.005, psize=0.003),
#     D(EME2, 0.025,   1.525,  1.9,     {'gam10' : 0,                       # 9
#                                        None: 2}, 10, 0.005, psize=0.001,
#       minevents = {'ele10' : 1e9}),
#     D(EME2, 0.025,   1.9,  2.3,       {'ele100':1,                        #10
#                                        None:2}, 10, psize=0.0003),
#     D(EME2, 0.025,   2.3,  2.45,      2, 10, psize=0.0003,                #11
#       ignorepars = {'ele25' : [(4,0),(5,0),(6,0)],
#                     'ele50' : [(4,0),(5,0),(6,0)]}),
#     D(EME2, 0.025,   2.45, 2.475,     {'ele25': 0,                        #12
#                                        'gam25': 0,
#                                        None:1},  3, 0.005, psize=0.001,
#       outlier_sigma = {'ele25': 2,
#                        'gam25': 2,
#                        'ele50': 3,
#                        'gam50': 3,
#                        'ele75': 3,
#                        'gam75': 3,
#                        None: 5}),
#     D(EME2, 0.025,   2.475,2.5,       {'ele25-37' : 1,                    #13
#                                        'ele25-35' : 2,
#                                        'gam25' : 1,
#                                        'gam50large' : 1,
#                                        'ele100-37' : 2,
#                                        None : 3}, 0, 0.005,
#       outlier_sigma = 3,
#       ignorepars=[(2,0)]),
#     ]
